@@ -1,6 +1,6 @@
 [
   inputs: [
-    "lib/*/{lib,unicode,test}/**/*.{ex,exs}",
+    "lib/*/{lib,scripts,unicode,test}/**/*.{ex,exs}",
     "lib/*/*.exs",
     "lib/ex_unit/examples/*.exs",
     ".formatter.exs"
@@ -13,7 +13,9 @@
     assert_same: 2,
 
     # Errors tests
-    assert_eval_raise: 3
-  ],
-  normalize_bitstring_modifiers: false
+    assert_eval_raise: 3,
+
+    # Float tests
+    float_assert: 1
+  ]
 ]
