@@ -3,10 +3,6 @@ defmodule Code.Fragment do
   This module provides conveniences for analyzing fragments of
   textual code and extract available information whenever possible.
 
-  Most of the functions in this module provide a best-effort
-  and may not be accurate under all circumstances. Read each
-  documentation for more information.
-
   This module should be considered experimental.
   """
 
@@ -17,13 +13,13 @@ defmodule Code.Fragment do
 
   This function receives a string with an Elixir code fragment,
   representing a cursor position, and based on the string, it
-  provides contextual information about said position. The
-  return of this function can then be used to provide tips,
+  provides contextual information about the latest token.
+  The return of this function can then be used to provide tips,
   suggestions, and autocompletion functionality.
 
-  This function provides a best-effort detection and may not be
-  accurate under all circumstances. See the "Limitations"
-  section below.
+  This function performs its analyses on tokens. This means
+  it does not understand how constructs are nested within each
+  other. See the "Limitations" section below.
 
   Consider adding a catch-all clause when handling the return
   type of this function as new cursor information may be added
@@ -35,7 +31,7 @@ defmodule Code.Fragment do
       :expr
 
       iex> Code.Fragment.cursor_context("hello_wor")
-      {:local_or_var, 'hello_wor'}
+      {:local_or_var, ~c"hello_wor"}
 
   ## Return values
 
@@ -82,6 +78,9 @@ defmodule Code.Fragment do
     * `{:local_call, charlist}` - the context is a local (import or local)
       call, such as `hello_world(` and `hello_world `
 
+    * `{:anonymous_call, inside_caller}` - the context is an anonymous
+      call, such as `fun.(` and `@fun.(`.
+
     * `{:module_attribute, charlist}` - the context is a module attribute,
       such as `@hello_wor`
 
@@ -110,11 +109,28 @@ defmodule Code.Fragment do
     * `{:unquoted_atom, charlist}` - the context is an unquoted atom. This
       can be any atom or an atom representing a module
 
+  We recommend looking at the test suite of this function for a complete list
+  of examples and their return values.
+
   ## Limitations
 
-  The current algorithm only considers the last line of the input. This means
-  it will also show suggestions inside strings, heredocs, etc, which is
-  intentional as it helps with doctests, references, and more.
+  The analysis is based on the current token, by analysing the last line of
+  the input. For example, this code:
+
+      iex> Code.Fragment.cursor_context("%URI{")
+      :expr
+
+  returns `:expr`, which suggests any variable, local function or alias
+  could be used. However, given we are inside a struct, the best suggestion
+  would be a struct field. In such cases, you can use
+  `container_cursor_to_quoted`, which will return the container of the AST
+  the cursor is currently within. You can then analyse this AST to provide
+  completion of field names.
+
+  As a consequence of its token-based implementation, this function considers
+  only the last line of the input. This means it will show suggestions inside
+  strings, heredocs, etc, which is intentional as it helps with doctests,
+  references, and more.
   """
   @doc since: "1.13.0"
   @spec cursor_context(List.Chars.t(), keyword()) ::
@@ -127,6 +143,7 @@ defmodule Code.Fragment do
           | {:local_or_var, charlist}
           | {:local_arity, charlist}
           | {:local_call, charlist}
+          | {:anonymous_call, inside_caller}
           | {:module_attribute, charlist}
           | {:operator, charlist}
           | {:operator_arity, charlist}
@@ -141,7 +158,8 @@ defmodule Code.Fragment do
                | {:dot, inside_dot, charlist}
                | {:module_attribute, charlist}
                | {:unquoted_atom, charlist}
-               | {:var, charlist},
+               | {:var, charlist}
+               | :expr,
              inside_alias:
                {:local_or_var, charlist}
                | {:module_attribute, charlist},
@@ -150,7 +168,8 @@ defmodule Code.Fragment do
                | {:alias, inside_alias, charlist}
                | {:local_or_var, charlist}
                | {:module_attribute, charlist}
-               | {:dot, inside_dot, charlist}
+               | {:dot, inside_dot, charlist},
+             inside_caller: {:var, charlist} | {:module_attribute, charlist}
   def cursor_context(fragment, opts \\ [])
 
   def cursor_context(fragment, opts)
@@ -166,12 +185,12 @@ defmodule Code.Fragment do
     cursor_context(to_charlist(other), opts)
   end
 
-  @operators '\\<>+-*/:=|&~^%!'
-  @starter_punctuation ',([{;'
-  @non_starter_punctuation ')]}"\'.$'
-  @space '\t\s'
-  @trailing_identifier '?!'
-  @tilde_op_prefix '<=~'
+  @operators ~c"\\<>+-*/:=|&~^%!"
+  @starter_punctuation ~c",([{;"
+  @non_starter_punctuation ~c")]}\"'.$"
+  @space ~c"\t\s"
+  @trailing_identifier ~c"?!"
+  @tilde_op_prefix ~c"<=~"
 
   @non_identifier @trailing_identifier ++
                     @operators ++ @starter_punctuation ++ @non_starter_punctuation ++ @space
@@ -186,9 +205,9 @@ defmodule Code.Fragment do
       # It is empty
       [] -> {:expr, 0}
       # Structs
-      [?%, ?:, ?: | _] -> {{:struct, ''}, 1}
-      [?%, ?: | _] -> {{:unquoted_atom, '%'}, 2}
-      [?% | _] -> {{:struct, ''}, 1}
+      [?%, ?:, ?: | _] -> {{:struct, ~c""}, 1}
+      [?%, ?: | _] -> {{:unquoted_atom, ~c"%"}, 2}
+      [?% | _] -> {{:struct, ~c""}, 1}
       # Token/AST only operators
       [?>, ?= | rest] when rest == [] or hd(rest) != ?: -> {:expr, 0}
       [?>, ?- | rest] when rest == [] or hd(rest) != ?: -> {:expr, 0}
@@ -198,7 +217,7 @@ defmodule Code.Fragment do
       [?: | rest] when rest == [] or hd(rest) != ?: -> unquoted_atom_or_expr(spaces)
       # Dots
       [?.] -> {:none, 0}
-      [?. | rest] when hd(rest) not in '.:' -> dot(rest, spaces + 1, '')
+      [?. | rest] when hd(rest) not in ~c".:" -> dot(rest, spaces + 1, ~c"")
       # It is a local or remote call with parens
       [?( | rest] -> call_to_cursor_context(strip_spaces(rest, spaces + 1))
       # A local arity definition
@@ -215,7 +234,7 @@ defmodule Code.Fragment do
   defp strip_spaces([h | rest], count) when h in @space, do: strip_spaces(rest, count + 1)
   defp strip_spaces(rest, count), do: {rest, count}
 
-  defp unquoted_atom_or_expr(0), do: {{:unquoted_atom, ''}, 1}
+  defp unquoted_atom_or_expr(0), do: {{:unquoted_atom, ~c""}, 1}
   defp unquoted_atom_or_expr(_), do: {:expr, 0}
 
   defp arity_to_cursor_context({reverse, spaces}) do
@@ -228,18 +247,29 @@ defmodule Code.Fragment do
   end
 
   defp call_to_cursor_context({reverse, spaces}) do
-    case identifier_to_cursor_context(reverse, spaces, true) do
-      {{:local_or_var, acc}, count} -> {{:local_call, acc}, count}
-      {{:dot, base, acc}, count} -> {{:dot_call, base, acc}, count}
-      {{:operator, acc}, count} -> {{:operator_call, acc}, count}
-      {_, _} -> {:none, 0}
+    with [?. | rest] <- reverse,
+         {rest, spaces} = strip_spaces(rest, spaces),
+         [h | _] when h not in @non_identifier <- rest do
+      case identifier_to_cursor_context(rest, spaces, true) do
+        {{:local_or_var, acc}, count} -> {{:anonymous_call, {:var, acc}}, count + 1}
+        {{:module_attribute, _} = attr, count} -> {{:anonymous_call, attr}, count + 1}
+        {_, _} -> {:none, 0}
+      end
+    else
+      _ ->
+        case identifier_to_cursor_context(reverse, spaces, true) do
+          {{:local_or_var, acc}, count} -> {{:local_call, acc}, count}
+          {{:dot, base, acc}, count} -> {{:dot_call, base, acc}, count}
+          {{:operator, acc}, count} -> {{:operator_call, acc}, count}
+          {_, _} -> {:none, 0}
+        end
     end
   end
 
-  defp identifier_to_cursor_context([?., ?., ?: | _], n, _), do: {{:unquoted_atom, '..'}, n + 3}
-  defp identifier_to_cursor_context([?., ?., ?. | _], n, _), do: {{:local_or_var, '...'}, n + 3}
-  defp identifier_to_cursor_context([?., ?: | _], n, _), do: {{:unquoted_atom, '.'}, n + 2}
-  defp identifier_to_cursor_context([?., ?. | _], n, _), do: {{:operator, '..'}, n + 2}
+  defp identifier_to_cursor_context([?., ?., ?: | _], n, _), do: {{:unquoted_atom, ~c".."}, n + 3}
+  defp identifier_to_cursor_context([?., ?., ?. | _], n, _), do: {{:local_or_var, ~c"..."}, n + 3}
+  defp identifier_to_cursor_context([?., ?: | _], n, _), do: {{:unquoted_atom, ~c"."}, n + 2}
+  defp identifier_to_cursor_context([?., ?. | _], n, _), do: {{:operator, ~c".."}, n + 2}
 
   defp identifier_to_cursor_context(reverse, count, call_op?) do
     case identifier(reverse, count) do
@@ -263,10 +293,10 @@ defmodule Code.Fragment do
 
       {:alias, rest, acc, count} ->
         case strip_spaces(rest, count) do
-          {'.' ++ rest, count} when rest == [] or hd(rest) != ?. ->
+          {~c"." ++ rest, count} when rest == [] or hd(rest) != ?. ->
             nested_alias(rest, count + 1, acc)
 
-          {'%' ++ _, count} ->
+          {~c"%" ++ _, count} ->
             {{:struct, acc}, count + 1}
 
           _ ->
@@ -278,18 +308,21 @@ defmodule Code.Fragment do
 
       {:identifier, [?%], acc, count} ->
         case identifier_to_cursor_context(acc |> Enum.reverse(), count, true) do
-          {{:local_or_var, _} = idenifier, _} -> {{:struct, idenifier}, count + 1}
+          {{:local_or_var, _} = identifier, _} -> {{:struct, identifier}, count + 1}
           _ -> {:none, 0}
         end
 
       {:identifier, rest, acc, count} ->
         case strip_spaces(rest, count) do
-          {'.' ++ rest, count} when rest == [] or hd(rest) != ?. ->
+          {~c"." ++ rest, count} when rest == [] or hd(rest) != ?. ->
             dot(rest, count + 1, acc)
 
           _ ->
             {{:local_or_var, acc}, count}
         end
+
+      {:capture_arg, acc, count} ->
+        {{:capture_arg, acc}, count}
     end
   end
 
@@ -310,7 +343,7 @@ defmodule Code.Fragment do
     case tokenize_identifier(rest, count, acc) do
       {:identifier, [?% | _rest], acc, count} -> {:struct, {:module_attribute, acc}, count}
       {:identifier, _rest, acc, count} -> {:module_attribute, acc, count}
-      :none when acc == [] -> {:module_attribute, '', count}
+      :none when acc == [] -> {:module_attribute, ~c"", count}
       _ -> :none
     end
   end
@@ -332,6 +365,14 @@ defmodule Code.Fragment do
     :none
   end
 
+  defp rest_identifier([?& | tail] = rest, count, acc) when tail == [] or hd(tail) != ?& do
+    if Enum.all?(acc, &(&1 in ?0..?9)) do
+      {:capture_arg, [?& | acc], count + 1}
+    else
+      tokenize_identifier(rest, count, acc)
+    end
+  end
+
   defp rest_identifier(rest, count, acc) do
     tokenize_identifier(rest, count, acc)
   end
@@ -363,19 +404,19 @@ defmodule Code.Fragment do
 
     case identifier_to_cursor_context(rest, count, true) do
       {{:struct, prev}, count} when is_list(prev) ->
-        {{:struct, prev ++ '.' ++ acc}, count}
+        {{:struct, prev ++ ~c"." ++ acc}, count}
 
       {{:struct, {:alias, parent, prev}}, count} ->
-        {{:struct, {:alias, parent, prev ++ '.' ++ acc}}, count}
+        {{:struct, {:alias, parent, prev ++ ~c"." ++ acc}}, count}
 
       {{:struct, prev}, count} ->
         {{:struct, {:alias, prev, acc}}, count}
 
       {{:alias, prev}, count} ->
-        {{:alias, prev ++ '.' ++ acc}, count}
+        {{:alias, prev ++ ~c"." ++ acc}, count}
 
       {{:alias, parent, prev}, count} ->
-        {{:alias, parent, prev ++ '.' ++ acc}, count}
+        {{:alias, parent, prev ++ ~c"." ++ acc}, count}
 
       {{:local_or_var, prev}, count} ->
         {{:alias, {:local_or_var, prev}, acc}, count}
@@ -416,6 +457,9 @@ defmodule Code.Fragment do
       {{:module_attribute, _} = prev, count} ->
         {{:dot, prev, acc}, count}
 
+      {:expr, count} ->
+        {{:dot, :expr, acc}, count}
+
       {_, _} ->
         {:none, 0}
     end
@@ -427,32 +471,36 @@ defmodule Code.Fragment do
 
   # If we are opening a sigil, ignore the operator.
   defp operator([letter, ?~ | rest], _count, [op], _call_op?)
-       when op in '<|/' and (letter in ?A..?Z or letter in ?a..?z) and
+       when op in ~c"<|/" and (letter in ?A..?Z or letter in ?a..?z) and
               (rest == [] or hd(rest) not in @tilde_op_prefix) do
     {:none, 0}
   end
 
-  defp operator(rest, count, '~', call_op?) do
+  defp operator(rest, count, ~c"~", call_op?) do
     {rest, _} = strip_spaces(rest, count)
 
     if call_op? or match?([?. | rest] when rest == [] or hd(rest) != ?., rest) do
       {:none, 0}
     else
-      {{:sigil, ''}, count}
+      {{:sigil, ~c""}, count}
     end
   end
 
+  defp operator([?) | rest], _, [], true) when hd(rest) != ?? do
+    {:expr, 0}
+  end
+
   defp operator(rest, count, acc, _call_op?) do
     case :elixir_tokenizer.tokenize(acc, 1, 1, []) do
-      {:ok, _, _, _, [{:atom, _, _}]} ->
+      {:ok, _, _, _, [{:atom, _, _}], []} ->
         {{:unquoted_atom, tl(acc)}, count}
 
-      {:ok, _, _, _, [{_, _, op}]} ->
+      {:ok, _, _, _, [{_, _, op}], []} ->
         {rest, dot_count} = strip_spaces(rest, count)
 
         cond do
           Code.Identifier.unary_op(op) == :error and Code.Identifier.binary_op(op) == :error ->
-            :none
+            {:none, 0}
 
           match?([?. | rest] when rest == [] or hd(rest) != ?., rest) ->
             dot(tl(rest), dot_count + 1, acc)
@@ -472,7 +520,8 @@ defmodule Code.Fragment do
   This function receives a string with an Elixir code fragment
   and a `position`. It returns a map containing the beginning
   and ending of the identifier alongside its context, or `:none`
-  if there is nothing with a known context.
+  if there is nothing with a known context. This is useful to
+  provide mouse-over and highlight functionality in editors.
 
   The difference between `cursor_context/2` and `surround_context/3`
   is that the former assumes the expression in the code fragment
@@ -501,20 +550,20 @@ defmodule Code.Fragment do
   The returned map contains the column the expression starts and the
   first column after the expression ends.
 
-  Similar to `cursor_context/2`, this function also provides a best-effort
-  detection and may not be accurate under all circumstances. See the
-  "Return values" and "Limitations" section under `cursor_context/2` for
-  more information.
+  Similar to `cursor_context/2`, this function is also token-based
+  and may not be accurate under all circumstances. See the
+  "Return values" and "Limitations" section under `cursor_context/2`
+  for more information.
 
   ## Examples
 
       iex> Code.Fragment.surround_context("foo", {1, 1})
-      %{begin: {1, 1}, context: {:local_or_var, 'foo'}, end: {1, 4}}
+      %{begin: {1, 1}, context: {:local_or_var, ~c"foo"}, end: {1, 4}}
 
   ## Differences to `cursor_context/2`
 
-  Because `surround_context/3` deals with complete code, it has some
-  difference to `cursor_context/2`:
+  Because `surround_context/3` attempts to capture complex expressions,
+  it has some differences to `cursor_context/2`:
 
     * `dot_call`/`dot_arity` and `operator_call`/`operator_arity`
       are collapsed into `dot` and `operator` contexts respectively
@@ -524,16 +573,18 @@ defmodule Code.Fragment do
       `local_call`/`local_arity` and `local_or_var`, since the latter can
       be a local or variable
 
-    * `@` when not followed by any identifier is returned as `{:operator, '@'}`
-      (in contrast to `{:module_attribute, ''}` in `cursor_context/2`
+    * `@` when not followed by any identifier is returned as `{:operator, ~c"@"}`
+      (in contrast to `{:module_attribute, ~c""}` in `cursor_context/2`
 
-    * This function never returns empty sigils `{:sigil, ''}` or empty structs
-      `{:struct, ''}` as context
+    * This function never returns empty sigils `{:sigil, ~c""}` or empty structs
+      `{:struct, ~c""}` as context
 
-    * This function returns keywords as `{:keyword, 'do'}`
+    * This function returns keywords as `{:keyword, ~c"do"}`
 
     * This function never returns `:expr`
 
+  We recommend looking at the test suite of this function for a complete list
+  of examples and their return values.
   """
   @doc since: "1.13.0"
   @spec surround_context(List.Chars.t(), position(), keyword()) ::
@@ -550,14 +601,17 @@ defmodule Code.Fragment do
                | {:sigil, charlist}
                | {:struct, inside_struct}
                | {:unquoted_atom, charlist}
-               | {:keyword, charlist},
+               | {:keyword, charlist}
+               | {:key, charlist}
+               | {:capture_arg, charlist},
              inside_dot:
                {:alias, charlist}
                | {:alias, inside_alias, charlist}
                | {:dot, inside_dot, charlist}
                | {:module_attribute, charlist}
                | {:unquoted_atom, charlist}
-               | {:var, charlist},
+               | {:var, charlist}
+               | :expr,
              inside_alias:
                {:local_or_var, charlist}
                | {:module_attribute, charlist},
@@ -599,7 +653,16 @@ defmodule Code.Fragment do
         maybe_operator(reversed_pre, post, line, opts)
 
       {:identifier, reversed_post, rest} ->
-        {rest, _} = strip_spaces(rest, 0)
+        {keyword_key?, rest} =
+          case rest do
+            [?: | tail] when tail == [] or hd(tail) in @space ->
+              {true, rest}
+
+            _ ->
+              {rest, _} = strip_spaces(rest, 0)
+              {false, rest}
+          end
+
         reversed = reversed_post ++ reversed_pre
 
         case codepoint_cursor_context(reversed, opts) do
@@ -615,6 +678,9 @@ defmodule Code.Fragment do
           {{:dot, _, [_ | _]} = dot, offset} ->
             build_surround(dot, reversed, line, offset)
 
+          {{:local_or_var, acc}, offset} when keyword_key? ->
+            build_surround({:key, acc}, reversed, line, offset)
+
           {{:local_or_var, acc}, offset} when hd(rest) == ?( ->
             build_surround({:local_call, acc}, reversed, line, offset)
 
@@ -630,8 +696,8 @@ defmodule Code.Fragment do
           {{:local_or_var, acc}, offset} ->
             build_surround({:local_or_var, acc}, reversed, line, offset)
 
-          {{:module_attribute, ''}, offset} ->
-            build_surround({:operator, '@'}, reversed, line, offset)
+          {{:module_attribute, ~c""}, offset} ->
+            build_surround({:operator, ~c"@"}, reversed, line, offset)
 
           {{:module_attribute, acc}, offset} ->
             build_surround({:module_attribute, acc}, reversed, line, offset)
@@ -642,6 +708,9 @@ defmodule Code.Fragment do
           {{:unquoted_atom, acc}, offset} ->
             build_surround({:unquoted_atom, acc}, reversed, line, offset)
 
+          {{:capture_arg, acc}, offset} ->
+            build_surround({:capture_arg, acc}, reversed, line, offset)
+
           _ ->
             maybe_operator(reversed_pre, post, line, opts)
         end
@@ -653,6 +722,9 @@ defmodule Code.Fragment do
           {{:alias, acc}, offset} ->
             build_surround({:alias, acc}, reversed, line, offset)
 
+          {{:alias, parent, acc}, offset} ->
+            build_surround({:alias, parent, acc}, reversed, line, offset)
+
           {{:struct, acc}, offset} ->
             build_surround({:struct, acc}, reversed, line, offset)
 
@@ -671,10 +743,20 @@ defmodule Code.Fragment do
         reversed = reversed_post ++ reversed_pre
 
         case codepoint_cursor_context(reversed, opts) do
+          {{:operator, ~c"&"}, offset} when hd(rest) in ?0..?9 ->
+            arg = Enum.take_while(rest, &(&1 in ?0..?9))
+
+            build_surround(
+              {:capture_arg, ~c"&" ++ arg},
+              :lists.reverse(arg, reversed),
+              line,
+              offset + length(arg)
+            )
+
           {{:operator, acc}, offset} ->
             build_surround({:operator, acc}, reversed, line, offset)
 
-          {{:sigil, ''}, offset} when hd(rest) in ?A..?Z or hd(rest) in ?a..?z ->
+          {{:sigil, ~c""}, offset} when hd(rest) in ?A..?Z or hd(rest) in ?a..?z ->
             build_surround({:sigil, [hd(rest)]}, [hd(rest) | reversed], line, offset + 1)
 
           {{:dot, _, [_ | _]} = dot, offset} ->
@@ -713,7 +795,7 @@ defmodule Code.Fragment do
     end
   end
 
-  defp take_alias([h | t], acc) when h not in @non_identifier,
+  defp take_alias([h | t], acc) when h in ?A..?Z or h in ?a..?z or h in ?0..?9 or h == ?_,
     do: take_alias(t, [h | acc])
 
   defp take_alias(rest, acc) do
@@ -750,7 +832,7 @@ defmodule Code.Fragment do
       _ ->
         case strip_spaces(reversed_pre, 0) do
           # If there is a dot to our left, make sure to move to the first character
-          {[?. | rest], _} when rest == [] or hd(rest) not in '.:' ->
+          {[?. | rest], _} when rest == [] or hd(rest) not in ~c".:" ->
             {post, reversed_pre} = move_spaces(post, reversed_pre)
             {reversed_pre, post}
 
@@ -798,9 +880,9 @@ defmodule Code.Fragment do
   defp last_line(charlist) when is_list(charlist) do
     [last_line | lines_reverse] =
       charlist
-      |> :string.replace('\r\n', '\n', :all)
-      |> :string.join('')
-      |> :string.split('\n', :all)
+      |> :string.replace(~c"\r\n", ~c"\n", :all)
+      |> :string.join(~c"")
+      |> :string.split(~c"\n", :all)
       |> Enum.reverse()
 
     prepend_cursor_lines(lines_reverse, last_line)
@@ -844,9 +926,9 @@ defmodule Code.Fragment do
 
   defp surround_line(charlist, line, column) when is_list(charlist) do
     charlist
-    |> :string.replace('\r\n', '\n', :all)
-    |> :string.join('')
-    |> :string.split('\n', :all)
+    |> :string.replace(~c"\r\n", ~c"\n", :all)
+    |> :string.join(~c"")
+    |> :string.split(~c"\n", :all)
     |> surround_lines(line, column)
   end
 
@@ -929,17 +1011,14 @@ defmodule Code.Fragment do
 
   @doc """
   Receives a string and returns a quoted expression
-  with a cursor at the nearest argument position.
+  with the cursor AST position within its parent expression.
 
   This function receives a string with an Elixir code fragment,
   representing a cursor position, and converts such string to
-  AST with the inclusion of special `__cursor__()` node based
-  on the position of the cursor with a container.
+  AST with the inclusion of special `__cursor__()` node representing
+  the cursor position within its container (i.e. its parent).
 
-  A container is any Elixir expression starting with `(`,
-  `{`, and `[`. This includes function calls, tuples, lists,
-  maps, and so on. For example, take this code, which would
-  be given as input:
+  For example, take this code, which would be given as input:
 
       max(some_value,
 
@@ -948,9 +1027,9 @@ defmodule Code.Fragment do
       max(some_value, __cursor__())
 
   In other words, this function is capable of closing any open
-  brackets and insert the cursor position. Any content at the
-  cursor position that is after a comma or an opening bracket
-  is discarded. For example, if this is given as input:
+  brackets and insert the cursor position. Other content at the
+  cursor position which is not a parent is discarded.
+  For example, if this is given as input:
 
       max(some_value, another_val
 
@@ -969,16 +1048,7 @@ defmodule Code.Fragment do
   Calls without parenthesis are also supported, as we assume the
   brackets are implicit.
 
-  Operators and anonymous functions are not containers, and therefore
-  will be discarded. The following will all return the same AST:
-
-      max(some_value,
-      max(some_value, fn x -> x end
-      max(some_value, 1 + another_val
-      max(some_value, 1 |> some_fun() |> another_fun
-
-  On the other hand, tuples, lists, maps, and binaries all retain the
-  cursor position:
+  Tuples, lists, maps, and binaries all retain the cursor position:
 
       max(some_value, [1, 2,
 
@@ -996,6 +1066,8 @@ defmodule Code.Fragment do
 
       if(some_value, do: __cursor__())
 
+  For multi-line blocks, all previous lines are preserved.
+
   The AST returned by this function is not safe to evaluate but
   it can be analyzed and expanded.
 
@@ -1011,10 +1083,28 @@ defmodule Code.Fragment do
       iex> Code.Fragment.container_cursor_to_quoted("[some, value")
       {:ok, [{:some, [line: 1], nil}, {:__cursor__, [line: 1], []}]}
 
-  For binaries, the `::` is exclusively kept as an operator:
+  If an expression is complete, then the whole expression is discarded
+  and only the parent is returned:
 
-      iex> Code.Fragment.container_cursor_to_quoted("<<some::integer")
-      {:ok, {:<<>>, [line: 1], [{:"::", [line: 1], [{:some, [line: 1], nil}, {:__cursor__, [line: 1], []}]}]}}
+      iex> Code.Fragment.container_cursor_to_quoted("if(is_atom(var)")
+      {:ok, {:if, [line: 1], [{:__cursor__, [line: 1], []}]}}
+
+  this means complete expressions themselves return only the cursor:
+
+      iex> Code.Fragment.container_cursor_to_quoted("if(is_atom(var))")
+      {:ok, {:__cursor__, [line: 1], []}}
+
+  Operators are also included from Elixir v1.15:
+
+      iex> Code.Fragment.container_cursor_to_quoted("foo +")
+      {:ok, {:+, [line: 1], [{:foo, [line: 1], nil}, {:__cursor__, [line: 1], []}]}}
+
+  In order to parse the left-side of `->` properly, which appears both
+  in anonymous functions and do-end blocks, the trailing fragment option
+  must be given with the rest of the contents:
+
+      iex> Code.Fragment.container_cursor_to_quoted("fn x", trailing_fragment: " -> :ok end")
+      {:ok, {:fn, [line: 1], [{:->, [line: 1], [[{:__cursor__, [line: 1], []}], :ok]}]}}
 
   ## Options
 
@@ -1038,14 +1128,105 @@ defmodule Code.Fragment do
     * `:literal_encoder` - a function to encode literals in the AST.
       See the documentation for `Code.string_to_quoted/2` for more information.
 
+    * `:trailing_fragment` (since v1.18.0) - the rest of the contents after
+      the cursor. This is necessary to correctly complete anonymous functions
+      and the left-hand side of `->`
+
   """
   @doc since: "1.13.0"
   @spec container_cursor_to_quoted(List.Chars.t(), keyword()) ::
           {:ok, Macro.t()} | {:error, {location :: keyword, binary | {binary, binary}, binary}}
   def container_cursor_to_quoted(fragment, opts \\ []) do
-    opts =
-      Keyword.take(opts, [:file, :line, :column, :columns, :token_metadata, :literal_encoder])
+    {trailing_fragment, opts} = Keyword.pop(opts, :trailing_fragment)
+    opts = Keyword.take(opts, [:columns, :token_metadata, :literal_encoder])
+    opts = [check_terminators: {:cursor, []}, emit_warnings: false] ++ opts
 
-    Code.string_to_quoted(fragment, [cursor_completion: true, emit_warnings: false] ++ opts)
+    file = Keyword.get(opts, :file, "nofile")
+    line = Keyword.get(opts, :line, 1)
+    column = Keyword.get(opts, :column, 1)
+
+    case :elixir_tokenizer.tokenize(to_charlist(fragment), line, column, opts) do
+      {:ok, line, column, _warnings, rev_tokens, rev_terminators}
+      when trailing_fragment == nil ->
+        {rev_tokens, rev_terminators} =
+          with [close, open, {_, _, :__cursor__} = cursor | rev_tokens] <- rev_tokens,
+               {_, [_ | after_fn]} <- Enum.split_while(rev_terminators, &(elem(&1, 0) != :fn)),
+               true <- maybe_missing_stab?(rev_tokens),
+               [_ | rev_tokens] <- Enum.drop_while(rev_tokens, &(elem(&1, 0) != :fn)) do
+            {[close, open, cursor | rev_tokens], after_fn}
+          else
+            _ -> {rev_tokens, rev_terminators}
+          end
+
+        tokens = reverse_tokens(line, column, rev_tokens, rev_terminators)
+        :elixir.tokens_to_quoted(tokens, file, opts)
+
+      {:ok, line, column, _warnings, rev_tokens, rev_terminators} ->
+        tokens =
+          with {before_start, [_ | _] = after_start} <-
+                 Enum.split_while(rev_terminators, &(elem(&1, 0) not in [:do, :fn])),
+               true <- maybe_missing_stab?(rev_tokens),
+               opts =
+                 Keyword.put(opts, :check_terminators, {:cursor, before_start}),
+               {:error, {meta, _, ~c"end"}, _rest, _warnings, trailing_rev_tokens} <-
+                 :elixir_tokenizer.tokenize(to_charlist(trailing_fragment), line, column, opts) do
+            trailing_tokens =
+              reverse_tokens(meta[:line], meta[:column], trailing_rev_tokens, after_start)
+
+            Enum.reverse(rev_tokens, drop_tokens(trailing_tokens, 0))
+          else
+            _ -> reverse_tokens(line, column, rev_tokens, rev_terminators)
+          end
+
+        :elixir.tokens_to_quoted(tokens, file, opts)
+
+      {:error, info, _rest, _warnings, _so_far} ->
+        {:error, :elixir.format_token_error(info)}
+    end
   end
+
+  defp reverse_tokens(line, column, tokens, terminators) do
+    {terminators, _} =
+      Enum.map_reduce(terminators, column, fn {start, _, _}, column ->
+        atom = :elixir_tokenizer.terminator(start)
+
+        {{atom, {line, column, nil}}, column + length(Atom.to_charlist(atom))}
+      end)
+
+    Enum.reverse(tokens, terminators)
+  end
+
+  defp drop_tokens([{:"}", _} | _] = tokens, 0), do: tokens
+  defp drop_tokens([{:"]", _} | _] = tokens, 0), do: tokens
+  defp drop_tokens([{:")", _} | _] = tokens, 0), do: tokens
+  defp drop_tokens([{:">>", _} | _] = tokens, 0), do: tokens
+  defp drop_tokens([{:end, _} | _] = tokens, 0), do: tokens
+  defp drop_tokens([{:",", _} | _] = tokens, 0), do: tokens
+  defp drop_tokens([{:stab_op, _, :->} | _] = tokens, 0), do: tokens
+
+  defp drop_tokens([{:"}", _} | tokens], counter), do: drop_tokens(tokens, counter - 1)
+  defp drop_tokens([{:"]", _} | tokens], counter), do: drop_tokens(tokens, counter - 1)
+  defp drop_tokens([{:")", _} | tokens], counter), do: drop_tokens(tokens, counter - 1)
+  defp drop_tokens([{:">>", _} | tokens], counter), do: drop_tokens(tokens, counter - 1)
+  defp drop_tokens([{:end, _} | tokens], counter), do: drop_tokens(tokens, counter - 1)
+
+  defp drop_tokens([{:"{", _} | tokens], counter), do: drop_tokens(tokens, counter + 1)
+  defp drop_tokens([{:"[", _} | tokens], counter), do: drop_tokens(tokens, counter + 1)
+  defp drop_tokens([{:"(", _} | tokens], counter), do: drop_tokens(tokens, counter + 1)
+  defp drop_tokens([{:"<<", _} | tokens], counter), do: drop_tokens(tokens, counter + 1)
+  defp drop_tokens([{:fn, _} | tokens], counter), do: drop_tokens(tokens, counter + 1)
+  defp drop_tokens([{:do, _} | tokens], counter), do: drop_tokens(tokens, counter + 1)
+
+  defp drop_tokens([_ | tokens], counter), do: drop_tokens(tokens, counter)
+  defp drop_tokens([], 0), do: []
+
+  defp maybe_missing_stab?([{:after, _} | _]), do: true
+  defp maybe_missing_stab?([{:do, _} | _]), do: true
+  defp maybe_missing_stab?([{:fn, _} | _]), do: true
+  defp maybe_missing_stab?([{:else, _} | _]), do: true
+  defp maybe_missing_stab?([{:catch, _} | _]), do: true
+  defp maybe_missing_stab?([{:rescue, _} | _]), do: true
+  defp maybe_missing_stab?([{:stab_op, _, :->} | _]), do: false
+  defp maybe_missing_stab?([_ | tail]), do: maybe_missing_stab?(tail)
+  defp maybe_missing_stab?([]), do: false
 end
