@@ -65,7 +65,7 @@ defmodule Module do
   For detailed documentation, see the
   [behaviour typespec documentation](typespecs.md#behaviours).
 
-  ### `@impl`
+  ### `@impl` (since v1.5.0)
 
   To aid in the correct implementation of behaviours, you may optionally declare
   `@impl` for implemented callbacks of a behaviour. This makes callbacks
@@ -237,7 +237,7 @@ defmodule Module do
   Accepts an atom, a tuple, or a list of atoms and tuples. For example:
 
       defmodule MyModule do
-        @dialyzer {:nowarn_function, my_fun: 1}
+        @dialyzer {:nowarn_function, [my_fun: 1]}
 
         def my_fun(arg) do
           M.not_a_function(arg)
@@ -261,10 +261,18 @@ defmodule Module do
   in case any of the external resources change, see for example:
   [`mix compile.elixir`](https://hexdocs.pm/mix/Mix.Tasks.Compile.Elixir.html).
 
+  The specified file path provided is interpreted as relative to
+  the folder containing the project's `mix.exs`, which is the
+  current working directory, not the file where `@external_resource`
+  is declared.
+
   If the external resource does not exist, the module still has
   a dependency on it, causing the module to be recompiled as soon
   as the file is added.
 
+  For more control over when a module is recompiled, see
+  [`__mix_recompile__?/0`](`m:Mix.Tasks.Compile.Elixir#module-__mix_recompile__-0`).
+
   ### `@file`
 
   Changes the filename used in stacktraces for the function or macro that
@@ -278,6 +286,24 @@ defmodule Module do
         end
       end
 
+  Note that this is only valid for exceptions/diagnostics that come from the
+  definition inner scope (which includes its patterns and guards). For example:
+
+      defmodule MyModule do # <---- module definition
+        @file "hello.ex"
+        defp unused(a) do # <---- function definition
+          "world" # <---- function scope
+        end
+
+        @file "bye.ex"
+        def unused(_), do: true
+      end
+
+  If you run this code with the second "unused" definition commented, you will
+  see that `hello.ex` is used as the stacktrace when reporting warnings, but if
+  you uncomment it you'll see that the error will not mention `bye.ex`, because
+  it's a module-level error rather than an expression-level error.
+
   ### `@moduledoc`
 
   Provides documentation for the current module.
@@ -300,6 +326,21 @@ defmodule Module do
   Once this module is compiled, this information becomes available via
   the `Code.fetch_docs/1` function.
 
+  ### `@nifs` (since v1.16.0)
+
+  A list of functions and their arities which will be overridden
+  by a native implementation (NIF).
+
+      defmodule MyLibrary.MyModule do
+        @nifs [foo: 1, bar: 2]
+
+        def foo(arg1), do: :erlang.nif_error(:not_loaded)
+        def bar(arg1, arg2), do: :erlang.nif_error(:not_loaded)
+      end
+
+  See the Erlang documentation for more information:
+  https://www.erlang.org/doc/man/erl_nif
+
   ### `@on_definition`
 
   A hook that will be invoked when each function or macro in the current
@@ -403,7 +444,8 @@ defmodule Module do
     * `@typep` - defines a private type to be used in `@spec`
     * `@opaque` - defines an opaque type to be used in `@spec`
     * `@spec` - provides a specification for a function
-    * `@callback` - provides a specification for a behaviour callback
+    * `@callback` - provides a specification for a behaviour callback (and generates
+      a `behaviour_info/1` function in the module, see below)
     * `@macrocallback` - provides a specification for a macro behaviour callback
     * `@optional_callbacks` - specifies which behaviour callbacks and macro
       behaviour callbacks are optional
@@ -500,11 +542,13 @@ defmodule Module do
   undefined functions, deprecations, etc. A module is always verified after
   it is compiled. In Mix projects, a module is also verified when any of its
   runtime dependencies change. Therefore this is useful to perform verification
-  of the current module while avoiding compile-time dependencies.
+  of the current module while avoiding compile-time dependencies. Given the
+  callback is invoked under different scenarios, Elixir provides no guarantees
+  of when in the compilation cycle nor in which process the callback runs.
 
   Accepts a module or a `{module, function_name}` tuple. The function
   must take one argument: the module name. When just a module is provided,
-  the function is assumed to be `__after_verify__/2`.
+  the function is assumed to be `__after_verify__/1`.
 
   Callbacks will run in the order they are registered.
 
@@ -532,7 +576,11 @@ defmodule Module do
       corresponding setting in `Code.get_compiler_option/1`
 
     * `@compile {:debug_info, false}` - disables `:debug_info` regardless
-      of the corresponding setting in `Code.get_compiler_option/1`
+      of the corresponding setting in `Code.get_compiler_option/1`. Note
+      disabling `:debug_info` is not recommended as it removes the ability
+      of the Elixir compiler and other tools to static analyse the code.
+      If you want to remove the `:debug_info` while deploying, tools like
+      `mix release` already do such by default.
 
     * `@compile {:inline, some_fun: 2, other_fun: 3}` - inlines the given
       name/arity pairs. Inlining is applied locally, calls from another
@@ -546,6 +594,78 @@ defmodule Module do
       `@compile {:no_warn_undefined, {Mod, fun, arity}}` - does not warn if
       the given module or the given `Mod.fun/arity` are not defined
 
+  ## Generated functions
+
+  Sometimes the compiler will generate public functions within modules. These
+  are documented below.
+
+  ### `behaviour_info/1`
+
+  This function is generated for modules that define a behaviour, that is,
+  that have one or more `@callback` definitions. The signature for this function,
+  expressed as a spec, is:
+
+      @spec behaviour_info(:callbacks) :: [function_info]
+        when function_info: {function_name :: atom(), arity :: non_neg_integer()}
+
+      @spec behaviour_info(:optional_callbacks) :: [function_info]
+        when function_info: {function_name :: atom(), arity :: non_neg_integer()}
+
+  `behaviour_info(:callbacks)` includes optional callbacks.
+
+  For example:
+
+      iex> Enum.sort(GenServer.behaviour_info(:callbacks))
+      [
+        code_change: 3,
+        format_status: 1,
+        format_status: 2,
+        handle_call: 3,
+        handle_cast: 2,
+        handle_continue: 2,
+        handle_info: 2,
+        init: 1,
+        terminate: 2
+      ]
+
+  ### `module_info/0`
+
+  This function is generated for all modules. It returns all the attributes
+  returned by `module_info/1` (see below), but as a single keyword list. See also the
+  [Erlang documentation](https://www.erlang.org/doc/system/modules.html#module_info-0).
+
+  ### `module_info/1`
+
+  This function is generated for all modules and returns
+  information about the module. The signature for this function,
+  expressed as a spec, is:
+
+      @spec module_info(:module) :: module() # Returns the module itself
+      @spec module_info(:attributes) :: keyword()
+      @spec module_info(:compile) :: keyword()
+      @spec module_info(:md5) :: binary()
+      @spec module_info(:nifs) :: module()
+      @spec module_info(:exports) :: [function_info]
+        when function_info: {function_name :: atom(), arity :: non_neg_integer()}
+      @spec module_info(:functions) :: [function_info]
+        when function_info: {function_name :: atom(), arity :: non_neg_integer()}
+
+  For example:
+
+      iex> URI.module_info(:module)
+      URI
+      iex> {:decode_www_form, 1} in URI.module_info(:exports)
+      true
+
+  For more information about `module_info/1`, also check out the [Erlang
+  documentation](https://www.erlang.org/doc/system/modules.html#module_info-1).
+
+  ### `__info__/1`
+
+  This function is generated for all modules. It's similar to `module_info/1` but
+  includes some additional Elixir-specific information, such as struct and macro
+  information. For documentation, see `c:Module.__info__/1`.
+
   '''
 
   @type definition :: {atom, arity}
@@ -575,7 +695,7 @@ defmodule Module do
 
     * `:module` - the module atom name
 
-    * `:struct` - if the module defines a struct and if so each field in order
+    * `:struct` - (since v1.14.0) if the module defines a struct and if so each field in order
 
   """
   @callback __info__(:attributes) :: keyword()
@@ -584,7 +704,8 @@ defmodule Module do
   @callback __info__(:macros) :: keyword()
   @callback __info__(:md5) :: binary()
   @callback __info__(:module) :: module()
-  @callback __info__(:struct) :: list(%{field: atom(), required: boolean()}) | nil
+  @callback __info__(:struct) ::
+              list(%{required(:field) => atom(), optional(:default) => term()}) | nil
 
   @doc """
   Returns information about module attributes used by Elixir.
@@ -602,6 +723,7 @@ defmodule Module do
 
   """
   @doc since: "1.12.0"
+  @spec reserved_attributes() :: map
   def reserved_attributes() do
     %{
       after_compile: %{
@@ -616,6 +738,25 @@ defmodule Module do
       behaviour: %{
         doc: "Specifies that the current module implements a given behaviour."
       },
+      enforce_keys: %{
+        doc:
+          "Ensures the given keys are always set when building the struct defined in the current module."
+      },
+      fallback_to_any: %{
+        doc:
+          "If set to `true` generates a default protocol implementation for all types (inside `defprotocol`)."
+      },
+      undefined_impl_description: %{
+        doc:
+          "A string with additional description to be used on `Protocol.UndefinedError` when looking up the implementation fails."
+      },
+      for: %{
+        doc:
+          "The current module/type a protocol implementation is being defined for (inside `defimpl`)."
+      },
+      protocol: %{
+        doc: "The current protocol being implemented (inside `defimpl`)."
+      },
       on_definition: %{
         doc:
           "A hook that will be invoked when each function or macro in the current module is defined."
@@ -693,50 +834,7 @@ defmodule Module do
     :elixir_module.is_open(module)
   end
 
-  @doc """
-  Evaluates the quoted contents in the given module's context.
-
-  A list of environment options can also be given as argument.
-  See `Code.eval_string/3` for more information.
-
-  Raises an error if the module was already compiled.
-
-  ## Examples
-
-      defmodule Foo do
-        contents =
-          quote do
-            def sum(a, b), do: a + b
-          end
-
-        Module.eval_quoted(__MODULE__, contents)
-      end
-
-      Foo.sum(1, 2)
-      #=> 3
-
-  For convenience, you can pass any `Macro.Env` struct, such
-  as  `__ENV__/0`, as the first argument or as options. Both
-  the module and all options will be automatically extracted
-  from the environment:
-
-      defmodule Foo do
-        contents =
-          quote do
-            def sum(a, b), do: a + b
-          end
-
-        Module.eval_quoted(__ENV__, contents)
-      end
-
-      Foo.sum(1, 2)
-      #=> 3
-
-  Note that if you pass a `Macro.Env` struct as first argument
-  while also passing `opts`, they will be merged with `opts`
-  having precedence.
-  """
-  @spec eval_quoted(module | Macro.Env.t(), Macro.t(), list, keyword | Macro.Env.t()) :: term
+  @deprecated "Use Code.eval_quoted/3 instead"
   def eval_quoted(module_or_env, quoted, binding \\ [], opts \\ [])
 
   def eval_quoted(%Macro.Env{} = env, quoted, binding, opts)
@@ -767,7 +865,8 @@ defmodule Module do
   the given quoted expressions.
 
   The line where the module is defined and its file **must**
-  be passed as options.
+  be passed as options. See `Code.env_for_eval/1` for a complete
+  list of options.
 
   It returns a tuple of shape `{:module, module, binary, term}`
   where `module` is the module name, `binary` is the module
@@ -807,18 +906,22 @@ defmodule Module do
   def create(module, quoted, opts)
 
   def create(module, quoted, %Macro.Env{} = env) when is_atom(module) do
-    create(module, quoted, Map.to_list(env))
+    create([line: env.line], module, quoted, env)
   end
 
   def create(module, quoted, opts) when is_atom(module) and is_list(opts) do
-    unless Keyword.has_key?(opts, :file) do
+    if not Keyword.has_key?(opts, :file) do
       raise ArgumentError, "expected :file to be given as option"
     end
 
-    next = :elixir_module.next_counter(nil)
     meta = Keyword.take(opts, [:line, :generated])
+    create(meta, module, quoted, :elixir.env_for_eval(opts))
+  end
+
+  defp create(meta, module, quoted, env_or_opts) do
+    next = :elixir_module.next_counter(nil)
     quoted = :elixir_quote.linify_with_context_counter(meta, {module, next}, quoted)
-    :elixir_module.compile(module, quoted, [], :elixir.env_for_eval(opts))
+    :elixir_module.compile(meta, module, quoted, [], false, :elixir.env_for_eval(env_or_opts))
   end
 
   @doc """
@@ -1074,7 +1177,7 @@ defmodule Module do
   Use `Kernel.function_exported?/3` and `Kernel.macro_exported?/3` to check for
   public functions and macros respectively in compiled modules.
 
-  Note that `defines?` returns false for functions and macros that have
+  Note that `defines?` returns `false` for functions and macros that have
   been defined but then marked as overridable and no other implementation
   has been provided. You can check the overridable status by calling
   `overridable?/2`.
@@ -1309,23 +1412,15 @@ defmodule Module do
   @doc """
   Deletes a definition from a module.
 
-  It returns true if the definition exists and it was removed,
-  otherwise it returns false.
+  It returns `true` if the definition exists and it was removed,
+  otherwise it returns `false`.
   """
   @doc since: "1.12.0"
   @spec delete_definition(module, definition) :: boolean()
   def delete_definition(module, {name, arity})
       when is_atom(module) and is_atom(name) and is_integer(arity) do
     assert_not_readonly!(__ENV__.function, module)
-
-    case :elixir_def.take_definition(module, {name, arity}) do
-      false ->
-        false
-
-      _ ->
-        :elixir_locals.yank({name, arity}, module)
-        true
-    end
+    :elixir_def.take_definition(module, {name, arity}) != false
   end
 
   @doc """
@@ -1354,8 +1449,7 @@ defmodule Module do
                     "overridable because it was not defined"
 
           clause ->
-            neighbours = :elixir_locals.yank(tuple, module)
-            :elixir_overridable.record_overridable(module, tuple, clause, neighbours)
+            :elixir_overridable.record_overridable(module, tuple, clause)
         end
 
       other ->
@@ -1379,11 +1473,7 @@ defmodule Module do
                 "to defoverridable/1 because #{error_explanation}"
     end
 
-    behaviour_callbacks =
-      for callback <- behaviour_info(behaviour, :callbacks) do
-        {pair, _kind} = normalize_macro_or_function_callback(callback)
-        pair
-      end
+    behaviour_callbacks = Module.Behaviour.callbacks(behaviour)
 
     tuples =
       for definition <- definitions_in(module),
@@ -1416,29 +1506,11 @@ defmodule Module do
     end
   end
 
-  defp normalize_macro_or_function_callback({function_name, arity}) do
-    case :erlang.atom_to_list(function_name) do
-      # Macros are always provided one extra argument in behaviour_info/1
-      'MACRO-' ++ tail ->
-        {{:erlang.list_to_atom(tail), arity - 1}, :defmacro}
-
-      _ ->
-        {{function_name, arity}, :def}
-    end
-  end
-
-  defp behaviour_info(module, key) do
-    case module.behaviour_info(key) do
-      list when is_list(list) -> list
-      :undefined -> []
-    end
-  end
-
   @doc """
   Returns `true` if `tuple` in `module` was marked as overridable
   at some point.
 
-  Note `overridable?/2` returns true even if the definition was
+  Note `overridable?/2` returns `true` even if the definition was
   already overridden. You can use `defines?/2` to see if a definition
   exists or one is pending.
   """
@@ -1502,10 +1574,42 @@ defmodule Module do
   """
   @spec get_attribute(module, atom, term) :: term
   def get_attribute(module, key, default \\ nil) when is_atom(module) and is_atom(key) do
-    case __get_attribute__(module, key, nil, true) do
-      nil -> default
-      value -> value
-    end
+    get_attribute(module, key, nil, true, false, default, {:get_attribute, 2})
+  end
+
+  @doc """
+  Gets the last set value of a given attribute from a module.
+
+  If the attribute was marked with `accumulate` with
+  `Module.register_attribute/3`, the previous value to have been set will be
+  returned. If the attribute does not accumulate, this call is the same as
+  calling `Module.get_attribute/3`.
+
+  This function can only be used on modules that have not yet been compiled.
+  Use the `c:Module.__info__/1` callback to get all persisted attributes, or
+  `Code.fetch_docs/1` to retrieve all documentation related attributes in
+  compiled modules.
+
+  ## Examples
+
+      defmodule Foo do
+        Module.put_attribute(__MODULE__, :value, 1)
+        Module.get_last_attribute(__MODULE__, :value) #=> 1
+
+        Module.get_last_attribute(__MODULE__, :not_found, :default) #=> :default
+
+        Module.register_attribute(__MODULE__, :acc, accumulate: true)
+        Module.put_attribute(__MODULE__, :acc, 1)
+        Module.get_last_attribute(__MODULE__, :acc) #=> 1
+        Module.put_attribute(__MODULE__, :acc, 2)
+        Module.get_last_attribute(__MODULE__, :acc) #=> 2
+      end
+
+  """
+  @doc since: "1.15.0"
+  @spec get_last_attribute(module, atom, term) :: term
+  def get_last_attribute(module, key, default \\ nil) when is_atom(module) and is_atom(key) do
+    get_attribute(module, key, nil, true, true, default, {:get_last_attribute, 2})
   end
 
   @doc """
@@ -1569,11 +1673,11 @@ defmodule Module do
 
     case :ets.lookup(set, key) do
       [{_, _, :accumulate, traces}] ->
-        trace_attribute(true, module, traces, set, key, [])
+        trace_attribute(true, module, traces, set, key, nil, [])
         reverse_values(:ets.take(bag, {:accumulate, key}), [])
 
       [{_, value, _, traces}] ->
-        trace_attribute(module, traces)
+        trace_attribute(module, key, nil, traces)
         :ets.delete(set, key)
         value
 
@@ -1820,22 +1924,12 @@ defmodule Module do
   defp args_count([], total, defaults), do: {total, defaults}
 
   @doc false
-  def check_derive_behaviours_and_impls(env, set, bag, all_definitions) do
+  def __check_attributes__(env, set, bag) do
     check_derive(env, set, bag)
+
     behaviours = bag_lookup_element(bag, {:accumulate, :behaviour}, 2)
-    impls = bag_lookup_element(bag, :impls, 2)
-    callbacks = check_behaviours(env, behaviours)
+    force_behaviour_dependencies(behaviours, env)
 
-    pending_callbacks =
-      if impls != [] do
-        {non_implemented_callbacks, contexts} = check_impls(env, behaviours, callbacks, impls)
-        warn_missing_impls(env, non_implemented_callbacks, contexts, all_definitions)
-        non_implemented_callbacks
-      else
-        callbacks
-      end
-
-    check_callbacks(env, pending_callbacks, all_definitions)
     :ok
   end
 
@@ -1846,7 +1940,7 @@ defmodule Module do
 
       _ ->
         message =
-          case :ets.lookup(set, :__struct__) do
+          case :ets.lookup(set, {:elixir, :struct}) do
             [] ->
               "warning: module attribute @derive was set but never used (it must come before defstruct)"
 
@@ -1858,266 +1952,37 @@ defmodule Module do
     end
   end
 
-  defp check_behaviours(env, behaviours) do
-    Enum.reduce(behaviours, %{}, fn behaviour, acc ->
-      cond do
-        not Code.ensure_loaded?(behaviour) ->
-          message =
-            "@behaviour #{inspect(behaviour)} does not exist (in module #{inspect(env.module)})"
+  # While `@behaviour MyBehaviour` will naturally introduce a runtime dependency,
+  # `@behaviour :"Elixir.MyBehaviour"` or similar would not.
+  # We force this dependency by adding the call to `MyBehaviour.behaviour_info/1`
+  defp force_behaviour_dependencies(behaviours, env) do
+    info_env = %{env | function: {:__info__, 1}}
 
-          IO.warn(message, env)
-          acc
-
-        not function_exported?(behaviour, :behaviour_info, 1) ->
-          message =
-            "module #{inspect(behaviour)} is not a behaviour (in module #{inspect(env.module)})"
-
-          IO.warn(message, env)
-          acc
-
-        true ->
-          :elixir_env.trace({:require, [from_macro: true], behaviour, []}, env)
-          optional_callbacks = behaviour_info(behaviour, :optional_callbacks)
-          callbacks = behaviour_info(behaviour, :callbacks)
-          Enum.reduce(callbacks, acc, &add_callback(&1, behaviour, env, optional_callbacks, &2))
-      end
-    end)
-  end
-
-  defp add_callback(original, behaviour, env, optional_callbacks, acc) do
-    {callback, kind} = normalize_macro_or_function_callback(original)
-
-    case acc do
-      %{^callback => {_kind, conflict, _optional?}} ->
-        message =
-          if conflict == behaviour do
-            "the behavior #{inspect(conflict)} has been declared twice " <>
-              "(conflict in #{format_definition(kind, callback)} in module #{inspect(env.module)})"
-          else
-            "conflicting behaviours found. #{format_definition(kind, callback)} is required by " <>
-              "#{inspect(conflict)} and #{inspect(behaviour)} (in module #{inspect(env.module)})"
-          end
-
-        IO.warn(message, env)
-
-      %{} ->
-        :ok
-    end
-
-    Map.put(acc, callback, {kind, behaviour, original in optional_callbacks})
-  end
-
-  defp check_callbacks(env, callbacks, all_definitions) do
-    for {callback, {kind, behaviour, optional?}} <- callbacks do
-      case :lists.keyfind(callback, 1, all_definitions) do
-        false when not optional? ->
-          message =
-            format_callback(callback, kind, behaviour) <>
-              " is not implemented (in module #{inspect(env.module)})"
-
-          IO.warn(message, env)
-
-        {_, wrong_kind, _, _} when kind != wrong_kind ->
-          message =
-            format_callback(callback, kind, behaviour) <>
-              " was implemented as \"#{wrong_kind}\" but should have been \"#{kind}\" " <>
-              "(in module #{inspect(env.module)})"
-
-          IO.warn(message, env)
-
-        _ ->
-          :ok
-      end
+    for behaviour <- behaviours do
+      :elixir_env.trace({:remote_function, [], behaviour, :behaviour_info, 1}, info_env)
     end
 
     :ok
   end
 
-  defp format_callback(callback, kind, module) do
-    protocol_or_behaviour = if protocol?(module), do: "protocol ", else: "behaviour "
-
-    format_definition(kind, callback) <>
-      " required by " <> protocol_or_behaviour <> inspect(module)
-  end
-
-  defp protocol?(module) do
-    Code.ensure_loaded?(module) and function_exported?(module, :__protocol__, 1) and
-      module.__protocol__(:module) == module
-  end
-
-  defp check_impls(env, behaviours, callbacks, impls) do
-    acc = {callbacks, %{}}
-
-    Enum.reduce(impls, acc, fn {fa, context, defaults, kind, line, file, value}, acc ->
-      case impl_behaviours(fa, defaults, kind, value, behaviours, callbacks) do
-        {:ok, impl_behaviours} ->
-          Enum.reduce(impl_behaviours, acc, fn {fa, behaviour}, {callbacks, contexts} ->
-            callbacks = Map.delete(callbacks, fa)
-            contexts = Map.update(contexts, behaviour, [context], &[context | &1])
-            {callbacks, contexts}
-          end)
-
-        {:error, message} ->
-          formatted = format_impl_warning(fa, kind, message)
-          IO.warn(formatted, %{env | line: line, file: file})
-          acc
-      end
-    end)
-  end
-
-  defp impl_behaviours({function, arity}, defaults, kind, value, behaviours, callbacks) do
-    impls = for n <- arity..(arity - defaults), do: {function, n}
-    impl_behaviours(impls, kind, value, behaviours, callbacks)
-  end
-
-  defp impl_behaviours(_, kind, _, _, _) when kind in [:defp, :defmacrop] do
-    {:error, :private_function}
-  end
-
-  defp impl_behaviours(_, _, value, [], _) do
-    {:error, {:no_behaviours, value}}
-  end
-
-  defp impl_behaviours(impls, _, false, _, callbacks) do
-    case callbacks_for_impls(impls, callbacks) do
-      [] -> {:ok, []}
-      [impl | _] -> {:error, {:impl_not_defined, impl}}
-    end
-  end
-
-  defp impl_behaviours(impls, _, true, _, callbacks) do
-    case callbacks_for_impls(impls, callbacks) do
-      [] -> {:error, {:impl_defined, callbacks}}
-      impls -> {:ok, impls}
-    end
-  end
-
-  defp impl_behaviours(impls, _, behaviour, behaviours, callbacks) do
-    filtered = behaviour_callbacks_for_impls(impls, behaviour, callbacks)
-
-    cond do
-      filtered != [] ->
-        {:ok, filtered}
-
-      behaviour not in behaviours ->
-        {:error, {:behaviour_not_declared, behaviour}}
-
-      true ->
-        {:error, {:behaviour_not_defined, behaviour, callbacks}}
-    end
-  end
-
-  defp behaviour_callbacks_for_impls([], _behaviour, _callbacks) do
-    []
-  end
-
-  defp behaviour_callbacks_for_impls([fa | tail], behaviour, callbacks) do
-    case callbacks[fa] do
-      {_, ^behaviour, _} ->
-        [{fa, behaviour} | behaviour_callbacks_for_impls(tail, behaviour, callbacks)]
-
-      _ ->
-        behaviour_callbacks_for_impls(tail, behaviour, callbacks)
-    end
-  end
-
-  defp callbacks_for_impls([], _) do
-    []
-  end
-
-  defp callbacks_for_impls([fa | tail], callbacks) do
-    case callbacks[fa] do
-      {_, behaviour, _} -> [{fa, behaviour} | callbacks_for_impls(tail, callbacks)]
-      nil -> callbacks_for_impls(tail, callbacks)
-    end
-  end
-
-  defp format_impl_warning(fa, kind, :private_function) do
-    "#{format_definition(kind, fa)} is private, @impl attribute is always discarded for private functions/macros"
-  end
-
-  defp format_impl_warning(fa, kind, {:no_behaviours, value}) do
-    "got \"@impl #{inspect(value)}\" for #{format_definition(kind, fa)} but no behaviour was declared"
-  end
-
-  defp format_impl_warning(_, kind, {:impl_not_defined, {fa, behaviour}}) do
-    "got \"@impl false\" for #{format_definition(kind, fa)} " <>
-      "but it is a callback specified in #{inspect(behaviour)}"
-  end
-
-  defp format_impl_warning(fa, kind, {:impl_defined, callbacks}) do
-    "got \"@impl true\" for #{format_definition(kind, fa)} " <>
-      "but no behaviour specifies such callback#{known_callbacks(callbacks)}"
-  end
-
-  defp format_impl_warning(fa, kind, {:behaviour_not_declared, behaviour}) do
-    "got \"@impl #{inspect(behaviour)}\" for #{format_definition(kind, fa)} " <>
-      "but this behaviour was not declared with @behaviour"
-  end
-
-  defp format_impl_warning(fa, kind, {:behaviour_not_defined, behaviour, callbacks}) do
-    "got \"@impl #{inspect(behaviour)}\" for #{format_definition(kind, fa)} " <>
-      "but this behaviour does not specify such callback#{known_callbacks(callbacks)}"
-  end
-
-  defp warn_missing_impls(_env, callbacks, _contexts, _defs) when map_size(callbacks) == 0 do
-    :ok
-  end
-
-  defp warn_missing_impls(env, non_implemented_callbacks, contexts, defs) do
-    for {pair, kind, meta, _clauses} <- defs,
-        kind in [:def, :defmacro] do
-      with {:ok, {_, behaviour, _}} <- Map.fetch(non_implemented_callbacks, pair),
-           true <- missing_impl_in_context?(meta, behaviour, contexts) do
-        message =
-          "module attribute @impl was not set for #{format_definition(kind, pair)} " <>
-            "callback (specified in #{inspect(behaviour)}). " <>
-            "This either means you forgot to add the \"@impl true\" annotation before the " <>
-            "definition or that you are accidentally overriding this callback"
-
-        IO.warn(message, %{env | line: :elixir_utils.get_line(meta)})
-      end
-    end
-
-    :ok
-  end
-
-  defp missing_impl_in_context?(meta, behaviour, contexts) do
-    case contexts do
-      %{^behaviour => known} -> Keyword.get(meta, :context) in known
-      %{} -> not Keyword.has_key?(meta, :context)
-    end
-  end
-
-  defp format_definition(kind, {name, arity}) do
-    format_definition(kind) <> " #{name}/#{arity}"
-  end
-
-  defp format_definition(:defmacro), do: "macro"
-  defp format_definition(:defmacrop), do: "macro"
-  defp format_definition(:def), do: "function"
-  defp format_definition(:defp), do: "function"
-
-  defp known_callbacks(callbacks) when map_size(callbacks) == 0 do
-    ". There are no known callbacks, please specify the proper @behaviour " <>
-      "and make sure it defines callbacks"
-  end
-
-  defp known_callbacks(callbacks) do
-    formatted_callbacks =
-      for {{name, arity}, {kind, module, _}} <- callbacks do
-        "\n  * " <> Exception.format_mfa(module, name, arity) <> " (#{format_definition(kind)})"
-      end
-
-    ". The known callbacks are:\n#{formatted_callbacks}\n"
-  end
-
   @doc false
   # Used internally by Kernel's @.
   # This function is private and must be used only internally.
   def __get_attribute__(module, key, caller_line, trace?) when is_atom(key) do
+    get_attribute(module, key, caller_line, trace?, false, nil, {:get_attribute, 2})
+  end
+
+  defp get_attribute(
+         module,
+         key,
+         caller_line,
+         trace?,
+         last_accumulated?,
+         default,
+         function_name_arity
+       ) do
     assert_not_compiled!(
-      {:get_attribute, 2},
+      function_name_arity,
       module,
       "Use the Module.__info__/1 callback or Code.fetch_docs/1 instead"
     )
@@ -2125,16 +1990,19 @@ defmodule Module do
     {set, bag} = data_tables_for(module)
 
     case :ets.lookup(set, key) do
+      [{_, _, :unset, _}] ->
+        default
+
       [{_, _, :accumulate, traces}] ->
-        trace_attribute(trace?, module, traces, set, key, [])
-        :lists.reverse(bag_lookup_element(bag, {:accumulate, key}, 2))
+        trace_attribute(trace?, module, traces, set, key, caller_line, [])
+        lookup_accumulate_attribute(bag, key, default, last_accumulated?)
 
       [{_, value, warn_line, traces}] when is_integer(warn_line) ->
-        trace_attribute(trace?, module, traces, set, key, [{3, :used}])
+        trace_attribute(trace?, module, traces, set, key, caller_line, [{3, :used}])
         value
 
       [{_, value, _, traces}] ->
-        trace_attribute(trace?, module, traces, set, key, [])
+        trace_attribute(trace?, module, traces, set, key, caller_line, [])
         value
 
       [] when is_integer(caller_line) ->
@@ -2144,19 +2012,33 @@ defmodule Module do
             "please remove access to @#{key} or explicitly set it before access"
 
         IO.warn(error_message, attribute_stack(module, caller_line))
-        nil
+        default
 
       [] ->
-        nil
+        default
     end
   end
 
-  defp trace_attribute(module, traces) do
+  defp lookup_accumulate_attribute(bag, key, _default, false) do
+    :lists.reverse(bag_lookup_element(bag, {:accumulate, key}, 2))
+  end
+
+  defp lookup_accumulate_attribute(bag, key, default, true) do
+    case :ets.select(bag, [{{{:accumulate, key}, :"$1"}, [], [:"$1"]}], 1) do
+      {[value], _} -> value
+      _ -> default
+    end
+  end
+
+  defp trace_attribute(module, key, caller_line, traces) do
     :lists.foreach(
       fn {line, lexical_tracker, tracers, aliases} ->
+        line = caller_line || line
+
         env = %{
           Macro.Env.__struct__()
           | line: line,
+            file: "@#{key}",
             lexical_tracker: lexical_tracker,
             module: module,
             tracers: tracers
@@ -2173,10 +2055,10 @@ defmodule Module do
     )
   end
 
-  defp trace_attribute(trace?, module, traces, set, key, updates) do
+  defp trace_attribute(trace?, module, traces, set, key, caller_line, updates) do
     updates =
       if trace? and traces != [] do
-        trace_attribute(module, traces)
+        trace_attribute(module, key, caller_line, traces)
         updates ++ [{4, []}]
       else
         updates
@@ -2289,8 +2171,7 @@ defmodule Module do
   end
 
   defp attribute_stack(module, line) do
-    file = String.to_charlist(Path.relative_to_cwd(:elixir_module.file(module)))
-    [{module, :__MODULE__, 0, file: file, line: line}]
+    struct!(Macro.Env, module: module, file: :elixir_module.file(module), line: line)
   end
 
   ## Helpers
@@ -2380,6 +2261,16 @@ defmodule Module do
     end
   end
 
+  defp preprocess_attribute(:nifs, value) do
+    if not function_arity_list?(value) do
+      raise ArgumentError,
+            "@nifs is a built-in module attribute for specifying a list " <>
+              "of functions and their arities that are NIFs, got: #{inspect(value)}"
+    end
+
+    value
+  end
+
   defp preprocess_attribute(:dialyzer, value) do
     # From https://github.com/erlang/otp/blob/master/lib/stdlib/src/erl_lint.erl
     :lists.foreach(
@@ -2398,17 +2289,22 @@ defmodule Module do
     value
   end
 
-  defp valid_dialyzer_attribute?({key, fun_arities}) when is_atom(key) do
-    (key == :nowarn_function or valid_dialyzer_attribute?(key)) and
+  defp function_arity_list?(fun_arities) do
+    is_list(fun_arities) and
       :lists.all(
         fn
           {fun, arity} when is_atom(fun) and is_integer(arity) -> true
           _ -> false
         end,
-        List.wrap(fun_arities)
+        fun_arities
       )
   end
 
+  defp valid_dialyzer_attribute?({key, fun_arities}) when is_atom(key) do
+    (key == :nowarn_function or valid_dialyzer_attribute?(key)) and
+      function_arity_list?(List.wrap(fun_arities))
+  end
+
   defp valid_dialyzer_attribute?(attr) do
     :lists.member(
       attr,
@@ -2416,7 +2312,9 @@ defmodule Module do
         [:no_match, :no_opaque, :no_fail_call, :no_contracts] ++
         [:no_behaviours, :no_undefined_callbacks, :unmatched_returns] ++
         [:error_handling, :race_conditions, :no_missing_calls] ++
-        [:specdiffs, :overspecs, :underspecs, :unknown, :no_underspecs]
+        [:specdiffs, :overspecs, :underspecs, :unknown, :no_underspecs] ++
+        [:extra_return, :no_extra_return, :no_missing_return] ++
+        [:missing_return, :no_unknown]
     )
   end
 
