@@ -178,7 +178,7 @@ defmodule Node do
 
   Inlined by the compiler.
   """
-  @spec spawn(t, (() -> any)) :: pid
+  @spec spawn(t, (-> any)) :: pid
   def spawn(node, fun) do
     :erlang.spawn(node, fun)
   end
@@ -193,7 +193,7 @@ defmodule Node do
 
   Inlined by the compiler.
   """
-  @spec spawn(t, (() -> any), Process.spawn_opts()) :: pid | {pid, reference}
+  @spec spawn(t, (-> any), Process.spawn_opts()) :: pid | {pid, reference}
   def spawn(node, fun, opts) do
     :erlang.spawn_opt(node, fun, opts)
   end
@@ -219,7 +219,7 @@ defmodule Node do
 
   If `node` does not exist, a useless PID is returned.
 
-  For the list of available options, see `:erlang.spawn/4`.
+  For the list of available options, see `:erlang.spawn_opt/5`.
 
   Inlined by the compiler.
   """
@@ -237,7 +237,7 @@ defmodule Node do
 
   Inlined by the compiler.
   """
-  @spec spawn_link(t, (() -> any)) :: pid
+  @spec spawn_link(t, (-> any)) :: pid
   def spawn_link(node, fun) do
     :erlang.spawn_link(node, fun)
   end
@@ -261,13 +261,10 @@ defmodule Node do
   Spawns the given function on a node, monitors it and returns its PID
   and monitoring reference.
 
-  This functionality was added on Erlang/OTP 23. Using this function to
-  communicate with nodes running on earlier versions will fail.
-
   Inlined by the compiler.
   """
   @doc since: "1.14.0"
-  @spec spawn_monitor(t, (() -> any)) :: {pid, reference}
+  @spec spawn_monitor(t, (-> any)) :: {pid, reference}
   def spawn_monitor(node, fun) do
     :erlang.spawn_monitor(node, fun)
   end
@@ -276,9 +273,6 @@ defmodule Node do
   Spawns the given module and function passing the given args on a node,
   monitors it and returns its PID and monitoring reference.
 
-  This functionality was added on Erlang/OTP 23. Using this function
-  to communicate with nodes running on earlier versions will fail.
-
   Inlined by the compiler.
   """
   @doc since: "1.14.0"
