@@ -6,7 +6,7 @@ defmodule Protocol do
   implementations. A protocol is defined with `Kernel.defprotocol/2`
   and its implementations with `Kernel.defimpl/3`.
 
-  ## A real case
+  ## Example
 
   In Elixir, we have two nouns for checking how many items there
   are in a data structure: `length` and `size`.  `length` means the
@@ -43,6 +43,13 @@ defmodule Protocol do
         def size(tuple), do: tuple_size(tuple)
       end
 
+  Finally, we can use the `Size` protocol to call the correct implementation:
+
+      Size.size({1, 2})
+      # => 2
+      Size.size(%{key: :value})
+      # => 1
+
   Note that we didn't implement it for lists as we don't have the
   `size` information on lists, rather its value needs to be
   computed with `length`.
@@ -95,8 +102,7 @@ defmodule Protocol do
   invoking the protocol will raise unless it is configured to
   fall back to `Any`. Conveniences for building implementations
   on top of existing ones are also available, look at `defstruct/1`
-  for more information about deriving
-  protocols.
+  for more information about deriving protocols.
 
   ## Fallback to `Any`
 
@@ -119,8 +125,8 @@ defmodule Protocol do
   Although the implementation above is arguably not a reasonable
   one. For example, it makes no sense to say a PID or an integer
   have a size of `0`. That's one of the reasons why `@fallback_to_any`
-  is an opt-in behaviour. For the majority of protocols, raising
-  an error when a protocol is not implemented is the proper behaviour.
+  is an opt-in behavior. For the majority of protocols, raising
+  an error when a protocol is not implemented is the proper behavior.
 
   ## Multiple implementations
 
@@ -158,56 +164,16 @@ defmodule Protocol do
   The `@spec` above expresses that all types allowed to implement the
   given protocol are valid argument types for the given function.
 
-  ## Reflection
+  ## Configuration
 
-  Any protocol module contains three extra functions:
+  The following module attributes are available to configure a protocol:
 
-    * `__protocol__/1` - returns the protocol information. The function takes
-      one of the following atoms:
+    * `@fallback_to_any` - when true, enables protocol dispatch to
+      fallback to any
 
-      * `:consolidated?` - returns whether the protocol is consolidated
-
-      * `:functions` - returns a keyword list of protocol functions and their arities
-
-      * `:impls` - if consolidated, returns `{:consolidated, modules}` with the list of modules
-        implementing the protocol, otherwise `:not_consolidated`
-
-      * `:module` - the protocol module atom name
-
-    * `impl_for/1` - returns the module that implements the protocol for the given argument,
-      `nil` otherwise
-
-    * `impl_for!/1` - same as above but raises `Protocol.UndefinedError` if an implementation is
-      not found
-
-  For example, for the `Enumerable` protocol we have:
-
-      iex> Enumerable.__protocol__(:functions)
-      [count: 1, member?: 2, reduce: 3, slice: 1]
-
-      iex> Enumerable.impl_for([])
-      Enumerable.List
-
-      iex> Enumerable.impl_for(42)
-      nil
-
-  In addition, every protocol implementation module contains the `__impl__/1`
-  function. The function takes one of the following atoms:
-
-    * `:for` - returns the module responsible for the data structure of the
-      protocol implementation
-
-    * `:protocol` - returns the protocol module for which this implementation
-    is provided
-
-  For example, the module implementing the `Enumerable` protocol for lists is
-  `Enumerable.List`. Therefore, we can invoke `__impl__/1` on this module:
-
-      iex(1)> Enumerable.List.__impl__(:for)
-      List
-
-      iex(2)> Enumerable.List.__impl__(:protocol)
-      Enumerable
+    * `@undefined_impl_description` - a string with additional description
+      to be used on `Protocol.UndefinedError` when looking up the implementation
+      fails. This option is only applied if `@fallback_to_any` is not set to true
 
   ## Consolidation
 
@@ -215,7 +181,8 @@ defmodule Protocol do
   are known up-front, typically after all Elixir code in a project is compiled,
   Elixir provides a feature called *protocol consolidation*. Consolidation directly
   links protocols to their implementations in a way that invoking a function from a
-  consolidated protocol is equivalent to invoking two remote functions.
+  consolidated protocol is equivalent to invoking two remote functions - one to
+  identify the correct implementation, and another to call the implementation.
 
   Protocol consolidation is applied by default to all Mix projects during compilation.
   This may be an issue during test. For instance, if you want to implement a protocol
@@ -261,6 +228,41 @@ defmodule Protocol do
   globally set.
   """
 
+  @doc """
+  A function available in all protocol definitions that returns protocol metadata.
+  """
+  @callback __protocol__(:consolidated?) :: boolean()
+  @callback __protocol__(:functions) :: [{atom(), arity()}]
+  @callback __protocol__(:impls) :: {:consolidated, [module()]} | :not_consolidated
+  @callback __protocol__(:module) :: module()
+
+  @doc """
+  A function available in all protocol definitions that returns the implementation
+  for the given `term` or nil.
+
+  If `@fallback_to_any` is true, `nil` is never returned.
+  """
+  @callback impl_for(term) :: module() | nil
+
+  @doc """
+  A function available in all protocol definitions that returns the implementation
+  for the given `term` or raises.
+
+  If `@fallback_to_any` is true, it never raises.
+  """
+  @callback impl_for!(term) :: module()
+
+  @doc """
+  An optional callback to be implemented by protocol authors for custom deriving.
+
+  It must return a quoted expression that implements the protocol for the given module.
+
+  See `Protocol.derive/3` for an example.
+  """
+  @macrocallback __deriving__(module(), term()) :: Macro.t()
+
+  @optional_callbacks __deriving__: 2
+
   @doc false
   defmacro def(signature)
 
@@ -373,52 +375,50 @@ defmodule Protocol do
   @doc """
   Derives the `protocol` for `module` with the given options.
 
-  If your implementation passes options or if you are generating
-  custom code based on the struct, you will also need to implement
-  a macro defined as `__deriving__(module, struct, options)`
-  to get the options that were passed.
+  Every time you derive a protocol, Elixir will verify if the protocol
+  has implemented the `c:Protocol.__deriving__/2` callback. If so,
+  the callback will be invoked and it should define the implementation
+  module. Otherwise an implementation that simply points to the `Any`
+  implementation is automatically derived.
 
   ## Examples
 
       defprotocol Derivable do
-        def ok(arg)
-      end
+        @impl true
+        defmacro __deriving__(module, options) do
+          # If you need to load struct metadata, you may call:
+          # struct_info = Macro.struct_info!(module, __CALLER__)
 
-      defimpl Derivable, for: Any do
-        defmacro __deriving__(module, struct, options) do
           quote do
             defimpl Derivable, for: unquote(module) do
               def ok(arg) do
-                {:ok, arg, unquote(Macro.escape(struct)), unquote(options)}
+                {:ok, arg, unquote(options)}
               end
             end
           end
         end
 
-        def ok(arg) do
-          {:ok, arg}
-        end
+        def ok(arg)
       end
 
+  Once the protocol is defined, there are two ways it can be
+  derived. The first is by using the `@derive` module attribute
+  by the time you define the struct:
+
       defmodule ImplStruct do
         @derive [Derivable]
         defstruct a: 0, b: 0
       end
 
       Derivable.ok(%ImplStruct{})
-      #=> {:ok, %ImplStruct{a: 0, b: 0}, %ImplStruct{a: 0, b: 0}, []}
+      #=> {:ok, %ImplStruct{a: 0, b: 0}, []}
 
-  Explicit derivations can now be called via `__deriving__/3`:
+  If the struct has already been defined, you can call this macro:
 
-      # Explicitly derived via `__deriving__/3`
-      Derivable.ok(%ImplStruct{a: 1, b: 1})
-      #=> {:ok, %ImplStruct{a: 1, b: 1}, %ImplStruct{a: 0, b: 0}, []}
-
-      # Explicitly derived by API via `__deriving__/3`
       require Protocol
       Protocol.derive(Derivable, ImplStruct, :oops)
       Derivable.ok(%ImplStruct{a: 1, b: 1})
-      #=> {:ok, %ImplStruct{a: 1, b: 1}, %ImplStruct{a: 0, b: 0}, :oops}
+      #=> {:ok, %ImplStruct{a: 1, b: 1}, :oops}
 
   """
   defmacro derive(protocol, module, options \\ []) do
@@ -441,7 +441,7 @@ defmodule Protocol do
   ## Examples
 
       # Get Elixir's ebin directory path and retrieve all protocols
-      iex> path = :code.lib_dir(:elixir, :ebin)
+      iex> path = Application.app_dir(:elixir, "ebin")
       iex> mods = Protocol.extract_protocols([path])
       iex> Enumerable in mods
       true
@@ -449,7 +449,7 @@ defmodule Protocol do
   """
   @spec extract_protocols([charlist | String.t()]) :: [atom]
   def extract_protocols(paths) do
-    extract_matching_by_attribute(paths, 'Elixir.', fn module, attributes ->
+    extract_matching_by_attribute(paths, [?E, ?l, ?i, ?x, ?i, ?r, ?.], fn module, attributes ->
       case attributes[:__protocol__] do
         [fallback_to_any: _] -> module
         _ -> nil
@@ -470,7 +470,7 @@ defmodule Protocol do
   ## Examples
 
       # Get Elixir's ebin directory path and retrieve all protocols
-      iex> path = :code.lib_dir(:elixir, :ebin)
+      iex> path = Application.app_dir(:elixir, "ebin")
       iex> mods = Protocol.extract_impls(Enumerable, [path])
       iex> List in mods
       true
@@ -478,7 +478,7 @@ defmodule Protocol do
   """
   @spec extract_impls(module, [charlist | String.t()]) :: [atom]
   def extract_impls(protocol, paths) when is_atom(protocol) do
-    prefix = Atom.to_charlist(protocol) ++ '.'
+    prefix = Atom.to_charlist(protocol) ++ [?.]
 
     extract_matching_by_attribute(paths, prefix, fn _mod, attributes ->
       case attributes[:__impl__] do
@@ -490,7 +490,8 @@ defmodule Protocol do
 
   defp extract_matching_by_attribute(paths, prefix, callback) do
     for path <- paths,
-        path = to_charlist(path),
+        # Do not use protocols as they may be consolidating
+        path = if(is_list(path), do: path, else: String.to_charlist(path)),
         file <- list_dir(path),
         mod = extract_from_file(path, file, prefix, callback),
         do: mod
@@ -504,7 +505,7 @@ defmodule Protocol do
   end
 
   defp extract_from_file(path, file, prefix, callback) do
-    if :lists.prefix(prefix, file) and :filename.extension(file) == '.beam' do
+    if :lists.prefix(prefix, file) and :filename.extension(file) == [?., ?b, ?e, ?a, ?m] do
       extract_from_beam(:filename.join(path, file), callback)
     end
   end
@@ -548,7 +549,7 @@ defmodule Protocol do
 
   This function does not load the protocol at any point
   nor loads the new bytecode for the compiled module.
-  However each implementation must be available and
+  However, each implementation must be available and
   it will be loaded.
   """
   @spec consolidate(module, [module]) ::
@@ -565,7 +566,7 @@ defmodule Protocol do
   end
 
   defp beam_protocol(protocol) do
-    chunk_ids = [:debug_info, 'Docs', 'ExCk']
+    chunk_ids = [:debug_info, [?D, ?o, ?c, ?s], [?E, ?x, ?C, ?k]]
     opts = [:allow_missing_chunks]
 
     case :beam_lib.chunks(beam_file(protocol), chunk_ids, opts) do
@@ -677,7 +678,7 @@ defmodule Protocol do
   end
 
   defp load_impl(protocol, for) do
-    Module.concat(protocol, for).__impl__(:target)
+    Module.concat(protocol, for)
   end
 
   # Finally compile the module and emit its bytecode.
@@ -692,25 +693,14 @@ defmodule Protocol do
   def __protocol__(name, do: block) do
     quote do
       defmodule unquote(name) do
+        @behaviour Protocol
         @before_compile Protocol
 
         # We don't allow function definition inside protocols
         import Kernel,
-          except: [
-            def: 1,
-            def: 2,
-            defp: 1,
-            defp: 2,
-            defdelegate: 2,
-            defguard: 1,
-            defguardp: 1,
-            defmacro: 1,
-            defmacro: 2,
-            defmacrop: 1,
-            defmacrop: 2
-          ]
+          except: [def: 1, def: 2, defdelegate: 2, defguard: 1, defguardp: 1]
 
-        # Import the new dsl that holds the new def
+        # Import the new `def` that is used by protocols
         import Protocol, only: [def: 1]
 
         # Compile with debug info for consolidation
@@ -729,19 +719,17 @@ defmodule Protocol do
     end
   end
 
-  defp callback_ast_to_fa({kind, {:"::", meta, [{name, _, args}, _return]}, _pos})
-       when kind in [:callback, :macrocallback] do
+  defp callback_ast_to_fa({_kind, {:"::", meta, [{name, _, args}, _return]}, _pos}) do
     [{{name, length(List.wrap(args))}, meta}]
   end
 
   defp callback_ast_to_fa(
-         {kind, {:when, _, [{:"::", meta, [{name, _, args}, _return]}, _vars]}, _pos}
-       )
-       when kind in [:callback, :macrocallback] do
+         {_kind, {:when, _, [{:"::", meta, [{name, _, args}, _return]}, _vars]}, _pos}
+       ) do
     [{{name, length(List.wrap(args))}, meta}]
   end
 
-  defp callback_ast_to_fa({kind, _, _pos}) when kind in [:callback, :macrocallback] do
+  defp callback_ast_to_fa({_kind, _clause, _pos}) do
     []
   end
 
@@ -759,13 +747,10 @@ defmodule Protocol do
   end
 
   defp warn(message, env, line) when is_integer(line) do
-    stacktrace = :maps.update(:line, line, env) |> Macro.Env.stacktrace()
-    IO.warn(message, stacktrace)
+    IO.warn(message, %{env | line: line})
   end
 
   def __before_compile__(env) do
-    callback_metas = callback_metas(env.module, :callback)
-    callbacks = :maps.keys(callback_metas)
     functions = Module.get_attribute(env.module, :__functions__)
 
     if functions == [] do
@@ -776,8 +761,11 @@ defmodule Protocol do
       )
     end
 
+    callback_metas = callback_metas(env.module, :callback)
+    callbacks = :maps.keys(callback_metas)
+
     # TODO: Convert the following warnings into errors in future Elixir versions
-    :lists.map(
+    :lists.foreach(
       fn {name, arity} = fa ->
         warn(
           "cannot define @callback #{name}/#{arity} inside protocol, use def/1 to outline your protocol definition",
@@ -792,7 +780,7 @@ defmodule Protocol do
     macrocallback_metas = callback_metas(env.module, :macrocallback)
     macrocallbacks = :maps.keys(macrocallback_metas)
 
-    :lists.map(
+    :lists.foreach(
       fn {name, arity} = fa ->
         warn(
           "cannot define @macrocallback #{name}/#{arity} inside protocol, use def/1 to outline your protocol definition",
@@ -819,7 +807,7 @@ defmodule Protocol do
     quote bind_quoted: [built_in: __built_in__()] do
       any_impl_for =
         if @fallback_to_any do
-          quote do: unquote(__MODULE__.Any).__impl__(:target)
+          __MODULE__.Any
         else
           nil
         end
@@ -846,11 +834,9 @@ defmodule Protocol do
           target = Module.concat(__MODULE__, mod)
 
           Kernel.def impl_for(data) when :erlang.unquote(guard)(data) do
-            try do
-              unquote(target).__impl__(:target)
-            rescue
-              UndefinedFunctionError ->
-                unquote(any_impl_for)
+            case Code.ensure_compiled(unquote(target)) do
+              {:module, module} -> module
+              {:error, _} -> unquote(any_impl_for)
             end
           end
         end,
@@ -867,6 +853,9 @@ defmodule Protocol do
         unquote(any_impl_for)
       end
 
+      undefined_impl_description =
+        Module.get_attribute(__MODULE__, :undefined_impl_description, "")
+
       @doc false
       @spec impl_for!(term) :: atom
       if any_impl_for do
@@ -875,26 +864,30 @@ defmodule Protocol do
         end
       else
         Kernel.def impl_for!(data) do
-          impl_for(data) || raise(Protocol.UndefinedError, protocol: __MODULE__, value: data)
+          impl_for(data) ||
+            raise(Protocol.UndefinedError,
+              protocol: __MODULE__,
+              value: data,
+              description: unquote(undefined_impl_description)
+            )
         end
       end
 
       # Internal handler for Structs
       Kernel.defp struct_impl_for(struct) do
-        target = Module.concat(__MODULE__, struct)
-
-        try do
-          target.__impl__(:target)
-        rescue
-          UndefinedFunctionError ->
-            unquote(any_impl_for)
+        case Code.ensure_compiled(Module.concat(__MODULE__, struct)) do
+          {:module, module} -> module
+          {:error, _} -> unquote(any_impl_for)
         end
       end
 
       # Inline struct implementation for performance
       @compile {:inline, struct_impl_for: 1}
 
-      unless Module.defines_type?(__MODULE__, {:t, 0}) do
+      if not Module.defines_type?(__MODULE__, {:t, 0}) do
+        @typedoc """
+        All the types that implement this protocol.
+        """
         @type t :: term
       end
 
@@ -905,9 +898,9 @@ defmodule Protocol do
 
       @doc false
       @spec __protocol__(:module) :: __MODULE__
-      @spec __protocol__(:functions) :: unquote(Protocol.__functions_spec__(@__functions__))
-      @spec __protocol__(:consolidated?) :: boolean
-      @spec __protocol__(:impls) :: :not_consolidated | {:consolidated, [module]}
+      @spec __protocol__(:functions) :: [{atom(), arity()}]
+      @spec __protocol__(:consolidated?) :: boolean()
+      @spec __protocol__(:impls) :: :not_consolidated | {:consolidated, [module()]}
       Kernel.def(__protocol__(:module), do: __MODULE__)
       Kernel.def(__protocol__(:functions), do: unquote(:lists.sort(@__functions__)))
       Kernel.def(__protocol__(:consolidated?), do: false)
@@ -915,12 +908,6 @@ defmodule Protocol do
     end
   end
 
-  @doc false
-  def __functions_spec__([]), do: []
-
-  def __functions_spec__([head | tail]),
-    do: [:lists.foldl(&{:|, [], [&1, &2]}, head, tail), quote(do: ...)]
-
   @doc false
   def __impl__(protocol, opts, do_block, env) do
     opts = Keyword.merge(opts, do_block)
@@ -931,7 +918,8 @@ defmodule Protocol do
           raise ArgumentError, "defimpl/3 expects a :for option when declared outside a module"
       end)
 
-    for = Macro.expand_literal(for, %{env | module: Kernel, function: {:defimpl, 3}})
+    for =
+      Macro.expand_literals(for, %{env | module: env.module || Elixir, function: {:__impl__, 1}})
 
     case opts do
       [] -> raise ArgumentError, "defimpl expects a do-end block"
@@ -951,10 +939,8 @@ defmodule Protocol do
     impl =
       quote unquote: false do
         @doc false
-        @spec __impl__(:target) :: __MODULE__
         @spec __impl__(:for) :: unquote(for)
         @spec __impl__(:protocol) :: unquote(protocol)
-        def __impl__(:target), do: __MODULE__
         def __impl__(:for), do: unquote(for)
         def __impl__(:protocol), do: unquote(protocol)
       end
@@ -968,6 +954,7 @@ defmodule Protocol do
       Protocol.__ensure_defimpl__(protocol, for, __ENV__)
 
       defmodule name do
+        @moduledoc false
         @behaviour protocol
         @protocol protocol
         @for for
@@ -984,14 +971,12 @@ defmodule Protocol do
 
   @doc false
   def __derive__(derives, for, %Macro.Env{} = env) when is_atom(for) do
-    struct = Macro.struct!(for, env)
-
     foreach = fn
       proto when is_atom(proto) ->
-        derive(proto, for, struct, [], env)
+        derive(proto, for, [], env)
 
       {proto, opts} when is_atom(proto) ->
-        derive(proto, for, struct, opts, env)
+        derive(proto, for, opts, env)
     end
 
     :lists.foreach(foreach, :lists.flatten(derives))
@@ -999,36 +984,54 @@ defmodule Protocol do
     :ok
   end
 
-  defp derive(protocol, for, struct, opts, env) do
+  defp derive(protocol, for, opts, env) do
     extra = ", cannot derive #{inspect(protocol)} for #{inspect(for)}"
     assert_protocol!(protocol, extra)
-    __ensure_defimpl__(protocol, for, env)
-    assert_impl!(protocol, Any, extra)
+
+    {mod, args} =
+      if macro_exported?(protocol, :__deriving__, 2) do
+        {protocol, [for, opts]}
+      else
+        # TODO: Deprecate this on Elixir v1.22+
+        assert_impl!(protocol, Any, extra)
+        {Module.concat(protocol, Any), [for, Macro.struct!(for, env), opts]}
+      end
 
     # Clean up variables from eval context
     env = :elixir_env.reset_vars(env)
-    args = [for, struct, opts]
-    impl = Module.concat(protocol, Any)
 
-    :elixir_module.expand_callback(env.line, impl, :__deriving__, args, env, fn mod, fun, args ->
+    :elixir_module.expand_callback(env.line, mod, :__deriving__, args, env, fn mod, fun, args ->
       if function_exported?(mod, fun, length(args)) do
         apply(mod, fun, args)
       else
+        __ensure_defimpl__(protocol, for, env)
+        assert_impl!(protocol, Any, extra)
+        impl = Module.concat(protocol, Any)
+
+        funs =
+          for {fun, arity} <- protocol.__protocol__(:functions) do
+            args = Macro.generate_arguments(arity, nil)
+
+            quote do
+              def unquote(fun)(unquote_splicing(args)),
+                do: unquote(impl).unquote(fun)(unquote_splicing(args))
+            end
+          end
+
         quoted =
           quote do
+            @behaviour unquote(protocol)
             Module.register_attribute(__MODULE__, :__impl__, persist: true)
             @__impl__ [protocol: unquote(protocol), for: unquote(for)]
 
             @doc false
-            @spec __impl__(:target) :: unquote(impl)
             @spec __impl__(:protocol) :: unquote(protocol)
             @spec __impl__(:for) :: unquote(for)
-            def __impl__(:target), do: unquote(impl)
             def __impl__(:protocol), do: unquote(protocol)
             def __impl__(:for), do: unquote(for)
           end
 
-        Module.create(Module.concat(protocol, for), quoted, Macro.Env.location(env))
+        Module.create(Module.concat(protocol, for), [quoted | funs], Macro.Env.location(env))
       end
     end)
   end
