@@ -42,11 +42,12 @@ defmodule System do
     * `system_time/0` - the VM view of the `os_time/0`. The system time and operating
       system time may not match in case of time warps although the VM works towards
       aligning them. This time is not monotonic (i.e., it may decrease)
-      as its behaviour is configured [by the VM time warp
+      as its behavior is configured [by the VM time warp
       mode](https://www.erlang.org/doc/apps/erts/time_correction.html#Time_Warp_Modes);
 
     * `monotonic_time/0` - a monotonically increasing time provided
-      by the Erlang VM.
+      by the Erlang VM. This is not strictly monotonically increasing. Multiple
+      sequential calls of the function may return the same value.
 
   The time functions in this module work in the `:native` unit
   (unless specified otherwise), which is operating system dependent. Most of
@@ -63,6 +64,12 @@ defmodule System do
   """
 
   defmodule EnvError do
+    @moduledoc """
+    An exception raised when a system environment variable is not set.
+
+    For example, see `System.fetch_env!/1`.
+    """
+
     defexception [:env]
 
     @impl true
@@ -141,11 +148,11 @@ defmodule System do
   defmacrop get_revision do
     null =
       case :os.type() do
-        {:win32, _} -> 'NUL'
-        _ -> '/dev/null'
+        {:win32, _} -> ~c"NUL"
+        _ -> ~c"/dev/null"
       end
 
-    'git rev-parse --short=7 HEAD 2> '
+    ~c"git rev-parse --short=7 HEAD 2> "
     |> Kernel.++(null)
     |> :os.cmd()
     |> strip
@@ -157,7 +164,7 @@ defmodule System do
   # Follows https://reproducible-builds.org/specs/source-date-epoch/
   defmacrop get_date do
     unix_epoch =
-      if source_date_epoch = :os.getenv('SOURCE_DATE_EPOCH') do
+      if source_date_epoch = :os.getenv(~c"SOURCE_DATE_EPOCH") do
         try do
           List.to_integer(source_date_epoch)
         rescue
@@ -374,8 +381,8 @@ defmodule System do
   """
   @spec tmp_dir() :: String.t() | nil
   def tmp_dir do
-    write_env_tmp_dir('TMPDIR') || write_env_tmp_dir('TEMP') || write_env_tmp_dir('TMP') ||
-      write_tmp_dir('/tmp') || write_cwd_tmp_dir()
+    write_env_tmp_dir(~c"TMPDIR") || write_env_tmp_dir(~c"TEMP") || write_env_tmp_dir(~c"TMP") ||
+      write_tmp_dir(~c"/tmp") || write_cwd_tmp_dir()
   end
 
   defp write_cwd_tmp_dir do
@@ -435,7 +442,7 @@ defmodule System do
   The function must receive the exit status code as an argument.
 
   If the VM terminates programmatically, via `System.stop/1`, `System.halt/1`,
-  or exit signals, the `at_exit/1` callbacks are not executed.
+  or exit signals, the `at_exit/1` callbacks are not guaranteed to be executed.
   """
   @spec at_exit((non_neg_integer -> any)) :: :ok
   def at_exit(fun) when is_function(fun, 1) do
@@ -472,8 +479,10 @@ defmodule System do
   @doc """
   Traps the given `signal` to execute the `fun`.
 
-  > **Important**: Trapping signals may have strong implications
-  > on how a system shuts down and behave in production and
+  > #### Avoid setting traps in libraries {: .warning}
+  >
+  > Trapping signals may have strong implications
+  > on how a system shuts down and behaves in production and
   > therefore it is extremely discouraged for libraries to
   > set their own traps. Instead, they should redirect users
   > to configure them themselves. The only cases where it is
@@ -497,7 +506,7 @@ defmodule System do
   in case trapping exists is not supported by the current OS.
 
   The first time a signal is trapped, it will override the
-  default behaviour from the operating system. If the same
+  default behavior from the operating system. If the same
   signal is trapped multiple times, subsequent functions
   given to `trap_signal` will execute *first*. In other
   words, you can consider each function is prepended to
@@ -511,7 +520,7 @@ defmodule System do
     * `:sigusr1` - halts the VM via status code of 1
 
   Therefore, if you add traps to the signals above, the
-  default behaviour above will be executed after all user
+  default behavior above will be executed after all user
   signals.
 
   ## Implementation notes
@@ -528,8 +537,8 @@ defmodule System do
   it may disable Elixir traps (or Elixir may override your configuration).
   """
   @doc since: "1.12.0"
-  @spec trap_signal(signal, (() -> :ok)) :: {:ok, reference()} | {:error, :not_sup}
-  @spec trap_signal(signal, id, (() -> :ok)) ::
+  @spec trap_signal(signal, (-> :ok)) :: {:ok, reference()} | {:error, :not_sup}
+  @spec trap_signal(signal, id, (-> :ok)) ::
           {:ok, id} | {:error, :already_registered} | {:error, :not_sup}
         when id: term()
   def trap_signal(signal, id \\ make_ref(), fun)
@@ -606,9 +615,6 @@ defmodule System do
     end
   end
 
-  # TODO: Remove this once we require Erlang/OTP 24+
-  @compile {:no_warn_undefined, {:os, :env, 0}}
-
   @doc """
   Returns all system environment variables.
 
@@ -617,17 +623,9 @@ defmodule System do
   """
   @spec get_env() :: %{optional(String.t()) => String.t()}
   def get_env do
-    if function_exported?(:os, :env, 0) do
-      Map.new(:os.env(), fn {k, v} ->
-        {IO.chardata_to_string(k), IO.chardata_to_string(v)}
-      end)
-    else
-      Enum.into(:os.getenv(), %{}, fn var ->
-        var = IO.chardata_to_string(var)
-        [k, v] = String.split(var, "=", parts: 2)
-        {k, v}
-      end)
-    end
+    Map.new(:os.env(), fn {k, v} ->
+      {IO.chardata_to_string(k), IO.chardata_to_string(v)}
+    end)
   end
 
   @doc """
@@ -651,7 +649,8 @@ defmodule System do
 
   """
   @doc since: "1.9.0"
-  @spec get_env(String.t(), String.t() | nil) :: String.t() | nil
+  @spec get_env(String.t(), String.t()) :: String.t()
+  @spec get_env(String.t(), nil) :: String.t() | nil
   def get_env(varname, default \\ nil)
       when is_binary(varname) and
              (is_binary(default) or is_nil(default)) do
@@ -697,7 +696,7 @@ defmodule System do
       "4000"
 
       iex> System.fetch_env!("NOT_SET")
-      ** (ArgumentError) could not fetch environment variable "NOT_SET" because it is not set
+      ** (System.EnvError) could not fetch environment variable "NOT_SET" because it is not set
 
   """
   @doc since: "1.9.0"
@@ -740,9 +739,12 @@ defmodule System do
   Sets multiple environment variables.
 
   Sets a new value for each environment variable corresponding
-  to each `{key, value}` pair in `enum`. Keys are automatically
-  converted to strings, values are sent as is. `nil` values erase
+  to each `{key, value}` pair in `enum`. Keys and non-nil values
+  are automatically converted to charlists. `nil` values erase
   the given keys.
+
+  Overall, this is a convenience wrapper around `put_env/2` and
+  `delete_env/2` with support for different key and value formats.
   """
   @spec put_env(Enumerable.t()) :: :ok
   def put_env(enum) do
@@ -862,7 +864,7 @@ defmodule System do
   defdelegate restart(), to: :init
 
   @doc """
-  Carefully stops the Erlang runtime system.
+  Asynchronously and carefully stops the Erlang runtime system.
 
   All applications are taken down smoothly, all code is unloaded, and all ports
   are closed before the system terminates by calling `halt/1`.
@@ -870,13 +872,16 @@ defmodule System do
   `status` must be a non-negative integer or a binary.
 
     * If an integer, the runtime system exits with the integer value which is
-      returned to the operating system.
+      returned to the operating system. On many platforms, only the status codes
+      0-255 are supported by the operating system.
 
     * If a binary, an Erlang crash dump is produced with status as slogan, and
       then the runtime system exits with status code 1.
 
-  Note that on many platforms, only the status codes 0-255 are supported
-  by the operating system.
+  Note this function is asynchronous and the current process will continue
+  executing after this function is invoked. In case you want to block the
+  current process until the system effectively shuts down, you can invoke
+  `Process.sleep(:infinity)`.
 
   ## Examples
 
@@ -885,14 +890,16 @@ defmodule System do
 
   """
   @doc since: "1.5.0"
-  @spec stop(non_neg_integer | binary) :: no_return
+  @spec stop(non_neg_integer | binary) :: :ok
   def stop(status \\ 0)
 
   def stop(status) when is_integer(status) do
+    at_exit(fn _ -> Process.sleep(:infinity) end)
     :init.stop(status)
   end
 
   def stop(status) when is_binary(status) do
+    at_exit(fn _ -> Process.sleep(:infinity) end)
     :init.stop(String.to_charlist(status))
   end
 
@@ -901,7 +908,9 @@ defmodule System do
 
   It uses `sh` for Unix-like systems and `cmd` for Windows.
 
-  > **Important:**: Use this function with care. In particular, **never
+  > #### Watch out {: .warning}
+  >
+  > Use this function with care. In particular, **never
   > pass untrusted user input to this function**, as the user would be
   > able to perform "command injection attacks" by executing any code
   > directly on the machine. Generally speaking, prefer to use `cmd/3`
@@ -920,19 +929,32 @@ defmodule System do
 
   ## Options
 
-  It accepts the same options as `cmd/3`, except for `arg0`.
+  It accepts the same options as `cmd/3` (except for `arg0`).
+  It also accepts the following exclusive options:
+
+    * `:close_stdin` (since v1.14.1) - if the stdin should be closed
+      on Unix systems, forcing any command that waits on stdin to
+      immediately terminate. Defaults to false.
   """
   @doc since: "1.12.0"
   @spec shell(binary, keyword) :: {Collectable.t(), exit_status :: non_neg_integer}
   def shell(command, opts \\ []) when is_binary(command) do
+    command |> String.trim() |> do_shell(opts)
+  end
+
+  defp do_shell("", _opts), do: {"", 0}
+
+  defp do_shell(command, opts) do
     assert_no_null_byte!(command, "System.shell/2")
+    {close_stdin?, opts} = Keyword.pop(opts, :close_stdin, false)
 
     # Finding shell command logic from :os.cmd in OTP
     # https://github.com/erlang/otp/blob/8deb96fb1d017307e22d2ab88968b9ef9f1b71d0/lib/kernel/src/os.erl#L184
     case :os.type() do
       {:unix, _} ->
-        shell_path = :os.find_executable('sh') || :erlang.error(:enoent, [command, opts])
-        command = "(#{command}\n) </dev/null"
+        shell_path = :os.find_executable(~c"sh") || :erlang.error(:enoent, [command, opts])
+        close_stdin = if close_stdin?, do: " </dev/null", else: ""
+        command = IO.iodata_to_binary(["(", command, "\n)", close_stdin])
         do_cmd({:spawn_executable, shell_path}, [args: ["-c", command]], opts)
 
       {:win32, osname} ->
@@ -940,9 +962,9 @@ defmodule System do
 
         command =
           case {System.get_env("COMSPEC"), osname} do
-            {nil, :windows} -> 'command.com /s /c ' ++ command
-            {nil, _} -> 'cmd /s /c ' ++ command
-            {cmd, _} -> '#{cmd} /s /c ' ++ command
+            {nil, :windows} -> ~c"command.com /s /c " ++ command
+            {nil, _} -> ~c"cmd /s /c " ++ command
+            {cmd, _} -> ~c"#{cmd} /s /c " ++ command
           end
 
         do_cmd({:spawn, command}, [], opts)
@@ -973,6 +995,25 @@ defmodule System do
   `Port` module describes this problem and possible solutions under
   the "Zombie processes" section.
 
+  > #### Windows argument splitting and untrusted arguments {: .warning}
+  >
+  > On Unix systems, arguments are passed to a new operating system
+  > process as an array of strings but on Windows it is up to the child
+  > process to parse them and some Windows programs may apply their own
+  > rules, which are inconsistent with the standard C runtime `argv` parsing
+  >
+  > This is particularly troublesome when invoking `.bat` or `.com` files
+  > as these run implicitly through `cmd.exe`, whose argument parsing is
+  > vulnerable to malicious input and can be used to run arbitrary shell
+  > commands.
+  >
+  > Therefore, if you are running on Windows and you execute batch
+  > files or `.com` applications, you must not pass untrusted input as
+  > arguments to the program. You may avoid accidentally executing them
+  > by explicitly passing the extension of the program you want to run,
+  > such as `.exe`, and double check the program is indeed not a batch
+  > file or `.com` application.
+
   ## Examples
 
       iex> System.cmd("echo", ["hello"])
@@ -987,23 +1028,43 @@ defmodule System do
       hello
       {%IO.Stream{}, 0}
 
+  If you want to read lines:
+
+      iex> System.cmd("echo", ["hello\nworld"], into: [], lines: 1024)
+      {["hello", "world"], 0}
+
   ## Options
 
     * `:into` - injects the result into the given collectable, defaults to `""`
+
+    * `:lines` - (since v1.15.0) reads the output by lines instead of in bytes. It expects a
+      number of maximum bytes to buffer internally (1024 is a reasonable default).
+      The collectable will be called with each finished line (regardless of buffer
+      size) and without the EOL character
+
     * `:cd` - the directory to run the command in
+
     * `:env` - an enumerable of tuples containing environment key-value as
       binary. The child process inherits all environment variables from its
       parent process, the Elixir application, except those overwritten or
       cleared using this option. Specify a value of `nil` to clear (unset) an
       environment variable, which is useful for preventing credentials passed
-      to the application from leaking into child processes.
+      to the application from leaking into child processes
+
     * `:arg0` - sets the command arg0
-    * `:stderr_to_stdout` - redirects stderr to stdout when `true`
+
+    * `:stderr_to_stdout` - redirects stderr to stdout when `true`, no effect
+      if `use_stdio` is `false`.
+
+    * `:use_stdio` - `true` by default, setting it to false allows direct
+      interaction with the terminal from the callee
+
     * `:parallelism` - when `true`, the VM will schedule port tasks to improve
       parallelism in the system. If set to `false`, the VM will try to perform
       commands immediately, improving latency at the expense of parallelism.
-      The default can be set on system startup by passing the "+spp" argument
-      to `--erl`.
+      The default is `false`, and can be set on system startup by passing the
+      [`+spp`](https://www.erlang.org/doc/man/erl.html#+spp) flag to `--erl`.
+      Use `:erlang.system_info(:port_parallelism)` to check if enabled.
 
   ## Error reasons
 
@@ -1040,7 +1101,7 @@ defmodule System do
   def cmd(command, args, opts \\ []) when is_binary(command) and is_list(args) do
     assert_no_null_byte!(command, "System.cmd/3")
 
-    unless Enum.all?(args, &is_binary/1) do
+    if not Enum.all?(args, &is_binary/1) do
       raise ArgumentError, "all arguments for System.cmd/3 must be binaries"
     end
 
@@ -1057,11 +1118,18 @@ defmodule System do
   end
 
   defp do_cmd(port_init, base_opts, opts) do
-    {into, opts} = cmd_opts(opts, [:use_stdio, :exit_status, :binary, :hide] ++ base_opts, "")
+    {use_stdio?, opts} = Keyword.pop(opts, :use_stdio, true)
+
+    {into, line, opts} =
+      cmd_opts(opts, [:exit_status, :binary, :hide] ++ base_opts, "", false, use_stdio?)
+
     {initial, fun} = Collectable.into(into)
 
     try do
-      do_port(Port.open(port_init, opts), initial, fun)
+      case line do
+        true -> do_port_line(Port.open(port_init, opts), initial, fun, [])
+        false -> do_port_byte(Port.open(port_init, opts), initial, fun)
+      end
     catch
       kind, reason ->
         fun.(initial, :halt)
@@ -1071,42 +1139,72 @@ defmodule System do
     end
   end
 
-  defp do_port(port, acc, fun) do
+  defp do_port_byte(port, acc, fun) do
     receive do
       {^port, {:data, data}} ->
-        do_port(port, fun.(acc, {:cont, data}), fun)
+        do_port_byte(port, fun.(acc, {:cont, data}), fun)
 
       {^port, {:exit_status, status}} ->
         {acc, status}
     end
   end
 
-  defp cmd_opts([{:into, any} | t], opts, _into),
-    do: cmd_opts(t, opts, any)
+  defp do_port_line(port, acc, fun, buffer) do
+    receive do
+      {^port, {:data, {:noeol, data}}} ->
+        do_port_line(port, acc, fun, [data | buffer])
 
-  defp cmd_opts([{:cd, bin} | t], opts, into) when is_binary(bin),
-    do: cmd_opts(t, [{:cd, bin} | opts], into)
+      {^port, {:data, {:eol, data}}} ->
+        data = [data | buffer] |> Enum.reverse() |> IO.iodata_to_binary()
+        do_port_line(port, fun.(acc, {:cont, data}), fun, [])
 
-  defp cmd_opts([{:arg0, bin} | t], opts, into) when is_binary(bin),
-    do: cmd_opts(t, [{:arg0, bin} | opts], into)
+      {^port, {:exit_status, status}} ->
+        # Data may arrive after exit status on line mode
+        receive do
+          {^port, {:data, {_, data}}} ->
+            data = [data | buffer] |> Enum.reverse() |> IO.iodata_to_binary()
+            {fun.(acc, {:cont, data}), status}
+        after
+          0 -> {acc, status}
+        end
+    end
+  end
 
-  defp cmd_opts([{:stderr_to_stdout, true} | t], opts, into),
-    do: cmd_opts(t, [:stderr_to_stdout | opts], into)
+  defp cmd_opts([{:into, any} | t], opts, _into, line, stdio?),
+    do: cmd_opts(t, opts, any, line, stdio?)
 
-  defp cmd_opts([{:stderr_to_stdout, false} | t], opts, into),
-    do: cmd_opts(t, opts, into)
+  defp cmd_opts([{:cd, bin} | t], opts, into, line, stdio?) when is_binary(bin),
+    do: cmd_opts(t, [{:cd, bin} | opts], into, line, stdio?)
 
-  defp cmd_opts([{:parallelism, bool} | t], opts, into) when is_boolean(bool),
-    do: cmd_opts(t, [{:parallelism, bool} | opts], into)
+  defp cmd_opts([{:arg0, bin} | t], opts, into, line, stdio?) when is_binary(bin),
+    do: cmd_opts(t, [{:arg0, bin} | opts], into, line, stdio?)
 
-  defp cmd_opts([{:env, enum} | t], opts, into),
-    do: cmd_opts(t, [{:env, validate_env(enum)} | opts], into)
+  defp cmd_opts([{:stderr_to_stdout, true} | t], opts, into, line, true),
+    do: cmd_opts(t, [:stderr_to_stdout | opts], into, line, true)
 
-  defp cmd_opts([{key, val} | _], _opts, _into),
+  defp cmd_opts([{:stderr_to_stdout, true} | _], _opts, _into, _line, false),
+    do: raise(ArgumentError, "cannot use \"stderr_to_stdout: true\" and \"use_stdio: false\"")
+
+  defp cmd_opts([{:stderr_to_stdout, false} | t], opts, into, line, stdio?),
+    do: cmd_opts(t, opts, into, line, stdio?)
+
+  defp cmd_opts([{:parallelism, bool} | t], opts, into, line, stdio?) when is_boolean(bool),
+    do: cmd_opts(t, [{:parallelism, bool} | opts], into, line, stdio?)
+
+  defp cmd_opts([{:env, enum} | t], opts, into, line, stdio?),
+    do: cmd_opts(t, [{:env, validate_env(enum)} | opts], into, line, stdio?)
+
+  defp cmd_opts([{:lines, max_line_length} | t], opts, into, _line, stdio?)
+       when is_integer(max_line_length) and max_line_length > 0,
+       do: cmd_opts(t, [{:line, max_line_length} | opts], into, true, stdio?)
+
+  defp cmd_opts([{key, val} | _], _opts, _into, _line, _stdio?),
     do: raise(ArgumentError, "invalid option #{inspect(key)} with value #{inspect(val)}")
 
-  defp cmd_opts([], opts, into),
-    do: {into, opts}
+  defp cmd_opts([], opts, into, line, stdio?) do
+    opt = if stdio?, do: :use_stdio, else: :nouse_stdio
+    {into, line, [opt | opts]}
+  end
 
   defp validate_env(enum) do
     Enum.map(enum, fn
@@ -1125,7 +1223,8 @@ defmodule System do
   Returns the current monotonic time in the `:native` time unit.
 
   This time is monotonically increasing and starts in an unspecified
-  point in time.
+  point in time. This is not strictly monotonically increasing. Multiple
+  sequential calls of the function may return the same value.
 
   Inlined by the compiler.
   """
@@ -1140,7 +1239,7 @@ defmodule System do
   This time is monotonically increasing and starts in an unspecified
   point in time.
   """
-  @spec monotonic_time(time_unit) :: integer
+  @spec monotonic_time(time_unit | :native) :: integer
   def monotonic_time(unit) do
     :erlang.monotonic_time(normalize_time_unit(unit))
   end
@@ -1166,7 +1265,7 @@ defmodule System do
   case of time warps although the VM works towards aligning
   them. This time is not monotonic.
   """
-  @spec system_time(time_unit) :: integer
+  @spec system_time(time_unit | :native) :: integer
   def system_time(unit) do
     :erlang.system_time(normalize_time_unit(unit))
   end
@@ -1217,7 +1316,7 @@ defmodule System do
   `monotonic_time/1`), gives the Erlang system time that corresponds
   to that monotonic time.
   """
-  @spec time_offset(time_unit) :: integer
+  @spec time_offset(time_unit | :native) :: integer
   def time_offset(unit) do
     :erlang.time_offset(normalize_time_unit(unit))
   end
@@ -1244,7 +1343,7 @@ defmodule System do
   This time may be adjusted forwards or backwards in time
   with no limitation and is not monotonic.
   """
-  @spec os_time(time_unit) :: integer
+  @spec os_time(time_unit | :native) :: integer
   @doc since: "1.3.0"
   def os_time(unit) do
     :os.system_time(normalize_time_unit(unit))
@@ -1343,8 +1442,10 @@ defmodule System do
   defp warn(unit, replacement_unit) do
     IO.warn_once(
       {__MODULE__, unit},
-      "deprecated time unit: #{inspect(unit)}. A time unit should be " <>
-        ":second, :millisecond, :microsecond, :nanosecond, or a positive integer",
+      fn ->
+        "deprecated time unit: #{inspect(unit)}. A time unit should be " <>
+          ":second, :millisecond, :microsecond, :nanosecond, or a positive integer"
+      end,
       _stacktrace_drop_levels = 4
     )
 
