@@ -3,18 +3,20 @@ defmodule Task.Supervised do
   @ref_timeout 5000
 
   def start(owner, callers, fun) do
-    {:ok, :proc_lib.spawn(__MODULE__, :noreply, [owner, callers, fun])}
+    {:ok, spawn(__MODULE__, :noreply, [owner, get_ancestors(), callers, fun])}
   end
 
   def start_link(owner, callers, fun) do
-    {:ok, :proc_lib.spawn_link(__MODULE__, :noreply, [owner, callers, fun])}
+    {:ok, spawn_link(__MODULE__, :noreply, [owner, get_ancestors(), callers, fun])}
   end
 
   def start_link(owner, monitor) do
-    {:ok, :proc_lib.spawn_link(__MODULE__, :reply, [owner, monitor])}
+    {:ok, spawn_link(__MODULE__, :reply, [owner, get_ancestors(), monitor])}
   end
 
-  def reply({_, _, owner_pid} = owner, monitor) do
+  def reply({_, _, owner_pid} = owner, ancestors, monitor) do
+    put_ancestors(ancestors)
+
     case monitor do
       :monitor ->
         mref = Process.monitor(owner_pid)
@@ -28,9 +30,9 @@ defmodule Task.Supervised do
   defp reply(owner, owner_pid, mref, timeout) do
     receive do
       {^owner_pid, ref, reply_to, callers, mfa} ->
-        initial_call(mfa)
+        put_initial_call(mfa)
         put_callers(callers)
-        _ = mref && Process.demonitor(mref, [:flush])
+        _ = is_reference(mref) && Process.demonitor(mref, [:flush])
         send(reply_to, {ref, invoke_mfa(owner, mfa)})
 
       {:DOWN, ^mref, _, _, reason} ->
@@ -60,24 +62,34 @@ defmodule Task.Supervised do
     end
   end
 
-  def noreply(owner, callers, mfa) do
-    initial_call(mfa)
+  def noreply(owner, ancestors, callers, mfa) do
+    put_initial_call(mfa)
+    put_ancestors(ancestors)
     put_callers(callers)
     invoke_mfa(owner, mfa)
   end
 
+  defp get_ancestors() do
+    case :erlang.get(:"$ancestors") do
+      ancestors when is_list(ancestors) -> [self() | ancestors]
+      _ -> [self()]
+    end
+  end
+
+  defp put_ancestors(ancestors) do
+    Process.put(:"$ancestors", ancestors)
+  end
+
   defp put_callers(callers) do
     Process.put(:"$callers", callers)
   end
 
-  defp initial_call(mfa) do
+  defp put_initial_call(mfa) do
     Process.put(:"$initial_call", get_initial_call(mfa))
   end
 
   defp get_initial_call({:erlang, :apply, [fun, []]}) when is_function(fun, 0) do
-    {:module, module} = Function.info(fun, :module)
-    {:name, name} = Function.info(fun, :name)
-    {module, name, 0}
+    :erlang.fun_info_mfa(fun)
   end
 
   defp get_initial_call({mod, fun, args}) do
@@ -105,7 +117,9 @@ defmodule Task.Supervised do
               starter: get_from(owner),
               function: fun,
               args: args,
-              reason: {log_value(kind, value), __STACKTRACE__}
+              reason: {log_value(kind, value), __STACKTRACE__},
+              # TODO use Process.get_label/0 when we require Erlang/OTP 27+
+              process_label: Process.get(:"$process_label", :undefined)
             }
           },
           %{
@@ -116,10 +130,14 @@ defmodule Task.Supervised do
           }
         )
 
-        :erlang.raise(kind, value, __STACKTRACE__)
+        :erlang.raise(:exit, exit_reason(kind, value, __STACKTRACE__), __STACKTRACE__)
     end
   end
 
+  defp exit_reason(:error, reason, stacktrace), do: {reason, stacktrace}
+  defp exit_reason(:exit, reason, _stacktrace), do: reason
+  defp exit_reason(:throw, reason, stacktrace), do: {{:nocatch, reason}, stacktrace}
+
   defp log_value(:throw, value), do: {:nocatch, value}
   defp log_value(_, value), do: value
 
@@ -131,16 +149,27 @@ defmodule Task.Supervised do
           starter: starter,
           function: fun,
           args: args,
-          reason: reason
+          reason: reason,
+          process_label: process_label
         }
       }) do
     message =
-      '** Task ~p terminating~n' ++
-        '** Started from ~p~n' ++
-        '** When function  == ~p~n' ++
-        '**      arguments == ~p~n' ++ '** Reason for termination == ~n' ++ '** ~p~n'
+      ~c"** Started from ~p~n" ++
+        ~c"** When function  == ~p~n" ++
+        ~c"**      arguments == ~p~n" ++ ~c"** Reason for termination == ~n" ++ ~c"** ~p~n"
 
-    {message, [starter, name, fun, args, get_reason(reason)]}
+    terms = [name, fun, args, get_reason(reason)]
+
+    {message, terms} =
+      case process_label do
+        :undefined -> {message, terms}
+        _ -> {~c"** Process Label == ~p~n" ++ message, [process_label | terms]}
+      end
+
+    message =
+      ~c"** Task ~p terminating~n" ++ message
+
+    {message, [starter | terms]}
   end
 
   defp get_from({node, pid_or_name, _pid}) when node == node(), do: pid_or_name
@@ -153,7 +182,7 @@ defmodule Task.Supervised do
   defp get_reason({:undef, [{mod, fun, args, _info} | _] = stacktrace} = reason)
        when is_atom(mod) and is_atom(fun) do
     cond do
-      :code.is_loaded(mod) === false ->
+      not Code.loaded?(mod) ->
         {:"module could not be loaded", stacktrace}
 
       is_list(args) and not function_exported?(mod, fun, length(args)) ->
@@ -173,18 +202,36 @@ defmodule Task.Supervised do
 
   ## Stream
 
-  def stream(enumerable, acc, reducer, callers, mfa, options, spawn) do
-    next = &Enumerable.reduce(enumerable, &1, fn x, acc -> {:suspend, [x | acc]} end)
-    max_concurrency = Keyword.get(options, :max_concurrency, System.schedulers_online())
+  def validate_stream_options(options) do
+    max_concurrency = Keyword.get_lazy(options, :max_concurrency, &System.schedulers_online/0)
+    on_timeout = Keyword.get(options, :on_timeout, :exit)
+    timeout = Keyword.get(options, :timeout, 5000)
+    ordered = Keyword.get(options, :ordered, true)
+    zip_input_on_exit = Keyword.get(options, :zip_input_on_exit, false)
 
-    unless is_integer(max_concurrency) and max_concurrency > 0 do
+    if not (is_integer(max_concurrency) and max_concurrency > 0) do
       raise ArgumentError, ":max_concurrency must be an integer greater than zero"
     end
 
-    ordered? = Keyword.get(options, :ordered, true)
-    timeout = Keyword.get(options, :timeout, 5000)
-    on_timeout = Keyword.get(options, :on_timeout, :exit)
-    zip_input_on_exit? = Keyword.get(options, :zip_input_on_exit, false)
+    if on_timeout not in [:exit, :kill_task] do
+      raise ArgumentError, ":on_timeout must be either :exit or :kill_task"
+    end
+
+    if not ((is_integer(timeout) and timeout >= 0) or timeout == :infinity) do
+      raise ArgumentError, ":timeout must be either a positive integer or :infinity"
+    end
+
+    %{
+      max_concurrency: max_concurrency,
+      on_timeout: on_timeout,
+      timeout: timeout,
+      ordered: ordered,
+      zip_input_on_exit: zip_input_on_exit
+    }
+  end
+
+  def stream(enumerable, acc, reducer, callers, mfa, options, spawn) when is_map(options) do
+    next = &Enumerable.reduce(enumerable, &1, fn x, acc -> {:suspend, [x | acc]} end)
     parent = self()
 
     {:trap_exit, trap_exit?} = Process.info(self(), :trap_exit)
@@ -196,7 +243,7 @@ defmodule Task.Supervised do
 
     {monitor_pid, monitor_ref} =
       Process.spawn(
-        fn -> stream_monitor(parent, spawn, trap_exit?, timeout) end,
+        fn -> stream_monitor(parent, spawn, trap_exit?, options.timeout) end,
         spawn_opts
       )
 
@@ -205,21 +252,21 @@ defmodule Task.Supervised do
     # about our reference to it.
     send(monitor_pid, {parent, monitor_ref})
 
-    config = %{
-      reducer: reducer,
-      monitor_pid: monitor_pid,
-      monitor_ref: monitor_ref,
-      ordered: ordered?,
-      timeout: timeout,
-      on_timeout: on_timeout,
-      zip_input_on_exit: zip_input_on_exit?,
-      callers: callers,
-      mfa: mfa
-    }
+    config =
+      Map.merge(
+        options,
+        %{
+          reducer: reducer,
+          monitor_pid: monitor_pid,
+          monitor_ref: monitor_ref,
+          callers: callers,
+          mfa: mfa
+        }
+      )
 
     stream_reduce(
       acc,
-      max_concurrency,
+      options.max_concurrency,
       _spawned = 0,
       _delivered = 0,
       _waiting = %{},
