@@ -6,6 +6,8 @@
 -export([start_link/0, init/1, handle_call/3, handle_cast/2]).
 -behaviour(gen_server).
 
+%% Persistent term
+
 static(Map) when is_map(Map) ->
   persistent_term:put(?MODULE, maps:merge(persistent_term:get(?MODULE, #{}), Map)).
 is_bootstrap() ->
@@ -13,6 +15,8 @@ is_bootstrap() ->
 identifier_tokenizer() ->
   maps:get(identifier_tokenizer, persistent_term:get(?MODULE, #{}), 'Elixir.String.Tokenizer').
 
+%% Key-value store (concurrent reads, serial writes)
+
 get(Key) ->
   [{_, Value}] = ets:lookup(?MODULE, Key),
   Value.
@@ -26,16 +30,26 @@ get(Key, Default) ->
   end.
 
 put(Key, Value) ->
-  gen_server:call(?MODULE, {put, Key, Value}).
+  gen_server:call(?MODULE, {put, Key, Value}, infinity).
 
 get_and_put(Key, Value) ->
-  gen_server:call(?MODULE, {get_and_put, Key, Value}).
+  gen_server:call(?MODULE, {get_and_put, Key, Value}, infinity).
 
 update(Key, Fun) ->
-  gen_server:call(?MODULE, {update, Key, Fun}).
+  gen_server:call(?MODULE, {update, Key, Fun}, infinity).
+
+new(Opts) ->
+  Tab = ets:new(?MODULE, [named_table, public, {read_concurrency, true}]),
+  true = ets:insert_new(?MODULE, Opts),
+  Tab.
+
+delete(?MODULE) ->
+  ets:delete(?MODULE).
+
+%% MISC
 
 serial(Fun) ->
-  gen_server:call(?MODULE, {serial, Fun}).
+  gen_server:call(?MODULE, {serial, Fun}, infinity).
 
 %% Used to guarantee warnings are emitted only once per caller.
 warn(Key, [{Mod, Fun, ArgsOrArity, _} | _]) ->
@@ -48,37 +62,27 @@ warn(_, _) ->
 to_arity(Args) when is_list(Args) -> length(Args);
 to_arity(Arity) -> Arity.
 
-%% ets life-cycle api
-
-new(Opts) ->
-  Tab = ets:new(?MODULE, [named_table, public, {read_concurrency, true}]),
-  true = ets:insert_new(?MODULE, Opts),
-  Tab.
-
-delete(?MODULE) ->
-  ets:delete(?MODULE).
-
 %% gen_server api
 
 start_link() ->
   gen_server:start_link({local, ?MODULE}, ?MODULE, ?MODULE, []).
 
-init(Tab) ->
-  {ok, Tab}.
+init(?MODULE) ->
+  {ok, []}.
 
-handle_call({serial, Fun}, _From, Tab) ->
-  {reply, Fun(), Tab};
-handle_call({put, Key, Value}, _From, Tab) ->
-  ets:insert(Tab, {Key, Value}),
-  {reply, ok, Tab};
-handle_call({update, Key, Fun}, _From, Tab) ->
+handle_call({serial, Fun}, _From, State) ->
+  {reply, Fun(), State};
+  handle_call({put, Key, Value}, _From, State) ->
+  ets:insert(?MODULE, {Key, Value}),
+  {reply, ok, State};
+handle_call({update, Key, Fun}, _From, State) ->
   Value = Fun(get(Key)),
-  ets:insert(Tab, {Key, Value}),
-  {reply, Value, Tab};
-handle_call({get_and_put, Key, Value}, _From, Tab) ->
+  ets:insert(?MODULE, {Key, Value}),
+  {reply, Value, State};
+handle_call({get_and_put, Key, Value}, _From, State) ->
   OldValue = get(Key),
-  ets:insert(Tab, {Key, Value}),
-  {reply, OldValue, Tab}.
+  ets:insert(?MODULE, {Key, Value}),
+  {reply, OldValue, State}.
 
 handle_cast(Cast, Tab) ->
   {stop, {bad_cast, Cast}, Tab}.
