@@ -1,7 +1,7 @@
 % Holds the logic responsible for function definitions (def(p) and defmacro(p)).
 -module(elixir_def).
 -export([setup/1, reset_last/1, local_for/5, external_for/5,
-  take_definition/2, store_definition/5, store_definition/9,
+  take_definition/2, store_definition/3, store_definition/9,
   fetch_definitions/2, format_error/1]).
 -include("elixir.hrl").
 -define(last_def, {elixir, last_def}).
@@ -37,6 +37,7 @@ fun_for(Meta, Module, Name, Arity, Kinds, External) ->
     {[{_, Kind, LocalMeta, _, _, _}], ClausesPairs} ->
       case (Kinds == all) orelse (lists:member(Kind, Kinds)) of
         true ->
+          (Kind == defmacrop) andalso track_defmacrop(Module, Tuple),
           Local = {value, fun(Fun, Args) -> invoke_local(Meta, Module, Fun, Args, External) end},
           Clauses = [Clause || {_, Clause} <- ClausesPairs],
           elixir_erl:definition_to_anonymous(Kind, LocalMeta, Clauses, Local, External);
@@ -60,6 +61,10 @@ invoke_local(Meta, Module, ErlName, Args, External) ->
       apply(Fun, Args)
   end.
 
+track_defmacrop(Module, FunArity) ->
+  {_, Bag} = elixir_module:data_tables(Module),
+  ets:insert(Bag, {used_private, FunArity}).
+
 invoke_external(Meta, Mod, Name, Args, E) ->
   is_map(E) andalso elixir_env:trace({require, Meta, Mod, []}, E),
   apply(Mod, Name, Args).
@@ -79,7 +84,7 @@ take_definition(Module, {Name, Arity} = Tuple) ->
 
 %% Fetch all available definitions
 
-fetch_definitions(File, Module) ->
+fetch_definitions(Module, E) ->
   {Set, Bag} = elixir_module:data_tables(Module),
 
   Entries = try
@@ -88,9 +93,9 @@ fetch_definitions(File, Module) ->
     error:badarg -> []
   end,
 
-  fetch_definition(Entries, File, Module, Set, Bag, [], []).
+  fetch_definition(Entries, E, Module, Set, Bag, [], []).
 
-fetch_definition([Tuple | T], File, Module, Set, Bag, All, Private) ->
+fetch_definition([Tuple | T], E, Module, Set, Bag, All, Private) ->
   [{_, Kind, Meta, _, Check, {MaxDefaults, _, Defaults}}] = ets:lookup(Set, {def, Tuple}),
 
   try ets:lookup_element(Bag, {clauses, Tuple}, 2) of
@@ -105,53 +110,65 @@ fetch_definition([Tuple | T], File, Module, Set, Bag, All, Private) ->
           false ->
             Private
         end,
-      fetch_definition(T, File, Module, Set, Bag, NewAll, NewPrivate)
+      fetch_definition(T, E, Module, Set, Bag, NewAll, NewPrivate)
   catch
     error:badarg ->
-      elixir_errors:form_error(Meta, File, ?MODULE, {function_head, Kind, Tuple}),
-      fetch_definition(T, File, Module, Set, Bag, All, Private)
+      elixir_errors:module_error(Meta, E, ?MODULE, {function_head, Kind, Tuple}),
+      fetch_definition(T, E, Module, Set, Bag, All, Private)
   end;
 
-fetch_definition([], _File, _Module, _Set, _Bag, All, Private) ->
+fetch_definition([], _E, _Module, _Set, _Bag, All, Private) ->
   {All, Private}.
 
 add_defaults_to_meta(0, Meta) -> Meta;
 add_defaults_to_meta(Defaults, Meta) -> [{defaults, Defaults} | Meta].
 
-head_and_definition_meta(true, Meta, 0, _All) ->
+head_and_definition_meta(none, _Meta, _HeadDefaults, _All) ->
+  false;
+head_and_definition_meta(_, Meta, 0, _All) ->
   Meta;
-head_and_definition_meta(true, _Meta, _HeadDefaults, [{_, _, HeadMeta, _} | _]) ->
-  HeadMeta;
-head_and_definition_meta(false, _Meta, _HeadDefaults, _All) ->
-  false.
+head_and_definition_meta(_, _Meta, _HeadDefaults, [{_, _, HeadMeta, _} | _]) ->
+  HeadMeta.
 
 %% Section for storing definitions
 
-store_definition(Kind, CheckClauses, Call, Body, Pos) ->
-  #{line := Line} = E = elixir_locals:get_cached_env(Pos),
+store_definition(Kind, {Call, Body}, Pos) ->
+  E = elixir_module:get_cached_env(Pos),
+  store_definition(Kind, false, Call, Body, E);
+store_definition(Kind, Key, Pos) ->
+  #{module := Module} = E = elixir_module:get_cached_env(Pos),
+  {Call, Body} = elixir_module:read_cache(Module, Key),
+  store_definition(Kind, true, Call, Body, E).
+
+store_definition(Kind, HasNoUnquote, Call, Body, #{line := Line} = E) ->
   {NameAndArgs, Guards} = elixir_utils:extract_guards(Call),
 
-  {Name, Args} = case NameAndArgs of
-    {N, _, A} when is_atom(N), is_atom(A) -> {N, []};
-    {N, _, A} when is_atom(N), is_list(A) -> {N, A};
-    _ -> elixir_errors:form_error([{line, Line}], E, ?MODULE, {invalid_def, Kind, NameAndArgs})
+  {Name, Meta, Args} = case NameAndArgs of
+    {N, M, A} when is_atom(N), is_atom(A) -> {N, M, []};
+    {N, M, A} when is_atom(N), is_list(A) -> {N, M, A};
+    _ -> elixir_errors:file_error([{line, Line}], E, ?MODULE, {invalid_def, Kind, NameAndArgs})
   end,
 
-  %% Now that we have verified the call format,
-  %% extract meta information like file and context.
-  {_, Meta, _} = Call,
-
   Context = case lists:keyfind(context, 1, Meta) of
     {context, _} = ContextPair -> [ContextPair];
     _ -> []
   end,
 
-  Generated = case lists:keyfind(generated, 1, Meta) of
-    {generated, true} -> ?generated(Context);
+  Column = case lists:keyfind(column, 1, Meta) of
+    {column, _} = ColumnPair -> [ColumnPair | Context];
     _ -> Context
   end,
 
-  DoCheckClauses = (Context == []) andalso (CheckClauses),
+  Generated = case lists:keyfind(generated, 1, Meta) of
+    {generated, true} = GeneratedPair -> [GeneratedPair | Column];
+    _ -> Column
+  end,
+
+  CheckClauses = if
+    Context /= [] -> none;
+    HasNoUnquote -> all;
+    true -> unused_only
+  end,
 
   %% Check if there is a file information in the definition.
   %% If so, we assume this come from another source and
@@ -184,7 +201,7 @@ store_definition(Kind, CheckClauses, Call, Body, Pos) ->
   run_with_location_change(File, E, fun(EL) ->
     assert_no_aliases_name(DefMeta, Name, Args, EL),
     assert_valid_name(DefMeta, Kind, Name, Args, EL),
-    store_definition(DefMeta, Kind, DoCheckClauses, Name, Arity,
+    store_definition(DefMeta, Kind, CheckClauses, Name, Arity,
                      LinifyArgs, LinifyGuards, LinifyBody, ?key(E, file), EL)
   end).
 
@@ -198,12 +215,11 @@ store_definition(Meta, Kind, CheckClauses, Name, Arity, DefaultsArgs, Guards, Bo
              Clause <- def_to_clauses(Kind, Meta, Args, Guards, Body, E)],
 
   DefaultsLength = length(Defaults),
-  elixir_locals:record_defaults(Tuple, Kind, Module, DefaultsLength, Meta),
   check_previous_defaults(Meta, Module, Name, Arity, Kind, DefaultsLength, E),
 
   store_definition(CheckClauses, Kind, Meta, Name, Arity, File,
                    Module, DefaultsLength, Clauses),
-  [store_definition(false, Kind, Meta, Name, length(DefaultArgs), File,
+  [store_definition(none, Kind, Meta, Name, length(DefaultArgs), File,
                     Module, 0, [Default]) || {_, DefaultArgs, _, _} = Default <- Defaults],
 
   run_on_definition_callbacks(Kind, Module, Name, DefaultsArgs, Guards, Body, E),
@@ -248,7 +264,7 @@ def_to_clauses(Kind, Meta, Args, Guards, Body, E) ->
     {do, _} ->
       [{Meta, Args, Guards, {'try', [{origin,  Kind} | Meta], [Body]}}];
     false ->
-      elixir_errors:form_error(Meta, E, elixir_expand, {missing_option, Kind, [do]})
+      elixir_errors:file_error(Meta, E, elixir_expand, {missing_option, Kind, [do]})
   end.
 
 run_on_definition_callbacks(Kind, Module, Name, Args, Guards, Body, E) ->
@@ -257,10 +273,12 @@ run_on_definition_callbacks(Kind, Module, Name, Args, Guards, Body, E) ->
   _ = [Mod:Fun(E, Kind, Name, Args, Guards, Body) || {Mod, Fun} <- lists:reverse(Callbacks)],
   ok.
 
-store_definition(Check, Kind, Meta, Name, Arity, File, Module, Defaults, Clauses) ->
+store_definition(CheckClauses, Kind, Meta, Name, Arity, File, Module, Defaults, Clauses)
+    when CheckClauses == all; CheckClauses == none; CheckClauses == unused_only ->
   {Set, Bag} = elixir_module:data_tables(Module),
   Tuple = {Name, Arity},
   HasBody = Clauses =/= [],
+  CheckAll = CheckClauses == all,
 
   if
     Defaults > 0 ->
@@ -274,7 +292,7 @@ store_definition(Check, Kind, Meta, Name, Arity, File, Module, Defaults, Clauses
       [{_, StoredKind, StoredMeta, StoredFile, StoredCheck, {StoredDefaults, LastHasBody, LastDefaults}}] ->
         check_valid_kind(Meta, File, Name, Arity, Kind, StoredKind, StoredFile, StoredMeta),
         check_valid_defaults(Meta, File, Name, Arity, Kind, Defaults, StoredDefaults, LastDefaults, HasBody, LastHasBody),
-        (Check and StoredCheck) andalso
+        (CheckAll and (StoredCheck == all)) andalso
           check_valid_clause(Meta, File, Name, Arity, Kind, Set, StoredMeta, StoredFile, Clauses),
 
         {max(Defaults, StoredDefaults), StoredMeta};
@@ -283,9 +301,9 @@ store_definition(Check, Kind, Meta, Name, Arity, File, Module, Defaults, Clauses
         {Defaults, Meta}
     end,
 
-  Check andalso ets:insert(Set, {?last_def, Tuple}),
+  CheckAll andalso ets:insert(Set, {?last_def, Tuple}),
   ets:insert(Bag, [{{clauses, Tuple}, Clause} || Clause <- Clauses]),
-  ets:insert(Set, {{def, Tuple}, Kind, FirstMeta, File, Check, {MaxDefaults, HasBody, Defaults}}).
+  ets:insert(Set, {{def, Tuple}, Kind, FirstMeta, File, CheckClauses, {MaxDefaults, HasBody, Defaults}}).
 
 %% Handling of defaults
 
@@ -332,7 +350,7 @@ default_var(Counter) ->
 
 check_valid_kind(_Meta, _File, _Name, _Arity, Kind, Kind, _StoredFile, _StoredMeta) -> ok;
 check_valid_kind(Meta, File, Name, Arity, Kind, StoredKind, StoredFile, StoredMeta) ->
-  elixir_errors:form_error(Meta, File, ?MODULE,
+  elixir_errors:file_error(Meta, File, ?MODULE,
     {changed_kind, {Name, Arity, StoredKind, Kind, StoredFile, ?line(StoredMeta)}}).
 
 check_valid_clause(Meta, File, Name, Arity, Kind, Set, StoredMeta, StoredFile, Clauses) ->
@@ -340,37 +358,37 @@ check_valid_clause(Meta, File, Name, Arity, Kind, Set, StoredMeta, StoredFile, C
     none ->
       ok;
     {Name, Arity} when Clauses == [] ->
-      elixir_errors:form_warn(Meta, File, ?MODULE,
+      elixir_errors:file_warn(Meta, File, ?MODULE,
         {late_function_head, {Kind, Name, Arity}});
     {Name, Arity} ->
       ok;
     {Name, _} ->
       Relative = elixir_utils:relative_to_cwd(StoredFile),
-      elixir_errors:form_warn(Meta, File, ?MODULE,
+      elixir_errors:file_warn(Meta, File, ?MODULE,
         {ungrouped_name, {Kind, Name, Arity, ?line(StoredMeta), Relative}});
     _ ->
       Relative = elixir_utils:relative_to_cwd(StoredFile),
-      elixir_errors:form_warn(Meta, File, ?MODULE,
+      elixir_errors:file_warn(Meta, File, ?MODULE,
         {ungrouped_arity, {Kind, Name, Arity, ?line(StoredMeta), Relative}})
   end.
 
 % Clause with defaults after clause with defaults
 check_valid_defaults(Meta, File, Name, Arity, Kind, Defaults, StoredDefaults, _, _, _)
     when Defaults > 0, StoredDefaults > 0 ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {duplicate_defaults, {Kind, Name, Arity}});
+  elixir_errors:file_error(Meta, File, ?MODULE, {duplicate_defaults, {Kind, Name, Arity}});
 % Clause with defaults after clause without defaults
 check_valid_defaults(Meta, File, Name, Arity, Kind, Defaults, 0, _, _, _) when Defaults > 0 ->
-  elixir_errors:form_warn(Meta, File, ?MODULE, {mixed_defaults, {Kind, Name, Arity}});
+  elixir_errors:file_warn(Meta, File, ?MODULE, {mixed_defaults, {Kind, Name, Arity}});
 % Clause without defaults directly after clause with defaults (bodiless does not count)
 check_valid_defaults(Meta, File, Name, Arity, Kind, 0, _, LastDefaults, true, true) when LastDefaults > 0 ->
-  elixir_errors:form_warn(Meta, File, ?MODULE, {mixed_defaults, {Kind, Name, Arity}});
+  elixir_errors:file_warn(Meta, File, ?MODULE, {mixed_defaults, {Kind, Name, Arity}});
 % Clause without defaults
 check_valid_defaults(_Meta, _File, _Name, _Arity, _Kind, 0, _StoredDefaults, _LastDefaults, _HasBody, _LastHasBody) ->
   ok.
 
 check_args_for_function_head(Meta, Args, E) ->
   [begin
-     elixir_errors:form_error(Meta, E, ?MODULE, invalid_args_for_function_head)
+     elixir_errors:module_error(Meta, E, ?MODULE, invalid_args_for_function_head)
    end || Arg <- Args, invalid_arg(Arg)].
 
 invalid_arg({Name, _, Kind}) when is_atom(Name), is_atom(Kind) -> false;
@@ -380,7 +398,7 @@ check_previous_defaults(Meta, Module, Name, Arity, Kind, Defaults, E) ->
   {_Set, Bag} = elixir_module:data_tables(Module),
   Matches = ets:lookup(Bag, {default, Name}),
   [begin
-     elixir_errors:form_error(Meta, E, ?MODULE,
+     elixir_errors:file_error(Meta, E, ?MODULE,
        {defs_with_defaults, Kind, Name, Arity, A})
    end || {_, A, D} <- Matches, A /= Arity, D /= 0, defaults_conflict(A, D, Arity, Defaults)].
 
@@ -389,18 +407,18 @@ defaults_conflict(A, D, Arity, Defaults) ->
     ((A >= (Arity - Defaults)) andalso (A < Arity)).
 
 assert_no_aliases_name(Meta, '__aliases__', [Atom], #{file := File}) when is_atom(Atom) ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {no_alias, Atom});
+  elixir_errors:file_error(Meta, File, ?MODULE, {no_alias, Atom});
 assert_no_aliases_name(_Meta, _Aliases, _Args, _S) ->
   ok.
 
 assert_valid_name(Meta, Kind, '__info__', [_], #{file := File}) ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {'__info__', Kind});
+  elixir_errors:file_error(Meta, File, ?MODULE, {'__info__', Kind});
 assert_valid_name(Meta, Kind, 'module_info', [], #{file := File}) ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {module_info, Kind, 0});
+  elixir_errors:file_error(Meta, File, ?MODULE, {module_info, Kind, 0});
 assert_valid_name(Meta, Kind, 'module_info', [_], #{file := File}) ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {module_info, Kind, 1});
+  elixir_errors:file_error(Meta, File, ?MODULE, {module_info, Kind, 1});
 assert_valid_name(Meta, Kind, is_record, [_, _], #{file := File}) when Kind == defp; Kind == def ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {is_record, Kind});
+  elixir_errors:file_error(Meta, File, ?MODULE, {is_record, Kind});
 assert_valid_name(_Meta, _Kind, _Name, _Args, _S) ->
   ok.
 
@@ -499,4 +517,4 @@ format_error({module_info, Kind, Arity}) ->
 
 format_error({is_record, Kind}) ->
   io_lib:format("cannot define ~ts is_record/2 due to compatibility "
-                "issues with the Erlang compiler (it is a known limitation)", [Kind]).
+                "with the Erlang compiler (it is a known limitation)", [Kind]).
