@@ -75,16 +75,16 @@ defmodule Code.Formatter.CallsTest do
 
     test "for heredocs" do
       assert_same """
-      foo('''
+      foo(~c'''
       bar
       ''')
       """
 
-      assert_same to_string('''
-                  foo("""
-                  bar
-                  """)
-                  ''')
+      assert_same ~S'''
+      foo("""
+      bar
+      """)
+      '''
 
       assert_same """
       foo(~S'''
@@ -165,6 +165,11 @@ defmodule Code.Formatter.CallsTest do
 
       assert_format bad, good, @short_length
     end
+
+    test "for maps" do
+      assert_same "a(%{x: 1})", @short_length
+      assert_format "ab(%{x: 1})", "ab(%{\n  x: 1\n})", @short_length
+    end
   end
 
   describe "local calls" do
@@ -590,6 +595,7 @@ defmodule Code.Formatter.CallsTest do
       assert_same ~S[Kernel.+(1, 2)]
       assert_same ~S[:erlang.+(1, 2)]
       assert_same ~S[foo."bar baz"(1, 2)]
+      assert_same ~S[foo."bar\nbaz"(1, 2)]
     end
 
     test "splits on arguments and dot on line limit" do
