@@ -177,7 +177,7 @@ defmodule Code.Formatter.IntegrationTest do
           str |> Model.Node.model_to_node_type()
 
         value, _ ->
-          Logger.warn("Could not extract node type from value: #{inspect(value)}")
+          Logger.warning("Could not extract node type from value: #{inspect(value)}")
           nil
       end)
     end
@@ -383,6 +383,26 @@ defmodule Code.Formatter.IntegrationTest do
     """
   end
 
+  test "keyword list at line limit" do
+    bad = """
+    pre()
+    config(arg, foo: bar)
+    post()
+    """
+
+    good = """
+    pre()
+
+    config(arg,
+      foo: bar
+    )
+
+    post()
+    """
+
+    assert_format bad, good, line_length: 20
+  end
+
   test "do at the end of the line with single argument" do
     bad = """
     defmodule Location do
@@ -484,12 +504,12 @@ defmodule Code.Formatter.IntegrationTest do
   test "newline after stab" do
     assert_same """
     capture_io(":erl. mof*,,l", fn ->
-      assert :io.scan_erl_form('>') == {:ok, [{:":", 1}, {:atom, 1, :erl}, {:dot, 1}], 1}
+      assert :io.scan_erl_form(~c">") == {:ok, [{:":", 1}, {:atom, 1, :erl}, {:dot, 1}], 1}
 
       expected_tokens = [{:atom, 1, :mof}, {:*, 1}, {:",", 1}, {:",", 1}, {:atom, 1, :l}]
-      assert :io.scan_erl_form('>') == {:ok, expected_tokens, 1}
+      assert :io.scan_erl_form(~c">") == {:ok, expected_tokens, 1}
 
-      assert :io.scan_erl_form('>') == {:eof, 1}
+      assert :io.scan_erl_form(~c">") == {:eof, 1}
     end)
     """
   end
@@ -678,4 +698,9 @@ defmodule Code.Formatter.IntegrationTest do
                 """,
                 line_length: :infinity
   end
+
+  test "functions without parentheses within do: keyword" do
+    assert_format ~S"defmodule Foo, do: foo bar, baz",
+                  ~S"defmodule Foo, do: foo(bar, baz)"
+  end
 end
