@@ -56,9 +56,9 @@ defmodule EnumTest do
     end
 
     test "mix and match" do
-      enums = [[1, 2], %{a: 3, b: 4}, [5, 6]]
+      enums = [[1, 2], 3..4, [5, 6]]
       result = Enum.zip_reduce(enums, [], fn elements, acc -> [List.to_tuple(elements) | acc] end)
-      assert result == [{2, {:b, 4}, 6}, {1, {:a, 3}, 5}]
+      assert result == [{2, 4, 6}, {1, 3, 5}]
     end
   end
 
@@ -94,12 +94,12 @@ defmodule EnumTest do
   end
 
   test "chunk/3" do
-    enum = Enum
+    enum = String.to_atom("Elixir.Enum")
     assert enum.chunk(1..5, 2, 1) == Enum.chunk_every(1..5, 2, 1, :discard)
   end
 
   test "chunk/4" do
-    enum = Enum
+    enum = String.to_atom("Elixir.Enum")
     assert enum.chunk(1..5, 2, 1, nil) == Enum.chunk_every(1..5, 2, 1, :discard)
   end
 
@@ -412,11 +412,25 @@ defmodule EnumTest do
     assert Enum.into([a: 1, b: 2], %{c: 3}) == %{a: 1, b: 2, c: 3}
     assert Enum.into(MapSet.new(a: 1, b: 2), %{}) == %{a: 1, b: 2}
     assert Enum.into(MapSet.new(a: 1, b: 2), %{c: 3}) == %{a: 1, b: 2, c: 3}
-    assert Enum.into(%{a: 1, b: 2}, []) == [a: 1, b: 2]
+    assert Enum.into(%{a: 1, b: 2}, []) |> Enum.sort() == [a: 1, b: 2]
     assert Enum.into(1..3, []) == [1, 2, 3]
     assert Enum.into(["H", "i"], "") == "Hi"
   end
 
+  test "into/2 exceptions" do
+    assert_raise ArgumentError,
+                 "collecting into a map requires {key, value} tuples, got: 1",
+                 fn -> Enum.into(1..10, %{}) end
+
+    assert_raise ArgumentError, "collecting into a binary requires a bitstring, got: 1", fn ->
+      Enum.into(1..10, <<>>)
+    end
+
+    assert_raise ArgumentError, "collecting into a bitstring requires a bitstring, got: 1", fn ->
+      Enum.into(1..10, <<1::1>>)
+    end
+  end
+
   test "into/3" do
     assert Enum.into([1, 2, 3], [], fn x -> x * 2 end) == [2, 4, 6]
     assert Enum.into([1, 2, 3], "numbers: ", &to_string/1) == "numbers: 123"
@@ -735,11 +749,11 @@ defmodule EnumTest do
 
     assert Enum.split_with(%{}, fn x -> rem(x, 2) == 0 end) == {[], []}
 
-    assert Enum.split_with(%{a: 1, b: 2, c: 3}, fn {_k, v} -> rem(v, 2) == 0 end) ==
-             {[b: 2], [a: 1, c: 3]}
+    assert Enum.split_with(%{a: 1, b: 2}, fn {_k, v} -> rem(v, 2) == 0 end) ==
+             {[b: 2], [a: 1]}
 
     assert Enum.split_with(%{b: 2, d: 4, f: 6}, fn {_k, v} -> rem(v, 2) == 0 end) ==
-             {[b: 2, d: 4, f: 6], []}
+             {Map.to_list(%{b: 2, d: 4, f: 6}), []}
   end
 
   test "random/1" do
@@ -851,7 +865,7 @@ defmodule EnumTest do
     end
 
     test "handles negative indices" do
-      make_negative_range = fn first..last, length ->
+      make_negative_range = fn first..last//1, length ->
         (first - length)..(last - length)//1
       end
 
@@ -890,7 +904,7 @@ defmodule EnumTest do
     end
 
     test "raises an error when the step is not exactly 1" do
-      slide_ranges_that_should_fail = [2..10//2, 8..-1, 10..2//-1, 10..4//-2, -1..-8//-1]
+      slide_ranges_that_should_fail = [2..10//2, 8..-1//-1, 10..2//-1, 10..4//-2, -1..-8//-1]
 
       for zero_to_20 <- [0..20, Enum.to_list(0..20)],
           range_that_should_fail <- slide_ranges_that_should_fail do
@@ -1008,7 +1022,7 @@ defmodule EnumTest do
   test "shuffle/1" do
     # set a fixed seed so the test can be deterministic
     :rand.seed(:exsss, {1374, 347_975, 449_264})
-    assert Enum.shuffle([1, 2, 3, 4, 5]) == [1, 3, 4, 5, 2]
+    assert Enum.shuffle([1, 2, 3, 4, 5]) == [2, 5, 4, 3, 1]
   end
 
   test "slice/2" do
@@ -1102,6 +1116,16 @@ defmodule EnumTest do
              [2, 1, 3]
   end
 
+  test "slice on MapSets" do
+    assert MapSet.new(1..10) |> Enum.slice(0, 2) |> Enum.count() == 2
+    assert MapSet.new(1..3) |> Enum.slice(0, 10) |> Enum.count() == 3
+    assert MapSet.new(1..10) |> Enum.slice(0..1) |> Enum.count() == 2
+    assert MapSet.new(1..3) |> Enum.slice(0..10) |> Enum.count() == 3
+
+    assert MapSet.new(1..10) |> Enum.slice(0..4//2) |> Enum.count() == 3
+    assert MapSet.new(1..10) |> Enum.slice(0..5//2) |> Enum.count() == 3
+  end
+
   test "sort/1" do
     assert Enum.sort([5, 3, 2, 4, 1]) == [1, 2, 3, 4, 5]
   end
@@ -1285,8 +1309,8 @@ defmodule EnumTest do
     assert Enum.sum([-3, -2, -1, 0, 1, 2, 3]) == 0
     assert Enum.sum(42..42) == 42
     assert Enum.sum(11..17) == 98
-    assert Enum.sum(17..11) == 98
-    assert Enum.sum(11..-17) == Enum.sum(-17..11)
+    assert Enum.sum(17..11//-1) == 98
+    assert Enum.sum(11..-17//-1) == Enum.sum(-17..11)
 
     assert_raise ArithmeticError, fn ->
       Enum.sum([{}])
@@ -1297,13 +1321,31 @@ defmodule EnumTest do
     end
   end
 
+  test "sum_by/2" do
+    assert Enum.sum_by([], &hd/1) == 0
+    assert Enum.sum_by([[1]], &hd/1) == 1
+    assert Enum.sum_by([[1], [2], [3]], &hd/1) == 6
+    assert Enum.sum_by([[1.1], [2.2], [3.3]], &hd/1) == 6.6
+    assert Enum.sum_by([[-3], [-2], [-1], [0], [1], [2], [3]], &hd/1) == 0
+
+    assert Enum.sum_by(1..3, &(&1 ** 2)) == 14
+
+    assert_raise ArithmeticError, fn ->
+      Enum.sum_by([[{}]], &hd/1)
+    end
+
+    assert_raise ArithmeticError, fn ->
+      Enum.sum_by([[1], [{}]], &hd/1)
+    end
+  end
+
   test "product/1" do
     assert Enum.product([]) == 1
     assert Enum.product([1]) == 1
     assert Enum.product([1, 2, 3, 4, 5]) == 120
     assert Enum.product([1, -2, 3, 4, 5]) == -120
     assert Enum.product(1..5) == 120
-    assert Enum.product(11..-17) == Enum.product(-17..11)
+    assert Enum.product(11..-17//-1) == Enum.product(-17..11)
 
     assert_raise ArithmeticError, fn ->
       Enum.product([{}])
@@ -1318,6 +1360,27 @@ defmodule EnumTest do
     end
   end
 
+  test "product_by/2" do
+    assert Enum.product_by([], &hd/1) == 1
+    assert Enum.product_by([[1]], &hd/1) == 1
+    assert Enum.product_by([[1], [2], [3], [4], [5]], &hd/1) == 120
+    assert Enum.product_by([[1], [-2], [3], [4], [5]], &hd/1) == -120
+    assert Enum.product_by(1..5, & &1) == 120
+    assert Enum.product_by(11..-17//-1, & &1) == 0
+
+    assert_raise ArithmeticError, fn ->
+      Enum.product_by([[{}]], &hd/1)
+    end
+
+    assert_raise ArithmeticError, fn ->
+      Enum.product_by([[1], [{}]], &hd/1)
+    end
+
+    assert_raise ArithmeticError, fn ->
+      Enum.product_by(%{a: 1, b: 2}, & &1)
+    end
+  end
+
   test "take/2" do
     assert Enum.take([1, 2, 3], 0) == []
     assert Enum.take([1, 2, 3], 1) == [1]
@@ -1367,16 +1430,16 @@ defmodule EnumTest do
     seed1 = {1406, 407_414, 139_258}
     seed2 = {1406, 421_106, 567_597}
     :rand.seed(:exsss, seed1)
-    assert Enum.take_random([1, 2, 3], 1) == [3]
-    assert Enum.take_random([1, 2, 3], 2) == [3, 2]
+    assert Enum.take_random([1, 2, 3], 1) == [2]
+    assert Enum.take_random([1, 2, 3], 2) == [2, 3]
     assert Enum.take_random([1, 2, 3], 3) == [3, 1, 2]
-    assert Enum.take_random([1, 2, 3], 4) == [1, 3, 2]
+    assert Enum.take_random([1, 2, 3], 4) == [2, 3, 1]
     :rand.seed(:exsss, seed2)
     assert Enum.take_random([1, 2, 3], 1) == [1]
     assert Enum.take_random([1, 2, 3], 2) == [3, 1]
-    assert Enum.take_random([1, 2, 3], 3) == [3, 1, 2]
-    assert Enum.take_random([1, 2, 3], 4) == [2, 1, 3]
-    assert Enum.take_random([1, 2, 3], 129) == [2, 3, 1]
+    assert Enum.take_random([1, 2, 3], 3) == [2, 3, 1]
+    assert Enum.take_random([1, 2, 3], 4) == [3, 2, 1]
+    assert Enum.take_random([1, 2, 3], 129) == [2, 1, 3]
 
     # assert that every item in the sample comes from the input list
     list = for _ <- 1..100, do: make_ref()
@@ -1420,7 +1483,7 @@ defmodule EnumTest do
   test "unzip/1" do
     assert Enum.unzip([{:a, 1}, {:b, 2}, {:c, 3}]) == {[:a, :b, :c], [1, 2, 3]}
     assert Enum.unzip([]) == {[], []}
-    assert Enum.unzip(%{a: 1, b: 2}) == {[:a, :b], [1, 2]}
+    assert Enum.unzip(%{a: 1}) == {[:a], [1]}
     assert Enum.unzip(foo: "a", bar: "b") == {[:foo, :bar], ["a", "b"]}
 
     assert_raise FunctionClauseError, fn -> Enum.unzip([{:a, 1}, {:b, 2, "foo"}]) end
@@ -1503,14 +1566,11 @@ defmodule EnumTest do
     assert result == [2, 4, 6, 8, 10]
 
     # Maps
-    result = Enum.zip_with(%{a: 7, c: 9}, 3..4, fn {key, value}, b -> {key, value + b} end)
-    assert result == [a: 10, c: 13]
+    result = Enum.zip_with(%{a: 7}, 3..4, fn {key, value}, b -> {key, value + b} end)
+    assert result == [a: 10]
 
-    colour_1 = %{r: 176, g: 175, b: 255}
-    colour_2 = %{r: 12, g: 176, b: 176}
-
-    result = Enum.zip_with(colour_1, colour_2, fn {k, left}, {k, right} -> {k, left + right} end)
-    assert result == [b: 431, g: 351, r: 188]
+    result = Enum.zip_with(3..4, %{a: 7}, fn a, {key, value} -> {key, value + a} end)
+    assert result == [a: 10]
   end
 
   test "zip_with/2" do
@@ -1518,8 +1578,10 @@ defmodule EnumTest do
     result = Enum.zip_with([[:a, :b], [1, 2], ["foo", "bar"]], zip_fun)
     assert result == [{:a, 1, "foo"}, {:b, 2, "bar"}]
 
-    lots = Enum.zip_with([[:a, :b], [1, 2], ["foo", "bar"], %{a: :b, c: :d}], zip_fun)
-    assert lots == [{:a, 1, "foo", {:a, :b}}, {:b, 2, "bar", {:c, :d}}]
+    map = %{a: :b, c: :d}
+    [x1, x2] = Map.to_list(map)
+    lots = Enum.zip_with([[:a, :b], [1, 2], ["foo", "bar"], map], zip_fun)
+    assert lots == [{:a, 1, "foo", x1}, {:b, 2, "bar", x2}]
 
     assert Enum.zip_with([[:a, :b], [1, 2, 3, 4], ["foo", "bar", "baz", "qux"]], zip_fun) ==
              [{:a, 1, "foo"}, {:b, 2, "bar"}]
@@ -1556,21 +1618,11 @@ defmodule EnumTest do
     assert result == [2, 4, 6, 8, 10]
 
     # Maps
-    result = Enum.zip_with([%{a: 7, c: 9}, 3..4], fn [{key, value}, b] -> {key, value + b} end)
-    assert result == [a: 10, c: 13]
+    result = Enum.zip_with([%{a: 7}, 3..4], fn [{key, value}, b] -> {key, value + b} end)
+    assert result == [a: 10]
 
-    colour_1 = %{r: 176, g: 175, b: 255}
-    colour_2 = %{r: 12, g: 176, b: 176}
-
-    result =
-      Enum.zip_with([colour_1, colour_2], fn [{k, left}, {k, right}] -> {k, left + right} end)
-
-    assert result == [b: 431, g: 351, r: 188]
-
-    assert Enum.zip_with([%{a: :b, c: :d}, %{e: :f, g: :h}], & &1) == [
-             [a: :b, e: :f],
-             [c: :d, g: :h]
-           ]
+    result = Enum.zip_with([%{a: 7}, 3..4], fn [{key, value}, b] -> {key, value + b} end)
+    assert result == [a: 10]
   end
 end
 
@@ -1580,7 +1632,7 @@ defmodule EnumTest.Range do
 
   test "all?/2" do
     assert Enum.all?(0..1)
-    assert Enum.all?(1..0)
+    assert Enum.all?(1..0//-1)
     refute Enum.all?(0..5, fn x -> rem(x, 2) == 0 end)
     assert Enum.all?(0..1, fn x -> x < 2 end)
 
@@ -1590,7 +1642,7 @@ defmodule EnumTest.Range do
   end
 
   test "any?/2" do
-    assert Enum.any?(1..0)
+    assert Enum.any?(1..0//-1)
     refute Enum.any?(0..5, &(&1 > 10))
     assert Enum.any?(0..5, &(&1 > 3))
 
@@ -1692,7 +1744,7 @@ defmodule EnumTest.Range do
     assert Enum.drop(1..3, -1) == [1, 2]
     assert Enum.drop(1..3, -2) == [1]
     assert Enum.drop(1..3, -4) == []
-    assert Enum.drop(1..0, 3) == []
+    assert Enum.drop(1..0//-1, 3) == []
 
     assert Enum.drop(1..9//2, 2) == [5, 7, 9]
     assert Enum.drop(1..9//2, -2) == [1, 3, 5]
@@ -1721,12 +1773,12 @@ defmodule EnumTest.Range do
     assert Enum.drop_while(0..6, fn x -> x <= 3 end) == [4, 5, 6]
     assert Enum.drop_while(0..6, fn _ -> false end) == [0, 1, 2, 3, 4, 5, 6]
     assert Enum.drop_while(0..3, fn x -> x <= 3 end) == []
-    assert Enum.drop_while(1..0, fn _ -> nil end) == [1, 0]
+    assert Enum.drop_while(1..0//-1, fn _ -> nil end) == [1, 0]
   end
 
   test "each/2" do
     try do
-      assert Enum.each(1..0, fn x -> x end) == :ok
+      assert Enum.each(1..0//-1, fn x -> x end) == :ok
       assert Enum.each(1..3, fn x -> Process.put(:enum_test_each, x * 2) end) == :ok
       assert Process.get(:enum_test_each) == 6
     after
@@ -1734,7 +1786,7 @@ defmodule EnumTest.Range do
     end
 
     try do
-      assert Enum.each(-1..-3, fn x -> Process.put(:enum_test_each, x * 2) end) == :ok
+      assert Enum.each(-1..-3//-1, fn x -> Process.put(:enum_test_each, x * 2) end) == :ok
       assert Process.get(:enum_test_each) == -6
     after
       Process.delete(:enum_test_each)
@@ -1742,7 +1794,7 @@ defmodule EnumTest.Range do
   end
 
   test "empty?/1" do
-    refute Enum.empty?(1..0)
+    refute Enum.empty?(1..0//-1)
     refute Enum.empty?(1..2)
     refute Enum.empty?(1..2//2)
     assert Enum.empty?(1..2//-2)
@@ -1763,17 +1815,17 @@ defmodule EnumTest.Range do
     assert Enum.fetch(-10..20, -32) == :error
 
     # descending order
-    assert Enum.fetch(20..-10, 4) == {:ok, 16}
-    assert Enum.fetch(20..-10, -4) == {:ok, -7}
+    assert Enum.fetch(20..-10//-1, 4) == {:ok, 16}
+    assert Enum.fetch(20..-10//-1, -4) == {:ok, -7}
     # descending order, first
-    assert Enum.fetch(20..-10, 0) == {:ok, 20}
-    assert Enum.fetch(20..-10, -31) == {:ok, 20}
+    assert Enum.fetch(20..-10//-1, 0) == {:ok, 20}
+    assert Enum.fetch(20..-10//-1, -31) == {:ok, 20}
     # descending order, last
-    assert Enum.fetch(20..-10, -1) == {:ok, -10}
-    assert Enum.fetch(20..-10, 30) == {:ok, -10}
+    assert Enum.fetch(20..-10//-1, -1) == {:ok, -10}
+    assert Enum.fetch(20..-10//-1, 30) == {:ok, -10}
     # descending order, out of bound
-    assert Enum.fetch(20..-10, 31) == :error
-    assert Enum.fetch(20..-10, -32) == :error
+    assert Enum.fetch(20..-10//-1, 31) == :error
+    assert Enum.fetch(20..-10//-1, -32) == :error
 
     # edge cases
     assert Enum.fetch(42..42, 0) == {:ok, 42}
@@ -1792,15 +1844,15 @@ defmodule EnumTest.Range do
     assert Enum.fetch!(2..6, 4) == 6
     assert Enum.fetch!(2..6, -1) == 6
     assert Enum.fetch!(2..6, -2) == 5
-    assert Enum.fetch!(-2..-6, 0) == -2
-    assert Enum.fetch!(-2..-6, 4) == -6
+    assert Enum.fetch!(-2..-6//-1, 0) == -2
+    assert Enum.fetch!(-2..-6//-1, 4) == -6
 
     assert_raise Enum.OutOfBoundsError, fn ->
       Enum.fetch!(2..6, 8)
     end
 
     assert_raise Enum.OutOfBoundsError, fn ->
-      Enum.fetch!(-2..-6, 8)
+      Enum.fetch!(-2..-6//-1, 8)
     end
 
     assert_raise Enum.OutOfBoundsError, fn ->
@@ -1850,7 +1902,7 @@ defmodule EnumTest.Range do
   end
 
   test "intersperse/2" do
-    assert Enum.intersperse(1..0, true) == [1, true, 0]
+    assert Enum.intersperse(1..0//-1, true) == [1, true, 0]
     assert Enum.intersperse(1..3, false) == [1, false, 2, false, 3]
   end
 
@@ -1864,20 +1916,23 @@ defmodule EnumTest.Range do
   end
 
   test "join/2" do
-    assert Enum.join(1..0, " = ") == "1 = 0"
+    assert Enum.join(1..0//-1, " = ") == "1 = 0"
     assert Enum.join(1..3, " = ") == "1 = 2 = 3"
     assert Enum.join(1..3) == "123"
   end
 
   test "map/2" do
     assert Enum.map(1..3, fn x -> x * 2 end) == [2, 4, 6]
-    assert Enum.map(-1..-3, fn x -> x * 2 end) == [-2, -4, -6]
+    assert Enum.map(-1..-3//-1, fn x -> x * 2 end) == [-2, -4, -6]
+    assert Enum.map(1..10//2, fn x -> x * 2 end) == [2, 6, 10, 14, 18]
+    assert Enum.map(3..1//-2, fn x -> x * 2 end) == [6, 2]
+    assert Enum.map(0..1//-1, fn x -> x * 2 end) == []
   end
 
   test "map_every/3" do
     assert Enum.map_every(1..10, 2, fn x -> x * 2 end) == [2, 2, 6, 4, 10, 6, 14, 8, 18, 10]
 
-    assert Enum.map_every(-1..-10, 2, fn x -> x * 2 end) ==
+    assert Enum.map_every(-1..-10//-1, 2, fn x -> x * 2 end) ==
              [-2, -2, -6, -4, -10, -6, -14, -8, -18, -10]
 
     assert Enum.map_every(1..2, 2, fn x -> x * 2 end) == [2, 2]
@@ -1894,20 +1949,20 @@ defmodule EnumTest.Range do
   end
 
   test "map_join/3" do
-    assert Enum.map_join(1..0, " = ", &(&1 * 2)) == "2 = 0"
+    assert Enum.map_join(1..0//-1, " = ", &(&1 * 2)) == "2 = 0"
     assert Enum.map_join(1..3, " = ", &(&1 * 2)) == "2 = 4 = 6"
     assert Enum.map_join(1..3, &(&1 * 2)) == "246"
   end
 
   test "map_reduce/3" do
-    assert Enum.map_reduce(1..0, 1, fn x, acc -> {x * 2, x + acc} end) == {[2, 0], 2}
+    assert Enum.map_reduce(1..0//-1, 1, fn x, acc -> {x * 2, x + acc} end) == {[2, 0], 2}
     assert Enum.map_reduce(1..3, 1, fn x, acc -> {x * 2, x + acc} end) == {[2, 4, 6], 7}
   end
 
   test "max/1" do
     assert Enum.max(1..1) == 1
     assert Enum.max(1..3) == 3
-    assert Enum.max(3..1) == 3
+    assert Enum.max(3..1//-1) == 3
 
     assert Enum.max(1..9//2) == 9
     assert Enum.max(1..10//2) == 9
@@ -1965,7 +2020,7 @@ defmodule EnumTest.Range do
   test "min_max/1" do
     assert Enum.min_max(1..1) == {1, 1}
     assert Enum.min_max(1..3) == {1, 3}
-    assert Enum.min_max(3..1) == {1, 3}
+    assert Enum.min_max(3..1//-1) == {1, 3}
 
     assert Enum.min_max(1..9//2) == {1, 9}
     assert Enum.min_max(1..10//2) == {1, 9}
@@ -1997,7 +2052,7 @@ defmodule EnumTest.Range do
     :rand.seed(:exsss, seed1)
     assert Enum.random(1..2) == 1
     assert Enum.random(1..3) == 1
-    assert Enum.random(3..1) == 2
+    assert Enum.random(3..1//-1) == 2
 
     :rand.seed(:exsss, seed2)
     assert Enum.random(1..2) == 1
@@ -2014,7 +2069,7 @@ defmodule EnumTest.Range do
   end
 
   test "reduce/3" do
-    assert Enum.reduce(1..0, 1, fn x, acc -> x + acc end) == 2
+    assert Enum.reduce(1..0//-1, 1, fn x, acc -> x + acc end) == 2
     assert Enum.reduce(1..3, 1, fn x, acc -> x + acc end) == 7
     assert Enum.reduce(1..10//2, 1, fn x, acc -> x + acc end) == 26
     assert Enum.reduce(0..1//-1, 1, fn x, acc -> x + acc end) == 1
@@ -2069,8 +2124,8 @@ defmodule EnumTest.Range do
   test "shuffle/1" do
     # set a fixed seed so the test can be deterministic
     :rand.seed(:exsss, {1374, 347_975, 449_264})
-    assert Enum.shuffle(1..5) == [1, 3, 4, 5, 2]
-    assert Enum.shuffle(1..10//2) == [3, 9, 7, 1, 5]
+    assert Enum.shuffle(1..5) == [2, 5, 4, 3, 1]
+    assert Enum.shuffle(1..10//2) == [5, 1, 7, 9, 3]
   end
 
   test "slice/2" do
@@ -2078,13 +2133,13 @@ defmodule EnumTest.Range do
     assert Enum.slice(1..5, 0..1) == [1, 2]
     assert Enum.slice(1..5, 0..2) == [1, 2, 3]
     assert Enum.slice(1..5, 1..2) == [2, 3]
-    assert Enum.slice(1..5, 1..0) == []
+    assert Enum.slice(1..5, 1..0//1) == []
     assert Enum.slice(1..5, 2..5) == [3, 4, 5]
     assert Enum.slice(1..5, 2..6) == [3, 4, 5]
     assert Enum.slice(1..5, 4..4) == [5]
     assert Enum.slice(1..5, 5..5) == []
-    assert Enum.slice(1..5, 6..5) == []
-    assert Enum.slice(1..5, 6..0) == []
+    assert Enum.slice(1..5, 6..5//1) == []
+    assert Enum.slice(1..5, 6..0//1) == []
     assert Enum.slice(1..5, -3..0) == []
     assert Enum.slice(1..5, -3..1) == []
     assert Enum.slice(1..5, -6..0) == [1]
@@ -2118,34 +2173,34 @@ defmodule EnumTest.Range do
     assert Enum.slice(1..5, -4..-1//4) == [2]
     assert Enum.slice(1..5, -4..-1//5) == [2]
 
-    assert Enum.slice(5..1, 0..0) == [5]
-    assert Enum.slice(5..1, 0..1) == [5, 4]
-    assert Enum.slice(5..1, 0..2) == [5, 4, 3]
-    assert Enum.slice(5..1, 1..2) == [4, 3]
-    assert Enum.slice(5..1, 1..0) == []
-    assert Enum.slice(5..1, 2..5) == [3, 2, 1]
-    assert Enum.slice(5..1, 2..6) == [3, 2, 1]
-    assert Enum.slice(5..1, 4..4) == [1]
-    assert Enum.slice(5..1, 5..5) == []
-    assert Enum.slice(5..1, 6..5) == []
-    assert Enum.slice(5..1, 6..0) == []
-    assert Enum.slice(5..1, -6..0) == [5]
-    assert Enum.slice(5..1, -6..5) == [5, 4, 3, 2, 1]
-    assert Enum.slice(5..1, -6..-1) == [5, 4, 3, 2, 1]
-    assert Enum.slice(5..1, -5..-1) == [5, 4, 3, 2, 1]
-    assert Enum.slice(5..1, -5..-3) == [5, 4, 3]
+    assert Enum.slice(5..1//-1, 0..0) == [5]
+    assert Enum.slice(5..1//-1, 0..1) == [5, 4]
+    assert Enum.slice(5..1//-1, 0..2) == [5, 4, 3]
+    assert Enum.slice(5..1//-1, 1..2) == [4, 3]
+    assert Enum.slice(5..1//-1, 1..0//1) == []
+    assert Enum.slice(5..1//-1, 2..5) == [3, 2, 1]
+    assert Enum.slice(5..1//-1, 2..6) == [3, 2, 1]
+    assert Enum.slice(5..1//-1, 4..4) == [1]
+    assert Enum.slice(5..1//-1, 5..5) == []
+    assert Enum.slice(5..1//-1, 6..5//1) == []
+    assert Enum.slice(5..1//-1, 6..0//1) == []
+    assert Enum.slice(5..1//-1, -6..0) == [5]
+    assert Enum.slice(5..1//-1, -6..5) == [5, 4, 3, 2, 1]
+    assert Enum.slice(5..1//-1, -6..-1) == [5, 4, 3, 2, 1]
+    assert Enum.slice(5..1//-1, -5..-1) == [5, 4, 3, 2, 1]
+    assert Enum.slice(5..1//-1, -5..-3) == [5, 4, 3]
 
     assert Enum.slice(1..10//2, 0..0) == [1]
     assert Enum.slice(1..10//2, 0..1) == [1, 3]
     assert Enum.slice(1..10//2, 0..2) == [1, 3, 5]
     assert Enum.slice(1..10//2, 1..2) == [3, 5]
-    assert Enum.slice(1..10//2, 1..0) == []
+    assert Enum.slice(1..10//2, 1..0//1) == []
     assert Enum.slice(1..10//2, 2..5) == [5, 7, 9]
     assert Enum.slice(1..10//2, 2..6) == [5, 7, 9]
     assert Enum.slice(1..10//2, 4..4) == [9]
     assert Enum.slice(1..10//2, 5..5) == []
-    assert Enum.slice(1..10//2, 6..5) == []
-    assert Enum.slice(1..10//2, 6..0) == []
+    assert Enum.slice(1..10//2, 6..5//1) == []
+    assert Enum.slice(1..10//2, 6..0//1) == []
     assert Enum.slice(1..10//2, -3..0) == []
     assert Enum.slice(1..10//2, -3..1) == []
     assert Enum.slice(1..10//2, -6..0) == [1]
@@ -2187,19 +2242,19 @@ defmodule EnumTest.Range do
       Enum.slice(1..5, 0, 0.99)
     end
 
-    assert Enum.slice(5..1, 0, 0) == []
-    assert Enum.slice(5..1, 0, 1) == [5]
-    assert Enum.slice(5..1, 0, 2) == [5, 4]
-    assert Enum.slice(5..1, 1, 2) == [4, 3]
-    assert Enum.slice(5..1, 1, 0) == []
-    assert Enum.slice(5..1, 2, 3) == [3, 2, 1]
-    assert Enum.slice(5..1, 2, 6) == [3, 2, 1]
-    assert Enum.slice(5..1, 4, 4) == [1]
-    assert Enum.slice(5..1, 5, 5) == []
-    assert Enum.slice(5..1, 6, 5) == []
-    assert Enum.slice(5..1, 6, 0) == []
-    assert Enum.slice(5..1, -6, 0) == []
-    assert Enum.slice(5..1, -6, 5) == [5, 4, 3, 2, 1]
+    assert Enum.slice(5..1//-1, 0, 0) == []
+    assert Enum.slice(5..1//-1, 0, 1) == [5]
+    assert Enum.slice(5..1//-1, 0, 2) == [5, 4]
+    assert Enum.slice(5..1//-1, 1, 2) == [4, 3]
+    assert Enum.slice(5..1//-1, 1, 0) == []
+    assert Enum.slice(5..1//-1, 2, 3) == [3, 2, 1]
+    assert Enum.slice(5..1//-1, 2, 6) == [3, 2, 1]
+    assert Enum.slice(5..1//-1, 4, 4) == [1]
+    assert Enum.slice(5..1//-1, 5, 5) == []
+    assert Enum.slice(5..1//-1, 6, 5) == []
+    assert Enum.slice(5..1//-1, 6, 0) == []
+    assert Enum.slice(5..1//-1, -6, 0) == []
+    assert Enum.slice(5..1//-1, -6, 5) == [5, 4, 3, 2, 1]
 
     assert Enum.slice(1..10//2, 0, 0) == []
     assert Enum.slice(1..10//2, 0, 1) == [1]
@@ -2218,24 +2273,24 @@ defmodule EnumTest.Range do
   end
 
   test "sort/1" do
-    assert Enum.sort(3..1) == [1, 2, 3]
-    assert Enum.sort(2..1) == [1, 2]
+    assert Enum.sort(3..1//-1) == [1, 2, 3]
+    assert Enum.sort(2..1//-1) == [1, 2]
     assert Enum.sort(1..1) == [1]
   end
 
   test "sort/2" do
-    assert Enum.sort(3..1, &(&1 > &2)) == [3, 2, 1]
-    assert Enum.sort(2..1, &(&1 > &2)) == [2, 1]
+    assert Enum.sort(3..1//-1, &(&1 > &2)) == [3, 2, 1]
+    assert Enum.sort(2..1//-1, &(&1 > &2)) == [2, 1]
     assert Enum.sort(1..1, &(&1 > &2)) == [1]
 
-    assert Enum.sort(3..1, :asc) == [1, 2, 3]
-    assert Enum.sort(3..1, :desc) == [3, 2, 1]
+    assert Enum.sort(3..1//-1, :asc) == [1, 2, 3]
+    assert Enum.sort(3..1//-1, :desc) == [3, 2, 1]
   end
 
   test "sort_by/2" do
-    assert Enum.sort_by(3..1, & &1) == [1, 2, 3]
-    assert Enum.sort_by(3..1, & &1, :asc) == [1, 2, 3]
-    assert Enum.sort_by(3..1, & &1, :desc) == [3, 2, 1]
+    assert Enum.sort_by(3..1//-1, & &1) == [1, 2, 3]
+    assert Enum.sort_by(3..1//-1, & &1, :asc) == [1, 2, 3]
+    assert Enum.sort_by(3..1//-1, & &1, :desc) == [3, 2, 1]
   end
 
   test "split/2" do
@@ -2248,7 +2303,7 @@ defmodule EnumTest.Range do
     assert Enum.split(1..3, -2) == {[1], [2, 3]}
     assert Enum.split(1..3, -3) == {[], [1, 2, 3]}
     assert Enum.split(1..3, -10) == {[], [1, 2, 3]}
-    assert Enum.split(1..0, 3) == {[1, 0], []}
+    assert Enum.split(1..0//-1, 3) == {[1, 0], []}
   end
 
   test "split_while/2" do
@@ -2259,7 +2314,7 @@ defmodule EnumTest.Range do
     assert Enum.split_while(1..3, fn x -> x > 3 end) == {[], [1, 2, 3]}
     assert Enum.split_while(1..3, fn x -> x < 3 end) == {[1, 2], [3]}
     assert Enum.split_while(1..3, fn x -> x end) == {[1, 2, 3], []}
-    assert Enum.split_while(1..0, fn _ -> true end) == {[1, 0], []}
+    assert Enum.split_while(1..0//-1, fn _ -> true end) == {[1, 0], []}
   end
 
   test "sum/1" do
@@ -2268,8 +2323,8 @@ defmodule EnumTest.Range do
     assert Enum.sum(1..3) == 6
     assert Enum.sum(0..100) == 5050
     assert Enum.sum(10..100) == 5005
-    assert Enum.sum(100..10) == 5005
-    assert Enum.sum(-10..-20) == -165
+    assert Enum.sum(100..10//-1) == 5005
+    assert Enum.sum(-10..-20//-1) == -165
     assert Enum.sum(-10..2) == -52
 
     assert Enum.sum(0..1//-1) == 0
@@ -2287,7 +2342,7 @@ defmodule EnumTest.Range do
     assert Enum.take(1..3, -1) == [3]
     assert Enum.take(1..3, -2) == [2, 3]
     assert Enum.take(1..3, -4) == [1, 2, 3]
-    assert Enum.take(1..0, 3) == [1, 0]
+    assert Enum.take(1..0//-1, 3) == [1, 0]
     assert Enum.take(1..0//1, -3) == []
   end
 
@@ -2314,7 +2369,7 @@ defmodule EnumTest.Range do
     seed1 = {1406, 407_414, 139_258}
     seed2 = {1406, 421_106, 567_597}
     :rand.seed(:exsss, seed1)
-    assert Enum.take_random(1..3, 1) == [3]
+    assert Enum.take_random(1..3, 1) == [2]
     :rand.seed(:exsss, seed1)
     assert Enum.take_random(1..3, 2) == [3, 1]
     :rand.seed(:exsss, seed1)
@@ -2322,22 +2377,19 @@ defmodule EnumTest.Range do
     :rand.seed(:exsss, seed1)
     assert Enum.take_random(1..3, 4) == [3, 1, 2]
     :rand.seed(:exsss, seed1)
-    assert Enum.take_random(3..1, 1) == [1]
+    assert Enum.take_random(1..3, 5) == [3, 1, 2]
+    :rand.seed(:exsss, seed1)
+    assert Enum.take_random(3..1//-1, 1) == [2]
     :rand.seed(:exsss, seed2)
     assert Enum.take_random(1..3, 1) == [1]
     :rand.seed(:exsss, seed2)
-    assert Enum.take_random(1..3, 2) == [1, 3]
+    assert Enum.take_random(1..3, 2) == [3, 2]
     :rand.seed(:exsss, seed2)
     assert Enum.take_random(1..3, 3) == [1, 3, 2]
     :rand.seed(:exsss, seed2)
     assert Enum.take_random(1..3, 4) == [1, 3, 2]
-
-    # make sure optimizations don't change fixed seeded tests
-    :rand.seed(:exsss, {101, 102, 103})
-    one = Enum.take_random(1..100, 1)
-    :rand.seed(:exsss, {101, 102, 103})
-    two = Enum.take_random(1..100, 2)
-    assert hd(one) == hd(two)
+    :rand.seed(:exsss, seed2)
+    assert Enum.take_random(1..3, 5) == [1, 3, 2]
   end
 
   test "take_while/2" do
@@ -2396,16 +2448,17 @@ defmodule EnumTest.Map do
 
   test "random/1" do
     map = %{a: 1, b: 2, c: 3}
+    [x1, x2, x3] = Map.to_list(map)
     seed1 = {1406, 407_414, 139_258}
     seed2 = {1406, 421_106, 567_597}
     :rand.seed(:exsss, seed1)
-    assert Enum.random(map) == {:c, 3}
-    assert Enum.random(map) == {:a, 1}
-    assert Enum.random(map) == {:b, 2}
+    assert Enum.random(map) == x3
+    assert Enum.random(map) == x1
+    assert Enum.random(map) == x2
 
     :rand.seed(:exsss, seed2)
-    assert Enum.random(map) == {:c, 3}
-    assert Enum.random(map) == {:b, 2}
+    assert Enum.random(map) == x3
+    assert Enum.random(map) == x2
   end
 
   test "take_random/2" do
@@ -2418,35 +2471,37 @@ defmodule EnumTest.Map do
     # set a fixed seed so the test can be deterministic
     # please note the order of following assertions is important
     map = %{a: 1, b: 2, c: 3}
+    [x1, x2, x3] = Map.to_list(map)
     seed1 = {1406, 407_414, 139_258}
     seed2 = {1406, 421_106, 567_597}
     :rand.seed(:exsss, seed1)
-    assert Enum.take_random(map, 1) == [c: 3]
+    assert Enum.take_random(map, 1) == [x2]
     :rand.seed(:exsss, seed1)
-    assert Enum.take_random(map, 2) == [c: 3, a: 1]
+    assert Enum.take_random(map, 2) == [x3, x1]
     :rand.seed(:exsss, seed1)
-    assert Enum.take_random(map, 3) == [c: 3, a: 1, b: 2]
+    assert Enum.take_random(map, 3) == [x3, x1, x2]
     :rand.seed(:exsss, seed1)
-    assert Enum.take_random(map, 4) == [c: 3, a: 1, b: 2]
+    assert Enum.take_random(map, 4) == [x3, x1, x2]
     :rand.seed(:exsss, seed2)
-    assert Enum.take_random(map, 1) == [a: 1]
+    assert Enum.take_random(map, 1) == [x1]
     :rand.seed(:exsss, seed2)
-    assert Enum.take_random(map, 2) == [a: 1, c: 3]
+    assert Enum.take_random(map, 2) == [x3, x2]
     :rand.seed(:exsss, seed2)
-    assert Enum.take_random(map, 3) == [a: 1, c: 3, b: 2]
+    assert Enum.take_random(map, 3) == [x1, x3, x2]
     :rand.seed(:exsss, seed2)
-    assert Enum.take_random(map, 4) == [a: 1, c: 3, b: 2]
+    assert Enum.take_random(map, 4) == [x1, x3, x2]
   end
 
   test "reverse/1" do
     assert Enum.reverse(%{}) == []
     assert Enum.reverse(MapSet.new()) == []
-    assert Enum.reverse(%{a: 1, b: 2, c: 3}) == [c: 3, b: 2, a: 1]
+
+    map = %{a: 1, b: 2, c: 3}
+    assert Enum.reverse(map) == Map.to_list(map) |> Enum.reverse()
   end
 
   test "reverse/2" do
-    assert Enum.reverse([a: 1, b: 2, c: 3, a: 1], %{x: 1, y: 2, z: 3}) ==
-             [a: 1, c: 3, b: 2, a: 1, x: 1, y: 2, z: 3]
+    assert Enum.reverse([a: 1, b: 2, c: 3, a: 1], %{x: 1}) == [a: 1, c: 3, b: 2, a: 1, x: 1]
 
     assert Enum.reverse([], %{a: 1}) == [a: 1]
     assert Enum.reverse([], %{}) == []
@@ -2456,44 +2511,49 @@ defmodule EnumTest.Map do
 
   test "fetch/2" do
     map = %{a: 1, b: 2, c: 3, d: 4, e: 5}
-    assert Enum.fetch(map, 0) == {:ok, {:a, 1}}
-    assert Enum.fetch(map, -2) == {:ok, {:d, 4}}
+    [x1, _x2, _x3, x4, x5] = Map.to_list(map)
+    assert Enum.fetch(map, 0) == {:ok, x1}
+    assert Enum.fetch(map, -2) == {:ok, x4}
     assert Enum.fetch(map, -6) == :error
     assert Enum.fetch(map, 5) == :error
     assert Enum.fetch(%{}, 0) == :error
 
     assert Stream.take(map, 3) |> Enum.fetch(3) == :error
-    assert Stream.take(map, 5) |> Enum.fetch(4) == {:ok, {:e, 5}}
+    assert Stream.take(map, 5) |> Enum.fetch(4) == {:ok, x5}
   end
 
   test "map_intersperse/3" do
     assert Enum.map_intersperse(%{}, :a, & &1) == []
     assert Enum.map_intersperse(%{foo: :bar}, :a, & &1) == [{:foo, :bar}]
 
-    assert Enum.map_intersperse(%{foo: :bar, baz: :bat}, :a, & &1) ==
-             [{:baz, :bat}, :a, {:foo, :bar}]
+    map = %{foo: :bar, baz: :bat}
+    [x1, x2] = Map.to_list(map)
+
+    assert Enum.map_intersperse(map, :a, & &1) == [x1, :a, x2]
   end
 
   test "slice/2" do
     map = %{a: 1, b: 2, c: 3, d: 4, e: 5}
-    assert Enum.slice(map, 0..0) == [a: 1]
-    assert Enum.slice(map, 0..1) == [a: 1, b: 2]
-    assert Enum.slice(map, 0..2) == [a: 1, b: 2, c: 3]
+    [x1, x2, x3 | _] = Map.to_list(map)
+    assert Enum.slice(map, 0..0) == [x1]
+    assert Enum.slice(map, 0..1) == [x1, x2]
+    assert Enum.slice(map, 0..2) == [x1, x2, x3]
   end
 
   test "slice/3" do
     map = %{a: 1, b: 2, c: 3, d: 4, e: 5}
-    assert Enum.slice(map, 1, 2) == [b: 2, c: 3]
+    [x1, x2, x3, x4, x5] = Map.to_list(map)
+    assert Enum.slice(map, 1, 2) == [x2, x3]
     assert Enum.slice(map, 1, 0) == []
-    assert Enum.slice(map, 2, 5) == [c: 3, d: 4, e: 5]
-    assert Enum.slice(map, 2, 6) == [c: 3, d: 4, e: 5]
+    assert Enum.slice(map, 2, 5) == [x3, x4, x5]
+    assert Enum.slice(map, 2, 6) == [x3, x4, x5]
     assert Enum.slice(map, 5, 5) == []
     assert Enum.slice(map, 6, 5) == []
     assert Enum.slice(map, 6, 0) == []
     assert Enum.slice(map, -6, 0) == []
-    assert Enum.slice(map, -6, 5) == [a: 1, b: 2, c: 3, d: 4, e: 5]
-    assert Enum.slice(map, -2, 5) == [d: 4, e: 5]
-    assert Enum.slice(map, -3, 1) == [c: 3]
+    assert Enum.slice(map, -6, 5) == [x1, x2, x3, x4, x5]
+    assert Enum.slice(map, -2, 5) == [x4, x5]
+    assert Enum.slice(map, -3, 1) == [x3]
 
     assert_raise FunctionClauseError, fn ->
       Enum.slice(map, 0, -1)
@@ -2508,19 +2568,19 @@ defmodule EnumTest.Map do
     end
 
     assert Enum.slice(map, 0, 0) == []
-    assert Enum.slice(map, 0, 1) == [a: 1]
-    assert Enum.slice(map, 0, 2) == [a: 1, b: 2]
-    assert Enum.slice(map, 1, 2) == [b: 2, c: 3]
+    assert Enum.slice(map, 0, 1) == [x1]
+    assert Enum.slice(map, 0, 2) == [x1, x2]
+    assert Enum.slice(map, 1, 2) == [x2, x3]
     assert Enum.slice(map, 1, 0) == []
-    assert Enum.slice(map, 2, 5) == [c: 3, d: 4, e: 5]
-    assert Enum.slice(map, 2, 6) == [c: 3, d: 4, e: 5]
+    assert Enum.slice(map, 2, 5) == [x3, x4, x5]
+    assert Enum.slice(map, 2, 6) == [x3, x4, x5]
     assert Enum.slice(map, 5, 5) == []
     assert Enum.slice(map, 6, 5) == []
     assert Enum.slice(map, 6, 0) == []
     assert Enum.slice(map, -6, 0) == []
-    assert Enum.slice(map, -6, 5) == [a: 1, b: 2, c: 3, d: 4, e: 5]
-    assert Enum.slice(map, -2, 5) == [d: 4, e: 5]
-    assert Enum.slice(map, -3, 1) == [c: 3]
+    assert Enum.slice(map, -6, 5) == [x1, x2, x3, x4, x5]
+    assert Enum.slice(map, -2, 5) == [x4, x5]
+    assert Enum.slice(map, -3, 1) == [x3]
 
     assert_raise FunctionClauseError, fn ->
       Enum.slice(map, 0, -1)
