@@ -83,9 +83,14 @@ defmodule Kernel.BinaryTest do
     assert x == "foo"
 
     size = 3
-    <<x::binary-size(size)>> <> _ = "foobar"
+    <<x::binary-size(^size)>> <> _ = "foobar"
     assert x == "foo"
 
+    size = 3
+    <<size, x::binary-size(^size)>> <> _ = <<10, "foobar">>
+    assert x == "foo"
+    assert size == 10
+
     <<x::6*4-binary>> <> _ = "foobar"
     assert x == "foo"
 
@@ -114,7 +119,7 @@ defmodule Kernel.BinaryTest do
       Code.eval_string(~s["foo" <> 1])
     end
 
-    message = ~r"left argument of <> operator inside a match"
+    message = ~r"cannot perform prefix match because the left operand of <> has unknown size."
 
     assert_raise ArgumentError, message, fn ->
       Code.eval_string(~s[a <> "b" = "ab"])
@@ -123,20 +128,6 @@ defmodule Kernel.BinaryTest do
     assert_raise ArgumentError, message, fn ->
       Code.eval_string(~s["a" <> b <> "c" = "abc"])
     end
-
-    assert_raise ArgumentError, message, fn ->
-      Code.eval_string(~s[
-        a = "a"
-        ^a <> "b" = "ab"
-      ])
-    end
-
-    assert_raise ArgumentError, message, fn ->
-      Code.eval_string(~s[
-        b = "b"
-        "a" <> ^b <> "c" = "abc"
-      ])
-    end
   end
 
   test "hex" do
@@ -164,7 +155,7 @@ defmodule Kernel.BinaryTest do
 
   test "pattern match" do
     s = 16
-    assert <<_a, _b::size(s)>> = "foo"
+    assert <<_a, _b::size(^s)>> = "foo"
   end
 
   test "pattern match with splice" do
@@ -192,21 +183,15 @@ defmodule Kernel.BinaryTest do
   end
 
   test "literal errors" do
-    message = ~r"conflicting type specification for bit field"
+    message = "conflicting type specification for bit field"
 
-    assert_raise CompileError, message, fn ->
+    assert_compile_error(message, fn ->
       Code.eval_string(~s[<<"foo"::integer>>])
-    end
+    end)
 
-    assert_raise CompileError, message, fn ->
+    assert_compile_error(message, fn ->
       Code.eval_string(~s[<<"foo"::float>>])
-    end
-
-    message = ~r"invalid literal 'foo'"
-
-    assert_raise CompileError, message, fn ->
-      Code.eval_string(~s[<<'foo'::binary>>])
-    end
+    end)
   end
 
   @bitstring <<"foo", 16::4>>
@@ -240,8 +225,8 @@ defmodule Kernel.BinaryTest do
 
   test "bitsyntax variable size" do
     x = 8
-    assert <<_, _::size(x)>> = <<?a, ?b>>
-    assert (fn <<_, _::size(x)>> -> true end).(<<?a, ?b>>)
+    assert <<_, _::size(^x)>> = <<?a, ?b>>
+    assert (fn <<_, _::size(^x)>> -> true end).(<<?a, ?b>>)
   end
 
   test "bitsyntax size using expressions" do
@@ -250,27 +235,55 @@ defmodule Kernel.BinaryTest do
 
     foo = %{bar: 5}
     assert <<1::size(foo.bar)>>
-    assert <<1::size(length('abcd'))>>
+    assert <<1::size(length(~c"abcd"))>>
     assert <<255::size(hd(List.flatten([3])))>>
   end
 
   test "bitsyntax size using guard expressions in match context" do
     x = 8
-    assert <<1::size(x - 5)>> = <<1::3>>
-    assert <<1::size(x - 5)-unit(8)>> = <<1::3*8>>
-    assert <<1::size(length('abcd'))>> = <<1::4>>
+    assert <<1::size(^x - 5)>> = <<1::3>>
+    assert <<1::size(^x - 5)-unit(8)>> = <<1::3*8>>
+    assert <<1::size(length(~c"abcd"))>> = <<1::4>>
 
     foo = %{bar: 5}
-    assert <<1::size(foo.bar)>> = <<1::5>>
+    assert <<1::size((^foo).bar)>> = <<1::5>>
   end
 
-  defmacrop signed_16 do
+  test "bitsyntax size with pinned integer" do
+    a = 1
+    b = <<2, 3>>
+    assert <<^a, ^b::binary>> = <<1, 2, 3>>
+  end
+
+  test "automatic size computation of matched bitsyntax variable" do
+    var = "foo"
+    <<^var::binary, rest::binary>> = "foobar"
+    assert rest == "bar"
+
+    <<^var::bytes, rest::bytes>> = "foobar"
+    assert rest == "bar"
+
+    ^var <> rest = "foobar"
+    assert rest == "bar"
+
+    var = <<0, 1>>
+    <<^var::bitstring, rest::bitstring>> = <<0, 1, 2, 3>>
+    assert rest == <<2, 3>>
+
+    <<^var::bits, rest::bits>> = <<0, 1, 2, 3>>
+    assert rest == <<2, 3>>
+
+    ^var <> rest = <<0, 1, 2, 3>>
+    assert rest == <<2, 3>>
+  end
+
+  defmacro signed_16 do
     quote do
       big - signed - integer - unit(16)
     end
   end
 
-  defmacrop refb_spec do
+  defmacro refb_spec do
     quote do
       1 * 8 - big - signed - integer
     end
@@ -281,10 +294,40 @@ defmodule Kernel.BinaryTest do
     sec_data = "another"
 
     <<
-      byte_size(refb)::refb_spec,
+      byte_size(refb)::refb_spec(),
       refb::binary,
-      byte_size(sec_data)::size(1)-signed_16,
+      byte_size(sec_data)::size(1)-signed_16(),
       sec_data::binary
     >>
   end
+
+  test "bitsyntax macro is expanded with a warning" do
+    assert capture_err(fn ->
+             Code.eval_string("<<1::refb_spec>>", [], __ENV__)
+           end) =~
+             "bitstring specifier \"refb_spec\" does not exist and is being expanded to \"refb_spec()\""
+
+    assert capture_err(fn ->
+             Code.eval_string("<<1::size(1)-signed_16>>", [], __ENV__)
+           end) =~
+             "bitstring specifier \"signed_16\" does not exist and is being expanded to \"signed_16()\""
+  end
+
+  test "bitsyntax with extra parentheses warns" do
+    assert capture_err(fn ->
+             Code.eval_string("<<1::big()>>")
+           end) =~ "extra parentheses on a bitstring specifier \"big()\" have been deprecated"
+
+    assert capture_err(fn ->
+             Code.eval_string("<<1::size(8)-integer()>>")
+           end) =~ "extra parentheses on a bitstring specifier \"integer()\" have been deprecated"
+  end
+
+  defp capture_err(fun) do
+    ExUnit.CaptureIO.capture_io(:stderr, fun)
+  end
+
+  defp assert_compile_error(message, fun) do
+    assert capture_err(fn -> assert_raise CompileError, fun end) =~ message
+  end
 end
