@@ -36,12 +36,11 @@ defmodule Kernel.CLITest do
   use ExUnit.Case, async: true
 
   import ExUnit.CaptureIO
-  import Retry
 
   defp run(argv) do
-    {config, argv} = Kernel.CLI.parse_argv(argv)
+    {config, argv} = Kernel.CLI.parse_argv(Enum.map(argv, &String.to_charlist/1))
     assert Kernel.CLI.process_commands(config) == []
-    argv
+    Enum.map(argv, &IO.chardata_to_string/1)
   end
 
   test "argv handling" do
@@ -60,87 +59,157 @@ defmodule Kernel.CLITest do
                       ["sample.exs", "-o", "1", "2"]
            end)
   end
+end
 
-  stderr_test "--help smoke test" do
-    output = elixir('--help')
-    assert output =~ "Usage: elixir"
+test_parameters =
+  if(PathHelpers.windows?(),
+    do: [%{cli_extension: ".bat"}],
+    else:
+      [%{cli_extension: ""}] ++
+        if(System.find_executable("pwsh"), do: [%{cli_extension: ".ps1"}], else: [])
+  )
+
+defmodule Kernel.CLI.ExecutableTest do
+  use ExUnit.Case,
+    async: true,
+    parameterize: test_parameters
+
+  import Retry
+
+  @tag :tmp_dir
+  test "file smoke test", context do
+    file = Path.join(context.tmp_dir, "hello_world!.exs")
+    File.write!(file, "IO.puts :hello_world123")
+    {output, 0} = System.cmd(elixir_executable(context.cli_extension), [file])
+    assert output =~ "hello_world123"
   end
 
-  test "--version smoke test" do
-    output = elixir('--version')
+  test "--eval smoke test", context do
+    {output, 0} =
+      System.cmd(elixir_executable(context.cli_extension), ["--eval", "IO.puts :hello_world123"])
+
+    assert output =~ "hello_world123"
+
+    # Check for -e and exclamation mark handling on Windows
+    assert {_output, 0} =
+             System.cmd(elixir_executable(context.cli_extension), ["-e", "Time.new!(0, 0, 0)"])
+
+    # TODO: remove this once we bump CI to 26.3
+    if not (windows?() and System.otp_release() == "26") do
+      {output, 0} =
+        System.cmd(iex_executable(context.cli_extension), [
+          "--eval",
+          "IO.puts :hello_world123; System.halt()"
+        ])
+
+      assert output =~ "hello_world123"
+
+      {output, 0} =
+        System.cmd(iex_executable(context.cli_extension), [
+          "-e",
+          "IO.puts :hello_world123; System.halt()"
+        ])
+
+      assert output =~ "hello_world123"
+    end
+  end
+
+  test "--version smoke test", %{cli_extension: cli_extension} do
+    output = elixir(~c"--version", cli_extension)
     assert output =~ "Erlang/OTP #{System.otp_release()}"
     assert output =~ "Elixir #{System.version()}"
 
-    output = iex('--version')
+    output = iex(~c"--version", cli_extension)
     assert output =~ "Erlang/OTP #{System.otp_release()}"
     assert output =~ "IEx #{System.version()}"
 
-    output = elixir('--version -e "IO.puts(:test_output)"')
+    output = elixir(~c"--version -e \"IO.puts(:test_output)\"", cli_extension)
     assert output =~ "Erlang/OTP #{System.otp_release()}"
     assert output =~ "Elixir #{System.version()}"
     assert output =~ "Standalone options can't be combined with other options"
   end
 
-  test "--short-version smoke test" do
-    output = elixir('--short-version')
+  test "--short-version smoke test", %{cli_extension: cli_extension} do
+    output = elixir(~c"--short-version", cli_extension)
     assert output =~ System.version()
     refute output =~ "Erlang"
   end
 
-  stderr_test "combining --help results in error" do
-    output = elixir('-e 1 --help')
+  stderr_test "--help smoke test", %{cli_extension: cli_extension} do
+    output = elixir(~c"--help", cli_extension)
+    assert output =~ "Usage: elixir"
+  end
+
+  stderr_test "combining --help results in error", %{cli_extension: cli_extension} do
+    output = elixir(~c"-e 1 --help", cli_extension)
     assert output =~ "--help : Standalone options can't be combined with other options"
 
-    output = elixir('--help -e 1')
+    output = elixir(~c"--help -e 1", cli_extension)
     assert output =~ "--help : Standalone options can't be combined with other options"
   end
 
-  stderr_test "combining --short-version results in error" do
-    output = elixir('--short-version -e 1')
+  stderr_test "combining --short-version results in error", %{cli_extension: cli_extension} do
+    output = elixir(~c"--short-version -e 1", cli_extension)
     assert output =~ "--short-version : Standalone options can't be combined with other options"
 
-    output = elixir('-e 1 --short-version')
+    output = elixir(~c"-e 1 --short-version", cli_extension)
     assert output =~ "--short-version : Standalone options can't be combined with other options"
   end
 
-  test "properly parses paths" do
+  test "parses paths", %{cli_extension: cli_extension} do
     root = fixture_path("../../..") |> to_charlist
 
     args =
-      '-pa "#{root}/*" -pz "#{root}/lib/*" -e "IO.inspect(:code.get_path(), limit: :infinity)"'
+      ~c"-pa \"#{root}/*\" -pz \"#{root}/lib/*\" -e \"IO.inspect(:code.get_path(), limit: :infinity)\""
 
-    list = elixir(args)
+    list = elixir(args, cli_extension)
     {path, _} = Code.eval_string(list, [])
 
     # pa
-    assert to_charlist(Path.expand('ebin', root)) in path
-    assert to_charlist(Path.expand('lib', root)) in path
-    assert to_charlist(Path.expand('src', root)) in path
+    assert to_charlist(Path.expand(~c"ebin", root)) in path
+    assert to_charlist(Path.expand(~c"lib", root)) in path
+    assert to_charlist(Path.expand(~c"src", root)) in path
 
     # pz
-    assert to_charlist(Path.expand('lib/list', root)) in path
+    assert to_charlist(Path.expand(~c"lib/list", root)) in path
   end
 
-  stderr_test "properly formats errors" do
-    assert String.starts_with?(elixir('-e ":erlang.throw 1"'), "** (throw) 1")
-    assert String.starts_with?(elixir('-e ":erlang.error 1"'), "** (ErlangError) Erlang error: 1")
-    assert String.starts_with?(elixir('-e "1 +"'), "** (TokenMissingError)")
+  stderr_test "formats errors", %{cli_extension: cli_extension} do
+    assert String.starts_with?(elixir(~c"-e \":erlang.throw 1\"", cli_extension), "** (throw) 1")
 
-    assert elixir('-e "Task.async(fn -> raise ArgumentError end) |> Task.await"') =~
+    assert String.starts_with?(
+             elixir(~c"-e \":erlang.error 1\"", cli_extension),
+             "** (ErlangError) Erlang error: 1"
+           )
+
+    assert String.starts_with?(elixir(~c"-e \"1 +\"", cli_extension), "** (TokenMissingError)")
+
+    assert elixir(
+             ~c"-e \"Task.async(fn -> raise ArgumentError end) |> Task.await\"",
+             cli_extension
+           ) =~
              "an exception was raised:\n    ** (ArgumentError) argument error"
 
-    assert elixir('-e "IO.puts(Process.flag(:trap_exit, false)); exit({:shutdown, 1})"') ==
+    assert elixir(
+             ~c"-e \"IO.puts(Process.flag(:trap_exit, false)); exit({:shutdown, 1})\"",
+             cli_extension
+           ) ==
              "false\n"
   end
 
-  stderr_test "blames exceptions" do
-    error = elixir('-e "Access.fetch :foo, :bar"')
+  stderr_test "blames exceptions", %{cli_extension: cli_extension} do
+    error = elixir(~c"-e \"Access.fetch :foo, :bar\"", cli_extension)
     assert error =~ "** (FunctionClauseError) no function clause matching in Access.fetch/2"
     assert error =~ "The following arguments were given to Access.fetch/2"
     assert error =~ ":foo"
     assert error =~ "def fetch(-%module{} = container-, +key+)"
     assert error =~ ~r"\(elixir #{System.version()}\) lib/access\.ex:\d+: Access\.fetch/2"
   end
+
+  test "invokes at_exit callbacks" do
+    assert elixir(fixture_path("at_exit.exs") |> to_charlist()) ==
+             "goodbye cruel world with status 1\n"
+  end
 end
 
 defmodule Kernel.CLI.RPCTest do
@@ -150,7 +219,7 @@ defmodule Kernel.CLI.RPCTest do
 
   defp rpc_eval(command) do
     node = "cli-rpc#{System.unique_integer()}@127.0.0.1"
-    elixir('--name #{node} --rpc-eval #{node} "#{command}"')
+    elixir(~c"--name #{node} --rpc-eval #{node} \"#{command}\"")
   end
 
   test "invokes command on remote node" do
@@ -159,14 +228,14 @@ defmodule Kernel.CLI.RPCTest do
 
   test "invokes command on remote node without host and --name after --rpc-eval" do
     node = "cli-rpc#{System.unique_integer()}"
-    assert elixir('--rpc-eval #{node} "IO.puts :ok" --name #{node}@127.0.0.1 ') == "ok\n"
+    assert elixir(~c"--rpc-eval #{node} \"IO.puts :ok\" --name #{node}@127.0.0.1 ") == "ok\n"
   end
 
   test "can be invoked multiple times" do
     node = "cli-rpc#{System.unique_integer()}"
 
     assert elixir(
-             '--name #{node}@127.0.0.1 --rpc-eval #{node} "IO.puts :foo" --rpc-eval #{node} "IO.puts :bar"'
+             ~c"--name #{node}@127.0.0.1 --rpc-eval #{node} \"IO.puts :foo\" --rpc-eval #{node} \"IO.puts :bar\""
            ) == "foo\nbar\n"
   end
 
@@ -175,10 +244,10 @@ defmodule Kernel.CLI.RPCTest do
   test "fails on wrong arguments" do
     node = "cli-rpc#{System.unique_integer()}"
 
-    assert elixir('--name #{node}@127.0.0.1 --rpc-eval') ==
+    assert elixir(~c"--name #{node}@127.0.0.1 --rpc-eval") ==
              "--rpc-eval : wrong number of arguments\n"
 
-    assert elixir('--name #{node}@127.0.0.1 --rpc-eval #{node}') ==
+    assert elixir(~c"--name #{node}@127.0.0.1 --rpc-eval #{node}") ==
              "--rpc-eval : wrong number of arguments\n"
   end
 
@@ -195,17 +264,10 @@ defmodule Kernel.CLI.RPCTest do
   end
 end
 
-defmodule Kernel.CLI.AtExitTest do
-  use ExUnit.Case, async: true
-
-  test "invokes at_exit callbacks" do
-    assert elixir(fixture_path("at_exit.exs") |> to_charlist) ==
-             "goodbye cruel world with status 1\n"
-  end
-end
-
 defmodule Kernel.CLI.CompileTest do
-  use ExUnit.Case, async: true
+  use ExUnit.Case,
+    async: true,
+    parameterize: test_parameters
 
   import Retry
   @moduletag :tmp_dir
@@ -217,7 +279,7 @@ defmodule Kernel.CLI.CompileTest do
   end
 
   test "compiles code", context do
-    assert elixirc('#{context.fixture} -o #{context.tmp_dir}') == ""
+    assert elixirc(~c"#{context.fixture} -o #{context.tmp_dir}", context.cli_extension) == ""
     assert File.regular?(context.beam_file_path)
 
     # Assert that the module is loaded into memory with the proper destination for the BEAM file.
@@ -234,7 +296,7 @@ defmodule Kernel.CLI.CompileTest do
     try do
       fixture = String.replace(context.fixture, "/", "\\")
       tmp_dir_path = String.replace(context.tmp_dir, "/", "\\")
-      assert elixirc('#{fixture} -o #{tmp_dir_path}') == ""
+      assert elixirc(~c"#{fixture} -o #{tmp_dir_path}", context.cli_extension) == ""
       assert File.regular?(context[:beam_file_path])
 
       # Assert that the module is loaded into memory with the proper destination for the BEAM file.
@@ -250,16 +312,18 @@ defmodule Kernel.CLI.CompileTest do
   end
 
   stderr_test "fails on missing patterns", context do
-    output = elixirc('#{context.fixture} non_existing.ex -o #{context.tmp_dir}')
+    output =
+      elixirc(~c"#{context.fixture} non_existing.ex -o #{context.tmp_dir}", context.cli_extension)
+
     assert output =~ "non_existing.ex"
     refute output =~ "compile_sample.ex"
     refute File.exists?(context.beam_file_path)
   end
 
   stderr_test "fails on missing write access to .beam file", context do
-    compilation_args = '#{context.fixture} -o #{context.tmp_dir}'
+    compilation_args = ~c"#{context.fixture} -o #{context.tmp_dir}"
 
-    assert elixirc(compilation_args) == ""
+    assert elixirc(compilation_args, context.cli_extension) == ""
     assert File.regular?(context.beam_file_path)
 
     # Set the .beam file to read-only
@@ -268,7 +332,7 @@ defmodule Kernel.CLI.CompileTest do
 
     # Can only assert when read-only applies to the user
     if access != :read_write do
-      output = elixirc(compilation_args)
+      output = elixirc(compilation_args, context.cli_extension)
 
       expected =
         "(File.Error) could not write to file #{inspect(context.beam_file_path)}: permission denied"
