@@ -21,7 +21,7 @@ defmodule Kernel.QuoteTest do
     line = __ENV__.line + 2
 
     assert quote(location: :keep, do: bar(1, 2, 3)) ==
-             {:bar, [keep: {Path.relative_to_cwd(__ENV__.file), line}], [1, 2, 3]}
+             {:bar, [keep: {__ENV__.file, line}], [1, 2, 3]}
   end
 
   test "fixed line" do
@@ -30,6 +30,17 @@ defmodule Kernel.QuoteTest do
     assert quote(line: true, do: bar(1, 2, 3)) == {:bar, [line: __ENV__.line], [1, 2, 3]}
   end
 
+  test "file line" do
+    assert quote(file: "foo", line: 3, do: bar(1, 2, 3)) ==
+             {:bar, [keep: {"foo", 3}], [1, 2, 3]}
+
+    assert quote(file: "foo", line: false, do: bar(1, 2, 3)) ==
+             {:bar, [keep: {"foo", 0}], [1, 2, 3]}
+
+    assert quote(file: "foo", line: true, do: bar(1, 2, 3)) ==
+             {:bar, [keep: {"foo", __ENV__.line - 1}], [1, 2, 3]}
+  end
+
   test "quote line var" do
     line = __ENV__.line
     assert quote(line: line, do: bar(1, 2, 3)) == {:bar, [line: line], [1, 2, 3]}
@@ -60,6 +71,14 @@ defmodule Kernel.QuoteTest do
     end
   end
 
+  test "quote context bind_quoted" do
+    assert {:__block__, _,
+            [{:=, [], [{:some_var, _, :fallback}, 321]}, {:some_var, _, :fallback}]} =
+             (quote bind_quoted: [some_var: 321], context: __ENV__.context || :fallback do
+                some_var
+              end)
+  end
+
   test "operator precedence" do
     assert {:+, _, [{:+, _, [1, _]}, 1]} = quote(do: 1 + Foo.l() + 1)
     assert {:+, _, [1, {_, _, [{:+, _, [1]}]}]} = quote(do: 1 + Foo.l(+1))
@@ -120,6 +139,12 @@ defmodule Kernel.QuoteTest do
     assert {:quote, _, [[do: {:unquote, _, _}]]} = quote(do: quote(do: unquote(x)))
   end
 
+  test "import inside nested quote" do
+    # Check that we can evaluate imports from quote inside quote
+    assert {{:to_string, meta, [123]}, _} = Code.eval_quoted(quote(do: quote(do: to_string(123))))
+    assert meta[:imports] == [{1, Kernel}]
+  end
+
   defmacrop nested_quote_in_macro do
     x = 1
 
@@ -203,15 +228,6 @@ defmodule Kernel.QuoteTest do
 
     map = %{foo: :default}
     assert %{map | unquote_splicing(foo: :bar)} == %{foo: :bar}
-
-    assert Code.eval_string("quote do: %{unquote_splicing foo: :bar}") ==
-             {{:%{}, [], [foo: :bar]}, []}
-
-    assert Code.eval_string("quote do: %{:baz => :bat, unquote_splicing foo: :bar}") ==
-             {{:%{}, [], [{:baz, :bat}, {:foo, :bar}]}, []}
-
-    assert Code.eval_string("quote do: %{foo bar | baz}") ==
-             {{:%{}, [], [{:foo, [], [{:|, [], [{:bar, [], Elixir}, {:baz, [], Elixir}]}]}]}, []}
   end
 
   test "when" do
@@ -227,7 +243,7 @@ defmodule Kernel.QuoteTest do
                 () -> 1
               end)
 
-    assert [{:->, _, [[], 1]}] = quote(do: (() -> 1))
+    assert [{:->, _, [[], 1]}] = quote(do: (-> 1))
   end
 
   test "empty block" do
@@ -322,9 +338,21 @@ defmodule Kernel.QuoteTest do
                 end
               end)
   end
+
+  test "capture" do
+    assert Code.string_to_quoted!("&1[:foo]") == Code.string_to_quoted!("(&1)[:foo]")
+    assert Code.string_to_quoted!("&1 [:foo]") == Code.string_to_quoted!("(&1)[:foo]")
+    assert Code.string_to_quoted!("& 1[:foo]") == Code.string_to_quoted!("&(1[:foo])")
+  end
+
+  test "not and ! as rearrange ops" do
+    assert {:__block__, _, [{:not, [line: 1], [true]}]} = Code.string_to_quoted!("(not true)")
+
+    assert {:fn, _, [{:->, _, [[], {:not, _, [true]}]}]} =
+             Code.string_to_quoted!("fn -> not true end")
+  end
 end
 
-# DO NOT MOVE THIS LINE
 defmodule Kernel.QuoteTest.Errors do
   def line, do: __ENV__.line + 4
 
@@ -355,8 +383,6 @@ defmodule Kernel.QuoteTest.ErrorsTest do
         mod = Kernel.QuoteTest.ErrorsTest
         file = __ENV__.file |> Path.relative_to_cwd() |> String.to_charlist()
         assert [{^mod, :will_raise, 2, [file: ^file, line: @line] ++ _} | _] = __STACKTRACE__
-    else
-      _ -> flunk("expected failure")
     end
   end
 
@@ -364,13 +390,12 @@ defmodule Kernel.QuoteTest.ErrorsTest do
   test "outside function error" do
     try do
       will_raise()
+      flunk("expected failure")
     rescue
       RuntimeError ->
         mod = Kernel.QuoteTest.ErrorsTest
         file = __ENV__.file |> Path.relative_to_cwd() |> String.to_charlist()
         assert [{^mod, _, _, [file: ^file, line: @line] ++ _} | _] = __STACKTRACE__
-    else
-      _ -> flunk("expected failure")
     end
   end
 end
@@ -562,25 +587,25 @@ defmodule Kernel.QuoteTest.ImportsHygieneTest do
 
   defmacrop get_list_length do
     quote do
-      length('hello')
+      length(~c"hello")
     end
   end
 
   defmacrop get_list_length_with_pipe do
     quote do
-      'hello' |> length()
+      ~c"hello" |> length()
     end
   end
 
   defmacrop get_list_length_with_partial do
     quote do
-      (&length(&1)).('hello')
+      (&length(&1)).(~c"hello")
     end
   end
 
   defmacrop get_list_length_with_function do
     quote do
-      (&length/1).('hello')
+      (&length/1).(~c"hello")
     end
   end
 
@@ -619,7 +644,7 @@ defmodule Kernel.QuoteTest.ImportsHygieneTest do
     quote do
       import Kernel, except: [length: 1]
       import String, only: [length: 1]
-      length('hello')
+      length(~c"hello")
     end
   end
 
@@ -641,3 +666,168 @@ defmodule Kernel.QuoteTest.ImportsHygieneTest do
     assert meta[:imports] == [{0, BinaryUtils}]
   end
 end
+
+defmodule Kernel.QuoteTest.HasUnquoteTest do
+  use ExUnit.Case, async: true
+
+  test "expression without unquote returns false" do
+    ast =
+      quote unquote: false do
+        opts = [x: 5]
+        x = Keyword.fetch!(opts, :x)
+        x + 1
+      end
+
+    refute :elixir_quote.has_unquotes(ast)
+  end
+
+  test "expression with unquote returns true" do
+    ast =
+      quote unquote: false do
+        [x: unquote(x)]
+      end
+
+    assert :elixir_quote.has_unquotes(ast)
+
+    ast =
+      quote unquote: false do
+        unquote(module).fun(x)
+      end
+
+    assert :elixir_quote.has_unquotes(ast)
+
+    ast =
+      quote unquote: false do
+        module.unquote(fun)(x)
+      end
+
+    assert :elixir_quote.has_unquotes(ast)
+
+    ast =
+      quote unquote: false do
+        module.fun(unquote(x))
+      end
+
+    assert :elixir_quote.has_unquotes(ast)
+
+    ast =
+      quote unquote: false do
+        module.fun(unquote_splicing(args))
+      end
+
+    assert :elixir_quote.has_unquotes(ast)
+  end
+
+  test "expression with unquote within quote returns false" do
+    ast =
+      quote unquote: false do
+        quote do
+          x + unquote(y)
+        end
+      end
+
+    refute :elixir_quote.has_unquotes(ast)
+
+    ast =
+      quote unquote: false do
+        quote do
+          foo = bar(unquote_splicing(args))
+        end
+      end
+
+    refute :elixir_quote.has_unquotes(ast)
+  end
+
+  test "expression with unquote within unquote within quote returns true" do
+    ast =
+      quote unquote: false do
+        quote do
+          x + unquote(unquote(y))
+        end
+      end
+
+    assert :elixir_quote.has_unquotes(ast)
+
+    ast =
+      quote unquote: false do
+        quote do
+          foo = bar(unquote_splicing(unquote(args)))
+        end
+      end
+
+    assert :elixir_quote.has_unquotes(ast)
+
+    ast =
+      quote unquote: false do
+        quote do
+          foo = bar(unquote(unquote_splicing(args)))
+        end
+      end
+
+    assert :elixir_quote.has_unquotes(ast)
+  end
+
+  test "expression within quote disabling unquotes always returns false" do
+    ast =
+      quote unquote: false do
+        quote unquote: false do
+          x + unquote(unquote(y))
+        end
+      end
+
+    refute :elixir_quote.has_unquotes(ast)
+
+    ast =
+      quote unquote: false do
+        quote bind_quoted: [x: x] do
+          x + unquote(unquote(y))
+        end
+      end
+
+    refute :elixir_quote.has_unquotes(ast)
+  end
+
+  test "unquote with invalid AST (shallow check)" do
+    for term <- [
+          %{unescaped: :map},
+          1..10,
+          {:bad_meta, nil, []},
+          {:bad_arg, nil, 1},
+          {:bad_tuple},
+          make_ref(),
+          [:improper | :list],
+          [nested: {}]
+        ] do
+      message = """
+      tried to unquote invalid AST: #{inspect(term)}
+      Did you forget to escape term using Macro.escape/1?\
+      """
+
+      assert_raise ArgumentError, message, fn -> quote do: unquote(term) end
+    end
+  end
+
+  test "unquote with invalid AST is not checked deeply" do
+    assert quote do: unquote(foo: [1 | 2]) == [foo: [1 | 2]]
+    assert quote do: unquote(foo: [bar: %{}]) == [foo: [bar: %{}]]
+  end
+
+  test "unquote_splicing with invalid AST" do
+    for args <- [
+          "not_a_list",
+          [:improper | :list],
+          [%{unescaped: :map}],
+          [1..10],
+          [{:bad_meta, nil, []}],
+          [{:bad_arg, nil, 1}],
+          [{:bad_tuple}],
+          [make_ref()],
+          [nested: {}]
+        ] do
+      message =
+        "expected a list with quoted expressions in unquote_splicing/1, got: #{inspect(args)}"
+
+      assert_raise ArgumentError, message, fn -> quote do: [unquote_splicing(args)] end
+    end
+  end
+end
