@@ -3,16 +3,19 @@ Code.require_file("test_helper.exs", __DIR__)
 defmodule KernelTest do
   use ExUnit.Case, async: true
 
-  doctest Kernel
+  # Skip these doctests are they emit warnings
+  doctest Kernel,
+    except:
+      [===: 2, !==: 2, and: 2, or: 2] ++
+        [is_exception: 1, is_exception: 2, is_nil: 1, is_struct: 1, is_non_struct_map: 1]
 
   def id(arg), do: arg
   def id(arg1, arg2), do: {arg1, arg2}
-  def empty_list(), do: []
   def empty_map, do: %{}
 
   defp purge(module) do
-    :code.delete(module)
     :code.purge(module)
+    :code.delete(module)
   end
 
   defp assert_eval_raise(error, msg, string) do
@@ -75,8 +78,8 @@ defmodule KernelTest do
       [1, 2] ++ var = [1, 2, 3]
       assert var == [3]
 
-      'ab' ++ var = 'abc'
-      assert var == 'c'
+      ~c"ab" ++ var = ~c"abc"
+      assert var == ~c"c"
 
       [:a, :b] ++ var = [:a, :b, :c]
       assert var == [:c]
@@ -86,24 +89,19 @@ defmodule KernelTest do
   test "=~/2" do
     assert "abcd" =~ ~r/c(d)/ == true
     assert "abcd" =~ ~r/e/ == false
-    assert "abcd" =~ ~R/c(d)/ == true
-    assert "abcd" =~ ~R/e/ == false
 
     string = "^ab+cd*$"
     assert string =~ "ab+" == true
     assert string =~ "bb" == false
 
     assert "abcd" =~ ~r// == true
-    assert "abcd" =~ ~R// == true
     assert "abcd" =~ "" == true
 
     assert "" =~ ~r// == true
-    assert "" =~ ~R// == true
     assert "" =~ "" == true
 
     assert "" =~ "abcd" == false
     assert "" =~ ~r/abcd/ == false
-    assert "" =~ ~R/abcd/ == false
 
     assert_raise FunctionClauseError, "no function clause matching in Kernel.=~/2", fn ->
       1234 =~ "hello"
@@ -114,7 +112,7 @@ defmodule KernelTest do
     end
 
     assert_raise FunctionClauseError, "no function clause matching in Kernel.=~/2", fn ->
-      1234 =~ ~R"hello"
+      1234 =~ ~r"hello"
     end
 
     assert_raise FunctionClauseError, "no function clause matching in Kernel.=~/2", fn ->
@@ -302,8 +300,8 @@ defmodule KernelTest do
     assert (false and true) == false
     assert (false and 0) == false
     assert (false and raise("oops")) == false
-    assert ((x = true) and not x) == false
-    assert_raise BadBooleanError, fn -> 0 and 1 end
+    assert ((x = Process.get(:unused, true)) and not x) == false
+    assert_raise BadBooleanError, fn -> Process.get(:unused, 0) and 1 end
   end
 
   test "or/2" do
@@ -314,25 +312,27 @@ defmodule KernelTest do
     assert (false or false) == false
     assert (false or true) == true
     assert (false or 0) == 0
-    assert ((x = false) or not x) == true
-    assert_raise BadBooleanError, fn -> 0 or 1 end
+    assert ((x = Process.get(:unused, false)) or not x) == true
+    assert_raise BadBooleanError, fn -> Process.get(:unused, 0) or 1 end
   end
 
-  defp struct?(arg) when is_struct(arg), do: true
-  defp struct?(_arg), do: false
+  defp delegate_is_struct(arg), do: is_struct(arg)
+
+  defp guarded_is_struct(arg) when is_struct(arg), do: true
+  defp guarded_is_struct(_arg), do: false
 
   defp struct_or_map?(arg) when is_struct(arg) or is_map(arg), do: true
   defp struct_or_map?(_arg), do: false
 
   test "is_struct/1" do
-    assert is_struct(%{}) == false
-    assert is_struct([]) == false
-    assert is_struct(%Macro.Env{}) == true
-    assert is_struct(%{__struct__: "foo"}) == false
-    assert struct?(%Macro.Env{}) == true
-    assert struct?(%{__struct__: "foo"}) == false
-    assert struct?([]) == false
-    assert struct?(%{}) == false
+    assert delegate_is_struct(%{}) == false
+    assert delegate_is_struct([]) == false
+    assert delegate_is_struct(%Macro.Env{}) == true
+    assert delegate_is_struct(%{__struct__: "foo"}) == false
+    assert guarded_is_struct(%Macro.Env{}) == true
+    assert guarded_is_struct(%{__struct__: "foo"}) == false
+    assert guarded_is_struct([]) == false
+    assert guarded_is_struct(%{}) == false
   end
 
   test "is_struct/1 and other match works" do
@@ -341,8 +341,10 @@ defmodule KernelTest do
     assert struct_or_map?(10) == false
   end
 
-  defp struct?(arg, name) when is_struct(arg, name), do: true
-  defp struct?(_arg, _name), do: false
+  defp delegate_is_struct(arg, name), do: is_struct(arg, name)
+
+  defp guarded_is_struct(arg, name) when is_struct(arg, name), do: true
+  defp guarded_is_struct(_arg, _name), do: false
 
   defp struct_or_map?(arg, name) when is_struct(arg, name) or is_map(arg), do: true
   defp struct_or_map?(_arg, _name), do: false
@@ -350,16 +352,16 @@ defmodule KernelTest do
   defp not_atom(), do: "not atom"
 
   test "is_struct/2" do
-    assert is_struct(%{}, Macro.Env) == false
-    assert is_struct([], Macro.Env) == false
-    assert is_struct(%Macro.Env{}, Macro.Env) == true
-    assert is_struct(%Macro.Env{}, URI) == false
-    assert struct?(%Macro.Env{}, Macro.Env) == true
-    assert struct?(%Macro.Env{}, URI) == false
-    assert struct?(%{__struct__: "foo"}, "foo") == false
-    assert struct?(%{__struct__: "foo"}, Macro.Env) == false
-    assert struct?([], Macro.Env) == false
-    assert struct?(%{}, Macro.Env) == false
+    assert delegate_is_struct(%{}, Macro.Env) == false
+    assert delegate_is_struct([], Macro.Env) == false
+    assert delegate_is_struct(%Macro.Env{}, Macro.Env) == true
+    assert delegate_is_struct(%Macro.Env{}, URI) == false
+    assert guarded_is_struct(%Macro.Env{}, Macro.Env) == true
+    assert guarded_is_struct(%Macro.Env{}, URI) == false
+    assert guarded_is_struct(%{__struct__: "foo"}, "foo") == false
+    assert guarded_is_struct(%{__struct__: "foo"}, Macro.Env) == false
+    assert guarded_is_struct([], Macro.Env) == false
+    assert guarded_is_struct(%{}, Macro.Env) == false
 
     assert_raise ArgumentError, "argument error", fn ->
       is_struct(%{}, not_atom())
@@ -372,21 +374,48 @@ defmodule KernelTest do
     assert struct_or_map?(%Macro.Env{}, Macro.Env) == true
   end
 
-  defp exception?(arg) when is_exception(arg), do: true
-  defp exception?(_arg), do: false
+  defp delegate_is_non_struct_map(arg), do: is_non_struct_map(arg)
+
+  defp guarded_is_non_struct_map(arg) when is_non_struct_map(arg), do: true
+  defp guarded_is_non_struct_map(_arg), do: false
+
+  defp non_struct_map_or_struct?(arg) when is_non_struct_map(arg) or is_struct(arg), do: true
+  defp non_struct_map_or_struct?(_arg), do: false
+
+  test "is_non_struct_map/1" do
+    assert delegate_is_non_struct_map(%{}) == true
+    assert delegate_is_non_struct_map([]) == false
+    assert delegate_is_non_struct_map(%Macro.Env{}) == false
+    assert delegate_is_non_struct_map(%{__struct__: "foo"}) == true
+    assert guarded_is_non_struct_map(%Macro.Env{}) == false
+    assert guarded_is_non_struct_map(%{__struct__: "foo"}) == true
+    assert guarded_is_non_struct_map([]) == false
+    assert guarded_is_non_struct_map(%{}) == true
+  end
+
+  test "is_non_struct_map/1 and other match works" do
+    assert non_struct_map_or_struct?(%Macro.Env{}) == true
+    assert non_struct_map_or_struct?(%{}) == true
+    assert non_struct_map_or_struct?(10) == false
+  end
+
+  defp delegate_is_exception(arg), do: is_exception(arg)
+
+  defp guarded_is_exception(arg) when is_exception(arg), do: true
+  defp guarded_is_exception(_arg), do: false
 
   defp exception_or_map?(arg) when is_exception(arg) or is_map(arg), do: true
   defp exception_or_map?(_arg), do: false
 
   test "is_exception/1" do
-    assert is_exception(%{}) == false
-    assert is_exception([]) == false
-    assert is_exception(%RuntimeError{}) == true
-    assert is_exception(%{__exception__: "foo"}) == false
-    assert exception?(%RuntimeError{}) == true
-    assert exception?(%{__exception__: "foo"}) == false
-    assert exception?([]) == false
-    assert exception?(%{}) == false
+    assert delegate_is_exception(%{}) == false
+    assert delegate_is_exception([]) == false
+    assert delegate_is_exception(%RuntimeError{}) == true
+    assert delegate_is_exception(%{__exception__: "foo"}) == false
+    assert guarded_is_exception(%RuntimeError{}) == true
+    assert guarded_is_exception(%{__exception__: "foo"}) == false
+    assert guarded_is_exception([]) == false
+    assert guarded_is_exception(%{}) == false
   end
 
   test "is_exception/1 and other match works" do
@@ -395,26 +424,28 @@ defmodule KernelTest do
     assert exception_or_map?(10) == false
   end
 
-  defp exception?(arg, name) when is_exception(arg, name), do: true
-  defp exception?(_arg, _name), do: false
+  defp delegate_is_exception(arg, name), do: is_exception(arg, name)
+
+  defp guarded_is_exception(arg, name) when is_exception(arg, name), do: true
+  defp guarded_is_exception(_arg, _name), do: false
 
   defp exception_or_map?(arg, name) when is_exception(arg, name) or is_map(arg), do: true
   defp exception_or_map?(_arg, _name), do: false
 
   test "is_exception/2" do
-    assert is_exception(%{}, RuntimeError) == false
-    assert is_exception([], RuntimeError) == false
-    assert is_exception(%RuntimeError{}, RuntimeError) == true
-    assert is_exception(%RuntimeError{}, Macro.Env) == false
-    assert exception?(%RuntimeError{}, RuntimeError) == true
-    assert exception?(%RuntimeError{}, Macro.Env) == false
-    assert exception?(%{__exception__: "foo"}, "foo") == false
-    assert exception?(%{__exception__: "foo"}, RuntimeError) == false
-    assert exception?([], RuntimeError) == false
-    assert exception?(%{}, RuntimeError) == false
+    assert delegate_is_exception(%{}, RuntimeError) == false
+    assert delegate_is_exception([], RuntimeError) == false
+    assert delegate_is_exception(%RuntimeError{}, RuntimeError) == true
+    assert delegate_is_exception(%RuntimeError{}, Macro.Env) == false
+    assert guarded_is_exception(%RuntimeError{}, RuntimeError) == true
+    assert guarded_is_exception(%RuntimeError{}, Macro.Env) == false
+    assert guarded_is_exception(%{__exception__: "foo"}, "foo") == false
+    assert guarded_is_exception(%{__exception__: "foo"}, RuntimeError) == false
+    assert guarded_is_exception([], RuntimeError) == false
+    assert guarded_is_exception(%{}, RuntimeError) == false
 
     assert_raise ArgumentError, "argument error", fn ->
-      is_exception(%{}, not_atom())
+      delegate_is_exception(%{}, not_atom())
     end
   end
 
@@ -539,28 +570,6 @@ defmodule KernelTest do
       assert fun_in(17.0) == :none
     end
 
-    def dynamic_in(x, y, z) when x in y..z, do: true
-    def dynamic_in(_x, _y, _z), do: false
-
-    test "in dynamic range function guard" do
-      assert dynamic_in(1, 1, 3)
-      assert dynamic_in(2, 1, 3)
-      assert dynamic_in(3, 1, 3)
-
-      assert dynamic_in(1, 3, 1)
-      assert dynamic_in(2, 3, 1)
-      assert dynamic_in(3, 3, 1)
-
-      refute dynamic_in(0, 1, 3)
-      refute dynamic_in(4, 1, 3)
-      refute dynamic_in(0, 3, 1)
-      refute dynamic_in(4, 3, 1)
-
-      refute dynamic_in(2, 1.0, 3)
-      refute dynamic_in(2, 1, 3.0)
-      refute dynamic_in(2.0, 1, 3)
-    end
-
     def dynamic_step_in(x, y, z, w) when x in y..z//w, do: true
     def dynamic_step_in(_x, _y, _z, _w), do: false
 
@@ -602,7 +611,7 @@ defmodule KernelTest do
       assert case_in(1, 1..3) == true
       assert case_in(2, 1..3) == true
       assert case_in(3, 1..3) == true
-      assert case_in(-3, -1..-3) == true
+      assert case_in(-3, -1..-3//-1) == true
     end
 
     def map_dot(map) when map.field, do: true
@@ -613,24 +622,6 @@ defmodule KernelTest do
       refute map_dot(%{})
       refute map_dot(%{field: false})
       assert map_dot(%{field: true})
-
-      message =
-        "cannot invoke remote function in guards. " <>
-          "If you want to do a map lookup instead, please remove parens from map.field()"
-
-      assert_raise CompileError, Regex.compile!(message), fn ->
-        defmodule MapDot do
-          def map_dot(map) when map.field(), do: true
-        end
-      end
-
-      message = ~r"cannot invoke remote function Module.fun/0 inside guards"
-
-      assert_raise CompileError, message, fn ->
-        defmodule MapDot do
-          def map_dot(map) when Module.fun(), do: true
-        end
-      end
     end
 
     test "performs all side-effects" do
@@ -752,14 +743,6 @@ defmodule KernelTest do
       assert result =~ ":lists.member(rand(), [1 | some_call()]"
     end
 
-    defp quote_case_in(left, right) do
-      quote do
-        case 0 do
-          _ when unquote(left) in unquote(right) -> true
-        end
-      end
-    end
-
     test "is optimized" do
       assert expand_to_string(quote(do: foo in [])) ==
                "_ = foo\nfalse"
@@ -813,7 +796,7 @@ defmodule KernelTest do
 
     test ":struct" do
       assert Kernel.__info__(:struct) == nil
-      assert hd(URI.__info__(:struct)) == %{field: :scheme, required: false}
+      assert [%{field: :scheme, default: nil} | _] = URI.__info__(:struct)
     end
 
     test "others" do
@@ -849,11 +832,25 @@ defmodule KernelTest do
     end
 
     test "matching attribute" do
-      assert_raise ArgumentError, ~r"invalid write attribute syntax", fn ->
+      assert_raise ArgumentError, ~r"invalid usage of module attributes", fn ->
         defmodule MatchAttributeInModule do
           @foo = 42
         end
       end
+
+      assert_raise ArgumentError, ~r"invalid usage of module attributes", fn ->
+        defmodule MatchAttributeInModule do
+          @foo 16
+          <<_::@foo>> = "ab"
+        end
+      end
+
+      assert_raise ArgumentError, ~r"invalid usage of module attributes", fn ->
+        defmodule MatchAttributeInModule do
+          @foo 16
+          <<_::size(@foo)>> = "ab"
+        end
+      end
     end
   end
 
@@ -935,7 +932,7 @@ defmodule KernelTest do
     test "expects atoms as module names" do
       msg = ~r"invalid module name: 3"
 
-      assert_raise CompileError, msg, fn ->
+      assert_raise ArgumentError, msg, fn ->
         defmodule 1 + 2, do: :ok
       end
     end
@@ -946,18 +943,18 @@ defmodule KernelTest do
       Enum.each(special_atoms, fn special_atom ->
         msg = ~r"invalid module name: #{inspect(special_atom)}"
 
-        assert_raise CompileError, msg, fn ->
+        assert_raise ArgumentError, msg, fn ->
           defmodule special_atom, do: :ok
         end
       end)
     end
 
     test "does not accept slashes in module names" do
-      assert_raise CompileError, ~r(invalid module name: :"foo/bar"), fn ->
+      assert_raise ArgumentError, ~r(invalid module name: :"foo/bar"), fn ->
         defmodule :"foo/bar", do: :ok
       end
 
-      assert_raise CompileError, ~r(invalid module name: :"foo\\\\bar"), fn ->
+      assert_raise ArgumentError, ~r(invalid module name: :"foo\\\\bar"), fn ->
         defmodule :"foo\\bar", do: :ok
       end
     end
@@ -968,6 +965,22 @@ defmodule KernelTest do
       defstruct [:foo, :bar]
     end
 
+    test "get_in/1" do
+      users = %{"john" => %{age: 27}, :meg => %{age: 23}}
+      assert get_in(users["john"][:age]) == 27
+      assert get_in(users["dave"][:age]) == nil
+      assert get_in(users["john"].age) == 27
+      assert get_in(users["dave"].age) == nil
+      assert get_in(users.meg[:age]) == 23
+      assert get_in(users.meg.age) == 23
+
+      is_nil = nil
+      assert get_in(is_nil.age) == nil
+
+      assert_raise KeyError, ~r"key :unknown not found", fn -> get_in(users.unknown) end
+      assert_raise KeyError, ~r"key :unknown not found", fn -> get_in(users.meg.unknown) end
+    end
+
     test "get_in/2" do
       users = %{"john" => %{age: 27}, "meg" => %{age: 23}}
       assert get_in(users, ["john", :age]) == 27
@@ -978,10 +991,6 @@ defmodule KernelTest do
       assert get_in(map, ["fruits", by_index(0)]) == "banana"
       assert get_in(map, ["fruits", by_index(3)]) == nil
       assert get_in(map, ["unknown", by_index(3)]) == nil
-
-      assert_raise FunctionClauseError, fn ->
-        get_in(users, [])
-      end
     end
 
     test "put_in/3" do
@@ -989,10 +998,6 @@ defmodule KernelTest do
 
       assert put_in(users, ["john", :age], 28) == %{"john" => %{age: 28}, "meg" => %{age: 23}}
 
-      assert_raise FunctionClauseError, fn ->
-        put_in(users, [], %{})
-      end
-
       assert_raise ArgumentError, "could not put/update key \"john\" on a nil value", fn ->
         put_in(nil, ["john", :age], 28)
       end
@@ -1025,10 +1030,6 @@ defmodule KernelTest do
       assert update_in(users, ["john", :age], &(&1 + 1)) ==
                %{"john" => %{age: 28}, "meg" => %{age: 23}}
 
-      assert_raise FunctionClauseError, fn ->
-        update_in(users, [], fn _ -> %{} end)
-      end
-
       assert_raise ArgumentError, "could not put/update key \"john\" on a nil value", fn ->
         update_in(nil, ["john", :age], fn _ -> %{} end)
       end
@@ -1077,10 +1078,6 @@ defmodule KernelTest do
 
       assert get_and_update_in(map, ["unknown", by_index(3)], &{&1, []}) ==
                {:oops, %{"fruits" => ["banana", "apple", "orange"], "unknown" => []}}
-
-      assert_raise FunctionClauseError, fn ->
-        update_in(users, [], fn _ -> %{} end)
-      end
     end
 
     test "get_and_update_in/2" do
@@ -1110,19 +1107,14 @@ defmodule KernelTest do
     test "pop_in/2" do
       users = %{"john" => %{age: 27}, "meg" => %{age: 23}}
 
-      assert pop_in(users, ["john", :age]) == {27, %{"john" => %{}, "meg" => %{age: 23}}}
+      assert pop_in(users, ["john", :age]) ==
+               {27, %{"john" => %{}, "meg" => %{age: 23}}}
 
-      assert pop_in(users, ["bob", :age]) == {nil, %{"john" => %{age: 27}, "meg" => %{age: 23}}}
+      assert pop_in(users, ["bob", :age]) ==
+               {nil, %{"john" => %{age: 27}, "meg" => %{age: 23}}}
 
-      assert pop_in([], [:foo, :bar]) == {nil, []}
-
-      assert_raise FunctionClauseError, fn ->
-        pop_in(users, [])
-      end
-
-      assert_raise FunctionClauseError, "no function clause matching in Kernel.pop_in/2", fn ->
-        pop_in(users, :not_a_list)
-      end
+      assert pop_in([], [:foo, :bar]) ==
+               {nil, []}
     end
 
     test "pop_in/2 with paths" do
@@ -1144,6 +1136,8 @@ defmodule KernelTest do
     test "pop_in/1" do
       users = %{"john" => %{age: 27}, "meg" => %{age: 23}}
 
+      assert pop_in(users["john"]) == {%{age: 27}, %{"meg" => %{age: 23}}}
+
       assert pop_in(users["john"][:age]) == {27, %{"john" => %{}, "meg" => %{age: 23}}}
       assert pop_in(users["john"][:name]) == {nil, %{"john" => %{age: 27}, "meg" => %{age: 23}}}
       assert pop_in(users["bob"][:age]) == {nil, %{"john" => %{age: 27}, "meg" => %{age: 23}}}
@@ -1366,7 +1360,7 @@ defmodule KernelTest do
     assert is_map_key(Map.new(a: 1), :a) == true
 
     assert_raise BadMapError, fn ->
-      is_map_key(empty_list(), :a)
+      is_map_key(Process.get(:unused, []), :a)
     end
 
     case Map.new(a: 1) do
@@ -1388,7 +1382,7 @@ defmodule KernelTest do
   test "tl/1" do
     assert tl([:one]) == []
     assert tl([1, 2, 3]) == [2, 3]
-    assert_raise ArgumentError, fn -> tl(empty_list()) end
+    assert_raise ArgumentError, fn -> tl(Process.get(:unused, [])) end
 
     assert tl([:a | :b]) == :b
     assert tl([:a, :b | :c]) == [:b | :c]
@@ -1396,7 +1390,7 @@ defmodule KernelTest do
 
   test "hd/1" do
     assert hd([1, 2, 3, 4]) == 1
-    assert_raise ArgumentError, fn -> hd(empty_list()) end
+    assert_raise ArgumentError, fn -> hd(Process.get(:unused, [])) end
     assert hd([1 | 2]) == 1
   end
 
@@ -1475,8 +1469,10 @@ defmodule KernelTest do
       output = capture_io(fn -> assert dbg(List.duplicate(:foo, 3)) == [:foo, :foo, :foo] end)
       assert output =~ "kernel_test.exs"
       assert output =~ "KernelTest"
-      assert output =~ "List.duplicate(:foo, 3)"
+      assert output =~ "List"
+      assert output =~ "duplicate"
       assert output =~ ":foo"
+      assert output =~ "3"
     end
 
     test "doesn't print any colors if :syntax_colors is []" do
@@ -1502,5 +1498,92 @@ defmodule KernelTest do
       assert output =~ "my_var:"
       assert output =~ "my_other_var:"
     end
+
+    test "is not allowed in guards" do
+      message = "invalid expression in guard, dbg is not allowed in guards"
+
+      assert_raise ArgumentError, Regex.compile!(message), fn ->
+        defmodule DbgGuard do
+          def dbg_guard() when dbg(1), do: true
+        end
+      end
+    end
+
+    test "is not allowed in pattern matches" do
+      message = "invalid expression in match, dbg is not allowed in patterns"
+
+      assert_eval_raise(ArgumentError, Regex.compile!(message), """
+      {:ok, dbg()} = make_ref()
+      """)
+    end
+  end
+
+  describe "to_timeout/1" do
+    test "works with keyword lists" do
+      assert to_timeout(hour: 2) == 1000 * 60 * 60 * 2
+      assert to_timeout(minute: 74) == 1000 * 60 * 74
+      assert to_timeout(second: 1293) == 1_293_000
+      assert to_timeout(millisecond: 1_234_123) == 1_234_123
+
+      assert to_timeout(hour: 2, minute: 30) == 1000 * 60 * 60 * 2 + 1000 * 60 * 30
+      assert to_timeout(minute: 30, hour: 2) == 1000 * 60 * 60 * 2 + 1000 * 60 * 30
+      assert to_timeout(minute: 74, second: 30) == 1000 * 60 * 74 + 1000 * 30
+    end
+
+    test "raises on invalid values with keyword lists" do
+      for unit <- [:hour, :minute, :second, :millisecond],
+          value <- [-1, 1.0, :not_an_int] do
+        message =
+          "timeout component #{inspect(unit)} must be a non-negative integer, " <>
+            "got: #{inspect(value)}"
+
+        assert_raise ArgumentError, message, fn -> to_timeout([{unit, value}]) end
+      end
+    end
+
+    test "raises on invalid keys with keyword lists" do
+      message =
+        "timeout component :not_a_unit is not a valid timeout component, valid values are: " <>
+          ":week, :day, :hour, :minute, :second, :millisecond"
+
+      assert_raise ArgumentError, message, fn -> to_timeout(minute: 3, not_a_unit: 1) end
+    end
+
+    test "raises on duplicated components with keyword lists" do
+      assert_raise ArgumentError, "timeout component :minute is duplicated", fn ->
+        to_timeout(minute: 3, hour: 2, minute: 1)
+      end
+    end
+
+    test "works with durations" do
+      assert to_timeout(Duration.new!(hour: 2)) == 1000 * 60 * 60 * 2
+      assert to_timeout(Duration.new!(minute: 74)) == 1000 * 60 * 74
+      assert to_timeout(Duration.new!(second: 1293)) == 1_293_000
+      assert to_timeout(Duration.new!(microsecond: {1_234_123, 4})) == 1_234
+
+      assert to_timeout(Duration.new!(hour: 2, minute: 30)) == 1000 * 60 * 60 * 2 + 1000 * 60 * 30
+      assert to_timeout(Duration.new!(minute: 30, hour: 2)) == 1000 * 60 * 60 * 2 + 1000 * 60 * 30
+      assert to_timeout(Duration.new!(minute: 74, second: 30)) == 1000 * 60 * 74 + 1000 * 30
+    end
+
+    test "raises on durations with non-zero months or days" do
+      message = "duration with a non-zero month cannot be reliably converted to timeouts"
+
+      assert_raise ArgumentError, message, fn ->
+        to_timeout(Duration.new!(month: 3))
+      end
+
+      message = "duration with a non-zero year cannot be reliably converted to timeouts"
+
+      assert_raise ArgumentError, message, fn ->
+        to_timeout(Duration.new!(year: 1))
+      end
+    end
+
+    test "works with timeouts" do
+      assert to_timeout(1_000) == 1_000
+      assert to_timeout(0) == 0
+      assert to_timeout(:infinity) == :infinity
+    end
   end
 end
