@@ -4,13 +4,13 @@ defmodule List.Chars.AtomTest do
   use ExUnit.Case, async: true
 
   test "basic" do
-    assert to_charlist(:foo) == 'foo'
+    assert to_charlist(:foo) == ~c"foo"
   end
 
   test "true false nil" do
-    assert to_charlist(false) == 'false'
-    assert to_charlist(true) == 'true'
-    assert to_charlist(nil) == ''
+    assert to_charlist(false) == ~c"false"
+    assert to_charlist(true) == ~c"true"
+    assert to_charlist(nil) == ~c""
   end
 end
 
@@ -18,7 +18,7 @@ defmodule List.Chars.BitStringTest do
   use ExUnit.Case, async: true
 
   test "basic" do
-    assert to_charlist("foo") == 'foo'
+    assert to_charlist("foo") == ~c"foo"
   end
 end
 
@@ -26,11 +26,11 @@ defmodule List.Chars.NumberTest do
   use ExUnit.Case, async: true
 
   test "integer" do
-    assert to_charlist(1) == '1'
+    assert to_charlist(1) == ~c"1"
   end
 
   test "float" do
-    assert to_charlist(1.0) == '1.0'
+    assert to_charlist(1.0) == ~c"1.0"
   end
 end
 
