@@ -36,25 +36,247 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      modules = compile(files)
+      modules = compile_modules(files)
 
-      assert read_chunk(modules[A]).exports == [
-               {{:c, 0}, %{deprecated_reason: nil, kind: :def}},
-               {{:d, 0}, %{deprecated_reason: nil, kind: :defmacro}},
-               {{:e, 0}, %{deprecated_reason: "oops", kind: :def}}
-             ]
+      assert [
+               {{:c, 0}, %{}},
+               {{:e, 0}, %{deprecated: "oops", sig: {:infer, _}}}
+             ] = read_chunk(modules[A]).exports
 
       assert read_chunk(modules[B]).exports == [
-               {{:behaviour_info, 1}, %{deprecated_reason: nil, kind: :def}}
+               {{:behaviour_info, 1}, %{sig: :none}}
              ]
 
       assert read_chunk(modules[C]).exports == [
-               {{:behaviour_info, 1}, %{deprecated_reason: nil, kind: :def}}
+               {{:behaviour_info, 1}, %{sig: :none}}
              ]
     end
   end
 
-  describe "undefined" do
+  describe "type checking" do
+    test "inferred remote calls" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          def fun(:ok), do: :doki
+          def fun(:error), do: :bad
+        end
+        """,
+        "b.ex" => """
+        defmodule B do
+          def badarg do
+            A.fun(:unknown)
+          end
+
+          def badmatch do
+            :doki = A.fun(:error)
+          end
+        end
+        """
+      }
+
+      warnings = [
+        """
+            warning: incompatible types given to A.fun/1:
+
+                A.fun(:unknown)
+        """,
+        """
+            but expected one of:
+
+                #1
+                dynamic(:ok)
+
+                #2
+                dynamic(:error)
+        """,
+        """
+            warning: the following pattern will never match:
+
+                :doki = A.fun(:error)
+
+            because the right-hand side has type:
+
+                dynamic(:bad)
+        """
+      ]
+
+      assert_warnings(files, warnings)
+    end
+
+    test "mismatched locals" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          def error(), do: private(raise "oops")
+          def public(x), do: private(List.to_tuple(x))
+          defp private(:ok), do: nil
+        end
+        """
+      }
+
+      warnings = [
+        """
+            warning: incompatible types given to private/1:
+
+                private(raise RuntimeError.exception("oops"))
+
+        """,
+        "the 1st argument is empty (often represented as none())",
+        """
+            typing violation found at:
+            │
+          2 │   def error(), do: private(raise "oops")
+            │                    ~
+            │
+            └─ a.ex:2:20: A.error/0
+        """,
+        """
+            warning: incompatible types given to private/1:
+
+                private(List.to_tuple(x))
+        """,
+        """
+            typing violation found at:
+            │
+          3 │   def public(x), do: private(List.to_tuple(x))
+            │                      ~
+            │
+            └─ a.ex:3:22: A.public/1
+        """
+      ]
+
+      assert_warnings(files, warnings)
+    end
+
+    test "unused private clauses" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          def public(x) do
+            private(List.to_tuple(x))
+          end
+
+          defp private(nil), do: nil
+          defp private("foo"), do: "foo"
+          defp private({:ok, ok}), do: ok
+          defp private({:error, error}), do: error
+          defp private("bar"), do: "bar"
+        end
+        """
+      }
+
+      warnings = [
+        """
+            warning: this clause of defp private/1 is never used
+            │
+          6 │   defp private(nil), do: nil
+            │        ~
+            │
+            └─ a.ex:6:8: A.private/1
+        """,
+        """
+            warning: this clause of defp private/1 is never used
+            │
+          7 │   defp private("foo"), do: "foo"
+            │        ~
+            │
+            └─ a.ex:7:8: A.private/1
+        """,
+        """
+            warning: this clause of defp private/1 is never used
+            │
+         10 │   defp private("bar"), do: "bar"
+            │        ~
+            │
+            └─ a.ex:10:8: A.private/1
+        """
+      ]
+
+      assert_warnings(files, warnings)
+    end
+
+    test "unused generated private clauses" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          use B
+          def public(x), do: private(List.to_tuple(x))
+        end
+        """,
+        "b.ex" => """
+        defmodule B do
+          defmacro __using__(_) do
+            quote generated: true do
+              defp private({:ok, ok}), do: ok
+              defp private(:error), do: :error
+            end
+          end
+        end
+        """
+      }
+
+      assert_no_warnings(files)
+    end
+
+    test "unused overridable private clauses" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          use B
+          def public(x), do: private(x)
+          defp private(x), do: super(List.to_tuple(x))
+        end
+        """,
+        "b.ex" => """
+        defmodule B do
+          defmacro __using__(_) do
+            quote do
+              defp private({:ok, ok}), do: ok
+              defp private(:error), do: :error
+              defoverridable private: 1
+            end
+          end
+        end
+        """
+      }
+
+      assert_no_warnings(files)
+    end
+
+    test "returns diagnostics with source and file" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          @file "generated.ex"
+          def fun(arg) do
+            :ok = List.to_tuple(arg)
+          end
+        end
+        """
+      }
+
+      {_modules, warnings} = with_compile_warnings(files)
+
+      assert [
+               %{
+                 message: "the following pattern will never match" <> _,
+                 file: file,
+                 source: source
+               }
+             ] = warnings.runtime_warnings
+
+      assert String.ends_with?(source, "a.ex")
+      assert Path.type(source) == :absolute
+      assert String.ends_with?(file, "generated.ex")
+      assert Path.type(file) == :absolute
+    after
+      :code.delete(A)
+      :code.purge(A)
+    end
+  end
+
+  describe "undefined warnings" do
     test "handles Erlang modules" do
       files = %{
         "a.ex" => """
@@ -65,16 +287,14 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: :not_a_module.no_module/0 is undefined (module :not_a_module is not available or is yet to be defined)
-        a.ex:2: A.a/0
+      warnings = [
+        ":not_a_module.no_module/0 is undefined (module :not_a_module is not available or is yet to be defined)",
+        "a.ex:2:28: A.a/0",
+        ":lists.no_func/0 is undefined or private",
+        "a.ex:3:21: A.b/0"
+      ]
 
-      warning: :lists.no_func/0 is undefined or private
-        a.ex:3: A.b/0
-
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
     test "handles built in functions" do
@@ -90,13 +310,12 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: Kernel.behaviour_info/1 is undefined or private
-        a.ex:6: A.e/0
+      warnings = [
+        "Kernel.behaviour_info/1 is undefined or private",
+        "a.ex:6:21: A.e/0"
+      ]
 
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
     test "handles module body conditionals" do
@@ -124,17 +343,14 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: List.old_flatten/1 is undefined or private. Did you mean:
+      warnings = [
+        "List.old_flatten/1 is undefined or private. Did you mean:",
+        "* flatten/1",
+        "* flatten/2",
+        "a.ex:15:33: A.flatten2/1"
+      ]
 
-            * flatten/1
-            * flatten/2
-
-        a.ex:15: A.flatten2/1
-
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
     test "reports missing functions" do
@@ -150,16 +366,14 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: A.no_func/0 is undefined or private
-        a.ex:2: A.a/0
+      warnings = [
+        "A.no_func/0 is undefined or private",
+        "a.ex:2:16: A.a/0",
+        "A.no_func/1 is undefined or private",
+        "external_source.ex:6:17: A.c/0"
+      ]
 
-      warning: A.no_func/1 is undefined or private
-        external_source.ex:6: A.c/0
-
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
     test "reports missing functions respecting arity" do
@@ -175,22 +389,16 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: A.a/1 is undefined or private. Did you mean:
+      warnings = [
+        "A.a/1 is undefined or private. Did you mean:",
+        "* a/0",
+        "a.ex:3:16: A.b/0",
+        "A.b/1 is undefined or private. Did you mean:",
+        "* b/0",
+        "external_source.ex:6:16: A.c/0"
+      ]
 
-            * a/0
-
-        a.ex:3: A.b/0
-
-      warning: A.b/1 is undefined or private. Did you mean:
-
-            * b/0
-
-        external_source.ex:6: A.c/0
-
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
     test "reports missing modules" do
@@ -207,19 +415,16 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: D.no_module/0 is undefined (module D is not available or is yet to be defined)
-        a.ex:2: A.a/0
+      warnings = [
+        "D.no_module/0 is undefined (module D is not available or is yet to be defined)",
+        "a.ex:2:16: A.a/0",
+        "E.no_module/0 is undefined (module E is not available or is yet to be defined)",
+        "external_source.ex:5:16: A.c/0",
+        "Io.puts/1 is undefined (module Io is not available or is yet to be defined)",
+        "a.ex:7:17: A.i/0"
+      ]
 
-      warning: E.no_module/0 is undefined (module E is not available or is yet to be defined)
-        external_source.ex:5: A.c/0
-
-      warning: Io.puts/1 is undefined (module Io is not available or is yet to be defined)
-        a.ex:7: A.i/0
-
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
     test "reports missing captures" do
@@ -234,19 +439,17 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: A.no_func/0 is undefined or private
-        a.ex:2: A.a/0
+      warnings = [
+        "A.no_func/0 is undefined or private",
+        "a.ex:2:17: A.a/0",
+        "A.no_func/1 is undefined or private",
+        "external_source.ex:5:17: A.c/0"
+      ]
 
-      warning: A.no_func/1 is undefined or private
-        external_source.ex:5: A.c/0
-
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
-    test "doesn't report missing funcs at compile time" do
+    test "doesn't report missing functions at compile time" do
       files = %{
         "a.ex" => """
         Enum.map([], fn _ -> BadReferencer.no_func4() end)
@@ -279,14 +482,10 @@ defmodule Module.Types.IntegrationTest do
       }
 
       warnings = [
-        """
-        warning: B.no_func/0 is undefined or private
-          a.ex:2: A.a/0
-        """,
-        """
-        warning: A.no_func/0 is undefined or private
-          b.ex:2: B.a/0
-        """
+        "B.no_func/0 is undefined or private",
+        "a.ex:2:16: A.a/0",
+        "A.no_func/0 is undefined or private",
+        "b.ex:2:16: B.a/0"
       ]
 
       assert_warnings(files, warnings)
@@ -307,57 +506,16 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: A2.no_func/0 is undefined (module A2 is not available or is yet to be defined)
-      Invalid call found at 2 locations:
-        a.ex:8: A.d/0
-        external_source.ex:5: A.b/0
+      warnings = [
+        "A2.no_func/0 is undefined (module A2 is not available or is yet to be defined)",
+        "└─ a.ex:8:17: A.d/0",
+        "└─ external_source.ex:5:17: A.b/0",
+        "A.no_func/0 is undefined or private",
+        "└─ a.ex:2:16: A.a/0",
+        "└─ a.ex:7:16: A.c/0"
+      ]
 
-      warning: A.no_func/0 is undefined or private
-      Invalid call found at 2 locations:
-        a.ex:2: A.a/0
-        a.ex:7: A.c/0
-
-      """
-
-      assert_warnings(files, warning)
-    end
-
-    test "protocols are checked, ignoring missing built-in impls" do
-      files = %{
-        "a.ex" => """
-        defprotocol AProtocol do
-          def func(arg)
-        end
-
-        defmodule AImplementation do
-          defimpl AProtocol do
-            def func(_), do: B.no_func()
-          end
-        end
-        """
-      }
-
-      warning = """
-      warning: B.no_func/0 is undefined (module B is not available or is yet to be defined)
-        a.ex:7: AProtocol.AImplementation.func/1
-
-      """
-
-      assert_warnings(files, warning)
-    end
-
-    test "handles Erlang ops" do
-      files = %{
-        "a.ex" => """
-        defmodule A do
-          def a(a, b), do: a and b
-          def b(a, b), do: a or b
-        end
-        """
-      }
-
-      assert_no_warnings(files)
+      assert_warnings(files, warnings)
     end
 
     test "hints exclude deprecated functions" do
@@ -374,19 +532,16 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: A.to_list/1 is undefined or private. Did you mean:
+      warnings = [
+        "A.to_list/1 is undefined or private. Did you mean:",
+        "* to_charlist/1",
+        "a.ex:7:19: A.c/1"
+      ]
 
-            * to_charlist/1
-
-        a.ex:7: A.c/1
-
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
-    test "do not warn for module defined in local context" do
+    test "do not warn for module defined in local (runtime) context" do
       files = %{
         "a.ex" => """
         defmodule A do
@@ -417,13 +572,12 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: you must require B before invoking the macro B.b/0
-        ab.ex:2: A.a/0
+      warnings = [
+        "Be sure to require B if you intend to invoke this macro",
+        "ab.ex:2:18: A.a/0"
+      ]
 
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
     test "excludes local no_warn_undefined" do
@@ -449,22 +603,100 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: MissingModule2.func/1 is undefined (module MissingModule2 is not available or is yet to be defined)
-        a.ex:7: A.c/0
+      warnings = [
+        "MissingModule2.func/1 is undefined (module MissingModule2 is not available or is yet to be defined)",
+        "a.ex:7:29: A.c/0",
+        "MissingModule3.func/2 is undefined (module MissingModule3 is not available or is yet to be defined)",
+        "a.ex:8:29: A.d/0",
+        "B.func/3 is undefined or private. Did you mean:",
+        "* func/1",
+        "a.ex:11:16: A.g/0"
+      ]
 
-      warning: MissingModule3.func/2 is undefined (module MissingModule3 is not available or is yet to be defined)
-        a.ex:8: A.d/0
+      assert_warnings(files, warnings)
+    end
 
-      warning: B.func/3 is undefined or private. Did you mean:
+    test "warn for external nested module" do
+      files = %{
+        "a.ex" => """
+        defmodule A.B do
+          def a, do: :ok
+        end
+        defmodule A do
+          alias A.B
+          def a, do: B.a()
+          def b, do: B.a(1)
+          def c, do: B.no_func()
+        end
+        """
+      }
 
-            * func/1
+      warnings = [
+        "A.B.a/1 is undefined or private. Did you mean:",
+        "* a/0",
+        " def b, do: B.a(1)",
+        "a.ex:7:16: A.b/0",
+        "A.B.no_func/0 is undefined or private",
+        "def c, do: B.no_func()",
+        "a.ex:8:16: A.c/0"
+      ]
 
-        a.ex:11: A.g/0
+      assert_warnings(files, warnings)
+    end
 
-      """
+    test "warn for compile time context module defined before calls" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          defmodule B do
+            def a, do: :ok
+          end
+          def a, do: B.a()
+          def b, do: B.a(1)
+          def c, do: B.no_func()
+        end
+        """
+      }
 
-      assert_warnings(files, warning)
+      warnings = [
+        "A.B.a/1 is undefined or private. Did you mean:",
+        "* a/0",
+        " def b, do: B.a(1)",
+        "a.ex:6:16: A.b/0",
+        "A.B.no_func/0 is undefined or private",
+        "def c, do: B.no_func()",
+        "a.ex:7:16: A.c/0"
+      ]
+
+      assert_warnings(files, warnings)
+    end
+
+    test "warn for compile time context module defined after calls and aliased" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          alias A.B
+          def a, do: B.a()
+          def b, do: B.a(1)
+          def c, do: B.no_func()
+          defmodule B do
+            def a, do: :ok
+          end
+        end
+        """
+      }
+
+      warnings = [
+        "A.B.a/1 is undefined or private. Did you mean:",
+        "* a/0",
+        " def b, do: B.a(1)",
+        "a.ex:4:16: A.b/0",
+        "A.B.no_func/0 is undefined or private",
+        "def c, do: B.no_func()",
+        "a.ex:5:16: A.c/0"
+      ]
+
+      assert_warnings(files, warnings)
     end
 
     test "excludes global no_warn_undefined" do
@@ -497,22 +729,17 @@ defmodule Module.Types.IntegrationTest do
           """
         }
 
-        warning = """
-        warning: MissingModule2.func/1 is undefined (module MissingModule2 is not available or is yet to be defined)
-          a.ex:7: A.c/0
+        warnings = [
+          "MissingModule2.func/1 is undefined (module MissingModule2 is not available or is yet to be defined)",
+          "a.ex:7:29: A.c/0",
+          "MissingModule3.func/2 is undefined (module MissingModule3 is not available or is yet to be defined)",
+          "a.ex:8:29: A.d/0",
+          "B.func/3 is undefined or private. Did you mean:",
+          "* func/1",
+          "a.ex:11:16: A.g/0"
+        ]
 
-        warning: MissingModule3.func/2 is undefined (module MissingModule3 is not available or is yet to be defined)
-          a.ex:8: A.d/0
-
-        warning: B.func/3 is undefined or private. Did you mean:
-
-              * func/1
-
-          a.ex:11: A.g/0
-
-        """
-
-        assert_warnings(files, warning)
+        assert_warnings(files, warnings)
       after
         Code.compiler_options(no_warn_undefined: no_warn_undefined)
       end
@@ -562,6 +789,41 @@ defmodule Module.Types.IntegrationTest do
     end
   end
 
+  describe "regressions" do
+    test "does not emit false positives from defguard" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          defguard is_non_nil_arity_function(fun, arity)
+                   when arity != nil and is_function(fun, arity)
+
+          def check(fun, args) do
+            is_non_nil_arity_function(fun, length(args))
+          end
+        end
+        """
+      }
+
+      assert_no_warnings(files)
+    end
+
+    test "do not parse binary segments as variables" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          def decode(byte) do
+            case byte do
+              enc when enc in [<<0x00>>, <<0x01>>] -> :ok
+            end
+          end
+        end
+        """
+      }
+
+      assert_no_warnings(files)
+    end
+  end
+
   describe "after_verify" do
     test "reports functions" do
       files = %{
@@ -576,10 +838,10 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: from after_verify
-
-      """
+      warning = [
+        "warning: ",
+        "from after_verify"
+      ]
 
       assert_warnings(files, warning)
     end
@@ -596,13 +858,12 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: A.a/0 is deprecated. oops
-        a.ex:3: A.a/0
+      warnings = [
+        "A.a/0 is deprecated. oops",
+        "a.ex:3:16: A.a/0"
+      ]
 
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
     test "reports imported functions" do
@@ -621,13 +882,12 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: A.a/0 is deprecated. oops
-        b.ex:3: B.b/0
+      warnings = [
+        "A.a/0 is deprecated. oops",
+        "b.ex:3:14: B.b/0"
+      ]
 
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
     test "reports structs" do
@@ -649,19 +909,12 @@ defmodule Module.Types.IntegrationTest do
       }
 
       warnings = [
-        """
-        warning: A.__struct__/0 is deprecated. oops
-        Invalid call found at 2 locations:
-          a.ex:4: A.match/1
-          a.ex:5: A.build/1
-        """,
-        """
-        warning: A.__struct__/0 is deprecated. oops
-        Invalid call found at 2 locations:
-          b.ex:2: B.match/1
-          b.ex:3: B.build/1
-
-        """
+        "A.__struct__/0 is deprecated. oops",
+        "└─ a.ex:4:13: A.match/1",
+        "└─ a.ex:5:23: A.build/1",
+        "A.__struct__/0 is deprecated. oops",
+        "└─ b.ex:2:13: B.match/1",
+        "└─ b.ex:3:23: B.build/1"
       ]
 
       assert_warnings(files, warnings)
@@ -683,13 +936,12 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: A.a/0 is deprecated. oops
-        b.ex:3: B
+      warnings = [
+        "A.a/0 is deprecated. oops",
+        "b.ex:3:5: B (module)"
+      ]
 
-      """
-
-      assert_warnings(files, warning)
+      assert_warnings(files, warnings)
     end
 
     test "reports macro" do
@@ -708,13 +960,35 @@ defmodule Module.Types.IntegrationTest do
         """
       }
 
-      warning = """
-      warning: A.a/0 is deprecated. oops
-        b.ex:3: B.b/0
+      warnings = [
+        "A.a/0 is deprecated. oops",
+        "b.ex:3:16: B.b/0"
+      ]
 
-      """
+      assert_warnings(files, warnings)
+    end
 
-      assert_warnings(files, warning)
+    test "reports unquote functions" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          @deprecated "oops"
+          def a, do: :ok
+        end
+        """,
+        "b.ex" => """
+        defmodule B do
+          def b, do: unquote(&A.a/0)
+        end
+        """
+      }
+
+      warnings = [
+        "A.a/0 is deprecated. oops",
+        "b.ex: B.b/0"
+      ]
+
+      assert_warnings(files, warnings)
     end
   end
 
@@ -737,28 +1011,41 @@ defmodule Module.Types.IntegrationTest do
   defp capture_compile_warnings(files) do
     in_tmp(fn ->
       paths = generate_files(files)
-      capture_io(:stderr, fn -> compile_files(paths) end)
+      capture_io(:stderr, fn -> compile_to_path(paths) end)
     end)
   end
 
-  defp compile(files) do
+  defp with_compile_warnings(files) do
     in_tmp(fn ->
       paths = generate_files(files)
-      compile_files(paths)
+      with_io(:stderr, fn -> compile_to_path(paths) end) |> elem(0)
     end)
   end
 
-  defp compile_files(paths) do
-    {:ok, modules, _warnings} = Kernel.ParallelCompiler.compile_to_path(paths, ".")
+  defp compile_modules(files) do
+    in_tmp(fn ->
+      paths = generate_files(files)
+      {modules, _warnings} = compile_to_path(paths)
 
-    Map.new(modules, fn module ->
-      {^module, binary, _filename} = :code.get_object_code(module)
-      :code.purge(module)
-      :code.delete(module)
-      {module, binary}
+      Map.new(modules, fn module ->
+        {^module, binary, _filename} = :code.get_object_code(module)
+        {module, binary}
+      end)
     end)
   end
 
+  defp compile_to_path(paths) do
+    {:ok, modules, warnings} =
+      Kernel.ParallelCompiler.compile_to_path(paths, ".", return_diagnostics: true)
+
+    for module <- modules do
+      :code.delete(module)
+      :code.purge(module)
+    end
+
+    {modules, warnings}
+  end
+
   defp generate_files(files) do
     for {file, contents} <- files do
       File.write!(file, contents)
@@ -767,7 +1054,7 @@ defmodule Module.Types.IntegrationTest do
   end
 
   defp read_chunk(binary) do
-    assert {:ok, {_module, [{'ExCk', chunk}]}} = :beam_lib.chunks(binary, ['ExCk'])
+    assert {:ok, {_module, [{~c"ExCk", chunk}]}} = :beam_lib.chunks(binary, [~c"ExCk"])
     assert {:elixir_checker_v1, map} = :erlang.binary_to_term(chunk)
     map
   end
