@@ -109,6 +109,15 @@ defmodule OptionParserTest do
     end
   end
 
+  test "parse!/2 raises an exception for an unknown option using strict when it is only off by underscores" do
+    msg = "1 error found!\n--docs_bar : Unknown option. Did you mean --docs-bar?"
+
+    assert_raise OptionParser.ParseError, msg, fn ->
+      argv = ["--source", "from_docs/", "--docs_bar", "show"]
+      OptionParser.parse!(argv, strict: [source: :string, docs_bar: :string])
+    end
+  end
+
   test "parse!/2 raises an exception when an option is of the wrong type" do
     assert_raise OptionParser.ParseError, fn ->
       argv = ["--bad", "opt", "foo", "-o", "bad", "bar"]
@@ -148,6 +157,23 @@ defmodule OptionParserTest do
              ) == {[source: "foo"], ["bar", "--", "-x"], []}
     end
 
+    test "return separators" do
+      assert OptionParser.parse_head(["--", "foo"],
+               switches: [],
+               return_separator: true
+             ) == {[], ["--", "foo"], []}
+
+      assert OptionParser.parse_head(["--no-halt", "--", "foo"],
+               switches: [halt: :boolean],
+               return_separator: true
+             ) == {[halt: false], ["--", "foo"], []}
+
+      assert OptionParser.parse_head(["foo.exs", "--no-halt", "--", "foo"],
+               switches: [halt: :boolean],
+               return_separator: true
+             ) == {[], ["foo.exs", "--no-halt", "--", "foo"], []}
+    end
+
     test "parses - as argument" do
       argv = ["--foo", "-", "-b", "-"]
       opts = [strict: [foo: :boolean, boo: :string], aliases: [b: :boo]]
