@@ -14,8 +14,8 @@ defmodule TypespecTest do
           unquote(block)
         end
 
-      :code.delete(TypespecSample)
       :code.purge(TypespecSample)
+      :code.delete(TypespecSample)
       bytecode
     end
   end
@@ -46,7 +46,7 @@ defmodule TypespecTest do
 
   describe "Kernel.Typespec errors" do
     test "invalid type specification" do
-      assert_raise CompileError, ~r"invalid type specification: my_type = 1", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid type specification: my_type = 1", fn ->
         test_module do
           @type my_type = 1
         end
@@ -54,21 +54,37 @@ defmodule TypespecTest do
     end
 
     test "unexpected expression in typespec" do
-      assert_raise CompileError, ~r"unexpected expression in typespec: \"foobar\"", fn ->
+      assert_raise Kernel.TypespecError, ~r"unexpected expression in typespec: \"foobar\"", fn ->
         test_module do
           @type my_type :: "foobar"
         end
       end
+
+      assert_raise Kernel.TypespecError,
+                   ~r"unexpected expression in typespec: integer\(\)\(\)",
+                   fn ->
+                     test_module do
+                       @type my_type :: integer()()
+                     end
+                   end
+
+      assert_raise Kernel.TypespecError,
+                   ~r"unexpected expression in typespec: %URI\.t\(\)\{\}",
+                   fn ->
+                     test_module do
+                       @type my_type :: %URI.t(){}
+                     end
+                   end
     end
 
     test "invalid function specification" do
-      assert_raise CompileError, ~r"invalid type specification: \"not a spec\"", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid type specification: \"not a spec\"", fn ->
         test_module do
           @spec "not a spec"
         end
       end
 
-      assert_raise CompileError, ~r"invalid type specification: 1 :: 2", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid type specification: 1 :: 2", fn ->
         test_module do
           @spec 1 :: 2
         end
@@ -76,26 +92,26 @@ defmodule TypespecTest do
     end
 
     test "undefined type" do
-      assert_raise CompileError, ~r"type foo/0 undefined", fn ->
+      assert_raise Kernel.TypespecError, ~r"type foo/0 undefined", fn ->
         test_module do
           @type omg :: foo
         end
       end
 
-      assert_raise CompileError, ~r"type foo/2 undefined", fn ->
+      assert_raise Kernel.TypespecError, ~r"type foo/2 undefined", fn ->
         test_module do
           @type omg :: foo(atom, integer)
         end
       end
 
-      assert_raise CompileError, ~r"type bar/0 undefined", fn ->
+      assert_raise Kernel.TypespecError, ~r"type bar/0 undefined", fn ->
         test_module do
           @spec foo(bar, integer) :: {atom, integer}
           def foo(var1, var2), do: {var1, var2}
         end
       end
 
-      assert_raise CompileError, ~r"type foo/0 undefined", fn ->
+      assert_raise Kernel.TypespecError, ~r"type foo/0 undefined", fn ->
         test_module do
           @type omg :: __MODULE__.foo()
         end
@@ -103,8 +119,8 @@ defmodule TypespecTest do
     end
 
     test "redefined type" do
-      assert_raise CompileError,
-                   ~r"type foo/0 is already defined in test/elixir/typespec_test.exs:110",
+      assert_raise Kernel.TypespecError,
+                   ~r"type foo/0 is already defined in .*test/elixir/typespec_test.exs:126",
                    fn ->
                      test_module do
                        @type foo :: atom
@@ -112,8 +128,8 @@ defmodule TypespecTest do
                      end
                    end
 
-      assert_raise CompileError,
-                   ~r"type foo/2 is already defined in test/elixir/typespec_test.exs:120",
+      assert_raise Kernel.TypespecError,
+                   ~r"type foo/2 is already defined in .*test/elixir/typespec_test.exs:136",
                    fn ->
                      test_module do
                        @type foo :: atom
@@ -122,8 +138,8 @@ defmodule TypespecTest do
                      end
                    end
 
-      assert_raise CompileError,
-                   ~r"type foo/0 is already defined in test/elixir/typespec_test.exs:129",
+      assert_raise Kernel.TypespecError,
+                   ~r"type foo/0 is already defined in .*test/elixir/typespec_test.exs:145",
                    fn ->
                      test_module do
                        @type foo :: atom
@@ -133,7 +149,7 @@ defmodule TypespecTest do
     end
 
     test "type variable unused (singleton type variable)" do
-      assert_raise CompileError, ~r"type variable x is used only once", fn ->
+      assert_raise Kernel.TypespecError, ~r"type variable x is used only once", fn ->
         test_module do
           @type foo(x) :: integer
         end
@@ -147,13 +163,13 @@ defmodule TypespecTest do
     end
 
     test "type variable named _" do
-      assert_raise CompileError, ~r"type variable '_' is invalid", fn ->
+      assert_raise Kernel.TypespecError, ~r"type variable '_' is invalid", fn ->
         test_module do
           @type foo(_) :: integer
         end
       end
 
-      assert_raise CompileError, ~r"type variable '_' is invalid", fn ->
+      assert_raise Kernel.TypespecError, ~r"type variable '_' is invalid", fn ->
         test_module do
           @type foo(_, _) :: integer
         end
@@ -161,15 +177,15 @@ defmodule TypespecTest do
     end
 
     test "spec for undefined function" do
-      assert_raise CompileError, ~r"spec for undefined function omg/0", fn ->
+      assert_compile_error(~r"spec for undefined function omg/0", fn ->
         test_module do
           @spec omg :: atom
         end
-      end
+      end)
     end
 
     test "spec variable used only once (singleton type variable)" do
-      assert_raise CompileError, ~r"type variable x is used only once", fn ->
+      assert_raise Kernel.TypespecError, ~r"type variable x is used only once", fn ->
         test_module do
           @spec foo(x, integer) :: integer when x: var
           def foo(x, y), do: x + y
@@ -177,51 +193,60 @@ defmodule TypespecTest do
       end
     end
 
-    test "invalid optional callback" do
-      assert_raise CompileError, ~r"invalid optional callback :foo", fn ->
+    test "spec with ... outside of fn and lists" do
+      assert_raise Kernel.TypespecError, ~r"... in typespecs is only allowed inside lists", fn ->
         test_module do
-          @optional_callbacks :foo
+          @spec foo(...) :: :ok
+          def foo(x), do: x
         end
       end
     end
 
+    test "invalid optional callback" do
+      assert_compile_error(~r"invalid optional callback :foo", fn ->
+        test_module do
+          @optional_callbacks :foo
+        end
+      end)
+    end
+
     test "unknown optional callback" do
-      assert_raise CompileError, ~r"unknown callback foo/1 given as optional callback", fn ->
+      assert_compile_error(~r"unknown callback foo/1 given as optional callback", fn ->
         test_module do
           @optional_callbacks foo: 1
         end
-      end
+      end)
     end
 
     test "repeated optional callback" do
       message = ~r"foo/1 has been specified as optional callback more than once"
 
-      assert_raise CompileError, message, fn ->
+      assert_compile_error(message, fn ->
         test_module do
           @callback foo(:ok) :: :ok
           @optional_callbacks foo: 1, foo: 1
         end
-      end
+      end)
     end
 
     test "behaviour_info/1 explicitly defined alongside @callback/@macrocallback" do
       message = ~r"cannot define @callback attribute for foo/1 when behaviour_info/1"
 
-      assert_raise CompileError, message, fn ->
+      assert_compile_error(message, fn ->
         test_module do
           @callback foo(:ok) :: :ok
           def behaviour_info(_), do: []
         end
-      end
+      end)
 
       message = ~r"cannot define @macrocallback attribute for foo/1 when behaviour_info/1"
 
-      assert_raise CompileError, message, fn ->
+      assert_compile_error(message, fn ->
         test_module do
           @macrocallback foo(:ok) :: :ok
           def behaviour_info(_), do: []
         end
-      end
+      end)
     end
 
     test "default is not supported" do
@@ -265,7 +290,7 @@ defmodule TypespecTest do
     test "@spec shows readable error message when return type is missing" do
       message = ~r"type specification missing return type: my_fun\(integer\)"
 
-      assert_raise CompileError, message, fn ->
+      assert_raise Kernel.TypespecError, message, fn ->
         test_module do
           @spec my_fun(integer)
         end
@@ -373,43 +398,43 @@ defmodule TypespecTest do
     end
 
     test "@type with invalid binary spec" do
-      assert_raise CompileError, ~r"invalid binary specification", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid binary specification", fn ->
         test_module do
           @type my_type :: <<_::3*8>>
         end
       end
 
-      assert_raise CompileError, ~r"invalid binary specification", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid binary specification", fn ->
         test_module do
-          @type my_type :: <<_::atom>>
+          @type my_type :: <<_::atom()>>
         end
       end
 
-      assert_raise CompileError, ~r"invalid binary specification", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid binary specification", fn ->
         test_module do
           @type my_type :: <<_::integer>>
         end
       end
 
-      assert_raise CompileError, ~r"invalid binary specification", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid binary specification", fn ->
         test_module do
           @type my_type :: <<_::(-4)>>
         end
       end
 
-      assert_raise CompileError, ~r"invalid binary specification", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid binary specification", fn ->
         test_module do
           @type my_type :: <<_::3, _::_*atom>>
         end
       end
 
-      assert_raise CompileError, ~r"invalid binary specification", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid binary specification", fn ->
         test_module do
           @type my_type :: <<_::3, _::_*(-8)>>
         end
       end
 
-      assert_raise CompileError, ~r"invalid binary specification", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid binary specification", fn ->
         test_module do
           @type my_type :: <<_::3, _::_*257>>
         end
@@ -433,7 +458,7 @@ defmodule TypespecTest do
     end
 
     test "@type with invalid range" do
-      assert_raise CompileError, ~r"invalid range specification", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid range specification", fn ->
         test_module do
           @type my_type :: atom..10
         end
@@ -522,13 +547,13 @@ defmodule TypespecTest do
     end
 
     test "@type with undefined struct" do
-      assert_raise CompileError, ~r"ThisModuleDoesNotExist.__struct__/0 is undefined", fn ->
+      assert_raise ArgumentError, ~r"ThisModuleDoesNotExist.__struct__/1 is undefined", fn ->
         test_module do
           @type my_type :: %ThisModuleDoesNotExist{}
         end
       end
 
-      assert_raise CompileError, ~r"cannot access struct TypespecTest.TypespecSample", fn ->
+      assert_raise ArgumentError, ~r"cannot access struct TypespecTest.TypespecSample", fn ->
         test_module do
           @type my_type :: %TypespecSample{}
         end
@@ -536,7 +561,7 @@ defmodule TypespecTest do
     end
 
     test "@type with a struct with undefined field" do
-      assert_raise CompileError,
+      assert_raise Kernel.TypespecError,
                    ~r"undefined field :no_field on struct TypespecTest.TypespecSample",
                    fn ->
                      test_module do
@@ -545,7 +570,7 @@ defmodule TypespecTest do
                      end
                    end
 
-      assert_raise CompileError,
+      assert_raise Kernel.TypespecError,
                    ~r"undefined field :no_field on struct TypespecTest.TypespecSample",
                    fn ->
                      test_module do
@@ -556,7 +581,7 @@ defmodule TypespecTest do
     end
 
     test "@type when overriding Elixir built-in" do
-      assert_raise CompileError, ~r"type struct/0 is a built-in type", fn ->
+      assert_raise Kernel.TypespecError, ~r"type struct/0 is a built-in type", fn ->
         test_module do
           @type struct :: :oops
         end
@@ -564,7 +589,7 @@ defmodule TypespecTest do
     end
 
     test "@type when overriding Erlang built-in" do
-      assert_raise CompileError, ~r"type list/0 is a built-in type", fn ->
+      assert_raise Kernel.TypespecError, ~r"type list/0 is a built-in type", fn ->
         test_module do
           @type list :: :oops
         end
@@ -620,7 +645,7 @@ defmodule TypespecTest do
     end
 
     test "@type with undefined record" do
-      assert_raise CompileError, ~r"unknown record :this_record_does_not_exist", fn ->
+      assert_raise Kernel.TypespecError, ~r"unknown record :this_record_does_not_exist", fn ->
         test_module do
           @type my_type :: record(:this_record_does_not_exist, [])
         end
@@ -628,7 +653,7 @@ defmodule TypespecTest do
     end
 
     test "@type with a record with undefined field" do
-      assert_raise CompileError, ~r"undefined field no_field on record :timestamp", fn ->
+      assert_raise Kernel.TypespecError, ~r"undefined field no_field on record :timestamp", fn ->
         test_module do
           require Record
           Record.defrecord(:timestamp, date: 1, time: 2)
@@ -638,7 +663,7 @@ defmodule TypespecTest do
     end
 
     test "@type with a record which declares the name as the type `atom` rather than an atom literal" do
-      assert_raise CompileError, ~r"expected the record name to be an atom literal", fn ->
+      assert_raise Kernel.TypespecError, ~r"expected the record name to be an atom literal", fn ->
         test_module do
           @type my_type :: record(atom, field: :foo)
         end
@@ -657,7 +682,7 @@ defmodule TypespecTest do
     end
 
     test "@type with an invalid map notation" do
-      assert_raise CompileError, ~r"invalid map specification", fn ->
+      assert_raise Kernel.TypespecError, ~r"invalid map specification", fn ->
         test_module do
           @type content :: %{atom | String.t() => term}
         end
@@ -685,7 +710,8 @@ defmodule TypespecTest do
           @type my_type :: (... -> any)
         end
 
-      assert [type: {:my_type, {:type, _, :fun, []}, []}] = types(bytecode)
+      assert [type: {:my_type, {:type, _, :fun, [{:type, _, :any}, {:type, _, :any, []}]}, []}] =
+               types(bytecode)
     end
 
     test "@type with a fun with multiple arguments and return type" do
@@ -703,7 +729,7 @@ defmodule TypespecTest do
     test "@type with a fun with no arguments and return type" do
       bytecode =
         test_module do
-          @type my_type :: (() -> integer)
+          @type my_type :: (-> integer)
         end
 
       assert [type: {:my_type, type, []}] = types(bytecode)
@@ -793,7 +819,7 @@ defmodule TypespecTest do
 
       bytecode =
         test_module do
-          Module.eval_quoted(__MODULE__, quoted)
+          Code.eval_quoted(quoted, [], module: __MODULE__)
         end
 
       assert [type: {:my_type, {:atom, _, :foo}, []}] = types(bytecode)
@@ -816,7 +842,7 @@ defmodule TypespecTest do
     end
 
     test "@type with a reserved signature" do
-      assert_raise CompileError,
+      assert_raise Kernel.TypespecError,
                    ~r"type required\/1 is a reserved type and it cannot be defined",
                    fn ->
                      test_module do
@@ -824,7 +850,7 @@ defmodule TypespecTest do
                      end
                    end
 
-      assert_raise CompileError,
+      assert_raise Kernel.TypespecError,
                    ~r"type optional\/1 is a reserved type and it cannot be defined",
                    fn ->
                      test_module do
@@ -832,7 +858,7 @@ defmodule TypespecTest do
                      end
                    end
 
-      assert_raise CompileError,
+      assert_raise Kernel.TypespecError,
                    ~r"type required\/1 is a reserved type and it cannot be defined",
                    fn ->
                      test_module do
@@ -840,7 +866,7 @@ defmodule TypespecTest do
                      end
                    end
 
-      assert_raise CompileError,
+      assert_raise Kernel.TypespecError,
                    ~r"type optional\/1 is a reserved type and it cannot be defined",
                    fn ->
                      test_module do
@@ -848,7 +874,7 @@ defmodule TypespecTest do
                      end
                    end
 
-      assert_raise CompileError,
+      assert_raise Kernel.TypespecError,
                    ~r"type required\/1 is a reserved type and it cannot be defined",
                    fn ->
                      test_module do
@@ -856,7 +882,7 @@ defmodule TypespecTest do
                      end
                    end
 
-      assert_raise CompileError,
+      assert_raise Kernel.TypespecError,
                    ~r"type optional\/1 is a reserved type and it cannot be defined",
                    fn ->
                      test_module do
@@ -866,7 +892,7 @@ defmodule TypespecTest do
     end
 
     test "invalid remote @type with module attribute that does not evaluate to a module" do
-      assert_raise CompileError, ~r/\(@foo is "bar"\)/, fn ->
+      assert_raise Kernel.TypespecError, ~r/\(@foo is "bar"\)/, fn ->
         test_module do
           @foo "bar"
           @type t :: @foo.t
@@ -1045,8 +1071,8 @@ defmodule TypespecTest do
                 {TypeModuleAttributes, _}}
              ] = TypeModuleAttributes.typep2()
     after
-      :code.delete(TypeModuleAttributes)
       :code.purge(TypeModuleAttributes)
+      :code.delete(TypeModuleAttributes)
     end
 
     test "@spec, @callback, and @macrocallback as module attributes" do
@@ -1104,8 +1130,8 @@ defmodule TypespecTest do
                 {SpecModuleAttributes, _}}
              ] = SpecModuleAttributes.macrocallback()
     after
-      :code.delete(SpecModuleAttributes)
       :code.purge(SpecModuleAttributes)
+      :code.delete(SpecModuleAttributes)
     end
 
     test "@callback" do
@@ -1140,7 +1166,7 @@ defmodule TypespecTest do
     test "block handling" do
       bytecode =
         test_module do
-          @spec foo((() -> [integer])) :: integer
+          @spec foo((-> [integer])) :: integer
           def foo(_), do: 1
         end
 
@@ -1171,9 +1197,7 @@ defmodule TypespecTest do
           quote(do: @type(binary_type2() :: <<_::3>>)),
           quote(do: @type(binary_type3() :: <<_::3, _::_*8>>)),
           quote(do: @type(tuple_type() :: {integer()})),
-          quote(
-            do: @type(ftype() :: (() -> any()) | (() -> integer()) | (integer() -> integer()))
-          ),
+          quote(do: @type(ftype() :: (-> any()) | (-> integer()) | (integer() -> integer()))),
           quote(do: @type(cl() :: charlist())),
           quote(do: @type(st() :: struct())),
           quote(do: @type(ab() :: as_boolean(term()))),
@@ -1200,7 +1224,7 @@ defmodule TypespecTest do
 
       bytecode =
         test_module do
-          Module.eval_quoted(__MODULE__, quoted)
+          Code.eval_quoted(quoted, [], module: __MODULE__)
         end
 
       types = types(bytecode)
@@ -1242,7 +1266,7 @@ defmodule TypespecTest do
 
           def foo(), do: 1
           def foo(arg), do: arg
-          Module.eval_quoted(__MODULE__, quote(do: (unquote_splicing(quoted))))
+          Code.eval_quoted(quote(do: (unquote_splicing(quoted))), [], module: __MODULE__)
         end
 
       specs =
@@ -1260,32 +1284,32 @@ defmodule TypespecTest do
     end
 
     test "spec_to_quoted with maps with __struct__ key" do
-      defmodule A do
+      defmodule StructA do
         defstruct [:key]
       end
 
-      defmodule B do
+      defmodule StructB do
         defstruct [:key]
       end
 
       bytecode =
         test_module do
-          @spec single_struct(%A{}) :: :ok
+          @spec single_struct(%StructA{}) :: :ok
           def single_struct(arg), do: {:ok, arg}
 
-          @spec single_struct_key(%{__struct__: A}) :: :ok
+          @spec single_struct_key(%{__struct__: StructA}) :: :ok
           def single_struct_key(arg), do: {:ok, arg}
 
           @spec single_struct_key_type(%{__struct__: atom()}) :: :ok
           def single_struct_key_type(arg), do: {:ok, arg}
 
-          @spec union_struct(%A{} | %B{}) :: :ok
+          @spec union_struct(%StructA{} | %StructB{}) :: :ok
           def union_struct(arg), do: {:ok, arg}
 
-          @spec union_struct_key(%{__struct__: A | B}) :: :ok
+          @spec union_struct_key(%{__struct__: StructA | StructB}) :: :ok
           def union_struct_key(arg), do: {:ok, arg}
 
-          @spec union_struct_key_type(%{__struct__: atom() | A | binary()}) :: :ok
+          @spec union_struct_key_type(%{__struct__: atom() | StructA | binary()}) :: :ok
           def union_struct_key_type(arg), do: {:ok, arg}
         end
 
@@ -1300,32 +1324,32 @@ defmodule TypespecTest do
 
       assert Code.Typespec.spec_to_quoted(:single_struct, ast_single_struct)
              |> Macro.to_string() ==
-               "single_struct(%TypespecTest.A{key: term()}) :: :ok"
+               "single_struct(%TypespecTest.StructA{key: term()}) :: :ok"
 
       assert Code.Typespec.spec_to_quoted(:single_struct_key, ast_single_struct_key)
              |> Macro.to_string() ==
-               "single_struct_key(%TypespecTest.A{}) :: :ok"
+               "single_struct_key(%TypespecTest.StructA{}) :: :ok"
 
       assert Code.Typespec.spec_to_quoted(:single_struct_key_type, ast_single_struct_key_type)
              |> Macro.to_string() ==
                "single_struct_key_type(%{__struct__: atom()}) :: :ok"
 
       assert Code.Typespec.spec_to_quoted(:union_struct, ast_union_struct) |> Macro.to_string() ==
-               "union_struct(%TypespecTest.A{key: term()} | %TypespecTest.B{key: term()}) :: :ok"
+               "union_struct(%TypespecTest.StructA{key: term()} | %TypespecTest.StructB{key: term()}) :: :ok"
 
       assert Code.Typespec.spec_to_quoted(:union_struct_key, ast_union_struct_key)
              |> Macro.to_string() ==
-               "union_struct_key(%{__struct__: TypespecTest.A | TypespecTest.B}) :: :ok"
+               "union_struct_key(%{__struct__: TypespecTest.StructA | TypespecTest.StructB}) :: :ok"
 
       assert Code.Typespec.spec_to_quoted(:union_struct_key_type, ast_union_struct_key_type)
              |> Macro.to_string() ==
-               "union_struct_key_type(%{__struct__: atom() | TypespecTest.A | binary()}) :: :ok"
+               "union_struct_key_type(%{__struct__: atom() | TypespecTest.StructA | binary()}) :: :ok"
     end
 
     test "non-variables are given as arguments" do
       msg = ~r/The type one_bad_variable\/1 has an invalid argument\(s\): String.t\(\)/
 
-      assert_raise CompileError, msg, fn ->
+      assert_raise Kernel.TypespecError, msg, fn ->
         test_module do
           @type one_bad_variable(String.t()) :: String.t()
         end
@@ -1333,7 +1357,7 @@ defmodule TypespecTest do
 
       msg = ~r/The type two_bad_variables\/2 has an invalid argument\(s\): :ok, Enumerable.t\(\)/
 
-      assert_raise CompileError, msg, fn ->
+      assert_raise Kernel.TypespecError, msg, fn ->
         test_module do
           @type two_bad_variables(:ok, Enumerable.t()) :: {:ok, []}
         end
@@ -1341,7 +1365,7 @@ defmodule TypespecTest do
 
       msg = ~r/The type one_bad_one_good\/2 has an invalid argument\(s\): \"\"/
 
-      assert_raise CompileError, msg, fn ->
+      assert_raise Kernel.TypespecError, msg, fn ->
         test_module do
           @type one_bad_one_good(input1, "") :: {:ok, input1}
         end
@@ -1402,7 +1426,7 @@ defmodule TypespecTest do
           quote(do: @type(literal_unit_1() :: <<_::_*1>>)),
           quote(do: @type(literal_size_1_unit_8() :: <<_::100, _::_*256>>)),
           quote(do: @type(literal_function_arity_any() :: (... -> integer()))),
-          quote(do: @type(literal_function_arity_0() :: (() -> integer()))),
+          quote(do: @type(literal_function_arity_0() :: (-> integer()))),
           quote(do: @type(literal_function_arity_2() :: (integer(), atom() -> integer()))),
           quote(do: @type(literal_list_type() :: [integer()])),
           quote(do: @type(literal_empty_list() :: [])),
@@ -1463,7 +1487,7 @@ defmodule TypespecTest do
 
       bytecode =
         test_module do
-          Module.eval_quoted(__MODULE__, quoted)
+          Code.eval_quoted(quoted, [], module: __MODULE__)
         end
 
       types = types(bytecode)
@@ -1497,9 +1521,6 @@ defmodule TypespecTest do
             assert ast_string ==
                      "@type literal_struct_all_fields_key_type() :: %TypespecTest.SomeStruct{key: integer()}"
 
-          {:built_in_fun, _, _} ->
-            assert ast_string == "@type built_in_fun() :: (... -> any())"
-
           {:built_in_nonempty_list, _, _} ->
             assert ast_string == "@type built_in_nonempty_list() :: [...]"
 
@@ -1559,23 +1580,29 @@ defmodule TypespecTest do
     """)
 
     [type: type] = types(:typespec_test_mod)
-    line = 5
 
     assert Code.Typespec.type_to_quoted(type) ==
-             {:"::", [], [{:t, [], [{:x, [line: line], nil}]}, [{:x, [line: line], nil}]]}
+             {:"::", [],
+              [
+                {:t, [], [{:x, meta(5, 9), nil}]},
+                [{:x, meta(5, 20), nil}]
+              ]}
 
     [{{:f, 1}, [spec]}] = specs(:typespec_test_mod)
-    line = 7
 
     assert Code.Typespec.spec_to_quoted(:f, spec) ==
-             {:when, [line: line],
+             {:when, meta(7, 8),
               [
-                {:"::", [line: line],
-                 [{:f, [line: line], [{:x, [line: line], nil}]}, {:x, [line: line], nil}]},
-                [x: {:var, [line: line], nil}]
+                {:"::", meta(7, 8),
+                 [{:f, meta(7, 8), [{:x, meta(7, 9), nil}]}, {:x, meta(7, 15), nil}]},
+                [x: {:var, meta(7, 8), nil}]
               ]}
   end
 
+  defp meta(line, column) do
+    [line: line, column: column]
+  end
+
   defp erlc(context, module, code) do
     dir = context.tmp_dir
 
@@ -1603,4 +1630,10 @@ defmodule TypespecTest do
 
     :ok
   end
+
+  defp assert_compile_error(message, fun) do
+    assert ExUnit.CaptureIO.capture_io(:stderr, fn ->
+             assert_raise CompileError, fun
+           end) =~ message
+  end
 end
