@@ -1,11 +1,3 @@
-defmodule ExUnit.DuplicateTestError do
-  defexception [:message]
-end
-
-defmodule ExUnit.DuplicateDescribeError do
-  defexception [:message]
-end
-
 defmodule ExUnit.Case do
   @moduledoc """
   Helpers for defining test cases.
@@ -20,11 +12,24 @@ defmodule ExUnit.Case do
       It should be enabled only if tests do not change any global state.
       Defaults to `false`.
 
-  This module automatically includes all callbacks defined in
-  `ExUnit.Callbacks`. See that module for more information on `setup`,
-  `start_supervised`, `on_exit` and the test process life cycle.
+    * `:group` (since v1.18.0) - configures the **group** this module belongs to.
+      Tests in the same group never run concurrently. Tests from different
+      groups (or with no groups) can run concurrently when `async: true`
+      is given. By default, this module belongs to no group (defaults to `nil`).
 
-  For grouping tests together, see `describe/2` in this module.
+    * `:register` - when `false`, does not register this module within
+      ExUnit server. This means the module won't run when ExUnit suite runs.
+
+    * `:parameterize` (since v1.18.0) - a list of maps to parameterize tests.
+      If both `:async` and `:parameterize` are given, the different parameters
+      run concurrently. See the "Parameterized tests" section below for more
+      information.
+
+  > #### `use ExUnit.Case` {: .info}
+  >
+  > When you `use ExUnit.Case`, it will import the functionality
+  > from `ExUnit.Assertions`, `ExUnit.Callbacks`, `ExUnit.DocTest`,
+  > and this module itself.
 
   ## Examples
 
@@ -135,26 +140,39 @@ defmodule ExUnit.Case do
   If the same key is set via `@tag`, the `@tag` value has higher
   precedence.
 
+  The `setup_all` blocks only receive tags that are set using `@moduletag`.
+
   ### Known tags
 
   The following tags are set automatically by ExUnit and are
   therefore reserved:
 
-    * `:module` - the module on which the test was defined
+    * `:async` - if the test case is in async mode
 
     * `:file` - the file on which the test was defined
 
     * `:line` - the line on which the test was defined
 
-    * `:test` - the test name
-
-    * `:async` - if the test case is in async mode
+    * `:module` - the module on which the test was defined
 
     * `:registered` - used for `ExUnit.Case.register_attribute/3` values
 
-    * `:describe` - the describe block the test belongs to
+    * `:test` - the test name
 
-    * `:describe_line` - the line the describe block begins on
+    * `:test_pid` - the PID of the testing process
+
+    * `:test_type` - the test type used when printing test results.
+      It is set by ExUnit to `:test`, `:doctest` and so on, but is customizable.
+
+    * `:describe` - the describe block the test belongs to (if in a describe)
+
+    * `:describe_line` - the line the describe block begins on (if in a describe)
+
+    * `:doctest` - the module or the file being doctested (if a doctest)
+
+    * `:doctest_data` - additional metadata about doctests (if a doctest)
+
+    * `:doctest_line` - the line the doctest was defined (if a doctest)
 
   The following tags customize how tests behave:
 
@@ -167,8 +185,40 @@ defmodule ExUnit.Case do
 
     * `:tmp_dir` - (since v1.11.0) see the "Tmp Dir" section below
 
-  The `:test_type` tag is automatically set by ExUnit, but is **not** reserved.
-  This tag is available for users to customize if they desire.
+  ## Parameterized tests
+
+  Sometimes you want to run the same tests but with different parameters.
+  In ExUnit, it is possible to do so by passing a `:parameterize` key to
+  `ExUnit.Case`. The value must be a list of maps which will be the
+  parameters merged into the test context.
+
+  For example, Elixir has a module called `Registry`, which can have type
+  `:unique` or `:duplicate`, and can control its concurrency factor using
+  the `:partitions` option. If you have a number of tests that *behave the
+  same* across all of those values, you can parameterize those tests with:
+
+      use ExUnit.Case,
+        async: true,
+        parameterize:
+          for(kind <- [:unique, :duplicate],
+              partitions <- [1, 8],
+              do: %{kind: kind, partitions: partitions})
+
+  Then, in your tests, you can access the parameters as part of the context:
+
+      test "starts a registry", %{kind: kind, partitions: partitions} do
+        ...
+      end
+
+  Use parameterized tests with care:
+
+    * Although parameterized tests run concurrently when `async: true` is also given,
+      abuse of parameterized tests may make your test suite slower
+
+    * If you use parameterized tests and then find yourself adding conditionals
+      in your tests to deal with different parameters, then parameterized tests
+      may be the wrong solution to your problem. Consider creating separated
+      tests and sharing logic between them using regular functions
 
   ## Filters
 
@@ -184,7 +234,7 @@ defmodule ExUnit.Case do
   set to `true`. This behaviour can be reversed with the `:include` option
   which is usually passed through the command line:
 
-      mix test --include external:true
+      $ mix test --include external:true
 
   Run `mix help test` for more information on how to run filters via Mix.
 
@@ -250,15 +300,13 @@ defmodule ExUnit.Case do
   """
 
   @type env :: module() | Macro.Env.t()
-
+  @compile {:no_warn_undefined, [IEx.Pry]}
   @reserved [:module, :file, :line, :test, :async, :registered, :describe]
 
   @doc false
   defmacro __using__(opts) do
     quote do
-      unless ExUnit.Case.__register__(__MODULE__, unquote(opts)) do
-        use ExUnit.Callbacks
-      end
+      ExUnit.Case.__register__(__MODULE__, unquote(opts))
 
       import ExUnit.Callbacks
       import ExUnit.Assertions
@@ -267,24 +315,34 @@ defmodule ExUnit.Case do
     end
   end
 
+  @keys [:async, :group, :parameterize, :register]
+
+  @doc false
+  def __keys__(opts), do: Keyword.take(opts, @keys)
+
   @doc false
   def __register__(module, opts) do
-    unless Keyword.keyword?(opts) do
+    if not Keyword.keyword?(opts) do
       raise ArgumentError,
             ~s(the argument passed to "use ExUnit.Case" must be a list of options, ) <>
               ~s(got: #{inspect(opts)})
     end
 
-    registered? = Module.has_attribute?(module, :ex_unit_tests)
+    {register?, opts} = Keyword.pop(opts, :register, true)
+    {next_opts, opts} = Keyword.split(opts, @keys)
 
-    unless registered? do
+    if opts != [] do
+      IO.warn("unknown options given to ExUnit.Case: #{inspect(opts)}")
+    end
+
+    if not Module.has_attribute?(module, :ex_unit_tests) do
       tag_check = Enum.any?([:moduletag, :describetag, :tag], &Module.has_attribute?(module, &1))
 
       if tag_check do
         raise "you must set @tag, @describetag, and @moduletag after the call to \"use ExUnit.Case\""
       end
 
-      attributes = [
+      accumulate_attributes = [
         :ex_unit_tests,
         :tag,
         :describetag,
@@ -294,27 +352,19 @@ defmodule ExUnit.Case do
         :ex_unit_registered_module_attributes
       ]
 
-      Enum.each(attributes, &Module.register_attribute(module, &1, accumulate: true))
+      Enum.each(accumulate_attributes, &Module.register_attribute(module, &1, accumulate: true))
 
-      persisted_attributes = [:ex_unit_async]
-      Enum.each(persisted_attributes, &Module.register_attribute(module, &1, persist: true))
+      if register? do
+        Module.put_attribute(module, :after_compile, ExUnit.Case)
+      end
 
-      attributes = [
-        before_compile: ExUnit.Case,
-        after_compile: ExUnit.Case,
-        ex_unit_async: false
-      ]
-
-      Enum.each(attributes, fn {k, v} -> Module.put_attribute(module, k, v) end)
+      ExUnit.Callbacks.__register__(module)
+      Module.put_attribute(module, :before_compile, ExUnit.Case)
     end
 
-    async? = opts[:async]
-
-    if is_boolean(async?) do
-      Module.put_attribute(module, :ex_unit_async, async?)
-    end
-
-    registered?
+    past_opts = Module.get_attribute(module, :ex_unit_module, [])
+    Module.put_attribute(module, :ex_unit_module, Keyword.merge(past_opts, next_opts))
+    :ok
   end
 
   @doc """
@@ -332,7 +382,7 @@ defmodule ExUnit.Case do
 
   """
   defmacro test(message, var \\ quote(do: _), contents) do
-    unless is_tuple(var) do
+    if not is_tuple(var) do
       IO.warn(
         "test context is always a map. The pattern " <>
           "#{inspect(Macro.to_string(var))} will never match",
@@ -341,34 +391,52 @@ defmodule ExUnit.Case do
     end
 
     contents =
-      case contents do
+      case annotate_test(contents, __CALLER__) do
         [do: block] ->
           quote do
             unquote(block)
             :ok
           end
 
-        _ ->
+        contents ->
           quote do
             try(unquote(contents))
             :ok
           end
       end
 
-    var = Macro.escape(var)
-    contents = Macro.escape(contents, unquote: true)
     %{module: mod, file: file, line: line} = __CALLER__
 
-    quote bind_quoted: [
-            var: var,
-            contents: contents,
-            message: message,
-            mod: mod,
-            file: file,
-            line: line
-          ] do
-      name = ExUnit.Case.register_test(mod, file, line, :test, message, [])
-      def unquote(name)(unquote(var)), do: unquote(contents)
+    name =
+      quote do
+        name =
+          ExUnit.Case.register_test(
+            unquote(mod),
+            unquote(file),
+            unquote(line),
+            :test,
+            unquote(message),
+            []
+          )
+      end
+
+    def =
+      {:def, [],
+       [
+         {{:unquote, [], [quote(do: name)]}, [], [var]},
+         [do: contents]
+       ]}
+
+    {:__block__, [], [name, def]}
+  end
+
+  defp annotate_test(contents, caller) do
+    if Application.get_env(:ex_unit, :breakpoints, false) and Keyword.keyword?(contents) do
+      for {key, expr} <- contents do
+        {key, IEx.Pry.annotate_quoted(expr, true, caller)}
+      end
+    else
+      contents
     end
   end
 
@@ -408,24 +476,24 @@ defmodule ExUnit.Case do
       defmodule StringTest do
         use ExUnit.Case, async: true
 
-        describe "String.capitalize/1" do
-          test "first grapheme is in uppercase" do
-            assert String.capitalize("hello") == "Hello"
+        describe "String.downcase/1" do
+          test "with ascii characters" do
+            assert String.downcase("HELLO") == "hello"
           end
 
-          test "converts remaining graphemes to lowercase" do
-            assert String.capitalize("HELLO") == "Hello"
+          test "with Unicode" do
+            assert String.downcase("HÉLLÒ") == "héllò"
           end
         end
       end
 
   When using Mix, you can run all tests in a describe block by name:
 
-      mix test --only describe:"String.capitalize/1"
+      $ mix test --only describe:"String.downcase/1"
 
   or by passing the exact line the describe block starts on:
 
-      mix test path/to/file:123
+      $ mix test path/to/file:123
 
   Note describe blocks cannot be nested. Instead of relying on hierarchy
   for composition, developers should build on top of named setups. For
@@ -455,6 +523,7 @@ defmodule ExUnit.Case do
   for the developer to glance at each describe block and know exactly the
   setup steps involved.
   """
+  @doc since: "1.3.0"
   defmacro describe(message, do: block) do
     definition =
       quote unquote: false do
@@ -477,10 +546,51 @@ defmodule ExUnit.Case do
   end
 
   @doc false
-  defmacro __before_compile__(_) do
+  defmacro __before_compile__(%{module: module}) do
+    tests =
+      module
+      |> Module.get_attribute(:ex_unit_tests)
+      |> Enum.reverse()
+      |> Macro.escape()
+
+    moduletag = Module.get_attribute(module, :moduletag)
+
+    tags =
+      moduletag
+      |> normalize_tags()
+      |> validate_tags()
+      |> Map.new()
+
+    opts = Module.get_attribute(module, :ex_unit_module, [])
+    async? = Keyword.get(opts, :async, false)
+    group = Keyword.get(opts, :group, nil)
+    parameterize = Keyword.get(opts, :parameterize, nil)
+
+    if not (parameterize == nil or (is_list(parameterize) and Enum.all?(parameterize, &is_map/1))) do
+      raise ArgumentError, ":parameterize must be a list of maps, got: #{inspect(parameterize)}"
+    end
+
+    {setup_all?, callbacks} = ExUnit.Callbacks.__callbacks__(module)
+
     quote do
+      unquote(callbacks)
+
       def __ex_unit__ do
-        %ExUnit.TestModule{file: __ENV__.file, name: __MODULE__, tests: @ex_unit_tests}
+        %ExUnit.TestModule{
+          file: __ENV__.file,
+          name: __MODULE__,
+          setup_all?: unquote(setup_all?),
+          tags: unquote(Macro.escape(tags)),
+          tests: unquote(tests)
+        }
+      end
+
+      def __ex_unit__(:config) do
+        %{
+          async?: unquote(async?),
+          group: unquote(Macro.escape(group)),
+          parameterize: unquote(Macro.escape(parameterize))
+        }
       end
     end
   end
@@ -488,17 +598,15 @@ defmodule ExUnit.Case do
   @doc false
   def __after_compile__(%{module: module}, _) do
     cond do
-      Process.whereis(ExUnit.Server) == nil ->
-        unless Code.can_await_module_compilation?() do
-          raise "cannot use ExUnit.Case without starting the ExUnit application, " <>
-                  "please call ExUnit.start() or explicitly start the :ex_unit app"
-        end
+      Process.whereis(ExUnit.Server) ->
+        ExUnit.Server.add_module(module, module.__ex_unit__(:config))
 
-      Module.get_attribute(module, :ex_unit_async) ->
-        ExUnit.Server.add_async_module(module)
+      Code.can_await_module_compilation?() ->
+        :ok
 
       true ->
-        ExUnit.Server.add_sync_module(module)
+        raise "cannot use ExUnit.Case without starting the ExUnit application, " <>
+                "please call ExUnit.start() or explicitly start the :ex_unit app"
     end
   end
 
@@ -514,8 +622,9 @@ defmodule ExUnit.Case do
   display. You can use `ExUnit.plural_rule/2` to set a custom
   pluralization.
   """
+  @doc since: "1.11.0"
   def register_test(mod, file, line, test_type, name, tags) do
-    unless Module.has_attribute?(mod, :ex_unit_tests) do
+    if not Module.has_attribute?(mod, :ex_unit_tests) do
       raise "cannot define #{test_type}. Please make sure you have invoked " <>
               "\"use ExUnit.Case\" in the current module"
     end
@@ -535,7 +644,6 @@ defmodule ExUnit.Case do
 
     moduletag = Module.get_attribute(mod, :moduletag)
     tag = Module.delete_attribute(mod, :tag)
-    async = Module.get_attribute(mod, :ex_unit_async)
 
     {name, describe, describe_line, describetag} =
       case Module.get_attribute(mod, :ex_unit_describe) do
@@ -549,7 +657,7 @@ defmodule ExUnit.Case do
       end
 
     if Module.defines?(mod, {name, 1}) do
-      raise ExUnit.DuplicateTestError, ~s("#{name}" is already defined in #{inspect(mod)})
+      raise ArgumentError, ~s("#{name}" is already defined in #{inspect(mod)})
     end
 
     tags =
@@ -560,7 +668,6 @@ defmodule ExUnit.Case do
         line: line,
         file: file,
         registered: registered,
-        async: async,
         describe: describe,
         describe_line: describe_line,
         test_type: test_type
@@ -582,7 +689,9 @@ defmodule ExUnit.Case do
   This function is deprecated in favor of `register_test/6` which performs
   better under tight loops by avoiding `__ENV__`.
   """
-  @doc deprecated: "Use register_test/6 instead"
+  # TODO: Remove me Elixir v2.0
+  @deprecated "Use register_test/6 instead"
+  @doc since: "1.3.0"
   def register_test(%{module: mod, file: file, line: line}, test_type, name, tags) do
     register_test(mod, file, line, test_type, name, tags)
   end
@@ -615,6 +724,7 @@ defmodule ExUnit.Case do
       end
 
   """
+  @doc since: "1.3.0"
   @spec register_attribute(env, atom, keyword) :: :ok
   def register_attribute(env, name, opts \\ [])
   def register_attribute(%{module: mod}, name, opts), do: register_attribute(mod, name, opts)
@@ -732,12 +842,33 @@ defmodule ExUnit.Case do
     end
   end
 
+  @doc """
+  Returns the most recently registered test case as an `%ExUnit.Test{}`
+  struct.
+
+  This is used by third-party utilities to allow compile-time configuration
+  using test tags without having to explicitly pass the test context at
+  run-time. It is intended to be invoked in macros before the test module
+  is compiled.
+
+  Raises if called with a module that has already been compiled.
+  """
+  @doc since: "1.15.0"
+  @spec get_last_registered_test(env) :: ExUnit.Test.t() | nil
+  def get_last_registered_test(%{module: mod}) do
+    get_last_registered_test(mod)
+  end
+
+  def get_last_registered_test(mod) when is_atom(mod) do
+    Module.get_last_attribute(mod, :ex_unit_tests)
+  end
+
   defp validate_tags(tags) do
     for tag <- @reserved, Map.has_key?(tags, tag) do
       raise "cannot set tag #{inspect(tag)} because it is reserved by ExUnit"
     end
 
-    unless is_atom(tags[:test_type]) do
+    if not is_atom(tags[:test_type]) do
       raise("value for tag \":test_type\" must be an atom")
     end
 
@@ -759,8 +890,9 @@ defmodule ExUnit.Case do
 
   defp normalize_tags(tags) do
     Enum.reduce(Enum.reverse(tags), %{}, fn
+      {key, value}, acc -> Map.put(acc, key, value)
       tag, acc when is_atom(tag) -> Map.put(acc, tag, true)
-      tag, acc when is_list(tag) -> tag |> Enum.into(acc)
+      tag, acc when is_list(tag) -> Enum.into(tag, acc)
     end)
   end
 end
