@@ -55,7 +55,7 @@ defmodule ExUnit.DocTest do
       iex> a = 1
       1
 
-      iex> a + 1 # will fail with a "undefined function a/0" error
+      iex> a + 1 # will fail with a `undefined variable "a"` error
       2
 
   If you don't want to assert for every result in a doctest, you can omit
@@ -96,9 +96,9 @@ defmodule ExUnit.DocTest do
   values are treated as comments in Elixir code due to the leading
   `#` sign, they require special care when being used in doctests.
 
-  Imagine you have a map that contains a MapSet and is printed as:
+  Imagine you have a map that contains a `DateTime` and is printed as:
 
-      %{users: #MapSet<[:foo, :bar]>}
+      %{datetime: #DateTime<2023-06-26 09:30:00+09:00 JST Asia/Tokyo>}
 
   If you try to match on such an expression, `doctest` will fail to compile.
   There are two ways to resolve this.
@@ -106,20 +106,20 @@ defmodule ExUnit.DocTest do
   The first is to rely on the fact that doctest can compare internal
   structures as long as they are at the root. So one could write:
 
-      iex> map = %{users: Enum.into([:foo, :bar], MapSet.new())}
-      iex> map.users
-      #MapSet<[:foo, :bar]>
+      iex> map = %{datetime: DateTime.from_naive!(~N[2023-06-26T09:30:00], "Asia/Tokyo")}
+      iex> map.datetime
+      #DateTime<2023-06-26 09:30:00+09:00 JST Asia/Tokyo>
 
   Whenever a doctest starts with "#Name<", `doctest` will perform a string
   comparison. For example, the above test will perform the following match:
 
-      inspect(map.users) == "#MapSet<[:foo, :bar]>"
+      inspect(map.datetime) == "#DateTime<2023-06-26 09:30:00+09:00 JST Asia/Tokyo>"
 
   Alternatively, since doctest results are actually evaluated, you can have
-  the MapSet building expression as the doctest result:
+  the `DateTime` building expression as the doctest result:
 
-      iex> %{users: Enum.into([:foo, :bar], MapSet.new())}
-      %{users: Enum.into([:foo, :bar], MapSet.new())}
+      iex> %{datetime: DateTime.from_naive!(~N[2023-06-26T09:30:00], "Asia/Tokyo")}
+      %{datetime: DateTime.from_naive!(~N[2023-06-26T09:30:00], "Asia/Tokyo")}
 
   The downside of this approach is that the doctest result is not really
   what users would see in the terminal.
@@ -152,6 +152,12 @@ defmodule ExUnit.DocTest do
   @opaque_type_regex ~r/#[\w\.]+</
 
   defmodule Error do
+    @moduledoc """
+    Exception raised when there's an error with the syntax or semantics of a doctest.
+    """
+
+    @typedoc since: "1.16.0"
+    @type t :: %__MODULE__{message: String.t()}
     defexception [:message]
 
     @impl true
@@ -159,19 +165,25 @@ defmodule ExUnit.DocTest do
       module = Keyword.fetch!(opts, :module)
       message = Keyword.fetch!(opts, :message)
 
-      file = module.module_info(:compile)[:source] |> Path.relative_to_cwd()
-      info = Exception.format_file_line(file, opts[:line])
-      %__MODULE__{message: info <> " " <> message}
+      message =
+        if source = module.module_info(:compile)[:source] do
+          file_line = Exception.format_file_line(Path.relative_to_cwd(source), opts[:line])
+          file_line <> " " <> message
+        else
+          message
+        end
+
+      %__MODULE__{message: message}
     end
   end
 
   @doc """
-  This macro is used to generate ExUnit test cases for doctests.
+  Generate test cases from module documentation.
 
   Calling `doctest(Module)` will generate tests for all doctests found
   in the `module`.
 
-  Options can also be given:
+  ## Options
 
     * `:except` - generates tests for all functions except those listed
       (list of `{function, arity}` tuples, and/or `:moduledoc`).
@@ -188,30 +200,35 @@ defmodule ExUnit.DocTest do
 
   ## Examples
 
-      doctest MyModule, except: [:moduledoc, trick_fun: 1]
+      defmodule MyModuleTest do
+        use ExUnit.Case
+        doctest MyModule, except: [:moduledoc, trick_fun: 1]
+      end
 
   This macro is auto-imported with every `ExUnit.Case`.
   """
   defmacro doctest(module, opts \\ []) do
+    caller = __CALLER__
+
     require =
-      if is_atom(Macro.expand(module, __CALLER__)) do
+      if is_atom(Macro.expand(module, caller)) do
         quote do
           require unquote(module)
         end
       end
 
     tests =
-      quote bind_quoted: [module: module, opts: opts] do
-        env = __ENV__
+      quote bind_quoted: [
+              module: module,
+              opts: opts,
+              env_line: caller.line,
+              env_file: caller.file
+            ] do
         file = ExUnit.DocTest.__file__(module)
 
-        for {name, test} <- ExUnit.DocTest.__doctests__(module, opts) do
-          if tags = Keyword.get(opts, :tags) do
-            @tag tags
-          end
-
+        for {name, test, tags} <- ExUnit.DocTest.__doctests__(module, opts) do
           @file file
-          doc = ExUnit.Case.register_test(env, :doctest, name, [])
+          doc = ExUnit.Case.register_test(__MODULE__, env_file, env_line, :doctest, name, tags)
           def unquote(doc)(_), do: unquote(test)
         end
       end
@@ -219,24 +236,76 @@ defmodule ExUnit.DocTest do
     [require, tests]
   end
 
+  @doc """
+  Generate test cases from a markdown file.
+
+  ## Options
+
+    * `:tags` - a list of tags to apply to all generated doctests.
+
+  ## Examples
+
+      defmodule ReadmeTest do
+        use ExUnit.Case
+        doctest_file "README.md"
+      end
+
+  This macro is auto-imported with every `ExUnit.Case`.
+  """
+  @doc since: "1.15.0"
+  defmacro doctest_file(file, opts \\ []) do
+    caller = __CALLER__
+
+    tests =
+      quote bind_quoted: [file: file, opts: opts, env_line: caller.line, env_file: caller.file] do
+        for {name, test, tags} <- ExUnit.DocTest.__doctest_file__(file, __MODULE__, opts) do
+          doc = ExUnit.Case.register_test(__MODULE__, env_file, env_line, :doctest, name, tags)
+          def unquote(doc)(_), do: unquote(test)
+        end
+      end
+
+    tests
+  end
+
+  @doc false
+  def __doctest_file__(file, module, opts) do
+    doc = File.read!(file)
+    file = Path.relative_to_cwd(file)
+    tags = [doctest: file] ++ Keyword.get(opts, :tags, [])
+
+    extract_tests(1, doc, module, :moduledoc)
+    |> Enum.with_index(fn test, acc ->
+      {"#{file} (#{acc + 1})", test_content(test, module, false, file), test_tags(test, tags)}
+    end)
+  end
+
   @doc false
   def __file__(module) do
-    source =
-      module.module_info(:compile)[:source] ||
-        raise "#{inspect(module)} does not have compile-time source information"
+    info =
+      if source = module.module_info(:compile)[:source] do
+        Path.relative_to_cwd(source)
+      else
+        inspect(module)
+      end
 
-    "(for doctest at) " <> Path.relative_to_cwd(source)
+    "(for doctest at) " <> info
   end
 
   @doc false
   def __doctests__(module, opts) do
-    do_import = Keyword.get(opts, :import, false)
+    tags = [doctest: module] ++ Keyword.get(opts, :tags, [])
+    import = Keyword.get(opts, :import, false)
+
+    maybe_source =
+      if source = module.module_info(:compile)[:source] do
+        Path.relative_to_cwd(source)
+      end
 
     extract(module)
     |> filter_by_opts(module, opts)
-    |> Stream.with_index()
-    |> Enum.map(fn {test, acc} ->
-      compile_test(test, module, do_import, acc + 1)
+    |> Enum.sort_by(& &1.line)
+    |> Enum.with_index(fn test, index ->
+      compile_test(test, module, import, index + 1, maybe_source, tags)
     end)
   end
 
@@ -244,14 +313,9 @@ defmodule ExUnit.DocTest do
     except = Keyword.get(opts, :except, [])
 
     case Keyword.fetch(opts, :only) do
-      {:ok, []} ->
-        []
-
-      {:ok, only} ->
-        filter_tests(module, tests, except, only)
-
-      :error ->
-        Stream.reject(tests, &(&1.fun_arity in except))
+      {:ok, []} -> []
+      {:ok, only} -> filter_tests(module, tests, except, only)
+      :error -> Stream.reject(tests, &(&1.fun_arity in except))
     end
   end
 
@@ -264,7 +328,7 @@ defmodule ExUnit.DocTest do
         {tests, fun_arities} -> {[test | tests], [test.fun_arity | fun_arities]}
       end
 
-    case only -- fun_arities do
+    case only -- [:moduledoc | fun_arities] do
       [] ->
         filtered_tests
 
@@ -287,8 +351,9 @@ defmodule ExUnit.DocTest do
 
   ## Compilation of extracted tests
 
-  defp compile_test(test, module, do_import, n) do
-    {test_name(test, module, n), test_content(test, module, do_import)}
+  defp compile_test(test, module, do_import, n, maybe_source, tags) do
+    {test_name(test, module, n), test_content(test, module, do_import, maybe_source),
+     test_tags(test, tags)}
   end
 
   defp test_name(%{fun_arity: :moduledoc}, m, n) do
@@ -299,11 +364,7 @@ defmodule ExUnit.DocTest do
     "#{inspect(m)}.#{f}/#{a} (#{n})"
   end
 
-  defp test_content(%{exprs: exprs, line: line}, module, do_import) do
-    file = module.module_info(:compile)[:source] |> Path.relative_to_cwd()
-    location = [line: line, file: file]
-    stack = Macro.escape([{module, :__MODULE__, 0, location}])
-
+  defp test_content(%{exprs: exprs, line: line}, module, do_import, maybe_source) do
     if multiple_exceptions?(exprs) do
       raise Error,
         line: line,
@@ -313,95 +374,113 @@ defmodule ExUnit.DocTest do
             "please separate your iex> prompts by multiple newlines to start new examples"
     end
 
-    tests =
-      Enum.map(exprs, fn {expr, expected, formatted} ->
-        test_case_content(expr, expected, location, stack, formatted)
-      end)
-
+    tests = Enum.map(exprs, fn expr -> test_case_content(expr, module, maybe_source) end)
     {:__block__, [], test_import(module, do_import) ++ tests}
   end
 
+  defp test_tags(test, tags) do
+    [doctest_line: test.line, doctest_data: %{end_line: test.end_line}] ++ tags
+  end
+
   defp multiple_exceptions?(exprs) do
     Enum.count(exprs, fn
-      {_, {:error, _, _}, _} -> true
+      %{expected: {:error, _, _}} -> true
       _ -> false
     end) > 1
   end
 
-  defp test_case_content(expr, :test, location, stack, formatted) do
-    string_to_quoted(location, stack, expr, "\n" <> formatted) |> insert_assertions()
+  defp test_case_content(%{expected: :test} = data, module, maybe_source) do
+    %{expr: expr, expr_line: expr_line, doctest: doctest} = data
+    string_to_quoted(module, maybe_source, expr_line, expr, doctest) |> insert_assertions()
   end
 
-  defp test_case_content(expr, {:test, expected}, location, stack, formatted) do
-    doctest = "\n" <> formatted <> "\n" <> expected
-    expr_ast = string_to_quoted(location, stack, expr, doctest) |> insert_assertions()
-    expected_ast = string_to_quoted(location, stack, expected, doctest)
+  defp test_case_content(%{expected: {:test, expected}} = data, module, maybe_source) do
+    %{expr: expr, expr_line: expr_line, expected_line: expected_line, doctest: doctest} = data
+
+    expr_ast =
+      string_to_quoted(module, maybe_source, expr_line, expr, doctest) |> insert_assertions()
+
+    expected_ast = string_to_quoted(module, maybe_source, expected_line, expected, doctest)
     last_expr = Macro.to_string(last_expr(expr_ast))
 
     quote do
+      # `expr_ast` may introduce variables that may be used
+      # within `expected_ast` so it needs to be unquoted here.
       value = unquote(expr_ast)
-      expected = unquote(expected_ast)
-      formatted = unquote(formatted)
-      last_expr = unquote(last_expr)
-      expected_expr = unquote(expected)
-      stack = unquote(stack)
 
-      ExUnit.DocTest.__test__(value, expected, formatted, last_expr, expected_expr, stack)
+      ExUnit.DocTest.__test__(
+        value,
+        unquote(expected_ast),
+        unquote(doctest),
+        unquote(last_expr),
+        unquote(expected),
+        unquote(module),
+        unquote(maybe_source),
+        unquote(expr_line)
+      )
     end
   end
 
-  defp test_case_content(expr, {:inspect, expected}, location, stack, formatted) do
-    doctest = "\n" <> formatted <> "\n" <> expected
-    expr_ast = string_to_quoted(location, stack, expr, doctest) |> insert_assertions()
-    expected_ast = string_to_quoted(location, stack, expected, doctest)
+  defp test_case_content(%{expected: {:inspect, expected}} = data, module, maybe_source) do
+    %{expr: expr, expr_line: expr_line, doctest: doctest} = data
+
+    expr_ast =
+      string_to_quoted(module, maybe_source, expr_line, expr, doctest) |> insert_assertions()
+
     last_expr = Macro.to_string(last_expr(expr_ast))
 
     quote do
-      value = unquote(expr_ast)
-      expected = unquote(expected_ast)
-      formatted = unquote(formatted)
-      last_expr = unquote(last_expr)
-      expected_expr = unquote(expected)
-      stack = unquote(stack)
-
-      ExUnit.DocTest.__inspect__(value, expected, formatted, last_expr, expected_expr, stack)
+      ExUnit.DocTest.__inspect__(
+        unquote(expr_ast),
+        unquote(expected),
+        unquote(doctest),
+        unquote(last_expr),
+        unquote(inspect(expected)),
+        unquote(module),
+        unquote(maybe_source),
+        unquote(expr_line)
+      )
     end
   end
 
-  defp test_case_content(expr, {:error, exception, message}, location, stack, formatted) do
-    doctest = "\n" <> formatted <> "\n** (#{inspect(exception)}) #{inspect(message)}"
-    expr_ast = string_to_quoted(location, stack, expr, doctest)
+  defp test_case_content(%{expected: {:error, exception, message}} = data, module, maybe_source) do
+    %{expr: expr, expr_line: expr_line, doctest: doctest} = data
+    expr_ast = string_to_quoted(module, maybe_source, expr_line, expr, doctest)
 
     quote do
-      stack = unquote(stack)
-      message = unquote(message)
-      formatted = unquote(formatted)
-      exception = unquote(exception)
-      ExUnit.DocTest.__error__(fn -> unquote(expr_ast) end, message, exception, formatted, stack)
+      ExUnit.DocTest.__error__(
+        fn -> unquote(expr_ast) end,
+        unquote(message),
+        unquote(exception),
+        unquote(doctest),
+        unquote(module),
+        unquote(maybe_source),
+        unquote(expr_line)
+      )
     end
   end
 
   @doc false
-  def __test__(value, expected, formatted, last_expr, expected_expr, stack) do
+  def __test__(value, expected, doctest, last_expr, expected_expr, module, maybe_source, line) do
     case value do
       ^expected ->
-        :ok
+        {:ok, value}
 
       _ ->
         error = [
           message: "Doctest failed",
-          doctest: "\n" <> formatted <> "\n" <> expected_expr,
+          doctest: doctest,
           expr: "#{last_expr} === #{String.trim(expected_expr)}",
           left: value,
           right: expected
         ]
 
-        reraise ExUnit.AssertionError, error, stack
+        reraise ExUnit.AssertionError, error, stack(module, maybe_source, line)
     end
   end
 
   @doc false
-  def __inspect__(value, expected, formatted, last_expr, expected_expr, parent_stack) do
+  def __inspect__(value, expected, doctest, last_expr, expected_expr, module, maybe_source, line) do
     result =
       try do
         inspect(value, safe: false)
@@ -416,18 +495,17 @@ defmodule ExUnit.DocTest do
 
     case result do
       :ok ->
-        :ok
+        {:ok, value}
 
       {extra, stack} ->
-        doctest = "\n" <> formatted <> "\n" <> expected_expr
         expr = "inspect(#{last_expr}) === #{String.trim(expected_expr)}"
         error = [doctest: doctest, expr: expr] ++ extra
-        reraise ExUnit.AssertionError, error, stack ++ parent_stack
+        reraise ExUnit.AssertionError, error, stack ++ stack(module, maybe_source, line)
     end
   end
 
   @doc false
-  def __error__(fun, message, exception, formatted, stack) do
+  def __error__(fun, message, exception, doctest, module, maybe_source, line) do
     try do
       fun.()
     rescue
@@ -452,24 +530,24 @@ defmodule ExUnit.DocTest do
           end
 
         if failed do
-          doctest = "\n" <> formatted <> "\n** (#{inspect(exception)}) #{inspect(message)}"
-          reraise ExUnit.AssertionError, [message: failed, doctest: doctest], stack
+          reraise ExUnit.AssertionError,
+                  [message: failed, doctest: doctest],
+                  stack(module, maybe_source, line)
         end
     else
       _ ->
-        doctest = "\n" <> formatted <> "\n** (#{inspect(exception)}) #{inspect(message)}"
         failed = "Doctest failed: expected exception #{inspect(exception)} but nothing was raised"
         error = [message: failed, doctest: doctest]
-        reraise ExUnit.AssertionError, error, stack
+        reraise ExUnit.AssertionError, error, stack(module, maybe_source, line)
     end
   end
 
   defp test_import(_mod, false), do: []
   defp test_import(mod, _), do: [quote(do: import(unquote(mod)))]
 
-  defp string_to_quoted(location, stack, expr, doctest) do
+  defp string_to_quoted(module, maybe_source, line, expr, doctest) when is_binary(expr) do
     try do
-      Code.string_to_quoted!(expr, location)
+      Code.string_to_quoted!(expr, file: maybe_source || inspect(module), line: line)
     rescue
       e ->
         ex_message = "(#{inspect(e.__struct__)}) #{Exception.message(e)}"
@@ -481,7 +559,7 @@ defmodule ExUnit.DocTest do
               """
               \nIf you are planning to assert on the result of an iex> expression \
               which contains a value inspected as #Name<...>, please make sure \
-              the inspected value is placed at the beginning of the expression; \
+              the inspected value is placed at the beginning of the expression, \
               otherwise Elixir will treat it as a comment due to the leading sign #.\
               """
           else
@@ -496,11 +574,18 @@ defmodule ExUnit.DocTest do
           end
 
         quote do
-          reraise ExUnit.AssertionError, unquote(opts), unquote(stack)
+          reraise ExUnit.AssertionError,
+                  unquote(opts),
+                  unquote(Macro.escape(stack(module, maybe_source, line)))
         end
     end
   end
 
+  defp stack(module, file, line) do
+    file_info = if file, do: [file: String.to_charlist(file)], else: []
+    [{module, :__MODULE__, 0, [line: line] ++ file_info}]
+  end
+
   ## Extraction of the tests
 
   defp extract(module) do
@@ -528,9 +613,7 @@ defmodule ExUnit.DocTest do
     do: "The documentation chunk in the module is invalid"
 
   defp extract_from_moduledoc(annotation, %{"en" => doc}, module) do
-    for test <- extract_tests(:erl_anno.line(annotation), doc, module) do
-      normalize_test(test, :moduledoc)
-    end
+    extract_tests(:erl_anno.line(annotation), doc, module, :moduledoc)
   end
 
   defp extract_from_moduledoc(_, _doc, _module), do: []
@@ -541,105 +624,80 @@ defmodule ExUnit.DocTest do
 
   defp extract_from_doc({{_, name, arity}, annotation, _, %{"en" => doc}, _}, module) do
     line = :erl_anno.line(annotation)
-
-    for test <- extract_tests(line, doc, module) do
-      normalize_test(test, {name, arity})
-    end
+    extract_tests(line, doc, module, {name, arity})
   end
 
   defp extract_from_doc(_doc, _module),
     do: []
 
-  defp extract_tests(line_no, doc, module) do
-    all_lines = String.split(doc, ["\r\n", "\n"], trim: false)
-    lines = adjust_indent(all_lines, line_no + 1, module)
-    extract_tests(lines, "", "", [], true, module, "")
-  end
-
   @iex_prompt ["iex>", "iex("]
   @dot_prompt ["...>", "...("]
+  @fences ["```", "~~~"]
 
   defp adjust_indent(lines, line_no, module) do
-    adjust_indent(:text, lines, line_no, [], 0, module)
+    adjust_text(lines, line_no, [], module)
   end
 
-  defp adjust_indent(_kind, [], _line_no, adjusted_lines, _indent, _module) do
-    Enum.reverse(adjusted_lines)
+  defp adjust_text([], line_no, adjusted_lines, _module) do
+    {Enum.reverse(adjusted_lines), line_no - 1}
   end
 
-  defp adjust_indent(:text, [line | rest], line_no, adjusted_lines, indent, module) do
+  defp adjust_text([line | rest], line_no, adjusted_lines, module) do
     case String.starts_with?(String.trim_leading(line), @iex_prompt) do
       true ->
-        line_indent = get_indent(line, indent)
-        adjust_indent(:prompt, [line | rest], line_no, adjusted_lines, line_indent, module)
+        {indent, _len} = :binary.match(line, "iex")
+        adjust_code(:prompt, [line | rest], line_no, adjusted_lines, indent, module)
 
       false ->
-        adjust_indent(:text, rest, line_no + 1, adjusted_lines, indent, module)
+        adjust_text(rest, line_no + 1, adjusted_lines, module)
     end
   end
 
-  defp adjust_indent(kind, [line | rest], line_no, adjusted_lines, indent, module)
-       when kind in [:prompt, :after_prompt] do
-    stripped_line = strip_indent(line, indent)
-
-    case String.trim_leading(line) do
-      "" ->
-        :ok
-
-      ^stripped_line ->
-        :ok
-
-      _ ->
-        n_spaces = if indent == 1, do: "#{indent} space", else: "#{indent} spaces"
-
-        raise Error,
-          line: line_no,
-          module: module,
-          message: """
-          indentation level mismatch on doctest line: #{inspect(line)}
-
-          If you are planning to assert on the result of an `iex>` expression, \
-          make sure the result is indented at the beginning of `iex>`, which \
-          in this case is exactly #{n_spaces}.
-
-          If instead you have an `iex>` expression that spans over multiple lines, \
-          please make sure that each line after the first one begins with `...>`.
-          """
-    end
-
-    adjusted_lines = [{stripped_line, line_no} | adjusted_lines]
-
-    next =
-      cond do
-        kind == :prompt -> :after_prompt
-        String.starts_with?(stripped_line, @iex_prompt ++ @dot_prompt) -> :after_prompt
-        true -> :code
-      end
-
-    adjust_indent(next, rest, line_no + 1, adjusted_lines, indent, module)
+  defp adjust_code(_kind, [], line_no, adjusted_lines, _indent, _module) do
+    {Enum.reverse(adjusted_lines), line_no - 1}
   end
 
-  defp adjust_indent(:code, [line | rest], line_no, adjusted_lines, indent, module) do
+  defp adjust_code(kind, [line | rest], line_no, adjusted_lines, indent, module) do
     stripped_line = strip_indent(line, indent)
+    trimmed_line = String.trim_leading(line)
+    done? = stripped_line == "" or String.starts_with?(stripped_line, @fences)
+
+    overrun? =
+      if kind == :code,
+        do: not done? and byte_size(trimmed_line) > byte_size(stripped_line),
+        else: byte_size(trimmed_line) != byte_size(stripped_line)
+
+    if overrun? do
+      n_spaces = if indent == 1, do: "#{indent} space", else: "#{indent} spaces"
+
+      raise Error,
+        line: line_no,
+        module: module,
+        message: """
+        indentation level mismatch on doctest line: #{inspect(line)}
+
+        If you are planning to assert on the result of an `iex>` expression, \
+        make sure the result is indented at the same level as `iex>`, which \
+        in this case is exactly #{n_spaces}.
+
+        If instead you have an `iex>` expression that spans over multiple lines, \
+        please make sure that each line after the first one begins with `...>`.
+        """
+    end
 
     cond do
-      stripped_line == "" ->
-        adjusted_lines = [{stripped_line, line_no} | adjusted_lines]
-        adjust_indent(:text, rest, line_no + 1, adjusted_lines, 0, module)
+      done? ->
+        adjusted_lines = [{"", line_no} | adjusted_lines]
+        adjust_text(rest, line_no + 1, adjusted_lines, module)
 
-      String.starts_with?(String.trim_leading(line), @iex_prompt) ->
-        adjust_indent(:prompt, [line | rest], line_no, adjusted_lines, indent, module)
+      kind == :prompt or String.starts_with?(trimmed_line, @iex_prompt) or
+          (kind == :maybe_prompt and String.starts_with?(trimmed_line, @dot_prompt)) ->
+        line = {adjust_prompt(stripped_line, line_no, module), line_no}
+        adjust_code(:maybe_prompt, rest, line_no + 1, [line | adjusted_lines], indent, module)
 
       true ->
         adjusted_lines = [{stripped_line, line_no} | adjusted_lines]
-        adjust_indent(:code, rest, line_no + 1, adjusted_lines, indent, module)
-    end
-  end
-
-  defp get_indent(line, current_indent) do
-    case :binary.match(line, "iex") do
-      {pos, _len} -> pos
-      :nomatch -> current_indent
+        adjust_code(:code, rest, line_no + 1, adjusted_lines, indent, module)
     end
   end
 
@@ -653,221 +711,132 @@ defmodule ExUnit.DocTest do
     end
   end
 
-  @fences ["```", "~~~"]
+  defp adjust_prompt("iex(" <> rest = line, line_no, module),
+    do: "iex>" <> skip_iex_number(rest, line_no, module, line)
 
-  defp extract_tests(lines, expr_acc, expected_acc, acc, new_test, module, formatted)
+  defp adjust_prompt("...(" <> rest = line, line_no, module),
+    do: "...>" <> skip_iex_number(rest, line_no, module, line)
 
-  defp extract_tests([], "", "", [], _, _, _) do
-    []
+  defp adjust_prompt(line, _line_no, _module),
+    do: line
+
+  defp skip_iex_number(string, line_no, module, line) do
+    case :binary.split(string, ")>") do
+      [_pre, post] ->
+        post
+
+      [_] ->
+        message =
+          "unknown IEx prompt: #{inspect(line)}.\nAccepted formats are: iex>, iex(1)>, ...>, ...(1)>}"
+
+        raise Error, line: line_no, module: module, message: message
+    end
   end
 
-  defp extract_tests([], "", "", acc, _, _, _) do
-    Enum.reverse(acc)
+  defp chunk_tests(lines, acc, last_no) do
+    case lines
+         |> Enum.drop_while(&(not test_started?(&1)))
+         |> Enum.split_while(&(not test_finished?(&1))) do
+      {[], []} -> Enum.reverse(acc)
+      {chunk, []} -> Enum.reverse([{chunk, last_no} | acc])
+      {[], [_empty_line | lines]} -> chunk_tests(lines, acc, last_no)
+      {chunk, [{_, line_no} | lines]} -> chunk_tests(lines, [{chunk, line_no} | acc], last_no)
+    end
   end
 
-  # End of input and we've still got a test pending.
-  defp extract_tests([], expr_acc, expected_acc, [test | rest], _, _, formatted) do
-    test = add_expr(test, expr_acc, expected_acc, formatted)
-    Enum.reverse([test | rest])
+  defp test_started?({"iex>" <> _, _}), do: true
+  defp test_started?(_), do: false
+
+  defp test_finished?({"", _}), do: true
+  defp test_finished?(_), do: false
+
+  defp extract_tests(line_no, doc, module, fun_arity) do
+    {lines, last_line} =
+      doc
+      |> String.split(["\r\n", "\n"], trim: false)
+      |> adjust_indent(line_no + 1, module)
+
+    lines
+    |> chunk_tests([], last_line)
+    |> Enum.map(&build_test(&1, fun_arity))
   end
 
-  # We've encountered the next test on an adjacent line. Put them into one group.
-  defp extract_tests(
-         [{"iex>" <> _, _} | _] = list,
-         expr_acc,
-         expected_acc,
-         [test | rest],
-         new_test,
-         module,
-         formatted
-       )
-       when expr_acc != "" and expected_acc != "" do
-    test = add_expr(test, expr_acc, expected_acc, formatted)
-    extract_tests(list, "", "", [test | rest], new_test, module, "")
+  defp build_test({[{"iex>" <> string = line, line_no} | lines], end_line_no}, fun_arity) do
+    exprs = build_test(lines, [string], [], [line], [], line_no)
+    %{line: line_no, end_line: end_line_no - 1, exprs: Enum.reverse(exprs), fun_arity: fun_arity}
   end
 
-  # Store expr_acc and start a new test case.
-  defp extract_tests(
-         [{"iex>" <> string = line, line_no} | lines],
-         "",
-         expected_acc,
+  # Started a new expression.
+  defp build_test(
+         [{"iex>" <> _, new_line_no} | _] = list,
+         [_ | _] = expr,
+         [_ | _] = expected,
+         formatted,
          acc,
-         true,
-         module,
-         _
+         line_no
        ) do
-    test = %{line: line_no, fun_arity: nil, exprs: []}
-    extract_tests(lines, string, expected_acc, [test | acc], false, module, line)
+    acc = add_expr(acc, expr, expected, formatted, line_no)
+    build_test(list, [], [], [], acc, new_line_no)
   end
 
-  # Store expr_acc.
-  defp extract_tests(
+  # Continuation of an expression.
+  defp build_test(
          [{"iex>" <> string = line, _} | lines],
-         "",
-         expected_acc,
+         expr,
+         expected,
+         formatted,
          acc,
-         false,
-         module,
-         _
+         line_no
        ) do
-    extract_tests(lines, string, expected_acc, acc, false, module, line)
+    expr = add_line(expr, string)
+    formatted = add_line(formatted, line)
+    build_test(lines, expr, expected, formatted, acc, line_no)
   end
 
-  # Still gathering expr_acc. Synonym for the next clause.
-  defp extract_tests(
-         [{"iex>" <> string = line, _} | lines],
-         expr_acc,
-         expected_acc,
-         acc,
-         new_test,
-         module,
-         formatted
-       ) do
-    extract_tests(
-      lines,
-      expr_acc <> "\n" <> string,
-      expected_acc,
-      acc,
-      new_test,
-      module,
-      formatted <> "\n" <> line
-    )
-  end
-
-  # Still gathering expr_acc. Synonym for the previous clause.
-  defp extract_tests(
+  # Continuation of an expression.
+  defp build_test(
          [{"...>" <> string = line, _} | lines],
-         expr_acc,
-         expected_acc,
+         expr,
+         expected,
+         formatted,
          acc,
-         new_test,
-         module,
-         formatted
-       )
-       when expr_acc != "" do
-    extract_tests(
-      lines,
-      expr_acc <> "\n" <> string,
-      expected_acc,
-      acc,
-      new_test,
-      module,
-      formatted <> "\n" <> line
-    )
-  end
-
-  # Expression numbers are simply skipped.
-  defp extract_tests(
-         [{<<"iex(", _>> <> string = line, line_no} | lines],
-         expr_acc,
-         expected_acc,
-         acc,
-         new_test,
-         module,
-         formatted
+         line_no
        ) do
-    new_line = {"iex" <> skip_iex_number(string, module, line_no, line), line_no}
-    extract_tests([new_line | lines], expr_acc, expected_acc, acc, new_test, module, formatted)
+    expr = add_line(expr, string)
+    formatted = add_line(formatted, line)
+    build_test(lines, expr, expected, formatted, acc, line_no)
   end
 
-  # Expression numbers are simply skipped redux.
-  defp extract_tests(
-         [{<<"...(", _>> <> string, line_no} = line | lines],
-         expr_acc,
-         expected_acc,
-         acc,
-         new_test,
-         module,
-         formatted
-       ) do
-    new_line = {"..." <> skip_iex_number(string, module, line_no, line), line_no}
-    extract_tests([new_line | lines], expr_acc, expected_acc, acc, new_test, module, formatted)
+  # Expected lines.
+  defp build_test([{line, _} | lines], expr, expected, formatted, acc, line_no) do
+    build_test(lines, expr, add_line(expected, line), formatted, acc, line_no)
   end
 
-  # Skip empty or documentation line.
-  defp extract_tests([_ | lines], "", "", acc, _, module, _formatted) do
-    extract_tests(lines, "", "", acc, true, module, "")
+  # We are done.
+  defp build_test([], [_ | _] = expr, expected, formatted, acc, line_no) do
+    add_expr(acc, expr, expected, formatted, line_no)
   end
 
-  # Encountered end of fenced code block, store pending test
-  defp extract_tests(
-         [{<<fence::3-bytes>> <> _, _} | lines],
-         expr_acc,
-         expected_acc,
-         [test | rest],
-         _new_test,
-         module,
-         formatted
-       )
-       when fence in @fences and expr_acc != "" do
-    test = add_expr(test, expr_acc, expected_acc, formatted)
-    extract_tests(lines, "", "", [test | rest], true, module, "")
+  defp add_line([], line), do: [line]
+  defp add_line(acc, line), do: [acc, [?\n, line]]
+
+  defp add_expr(exprs, expr_lines, expected_lines, formatted_lines, line_no) do
+    expected = IO.iodata_to_binary(expected_lines)
+    doctest = IO.iodata_to_binary([?\n, formatted_lines, ?\n, expected])
+
+    expr = %{
+      expr: IO.iodata_to_binary(expr_lines),
+      expr_line: line_no,
+      expected: tag_expected(expected),
+      expected_line: line_no + length(expr_lines),
+      doctest: doctest
+    }
+
+    [expr | exprs]
   end
 
-  # Encountered an empty line, store pending test
-  defp extract_tests(
-         [{"", _} | lines],
-         expr_acc,
-         expected_acc,
-         [test | rest],
-         _new_test,
-         module,
-         formatted
-       ) do
-    test = add_expr(test, expr_acc, expected_acc, formatted)
-    extract_tests(lines, "", "", [test | rest], true, module, "")
-  end
-
-  # Finally, parse expected_acc.
-  defp extract_tests([{expected, _} | lines], expr_acc, "", acc, new_test, module, formatted) do
-    extract_tests(lines, expr_acc, expected, acc, new_test, module, formatted)
-  end
-
-  defp extract_tests(
-         [{expected, _} | lines],
-         expr_acc,
-         expected_acc,
-         acc,
-         new_test,
-         module,
-         formatted
-       ) do
-    extract_tests(
-      lines,
-      expr_acc,
-      expected_acc <> "\n" <> expected,
-      acc,
-      new_test,
-      module,
-      formatted
-    )
-  end
-
-  defp skip_iex_number(")>" <> string, _module, _line_no, _line) do
-    ">" <> string
-  end
-
-  defp skip_iex_number("", module, line_no, line) do
-    message =
-      "unknown IEx prompt: #{inspect(line)}.\nAccepted formats are: iex>, iex(1)>, ...>, ...(1)>}"
-
-    raise Error, line: line_no, module: module, message: message
-  end
-
-  defp skip_iex_number(<<_>> <> string, module, line_no, line) do
-    skip_iex_number(string, module, line_no, line)
-  end
-
-  defp normalize_test(%{exprs: exprs} = test, fa) do
-    %{test | fun_arity: fa, exprs: Enum.reverse(exprs)}
-  end
-
-  defp add_expr(%{exprs: exprs} = test, expr, expected, formatted) do
-    %{test | exprs: [{expr, tag_expected(expected), formatted} | exprs]}
-  end
-
-  defp tag_expected(string) do
-    case string do
+  defp tag_expected(expected) do
+    case expected do
       "" ->
         :test
 
@@ -876,10 +845,10 @@ defmodule ExUnit.DocTest do
         {:error, Module.concat([mod]), String.trim_leading(message)}
 
       _ ->
-        if inspectable?(string) do
-          {:inspect, inspect(string)}
+        if inspectable?(expected) do
+          {:inspect, expected}
         else
-          {:test, string}
+          {:test, expected}
         end
     end
   end
