@@ -3,7 +3,6 @@ defmodule ExUnit.Runner do
 
   alias ExUnit.EventManager, as: EM
 
-  @rand_algorithm :exs1024
   @current_key __MODULE__
 
   def run(opts, load_us) when (is_integer(load_us) or is_nil(load_us)) and is_list(opts) do
@@ -43,7 +42,11 @@ defmodule ExUnit.Runner do
 
     start_time = System.monotonic_time()
     EM.suite_started(config.manager, opts)
-    async_stop_time = async_loop(config, %{}, false)
+
+    modules_to_restore =
+      if Keyword.fetch!(opts, :repeat_until_failure) > 0, do: {[], []}, else: nil
+
+    {async_stop_time, modules_to_restore} = async_loop(config, %{}, false, modules_to_restore)
     stop_time = System.monotonic_time()
 
     if max_failures_reached?(config) do
@@ -62,7 +65,7 @@ defmodule ExUnit.Runner do
     EM.stop(config.manager)
     after_suite_callbacks = Application.fetch_env!(:ex_unit, :after_suite)
     Enum.each(after_suite_callbacks, fn callback -> callback.(stats) end)
-    stats
+    {stats, modules_to_restore}
   end
 
   defp configure(opts, manager, runner_pid, stats_pid) do
@@ -76,6 +79,7 @@ defmodule ExUnit.Runner do
       max_cases: opts[:max_cases],
       max_failures: opts[:max_failures],
       only_test_ids: opts[:only_test_ids],
+      rand_algorithm: opts[:rand_algorithm],
       runner_pid: runner_pid,
       seed: opts[:seed],
       stats_pid: stats_pid,
@@ -92,22 +96,24 @@ defmodule ExUnit.Runner do
     |> Keyword.put(:include, include)
   end
 
-  defp async_loop(config, running, async_once?) do
+  defp async_loop(config, running, async_once?, modules_to_restore) do
     available = config.max_cases - map_size(running)
 
     cond do
       # No modules available, wait for one
       available <= 0 ->
         running = wait_until_available(config, running)
-        async_loop(config, running, async_once?)
+        async_loop(config, running, async_once?, modules_to_restore)
 
       # Slots are available, start with async modules
-      modules = ExUnit.Server.take_async_modules(available) ->
-        running = spawn_modules(config, modules, running)
-        async_loop(config, running, true)
+      async_modules = ExUnit.Server.take_async_modules(available) ->
+        running = spawn_modules(config, async_modules, true, running)
+        modules_to_restore = maybe_store_modules(modules_to_restore, :async, async_modules)
+        async_loop(config, running, true, modules_to_restore)
 
       true ->
         sync_modules = ExUnit.Server.take_sync_modules()
+        modules_to_restore = maybe_store_modules(modules_to_restore, :sync, sync_modules)
 
         # Wait for all async modules
         0 =
@@ -118,12 +124,12 @@ defmodule ExUnit.Runner do
         async_stop_time = if async_once?, do: System.monotonic_time(), else: nil
 
         # Run all sync modules directly
-        for module <- sync_modules do
-          running = spawn_modules(config, [module], %{})
+        for pair <- sync_modules do
+          running = spawn_modules(config, [{nil, [pair]}], false, %{})
           running != %{} and wait_until_available(config, running)
         end
 
-        async_stop_time
+        {async_stop_time, modules_to_restore}
     end
   end
 
@@ -151,19 +157,29 @@ defmodule ExUnit.Runner do
     end
   end
 
-  defp spawn_modules(_config, [], running) do
+  defp spawn_modules(_config, [], _async, running) do
     running
   end
 
-  defp spawn_modules(config, [module | modules], running) do
+  defp spawn_modules(config, [{_group, [_ | _] = modules} | groups], async?, running) do
     if max_failures_reached?(config) do
       running
     else
-      {pid, ref} = spawn_monitor(fn -> run_module(config, module) end)
-      spawn_modules(config, modules, Map.put(running, ref, pid))
+      {pid, ref} =
+        spawn_monitor(fn ->
+          Enum.each(modules, fn {module, params} ->
+            run_module(config, module, async?, params)
+          end)
+        end)
+
+      spawn_modules(config, groups, async?, Map.put(running, ref, pid))
     end
   end
 
+  defp maybe_store_modules(nil, _type, _modules), do: nil
+  defp maybe_store_modules({async, sync}, :async, modules), do: {async ++ modules, sync}
+  defp maybe_store_modules({async, sync}, :sync, modules), do: {async, sync ++ modules}
+
   ## Stacktrace
 
   # Assertions can pop-up in the middle of the stack
@@ -211,20 +227,20 @@ defmodule ExUnit.Runner do
 
   ## Running modules
 
-  defp run_module(config, module) do
-    test_module = module.__ex_unit__()
+  defp run_module(config, module, async?, params) do
+    test_module = %{module.__ex_unit__() | parameters: params}
     EM.module_started(config.manager, test_module)
 
     # Prepare tests, selecting which ones should be run or skipped
-    tests = prepare_tests(config, test_module.tests)
-    {excluded_and_skipped_tests, to_run_tests} = Enum.split_with(tests, & &1.state)
+    {to_run_tests, excluded_and_skipped_tests} = prepare_tests(config, async?, test_module.tests)
 
     for excluded_or_skipped_test <- excluded_and_skipped_tests do
       EM.test_started(config.manager, excluded_or_skipped_test)
       EM.test_finished(config.manager, excluded_or_skipped_test)
     end
 
-    {test_module, invalid_tests, finished_tests} = run_module(config, test_module, to_run_tests)
+    {test_module, invalid_tests, finished_tests} =
+      run_module_tests(config, test_module, async?, to_run_tests)
 
     pending_tests =
       case process_max_failures(config, test_module) do
@@ -251,82 +267,108 @@ defmodule ExUnit.Runner do
     end
   end
 
-  defp prepare_tests(config, tests) do
+  defp prepare_tests(config, async?, tests) do
     tests = shuffle(config, tests)
     include = config.include
     exclude = config.exclude
     test_ids = config.only_test_ids
 
-    for test <- tests, include_test?(test_ids, test) do
-      tags = Map.merge(test.tags, %{test: test.name, module: test.module})
+    {to_run, to_skip} =
+      for test <- tests, include_test?(test_ids, test), reduce: {[], []} do
+        {to_run, to_skip} ->
+          tags = Map.merge(test.tags, %{test: test.name, module: test.module, async: async?})
 
-      case ExUnit.Filters.eval(include, exclude, tags, tests) do
-        :ok -> %{test | tags: tags}
-        excluded_or_skipped -> %{test | state: excluded_or_skipped}
+          case ExUnit.Filters.eval(include, exclude, tags, tests) do
+            :ok -> {[%{test | tags: tags} | to_run], to_skip}
+            excluded_or_skipped -> {to_run, [%{test | state: excluded_or_skipped} | to_skip]}
+          end
       end
-    end
+
+    {Enum.reverse(to_run), Enum.reverse(to_skip)}
   end
 
   defp include_test?(test_ids, test) do
     test_ids == nil or MapSet.member?(test_ids, {test.module, test.name})
   end
 
-  defp run_module(_config, test_module, []) do
+  defp run_module_tests(_config, test_module, _async?, []) do
     {test_module, [], []}
   end
 
-  defp run_module(config, test_module, tests) do
-    {module_pid, module_ref} = run_setup_all(test_module, self())
+  defp run_module_tests(config, test_module, async?, tests) do
+    Process.put(@current_key, test_module)
+    %ExUnit.TestModule{name: module, tags: tags, parameters: params} = test_module
+    context = tags |> Map.merge(params) |> Map.merge(%{module: module, async: async?})
 
-    {test_module, invalid_tests, finished_tests} =
+    config
+    |> run_setup_all(test_module, context, fn context ->
+      if max_failures_reached?(config),
+        do: [],
+        else: run_tests(config, tests, test_module.parameters, context)
+    end)
+    |> case do
+      {{:ok, finished_tests}, test_module} ->
+        {test_module, [], finished_tests}
+
+      {:error, test_module} ->
+        {test_module, Enum.map(tests, &%{&1 | state: {:invalid, test_module}}), []}
+    end
+  end
+
+  defp run_setup_all(
+         _config,
+         %ExUnit.TestModule{setup_all?: false} = test_module,
+         context,
+         callback
+       ) do
+    {{:ok, callback.(context)}, test_module}
+  end
+
+  defp run_setup_all(config, %ExUnit.TestModule{name: module} = test_module, context, callback) do
+    parent_pid = self()
+
+    {module_pid, module_ref} =
+      spawn_monitor(fn ->
+        ExUnit.OnExitHandler.register(self())
+
+        result =
+          try do
+            {:ok, module.__ex_unit__(:setup_all, context)}
+          catch
+            kind, error ->
+              failed = failed(kind, error, prune_stacktrace(__STACKTRACE__))
+              {:error, %{test_module | state: failed}}
+          end
+
+        send(parent_pid, {self(), :setup_all, result})
+
+        # We keep the process alive so all of its resources
+        # stay alive until we run all tests in this case.
+        ref = Process.monitor(parent_pid)
+
+        receive do
+          {^parent_pid, :exit} -> :ok
+          {:DOWN, ^ref, _, _, _} -> :ok
+        end
+      end)
+
+    {ok_or_error, test_module} =
       receive do
         {^module_pid, :setup_all, {:ok, context}} ->
-          finished_tests =
-            if max_failures_reached?(config), do: [], else: run_tests(config, tests, context)
-
+          finished_tests = callback.(context)
           :ok = exit_setup_all(module_pid, module_ref)
-          {test_module, [], finished_tests}
+          {{:ok, finished_tests}, test_module}
 
         {^module_pid, :setup_all, {:error, test_module}} ->
-          invalid_tests = Enum.map(tests, &%{&1 | state: {:invalid, test_module}})
           :ok = exit_setup_all(module_pid, module_ref)
-          {test_module, invalid_tests, []}
+          {:error, test_module}
 
         {:DOWN, ^module_ref, :process, ^module_pid, error} ->
-          test_module = %{test_module | state: failed({:EXIT, module_pid}, error, [])}
-          {test_module, [], []}
+          {:error, %{test_module | state: failed({:EXIT, module_pid}, error, [])}}
       end
 
     timeout = get_timeout(config, %{})
-    {exec_on_exit(test_module, module_pid, timeout), invalid_tests, finished_tests}
-  end
-
-  defp run_setup_all(%ExUnit.TestModule{name: module} = test_module, parent_pid) do
-    Process.put(@current_key, test_module)
-
-    spawn_monitor(fn ->
-      ExUnit.OnExitHandler.register(self())
-
-      result =
-        try do
-          {:ok, module.__ex_unit__(:setup_all, %{module: module, case: module})}
-        catch
-          kind, error ->
-            failed = failed(kind, error, prune_stacktrace(__STACKTRACE__))
-            {:error, %{test_module | state: failed}}
-        end
-
-      send(parent_pid, {self(), :setup_all, result})
-
-      # We keep the process alive so all of its resources
-      # stay alive until we run all tests in this case.
-      ref = Process.monitor(parent_pid)
-
-      receive do
-        {^parent_pid, :exit} -> :ok
-        {:DOWN, ^ref, _, _, _} -> :ok
-      end
-    end)
+    {ok_or_error, exec_on_exit(test_module, module_pid, timeout)}
   end
 
   defp exit_setup_all(pid, ref) do
@@ -337,8 +379,9 @@ defmodule ExUnit.Runner do
     end
   end
 
-  defp run_tests(config, tests, context) do
+  defp run_tests(config, tests, params, context) do
     Enum.reduce_while(tests, [], fn test, acc ->
+      test = %{test | parameters: params}
       Process.put(@current_key, test)
 
       case run_test(config, test, context) do
@@ -348,12 +391,9 @@ defmodule ExUnit.Runner do
     end)
   end
 
-  defp run_test(config, %{tags: tags} = test, context) do
+  defp run_test(config, test, context) do
     EM.test_started(config.manager, test)
-
-    context = maybe_create_tmp_dir(test, context, tags)
-    capture_log = Map.get(tags, :capture_log, config.capture_log)
-    test = run_test_with_capture_log(capture_log, config, test, Map.merge(tags, context))
+    test = spawn_test(config, test, context)
 
     case process_max_failures(config, test) do
       :no ->
@@ -369,83 +409,6 @@ defmodule ExUnit.Runner do
     end
   end
 
-  defp maybe_create_tmp_dir(test, context, %{tmp_dir: true}) do
-    create_tmp_dir!(test, "", context)
-  end
-
-  defp maybe_create_tmp_dir(test, context, %{tmp_dir: path}) when is_binary(path) do
-    create_tmp_dir!(test, path, context)
-  end
-
-  defp maybe_create_tmp_dir(_, context, %{tmp_dir: false}) do
-    context
-  end
-
-  defp maybe_create_tmp_dir(_, _, %{tmp_dir: other}) do
-    raise ArgumentError, "expected :tmp_dir to be a boolean or a string, got: #{inspect(other)}"
-  end
-
-  defp maybe_create_tmp_dir(_, context, _) do
-    context
-  end
-
-  defp short_hash(module, test_name) do
-    (module <> "/" <> test_name)
-    |> :erlang.md5()
-    |> Base.encode16(case: :lower)
-    |> binary_slice(0..7)
-  end
-
-  defp create_tmp_dir!(test, extra_path, context) do
-    module_string = inspect(test.module)
-    name_string = to_string(test.name)
-
-    module = escape_path(module_string)
-    name = escape_path(name_string)
-    short_hash = short_hash(module_string, name_string)
-
-    path = ["tmp", module, "#{name}-#{short_hash}", extra_path] |> Path.join() |> Path.expand()
-    File.rm_rf!(path)
-    File.mkdir_p!(path)
-    Map.put(context, :tmp_dir, path)
-  end
-
-  @escape Enum.map(' [~#%&*{}\\:<>?/+|"]', &<<&1::utf8>>)
-
-  defp escape_path(path) do
-    String.replace(path, @escape, "-")
-  end
-
-  defp run_test_with_capture_log(true, config, test, context) do
-    run_test_with_capture_log([], config, test, context)
-  end
-
-  defp run_test_with_capture_log(false, config, test, context) do
-    spawn_test(config, test, context)
-  end
-
-  defp run_test_with_capture_log(capture_log_opts, config, test, context) do
-    ref = make_ref()
-
-    try do
-      ExUnit.CaptureLog.capture_log(capture_log_opts, fn ->
-        send(self(), {ref, spawn_test(config, test, context)})
-      end)
-    catch
-      :exit, :noproc ->
-        message =
-          "could not run test, it uses @tag :capture_log" <>
-            " but the :logger application is not running"
-
-        %{test | state: failed(:error, RuntimeError.exception(message), [])}
-    else
-      logged ->
-        receive do
-          {^ref, test} -> %{test | logs: logged}
-        end
-    end
-  end
-
   defp spawn_test(config, test, context) do
     parent_pid = self()
     timeout = get_timeout(config, test.tags)
@@ -454,24 +417,60 @@ defmodule ExUnit.Runner do
     exec_on_exit(test, test_pid, timeout)
   end
 
-  defp spawn_test_monitor(%{seed: seed}, test, parent_pid, context) do
+  defp spawn_test_monitor(
+         %{seed: seed, capture_log: capture_log, rand_algorithm: rand_algorithm},
+         test,
+         parent_pid,
+         context
+       ) do
     spawn_monitor(fn ->
       ExUnit.OnExitHandler.register(self())
-      generate_test_seed(seed, test)
+      generate_test_seed(seed, test, rand_algorithm)
+      context = context |> Map.merge(test.tags) |> Map.put(:test_pid, self())
+      capture_log = Map.get(context, :capture_log, capture_log)
 
       {time, test} =
-        :timer.tc(fn ->
-          case exec_test_setup(test, context) do
-            {:ok, test} -> exec_test(test)
-            {:error, test} -> test
-          end
-        end)
+        :timer.tc(
+          maybe_capture_log(capture_log, test, fn ->
+            context = maybe_create_tmp_dir(context, test)
+
+            case exec_test_setup(test, context) do
+              {:ok, context} -> exec_test(test, context)
+              {:error, test} -> test
+            end
+          end)
+        )
 
       send(parent_pid, {self(), :test_finished, %{test | time: time}})
       exit(:shutdown)
     end)
   end
 
+  defp maybe_capture_log(true, test, fun) do
+    maybe_capture_log([], test, fun)
+  end
+
+  defp maybe_capture_log(false, _test, fun) do
+    fun
+  end
+
+  defp maybe_capture_log(capture_log_opts, test, fun) do
+    fn ->
+      try do
+        ExUnit.CaptureLog.with_log(capture_log_opts, fun)
+      catch
+        :exit, :noproc ->
+          message =
+            "could not run test, it uses @tag :capture_log" <>
+              " but the :logger application is not running"
+
+          %{test | state: failed(:error, RuntimeError.exception(message), [])}
+      else
+        {test, logs} -> %{test | logs: logs}
+      end
+    end
+  end
+
   defp receive_test_reply(test, test_pid, test_ref, timeout) do
     receive do
       {^test_pid, :test_finished, test} ->
@@ -502,13 +501,13 @@ defmodule ExUnit.Runner do
   end
 
   defp exec_test_setup(%ExUnit.Test{module: module} = test, context) do
-    {:ok, %{test | tags: module.__ex_unit__(:setup, context)}}
+    {:ok, module.__ex_unit__(:setup, context)}
   catch
     kind, error ->
       {:error, %{test | state: failed(kind, error, prune_stacktrace(__STACKTRACE__))}}
   end
 
-  defp exec_test(%ExUnit.Test{module: module, name: name, tags: context} = test) do
+  defp exec_test(%ExUnit.Test{module: module, name: name} = test, context) do
     apply(module, name, [context])
     test
   catch
@@ -529,8 +528,8 @@ defmodule ExUnit.Runner do
 
   ## Helpers
 
-  defp generate_test_seed(seed, %ExUnit.Test{module: module, name: name}) do
-    :rand.seed(@rand_algorithm, {:erlang.phash2(module), :erlang.phash2(name), seed})
+  defp generate_test_seed(seed, %ExUnit.Test{module: module, name: name}, rand_algorithm) do
+    :rand.seed(rand_algorithm, {:erlang.phash2(module), :erlang.phash2(name), seed})
   end
 
   defp process_max_failures(%{max_failures: :infinity}, _), do: :no
@@ -571,11 +570,11 @@ defmodule ExUnit.Runner do
   end
 
   defp shuffle(%{seed: 0}, list) do
-    Enum.reverse(list)
+    list
   end
 
-  defp shuffle(%{seed: seed}, list) do
-    _ = :rand.seed(@rand_algorithm, {seed, seed, seed})
+  defp shuffle(%{seed: seed, rand_algorithm: rand_algorithm}, list) do
+    _ = :rand.seed(rand_algorithm, {seed, seed, seed})
     Enum.shuffle(list)
   end
 
@@ -591,4 +590,51 @@ defmodule ExUnit.Runner do
   defp failed(kind, reason, stack) do
     {:failed, [{kind, Exception.normalize(kind, reason, stack), stack}]}
   end
+
+  ## Tmp dir handling
+
+  defp maybe_create_tmp_dir(%{tmp_dir: true} = tags, test) do
+    create_tmp_dir!(test, "", tags)
+  end
+
+  defp maybe_create_tmp_dir(%{tmp_dir: path} = tags, test) when is_binary(path) do
+    create_tmp_dir!(test, path, tags)
+  end
+
+  defp maybe_create_tmp_dir(%{tmp_dir: other}, _test) when other != false do
+    raise ArgumentError, "expected :tmp_dir to be a boolean or a string, got: #{inspect(other)}"
+  end
+
+  defp maybe_create_tmp_dir(tags, _test) do
+    tags
+  end
+
+  defp short_hash(module, test_name, parameters) do
+    suffix = if parameters == %{}, do: "", else: :erlang.term_to_binary(parameters)
+
+    (module <> "/" <> test_name <> suffix)
+    |> :erlang.md5()
+    |> Base.encode16(case: :lower)
+    |> binary_slice(0..7)
+  end
+
+  defp create_tmp_dir!(test, extra_path, tags) do
+    module_string = inspect(test.module)
+    name_string = to_string(test.name)
+
+    module = escape_path(module_string)
+    name = escape_path(name_string)
+    short_hash = short_hash(module_string, name_string, test.parameters)
+
+    path = ["tmp", module, "#{name}-#{short_hash}", extra_path] |> Path.join() |> Path.expand()
+    File.rm_rf!(path)
+    File.mkdir_p!(path)
+    Map.put(tags, :tmp_dir, path)
+  end
+
+  @escape Enum.map(~c" [~#%&*{}\\:<>?/+|\"]", &<<&1::utf8>>)
+
+  defp escape_path(path) do
+    String.replace(path, @escape, "-")
+  end
 end
