@@ -78,7 +78,8 @@ defmodule ExUnit.AssertionsTest do
 
   test "assert with message when value is falsy" do
     try do
-      "This should never be tested" = assert Value.falsy(), "This should be truthy"
+      assert Value.falsy(), "This should be truthy"
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "This should be truthy" = error.message
@@ -87,7 +88,8 @@ defmodule ExUnit.AssertionsTest do
 
   test "assert when value evaluates to falsy" do
     try do
-      "This should never be tested" = assert Value.falsy()
+      assert Value.falsy()
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "assert Value.falsy()" = error.expr |> Macro.to_string()
@@ -110,7 +112,8 @@ defmodule ExUnit.AssertionsTest do
 
   test "assert arguments are not kept for operators" do
     try do
-      "This should never be tested" = assert !Value.truthy()
+      assert !Value.truthy()
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         false = is_list(error.args)
@@ -119,7 +122,8 @@ defmodule ExUnit.AssertionsTest do
 
   test "assert with equality" do
     try do
-      "This should never be tested" = assert 1 + 1 == 1
+      assert 1 + 1 == 1
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         1 = error.right
@@ -130,7 +134,8 @@ defmodule ExUnit.AssertionsTest do
 
   test "assert with equality in reverse" do
     try do
-      "This should never be tested" = assert 1 == 1 + 1
+      assert 1 == 1 + 1
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         1 = error.left
@@ -199,8 +204,18 @@ defmodule ExUnit.AssertionsTest do
              """)
            end) =~ "variable \"var\" is unused"
   after
-    :code.delete(ExSample)
     :code.purge(ExSample)
+    :code.delete(ExSample)
+  end
+
+  test "assert match with quote on left-side" do
+    assert quote(do: x in Alias) = quote(do: x in Alias)
+  end
+
+  defmacro quote_like(ast), do: Macro.escape(ast, prune_metadata: true)
+
+  test "assert match with quote-like on left-side" do
+    assert quote_like(x in Alias) = quote_like(x in Alias)
   end
 
   test "assert match expands argument in match context" do
@@ -246,16 +261,66 @@ defmodule ExUnit.AssertionsTest do
     end
   end
 
+  test "assert match with `when` in the pattern fails" do
+    message = """
+    invalid pattern in assert\/1:
+
+      x when is_map(x)
+
+    To assert with guards, use match?/2:
+
+      assert match?(x when is_map(x), %{})
+    """
+
+    assert_raise ArgumentError, message, fn ->
+      Code.eval_string("""
+      defmodule AssertGuard do
+        import ExUnit.Assertions
+
+        def run do
+          assert (x when is_map(x)) = %{}
+        end
+      end
+      """)
+    end
+  after
+    :code.purge(AssertGuard)
+    :code.delete(AssertGuard)
+  end
+
+  test "assert match with __ENV__ in the pattern" do
+    message =
+      ExUnit.CaptureIO.capture_io(:stderr, fn ->
+        assert_raise CompileError, fn ->
+          Code.eval_string("""
+          defmodule EnvMatch do
+            import ExUnit.Assertions
+
+            def run do
+              assert __ENV__ = %{}
+            end
+          end
+          """)
+        end
+      end)
+
+    assert message =~ "invalid pattern in match, __ENV__ is not allowed in matches"
+  after
+    :code.purge(EnvMatch)
+    :code.delete(EnvMatch)
+  end
+
   test "assert match?" do
     true = assert match?({2, 1}, Value.tuple())
 
     try do
-      "This should never be tested" = assert match?({:ok, _}, error(true))
+      assert match?({:ok, _}, Process.get(:unused, :ok))
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "match (match?) failed" = error.message
-        "assert match?({:ok, _}, error(true))" = Macro.to_string(error.expr)
-        "{:error, true}" = Macro.to_string(error.right)
+        "assert match?({:ok, _}, Process.get(:unused, :ok))" = Macro.to_string(error.expr)
+        ":ok" = Macro.to_string(error.right)
     end
   end
 
@@ -263,7 +328,8 @@ defmodule ExUnit.AssertionsTest do
     true = assert match?(tuple when is_tuple(tuple), Value.tuple())
 
     try do
-      "This should never be tested" = assert match?(tuple when not is_tuple(tuple), error(true))
+      assert match?(tuple when not is_tuple(tuple), error(true))
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "match (match?) failed" = error.message
@@ -278,7 +344,8 @@ defmodule ExUnit.AssertionsTest do
     false = refute match?({1, 1}, Value.tuple())
 
     try do
-      "This should never be tested" = refute match?({:error, _}, error(true))
+      refute match?({:error, _}, error(true))
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "match (match?) succeeded, but should have failed" = error.message
@@ -291,7 +358,8 @@ defmodule ExUnit.AssertionsTest do
     a = 1
 
     try do
-      "This should never be tested" = assert(match?({^a, 1}, Value.tuple()))
+      assert(match?({^a, 1}, Value.tuple()))
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "match (match?) failed\nThe following variables were pinned:\n  a = 1" = error.message
@@ -304,7 +372,8 @@ defmodule ExUnit.AssertionsTest do
     a = 2
 
     try do
-      "This should never be tested" = refute(match?({^a, 1}, Value.tuple()))
+      refute(match?({^a, 1}, Value.tuple()))
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         """
@@ -400,7 +469,8 @@ defmodule ExUnit.AssertionsTest do
     send(self(), {:status, :invalid})
 
     try do
-      "This should never be tested" = assert_received {:status, ^status}
+      assert_received {:status, ^status}
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         """
@@ -420,7 +490,8 @@ defmodule ExUnit.AssertionsTest do
     send(self(), {:status, :invalid, :invalid})
 
     try do
-      "This should never be tested" = assert_received {:status, ^status, ^status}
+      assert_received {:status, ^status, ^status}
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         """
@@ -442,7 +513,8 @@ defmodule ExUnit.AssertionsTest do
     send(self(), {:status, :invalid, :invalid})
 
     try do
-      "This should never be tested" = assert_received {:status, ^status, ^other_status}
+      assert_received {:status, ^status, ^other_status}
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         """
@@ -460,7 +532,8 @@ defmodule ExUnit.AssertionsTest do
 
   test "assert received when empty mailbox" do
     try do
-      "This should never be tested" = assert_received :hello
+      assert_received :hello
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "Assertion failed, no matching message after 0ms\nThe process mailbox is empty." =
@@ -474,7 +547,8 @@ defmodule ExUnit.AssertionsTest do
     send(self(), {:message, :not_expected, :at_all})
 
     try do
-      "This should never be tested" = assert_received :hello
+      assert_received :hello
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         """
@@ -491,7 +565,8 @@ defmodule ExUnit.AssertionsTest do
     for i <- 1..11, do: send(self(), {:message, i})
 
     try do
-      "This should never be tested" = assert_received x when x == :hello
+      assert_received x when x == :hello
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         """
@@ -529,7 +604,8 @@ defmodule ExUnit.AssertionsTest do
     send(self(), :hello)
 
     try do
-      "This should never be tested" = refute_received :hello
+      refute_received :hello
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "Unexpectedly received message :hello (which matched :hello)" = error.message
@@ -537,32 +613,34 @@ defmodule ExUnit.AssertionsTest do
   end
 
   test "assert in when member" do
-    true = assert 'foo' in ['foo', 'bar']
+    true = assert ~c"foo" in [~c"foo", ~c"bar"]
   end
 
   test "assert in when is not member" do
     try do
-      "This should never be tested" = assert 'foo' in 'bar'
+      assert ~c"foo" in ~c"bar"
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
-        'foo' = error.left
-        'bar' = error.right
-        "assert 'foo' in 'bar'" = Macro.to_string(error.expr)
+        ~c"foo" = error.left
+        ~c"bar" = error.right
+        ~S(assert ~c"foo" in ~c"bar") = Macro.to_string(error.expr)
     end
   end
 
   test "refute in when is not member" do
-    false = refute 'baz' in ['foo', 'bar']
+    false = refute ~c"baz" in [~c"foo", ~c"bar"]
   end
 
   test "refute in when is member" do
     try do
-      "This should never be tested" = refute 'foo' in ['foo', 'bar']
+      refute ~c"foo" in [~c"foo", ~c"bar"]
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
-        'foo' = error.left
-        ['foo', 'bar'] = error.right
-        "refute 'foo' in ['foo', 'bar']" = Macro.to_string(error.expr)
+        ~c"foo" = error.left
+        [~c"foo", ~c"bar"] = error.right
+        ~S(refute ~c"foo" in [~c"foo", ~c"bar"]) = Macro.to_string(error.expr)
     end
   end
 
@@ -609,7 +687,8 @@ defmodule ExUnit.AssertionsTest do
 
   test "refute match when no match" do
     try do
-      "This should never be tested" = refute _ = ok(true)
+      refute _ = ok(true)
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "refute _ = ok(true)" = Macro.to_string(error.expr)
@@ -623,7 +702,8 @@ defmodule ExUnit.AssertionsTest do
 
   test "assert regex match when no match" do
     try do
-      "This should never be tested" = assert "foo" =~ ~r/a/
+      assert "foo" =~ ~r/a/
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "foo" = error.left
@@ -637,7 +717,8 @@ defmodule ExUnit.AssertionsTest do
 
   test "refute regex match when match" do
     try do
-      "This should never be tested" = refute "foo" =~ ~r/o/
+      refute "foo" =~ ~r/o/
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "foo" = error.left
@@ -646,7 +727,8 @@ defmodule ExUnit.AssertionsTest do
   end
 
   test "assert raise with no error" do
-    "This should never be tested" = assert_raise ArgumentError, fn -> nil end
+    assert_raise ArgumentError, fn -> nil end
+    flunk("This should never be tested")
   rescue
     error in [ExUnit.AssertionError] ->
       "Expected exception ArgumentError but nothing was raised" = error.message
@@ -735,7 +817,8 @@ defmodule ExUnit.AssertionsTest do
   end
 
   test "assert greater-than operator error" do
-    "This should never be tested" = assert 1 > 2
+    assert 1 > 2
+    flunk("This should never be tested")
   rescue
     error in [ExUnit.AssertionError] ->
       1 = error.left
@@ -748,7 +831,8 @@ defmodule ExUnit.AssertionsTest do
   end
 
   test "assert less or equal than operator error" do
-    "This should never be tested" = assert 2 <= 1
+    assert 2 <= 1
+    flunk("This should never be tested")
   rescue
     error in [ExUnit.AssertionError] ->
       "assert 2 <= 1" = Macro.to_string(error.expr)
@@ -762,7 +846,8 @@ defmodule ExUnit.AssertionsTest do
   end
 
   test "assert operator with custom message" do
-    "This should never be tested" = assert 1 > 2, "assertion"
+    assert 1 > 2, "assertion"
+    flunk("This should never be tested")
   rescue
     error in [ExUnit.AssertionError] ->
       "assertion" = error.message
@@ -770,7 +855,8 @@ defmodule ExUnit.AssertionsTest do
 
   test "assert lack of equality" do
     try do
-      "This should never be tested" = assert "one" != "one"
+      assert "one" != "one"
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "Assertion with != failed, both sides are exactly equal" = error.message
@@ -778,7 +864,8 @@ defmodule ExUnit.AssertionsTest do
     end
 
     try do
-      "This should never be tested" = assert 2 != 2.0
+      assert 2 != 2.0
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "Assertion with != failed" = error.message
@@ -789,7 +876,8 @@ defmodule ExUnit.AssertionsTest do
 
   test "refute equality" do
     try do
-      "This should never be tested" = refute "one" == "one"
+      refute "one" == "one"
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "Refute with == failed, both sides are exactly equal" = error.message
@@ -797,7 +885,8 @@ defmodule ExUnit.AssertionsTest do
     end
 
     try do
-      "This should never be tested" = refute 2 == 2.0
+      refute 2 == 2.0
+      flunk("This should never be tested")
     rescue
       error in [ExUnit.AssertionError] ->
         "Refute with == failed" = error.message
@@ -821,7 +910,8 @@ defmodule ExUnit.AssertionsTest do
   end
 
   test "assert in delta error" do
-    "This should never be tested" = assert_in_delta(10, 12, 1)
+    assert_in_delta(10, 12, 1)
+    flunk("This should never be tested")
   rescue
     error in [ExUnit.AssertionError] ->
       "Expected the difference between 10 and 12 (2) to be less than or equal to 1" =
@@ -829,7 +919,8 @@ defmodule ExUnit.AssertionsTest do
   end
 
   test "assert in delta with message" do
-    "This should never be tested" = assert_in_delta(10, 12, 1, "test message")
+    assert_in_delta(10, 12, 1, "test message")
+    flunk("This should never be tested")
   rescue
     error in [ExUnit.AssertionError] ->
       "test message" = error.message
@@ -840,14 +931,16 @@ defmodule ExUnit.AssertionsTest do
   end
 
   test "refute in delta error" do
-    "This should never be tested" = refute_in_delta(10, 11, 2)
+    refute_in_delta(10, 11, 2)
+    flunk("This should never be tested")
   rescue
     error in [ExUnit.AssertionError] ->
       "Expected the difference between 10 and 11 (1) to be more than 2" = error.message
   end
 
   test "refute in delta with message" do
-    "This should never be tested" = refute_in_delta(10, 11, 2, "test message")
+    refute_in_delta(10, 11, 2, "test message")
+    flunk("This should never be tested")
   rescue
     error in [ExUnit.AssertionError] ->
       "test message (difference between 10 and 11 is less than 2)" = error.message
@@ -887,21 +980,24 @@ defmodule ExUnit.AssertionsTest do
   end
 
   test "flunk" do
-    "This should never be tested" = flunk()
+    flunk()
+    flunk("This should never be tested")
   rescue
     error in [ExUnit.AssertionError] ->
       "Flunked!" = error.message
   end
 
   test "flunk with message" do
-    "This should never be tested" = flunk("This should raise an error")
+    flunk("This should raise an error")
+    flunk("This should never be tested")
   rescue
     error in [ExUnit.AssertionError] ->
       "This should raise an error" = error.message
   end
 
   test "flunk with wrong argument type" do
-    "This should never be tested" = flunk(["flunk takes a binary, not a list"])
+    flunk(["flunk takes a binary, not a list"])
+    flunk("This should never be tested")
   rescue
     error ->
       "no function clause matching in ExUnit.Assertions.flunk/1" =
