@@ -10,6 +10,10 @@ defmodule ExUnit.DiffTest do
     defstruct [:name, :age]
   end
 
+  defmodule Customer do
+    defstruct [:address, :age, :first_name, :language, :last_name, :notifications]
+  end
+
   defmodule Person do
     defstruct [:age]
   end
@@ -26,6 +30,20 @@ defmodule ExUnit.DiffTest do
     end
   end
 
+  defmodule HTML do
+    defstruct [:string]
+
+    defimpl Inspect do
+      def inspect(%{string: string}, _) do
+        "~HTML[#{string}]"
+      end
+    end
+  end
+
+  defmacro sigil_HTML({:<<>>, _, [string]}, []) do
+    Macro.escape(%HTML{string: string})
+  end
+
   defmacrop one, do: 1
 
   defmacrop tuple(a, b) do
@@ -112,10 +130,11 @@ defmodule ExUnit.DiffTest do
     assert env_binding == expected_binding
   end
 
+  @terminal_width 80
   defp to_diff(side, sign) do
     side
     |> Diff.to_algebra(&diff_wrapper(&1, sign))
-    |> Algebra.format(:infinity)
+    |> Algebra.format(@terminal_width)
     |> IO.iodata_to_binary()
   end
 
@@ -148,7 +167,7 @@ defmodule ExUnit.DiffTest do
 
   test "pseudo vars" do
     assert_diff(__MODULE__ = ExUnit.DiffTest, [])
-    refute_diff(__MODULE__ = SomethingElse, "-__MODULE__-", "+SomethingElse+")
+    refute_diff(__MODULE__ = SomethingElse, "-ExUnit.DiffTest-", "+SomethingElse+")
   end
 
   test "integers" do
@@ -237,7 +256,12 @@ defmodule ExUnit.DiffTest do
     refute_diff([:a, [:c, :b]] = [:a, [:b, :c]], "[:a, [-:c-, :b]]", "[:a, [:b, +:c+]]")
     refute_diff(:a = [:a, [:b, :c]], "-:a-", "+[:a, [:b, :c]]+")
 
-    pins = %{{:a, nil} => :a, {:b, nil} => :b, {:list_ab, nil} => [:a, :b]}
+    pins = %{
+      {:a, nil} => :a,
+      {:b, nil} => :b,
+      {:list_ab, nil} => [:a, :b],
+      {:list_tuple, nil} => [{:foo}]
+    }
 
     assert_diff(x = [], [x: []], pins)
     assert_diff(x = [:a, :b], [x: [:a, :b]], pins)
@@ -268,6 +292,9 @@ defmodule ExUnit.DiffTest do
 
     refute_diff([:a, :b] = :a, "-[:a, :b]-", "+:a+")
     refute_diff([:foo] = [:foo, {:a, :b, :c}], "[:foo]", "[:foo, +{:a, :b, :c}+]")
+
+    refute_diff([{:foo}] = [{:bar}], "[{-:foo-}]", "[{+:bar+}]")
+    refute_diff(^list_tuple = [{:bar}], "-^list_tuple-", "[{+:bar+}]", pins)
   end
 
   test "improper lists" do
@@ -553,6 +580,12 @@ defmodule ExUnit.DiffTest do
     refute_diff(%{a: 1, b: 12} == %{a: 1, b: 2}, "%{a: 1, b: -1-2}", "%{a: 1, b: 2}")
     refute_diff(%{a: 1, b: 2} == %{a: 1, b: 12}, "%{a: 1, b: 2}", "%{a: 1, b: +1+2}")
     refute_diff(%{a: 1, b: 2} == %{a: 1, c: 2}, "%{a: 1, -b: 2-}", "%{a: 1, +c: 2+}")
+
+    refute_diff(
+      %{name: {:a, :b, :c}} == :error,
+      "-%{name: {:a, :b, :c}}",
+      "+:error+"
+    )
   end
 
   test "structs" do
@@ -642,6 +675,12 @@ defmodule ExUnit.DiffTest do
       ~s/-~D[2017-10-02]-/,
       ~s/+"2017-10-01"+/
     )
+
+    refute_diff(
+      ~HTML[hi] = ~HTML[bye],
+      "-~HTML[hi]-",
+      "~HTML[+bye+]"
+    )
   end
 
   test "structs with missing keys on match" do
@@ -654,13 +693,49 @@ defmodule ExUnit.DiffTest do
     refute_diff(
       %User{age: %Date{}} = struct,
       ~s/%ExUnit.DiffTest.User{age: %-Date-{}}/,
-      ~s/%ExUnit.DiffTest.User{age: %+DateTime+{calendar: Calendar.ISO, day: 30, hour: 13, microsecond: {253158, 6}, minute: 49, month: 7, second: 59, std_offset: 0, time_zone: "Etc\/UTC", utc_offset: 0, year: 2020, zone_abbr: "UTC"}, name: nil}/
+      """
+      %ExUnit.DiffTest.User{
+        age: %+DateTime+{
+          calendar: Calendar.ISO,
+          day: 30,
+          hour: 13,
+          microsecond: {253158, 6},
+          minute: 49,
+          month: 7,
+          second: 59,
+          std_offset: 0,
+          time_zone: "Etc\/UTC",
+          utc_offset: 0,
+          year: 2020,
+          zone_abbr: "UTC"
+        },
+        name: nil
+      }\
+      """
     )
 
     refute_diff(
       %{age: %Date{}} = struct,
       ~s/%{age: %-Date-{}}/,
-      ~s/%ExUnit.DiffTest.User{age: %+DateTime+{calendar: Calendar.ISO, day: 30, hour: 13, microsecond: {253158, 6}, minute: 49, month: 7, second: 59, std_offset: 0, time_zone: "Etc\/UTC", utc_offset: 0, year: 2020, zone_abbr: "UTC"}, name: nil}/
+      """
+      %ExUnit.DiffTest.User{
+        age: %+DateTime+{
+          calendar: Calendar.ISO,
+          day: 30,
+          hour: 13,
+          microsecond: {253158, 6},
+          minute: 49,
+          month: 7,
+          second: 59,
+          std_offset: 0,
+          time_zone: \"Etc/UTC\",
+          utc_offset: 0,
+          year: 2020,
+          zone_abbr: \"UTC\"
+        },
+        name: nil
+      }\
+      """
     )
   end
 
@@ -682,6 +757,12 @@ defmodule ExUnit.DiffTest do
       ~s/-~D[2017-10-02]-/,
       ~s/+"2017-10-01"+/
     )
+
+    refute_diff(
+      ~HTML[hi] == ~HTML[bye],
+      "~HTML[-hi-]",
+      "~HTML[+bye+]"
+    )
   end
 
   test "structs with same inspect but different inside match" do
@@ -802,6 +883,106 @@ defmodule ExUnit.DiffTest do
     )
   end
 
+  test "maps in lists" do
+    map = %{
+      "address" => %{
+        "street" => "123 Main St",
+        "zip" => "62701"
+      },
+      "age" => 30,
+      "first_name" => "John",
+      "language" => "en-US",
+      "last_name" => "Doe",
+      "notifications" => true
+    }
+
+    refute_diff(
+      [map] == [],
+      """
+      [
+        -%{
+          "address" => %{"street" => "123 Main St", "zip" => "62701"},
+          "age" => 30,
+          "first_name" => "John",
+          "language" => "en-US",
+          "last_name" => "Doe",
+          "notifications" => true
+        }-
+      ]\
+      """,
+      "[]"
+    )
+
+    refute_diff(
+      [] == [map],
+      "[]",
+      """
+      [
+        +%{
+          "address" => %{"street" => "123 Main St", "zip" => "62701"},
+          "age" => 30,
+          "first_name" => "John",
+          "language" => "en-US",
+          "last_name" => "Doe",
+          "notifications" => true
+        }+
+      ]\
+      """
+    )
+
+    assert_diff([map] == [map], [])
+  end
+
+  test "structs in lists" do
+    customer = %Customer{
+      address: %{
+        "street" => "123 Main St",
+        "zip" => "62701"
+      },
+      age: 30,
+      first_name: "John",
+      language: "en-US",
+      last_name: "Doe",
+      notifications: true
+    }
+
+    refute_diff(
+      [customer] == [],
+      """
+      [
+        -%ExUnit.DiffTest.Customer{
+          address: %{"street" => "123 Main St", "zip" => "62701"},
+          age: 30,
+          first_name: "John",
+          language: "en-US",
+          last_name: "Doe",
+          notifications: true
+        }-
+      ]\
+      """,
+      "[]"
+    )
+
+    refute_diff(
+      [] == [customer],
+      "[]",
+      """
+      [
+        +%ExUnit.DiffTest.Customer{
+          address: %{"street" => "123 Main St", "zip" => "62701"},
+          age: 30,
+          first_name: "John",
+          language: "en-US",
+          last_name: "Doe",
+          notifications: true
+        }+
+      ]\
+      """
+    )
+
+    assert_diff([customer] == [customer], [])
+  end
+
   test "maps and structs with escaped values" do
     refute_diff(
       %User{age: {1, 2, 3}} = %User{age: {1, 2, 4}},
@@ -880,9 +1061,16 @@ defmodule ExUnit.DiffTest do
   end
 
   test "concat binaries" do
+    assert_diff("fox hops" <> _ = "fox hops over the dog", [])
     assert_diff("fox hops" <> " over the dog" = "fox hops over the dog", [])
     assert_diff("fox hops " <> "over " <> "the dog" = "fox hops over the dog", [])
 
+    refute_diff(
+      "fox hops" <> _ = "dog hops over the fox",
+      ~s/"-f-o-x- hops" <> _/,
+      ~s/+d+o+g+ hops over the fox/
+    )
+
     refute_diff(
       "fox hops" <> " under the dog" = "fox hops over the dog",
       ~s/"fox hops" <> " -und-er the dog"/,
@@ -958,8 +1146,8 @@ defmodule ExUnit.DiffTest do
     assert_diff(one() = 1, [])
     assert_diff(tuple(x, x) = {1, 1}, x: 1)
 
-    refute_diff(one() = 2, "-one()-", "+2+")
-    refute_diff(tuple(x, x) = {1, 2}, "-tuple(x, x)-", "{1, +2+}")
+    refute_diff(one() = 2, "-1-", "+2+")
+    refute_diff(tuple(x, x) = {1, 2}, "{x, -x-}", "{1, +2+}")
 
     pins = %{{:x, nil} => 1}
     assert_diff(pin_x() = 1, [], pins)
@@ -1011,15 +1199,20 @@ defmodule ExUnit.DiffTest do
 
   test "charlists" do
     refute_diff(
-      'fox hops over \'the dog' = 'fox jumps over the lazy cat',
-      "'fox -ho-ps over -\\'-the -dog-'",
-      "'fox +jum+ps over the +lazy cat+'"
+      ~c"fox hops over 'the dog" = ~c"fox jumps over the lazy cat",
+      "~c\"fox -ho-ps over -'-the -dog-\"",
+      "~c\"fox +jum+ps over the +lazy cat+\""
     )
 
     refute_diff({[], :ok} = {[], [], :ok}, "{[], -:ok-}", "{[], +[]+, +:ok+}")
-    refute_diff({[], :ok} = {'foo', [], :ok}, "{'--', -:ok-}", "{'+foo+', +[]+, +:ok+}")
-    refute_diff({'foo', :ok} = {[], [], :ok}, "{'-foo-', -:ok-}", "{'++', +[]+, +:ok+}")
-    refute_diff({'foo', :ok} = {'bar', [], :ok}, "{'-foo-', -:ok-}", "{'+bar+', +[]+, +:ok+}")
+    refute_diff({[], :ok} = {~c"foo", [], :ok}, "{~c\"--\", -:ok-}", "{~c\"+foo+\", +[]+, +:ok+}")
+    refute_diff({~c"foo", :ok} = {[], [], :ok}, "{~c\"-foo-\", -:ok-}", "{~c\"++\", +[]+, +:ok+}")
+
+    refute_diff(
+      {~c"foo", :ok} = {~c"bar", [], :ok},
+      "{~c\"-foo-\", -:ok-}",
+      "{~c\"+bar+\", +[]+, +:ok+}"
+    )
   end
 
   test "refs" do
@@ -1036,8 +1229,18 @@ defmodule ExUnit.DiffTest do
 
     refute_diff(
       {ref1, ref2} == {ref2, ref1},
-      "{-#{inspect_ref1}-, -#{inspect_ref2}-}",
-      "{+#{inspect_ref2}+, +#{inspect_ref1}+}"
+      """
+      {
+        -#{inspect_ref1}-,
+        -#{inspect_ref2}-
+      }\
+      """,
+      """
+      {
+        +#{inspect_ref2}+,
+        +#{inspect_ref1}+
+      }\
+      """
     )
 
     refute_diff(
@@ -1054,7 +1257,16 @@ defmodule ExUnit.DiffTest do
 
     refute_diff(ref1 == :a, "-#{inspect_ref1}-", "+:a+")
     refute_diff({ref1, ref2} == :a, "-{#{inspect_ref1}, #{inspect_ref2}}", "+:a+")
-    refute_diff(%{ref1 => ref2} == :a, "-%{#{inspect_ref1} => #{inspect_ref2}}", "+:a+")
+
+    refute_diff(
+      %{ref1 => ref2} == :a,
+      """
+      -%{
+        #{inspect_ref1} => #{inspect_ref2}
+      }\
+      """,
+      "+:a+"
+    )
 
     refute_diff(
       %Opaque{data: ref1} == :a,
@@ -1083,6 +1295,8 @@ defmodule ExUnit.DiffTest do
     )
   end
 
+  @compile {:no_warn_undefined, String}
+
   test "functions" do
     identity = & &1
     inspect = inspect(identity)
@@ -1092,10 +1306,23 @@ defmodule ExUnit.DiffTest do
 
     refute_diff(identity == :a, "-#{inspect}-", "+:a+")
     refute_diff({identity, identity} == :a, "-{#{inspect}, #{inspect}}", "+:a+")
-
     refute_diff({identity, :a} == {:a, identity}, "{-#{inspect}-, -:a-}", "{+:a+, +#{inspect}+}")
 
-    refute_diff(%{identity => identity} == :a, "-%{#{inspect} => #{inspect}}", "+:a+")
+    refute_diff(
+      %{identity => identity} == :a,
+      """
+      -%{
+        #{inspect} => #{inspect}
+      }-\
+      """,
+      "+:a+"
+    )
+
+    refute_diff(
+      (&String.to_charlist/1) == (&String.unknown/1),
+      "-&String.to_charlist/1-",
+      "+&String.unknown/1"
+    )
 
     refute_diff(
       %Opaque{data: identity} == :a,
@@ -1104,6 +1331,24 @@ defmodule ExUnit.DiffTest do
     )
   end
 
+  defp closure(a), do: fn -> a end
+
+  test "functions with closure" do
+    closure1 = closure(1)
+    closure2 = closure(2)
+
+    fun_info = Function.info(closure1)
+    uniq = Integer.to_string(fun_info[:new_index]) <> "." <> Integer.to_string(fun_info[:uniq])
+
+    assert_diff(closure1 == closure1, [])
+
+    refute_diff(
+      closure1 == closure2,
+      "#Function<\n  #{uniq}/0 in ExUnit.DiffTest.closure/1\n  [-1-]\n>",
+      "#Function<\n  #{uniq}/0 in ExUnit.DiffTest.closure/1\n  [+2+]\n>"
+    )
+  end
+
   test "not supported" do
     refute_diff(
       <<147, 1, 2, 31>> = <<193, 1, 31>>,
