@@ -33,6 +33,8 @@ defmodule ExUnit.FormatterTest do
   defp diff_formatter(:diff_enabled?, _default), do: true
   defp diff_formatter(_key, value), do: value
 
+  defp kw_to_string(kw), do: for({k, v} <- kw, do: {k, IO.iodata_to_binary(v)})
+
   test "formats test case filters" do
     filters = [run: true, slow: false]
     assert format_filters(filters, :exclude) =~ "Excluding tags: [run: true, slow: false]"
@@ -228,6 +230,31 @@ defmodule ExUnit.FormatterTest do
            """
   end
 
+  test "formats test errors with parameters" do
+    failure = [{:error, catch_error(raise "oops"), []}]
+
+    assert format_test_failure(%{test() | parameters: %{foo: :bar}}, failure, 1, 80, &formatter/2) =~
+             """
+               1) world (Hello)
+                  Parameters: %{foo: :bar}
+                  test/ex_unit/formatter_test.exs:1
+                  ** (RuntimeError) oops
+             """
+
+    formatter = fn
+      :parameters_info, map -> Map.put(map, :more, :keys)
+      key, val -> formatter(key, val)
+    end
+
+    assert format_test_failure(%{test() | parameters: %{foo: :bar}}, failure, 1, 80, formatter) =~
+             """
+               1) world (Hello)
+                  Parameters: #{inspect(%{foo: :bar, more: :keys})}
+                  test/ex_unit/formatter_test.exs:1
+                  ** (RuntimeError) oops
+             """
+  end
+
   test "formats stacktraces" do
     stacktrace = [{Oops, :wrong, 1, [file: "formatter_test.exs", line: 1]}]
     failure = [{:error, catch_error(raise "oops"), stacktrace}]
@@ -242,7 +269,8 @@ defmodule ExUnit.FormatterTest do
   end
 
   test "formats assertions" do
-    failure = [{:error, catch_assertion(assert ExUnit.FormatterTest.falsy()), []}]
+    assertion_error = catch_assertion(assert ExUnit.FormatterTest.falsy())
+    failure = [{:error, assertion_error, []}]
 
     assert format_test_failure(test(), failure, 1, 80, &formatter/2) =~ """
              1) world (Hello)
@@ -250,10 +278,13 @@ defmodule ExUnit.FormatterTest do
                 Expected truthy, got false
                 code: assert ExUnit.FormatterTest.falsy()
            """
+
+    assert format_assertion_diff(assertion_error, 0, :infinity, &formatter/2) == []
   end
 
   test "formats assertions with patterns and values" do
-    failure = [{:error, catch_assertion(assert {1, 2, 3} > {1, 2, 3}), []}]
+    assertion_error = catch_assertion(assert {1, 2, 3} > {1, 2, 3})
+    failure = [{:error, assertion_error, []}]
 
     assert format_test_failure(test(), failure, 1, 80, &formatter/2) =~ """
              1) world (Hello)
@@ -263,7 +294,12 @@ defmodule ExUnit.FormatterTest do
                 left: {1, 2, 3}
            """
 
-    failure = [{:error, catch_assertion(assert {3, 2, 1} = {1, 2, 3}), []}]
+    assert format_assertion_diff(assertion_error, 0, :infinity, &diff_formatter/2)
+           |> kw_to_string() ==
+             [left: "{1, 2, 3}"]
+
+    assertion_error = catch_assertion(assert {3, 2, 1} = {1, 2, 3})
+    failure = [{:error, assertion_error, []}]
 
     assert format_test_failure(test(), failure, 1, 80, &formatter/2) =~ """
              1) world (Hello)
@@ -273,13 +309,18 @@ defmodule ExUnit.FormatterTest do
                 left:  {3, 2, 1}
                 right: {1, 2, 3}
            """
+
+    assert format_assertion_diff(assertion_error, 0, :infinity, &diff_formatter/2)
+           |> kw_to_string() ==
+             [{:left, "{3, 2, 1}"}, {:right, "{1, 2, 3}"}]
   end
 
   nfc_hello = String.normalize("héllo", :nfc)
   nfd_hello = String.normalize("héllo", :nfd)
 
   test "formats assertions with hints" do
-    failure = [{:error, catch_assertion(assert unquote(nfc_hello) == unquote(nfd_hello)), []}]
+    assertion_error = catch_assertion(assert unquote(nfc_hello) == unquote(nfd_hello))
+    failure = [{:error, assertion_error, []}]
 
     assert format_test_failure(test(), failure, 1, 80, &diff_formatter/2) =~ """
              1) world (Hello)
@@ -290,6 +331,15 @@ defmodule ExUnit.FormatterTest do
                 right: "#{unquote(nfd_hello)}"
                 hint:  you are comparing strings that have the same visual representation but are made of different Unicode codepoints
            """
+
+    assert format_assertion_diff(assertion_error, 0, :infinity, &diff_formatter/2)
+           |> kw_to_string() ==
+             [
+               left: inspect(unquote(nfc_hello)),
+               right: inspect(unquote(nfd_hello)),
+               hint:
+                 "you are comparing strings that have the same visual representation but are made of different Unicode codepoints"
+             ]
   end
 
   test "formats match error between pinned struct type and a non-struct" do
@@ -464,30 +514,27 @@ defmodule ExUnit.FormatterTest do
   end
 
   test "inspect failure" do
-    failure = [{:error, catch_assertion(assert :will_fail == %BadInspect{}), []}]
+    failure = [{:error, catch_assertion(assert :will_fail == struct!(BadInspect)), []}]
 
-    message = ~S'''
-      got FunctionClauseError with message:
-
-          """
-          no function clause matching in Inspect.ExUnit.FormatterTest.BadInspect.inspect/2
-          """
-
-      while inspecting:
-
-          %{__struct__: ExUnit.FormatterTest.BadInspect, key: 0}
-
-      Stacktrace:
-    '''
-
-    assert format_test_failure(test(), failure, 1, 80, &formatter/2) =~ """
+    assert format_test_failure(test(), failure, 1, 80, &formatter/2) =~ ~s'''
              1) world (Hello)
                 test/ex_unit/formatter_test.exs:1
                 Assertion with == failed
-                code:  assert :will_fail == %BadInspect{}
+                code:  assert :will_fail == struct!(BadInspect)
                 left:  :will_fail
-                right: #Inspect.Error<\n#{message}\
-           """
+                right: #Inspect.Error<
+             got FunctionClauseError with message:
+
+                 """
+                 no function clause matching in Inspect.ExUnit.FormatterTest.BadInspect.inspect/2
+                 """
+
+             while inspecting:
+
+                 #{inspect(%BadInspect{}, structs: false)}
+
+             Stacktrace:
+           '''
   end
 
   defmodule BadMessage do
