@@ -73,19 +73,32 @@ defmodule ExUnit.SupervisedTest do
 
   test "starts a supervised process with ID checks" do
     {:ok, pid} = start_supervised({MyAgent, 0})
+    assert is_pid(pid)
 
-    assert {:error, {:duplicate_child_name, ExUnit.SupervisedTest.MyAgent}} =
-             start_supervised({MyAgent, 0})
-
-    assert {:error, {{:already_started, ^pid}, _}} = start_supervised({MyAgent, 0}, id: :another)
+    assert {:error, _} = start_supervised({MyAgent, 0})
+    assert {:error, _} = start_supervised({MyAgent, 0}, id: :another)
 
     assert_raise RuntimeError, ~r"Reason: bad child specification", fn ->
       start_supervised!(%{id: 1, start: :oops})
     end
+  end
 
-    assert_raise RuntimeError, ~r"Reason: already started", fn ->
-      start_supervised!({MyAgent, 0}, id: :another)
-    end
+  test "starts a supervised process with correct :\"$callers\"" do
+    test_pid = self()
+    fun = fn -> send(test_pid, {:callers, Process.get(:"$callers")}) end
+    {:ok, _pid} = start_supervised({Task, fun})
+
+    assert_receive {:callers, callers}
+    assert List.last(callers) == test_pid
+  end
+
+  test "starts a supervised process with correct :\"$ancestors\"" do
+    test_pid = self()
+    fun = fn -> send(test_pid, {:ancestors, Process.get(:"$ancestors")}) end
+    {:ok, _pid} = start_supervised({Task, fun})
+
+    assert_receive {:ancestors, ancestors}
+    assert List.last(ancestors) == test_pid
   end
 
   test "stops a supervised process" do
@@ -136,7 +149,7 @@ defmodule ExUnit.SupervisedTest do
   end
 
   defp wait_until_registered(name) do
-    unless Process.whereis(name) do
+    if !Process.whereis(name) do
       wait_until_registered(name)
     end
   end
