@@ -25,26 +25,24 @@ defmodule IEx do
   aren't explicitly documented and where the function name is in the form
   of `__foo__`.
 
-  Autocomplete may not be available on some Windows shells. You may need
-  to pass the `--werl` option when starting IEx, such as `iex --werl`
-  (or `iex.bat --werl` if using PowerShell). `--werl` may be permanently
-  enabled by setting the `IEX_WITH_WERL` environment variable to `1`.
+  Autocomplete is available by default on Windows shells from Erlang/OTP 26.
 
   ## Encoding and coloring
 
   IEx expects inputs and outputs to be in UTF-8 encoding. This is the
   default for most Unix terminals but it may not be the case on Windows.
   If you are running on Windows and you see incorrect values printed,
-  you may need change the encoding of your current session by running
+  you may need to change the encoding of your current session by running
   `chcp 65001` before calling `iex` (or before calling `iex.bat` if using
   PowerShell).
 
   Similarly, ANSI coloring is enabled by default on most Unix terminals.
-  They are also available on Windows consoles from Windows 10, although
-  it must be explicitly enabled for the current user in the registry by
-  running the following command:
+  They are also available on Windows consoles from Windows 10 and on
+  Erlang/OTP 26 or later. For earlier Erlang/OTP versions, you can
+  explicitly enable it for the current user in the registry by running
+  the following command:
 
-      reg add HKCU\Console /v VirtualTerminalLevel /t REG_DWORD /d 1
+      $ reg add HKCU\Console /v VirtualTerminalLevel /t REG_DWORD /d 1
 
   After running the command above, you must restart your current console.
 
@@ -53,7 +51,7 @@ defmodule IEx do
   It is possible to get shell history by passing some options that enable it
   in the VM. This can be done on a per-need basis when starting IEx:
 
-      iex --erl "-kernel shell_history enabled"
+      $ iex --erl "-kernel shell_history enabled"
 
   If you would rather enable it on your system as a whole, you can use
   the `ERL_AFLAGS` environment variable and make sure that it is set
@@ -61,15 +59,15 @@ defmodule IEx do
 
   On Unix-like / Bash:
 
-      export ERL_AFLAGS="-kernel shell_history enabled"
+      $ export ERL_AFLAGS="-kernel shell_history enabled"
 
   On Windows:
 
-      set ERL_AFLAGS "-kernel shell_history enabled"
+      $ set ERL_AFLAGS "-kernel shell_history enabled"
 
   On Windows 10 / PowerShell:
 
-      $env:ERL_AFLAGS = "-kernel shell_history enabled"
+      $ $env:ERL_AFLAGS = "-kernel shell_history enabled"
 
   ## Expressions in IEx
 
@@ -113,7 +111,7 @@ defmodule IEx do
       iex(1)> [1, [2], 3]
       [1, [2], 3]
 
-  To prevent this behaviour breaking valid code where the subsequent line
+  To prevent this behavior breaking valid code where the subsequent line
   begins with a binary operator, such as `|>/2` or `++/2` , IEx automatically
   treats such lines as if they were prepended with `IEx.Helpers.v/0`, which
   returns the value of the previous expression, if available.
@@ -146,7 +144,7 @@ defmodule IEx do
         2 | |> IO.puts()
           | ^
 
-  Note however the above does not work for `+/2` and `-/2`, as they
+  Note, however, the above does not work for `+/2` and `-/2`, as they
   are ambiguous with the unary `+/1` and `-/1`:
 
       iex(1)> 1
@@ -173,7 +171,11 @@ defmodule IEx do
   ## `dbg` and breakpoints
 
   IEx integrates with `Kernel.dbg/2` and introduces a backend that
-  can pause code execution:
+  can pause code execution. To enable it, you must pass `--dbg pry`:
+
+      $ iex --dbg pry
+
+  For example, take the following function:
 
       def my_fun(arg1, arg2) do
         dbg(arg1 + arg2)
@@ -181,8 +183,8 @@ defmodule IEx do
       end
 
   When the code is executed with `iex` (most often by calling
-  `iex -S mix`), it will ask you permission to use "pry". If you
-  agree, it will start an IEx shell in the context of the function
+  `iex --dbg pry -S mix`), it will ask you permission to use "pry".
+  If you agree, it will start an IEx shell in the context of the function
   above, with access to its variables, imports, and aliases. However,
   you can only access existing values, it is not possible to access
   private functions nor change the execution itself (hence the name
@@ -194,7 +196,6 @@ defmodule IEx do
   of the steps but stay within the pried process. Type `respawn` when
   you want to leave the pried process and start a new shell.
 
-  You can disable "pry" as the `dbg/2` backend by calling `iex --no-pry`.
   Alternatively, you can start a pry session directly, without `dbg/2`
   by calling `IEx.pry/0`.
 
@@ -208,8 +209,8 @@ defmodule IEx do
   When using `dbg` or breakpoints with tests, remember to pass the
   `--trace` to `mix test` to avoid running into timeouts:
 
-      iex -S mix test --trace
-      iex -S mix test path/to/file:line --trace
+      $ iex -S mix test --trace
+      $ iex -S mix test path/to/file:line --trace
 
   ## The User switch command
 
@@ -221,7 +222,7 @@ defmodule IEx do
   alternate between them. Let's give it a try:
 
       User switch command
-       --> s 'Elixir.IEx'
+       --> s iex
        --> c
 
   The command above will start a new shell and connect to it.
@@ -237,7 +238,7 @@ defmodule IEx do
   Now, try to access the `hello` variable again:
 
       hello
-      ** (UndefinedFunctionError) undefined function hello/0
+      ** (CompileError) undefined variable "hello"
 
   The command above fails because we have switched shells.
   Since shells are isolated from each other, you can't access the
@@ -320,9 +321,10 @@ defmodule IEx do
 
   ## The .iex.exs file
 
-  When starting, IEx looks for a local `.iex.exs` file (located in the current
-  working directory), then a global one (located at `~/.iex.exs`) and loads the
-  first one it finds (if any).
+  When starting, IEx looks for a configured path, then for a local `.iex.exs` file
+  (located in the current working directory), then for a global `.iex.exs` file
+  located inside the directory pointed by the `IEX_HOME` environment variable
+  (which defaults to `~`) and loads the first one it finds (if any).
 
   The code in the chosen `.iex.exs` file is evaluated line by line in the shell's
   context, as if each line were being typed in the shell. For instance, any modules
@@ -354,8 +356,9 @@ defmodule IEx do
       iex(1)> value
       13
 
-  It is possible to load another file by supplying the `--dot-iex` option
-  to IEx. See `iex --help`.
+  It is possible to load another file by configuring the `iex` application's `dot_iex`
+  value (`config :iex, dot_iex: "PATH"` or `IEx.configure(dot_iex: "PATH")`)
+  or supplying the `--dot-iex` option to IEx. See `iex --help`.
 
   In case of remote nodes, the location of the `.iex.exs` files are taken
   relative to the user that started the application, not to the user that
@@ -388,15 +391,17 @@ defmodule IEx do
 
   The supported options are:
 
-    * `:colors`
-    * `:inspect`
-    * `:width`
-    * `:history_size`
-    * `:default_prompt`
-    * `:continuation_prompt`
-    * `:alive_prompt`
+    * `:auto_reload`
     * `:alive_continuation_prompt`
+    * `:alive_prompt`
+    * `:colors`
+    * `:continuation_prompt`
+    * `:default_prompt`
+    * `:dot_iex`
+    * `:history_size`
+    * `:inspect`
     * `:parser`
+    * `:width`
 
   They are discussed individually in the sections below.
 
@@ -502,6 +507,17 @@ defmodule IEx do
   or `{:incomplete, buffer}`.
 
   If the parser raises, the buffer is reset to an empty string.
+
+  ## `.iex`
+
+  Configure the file loaded into your IEx session when it starts.
+  See more information [in the `.iex.exs` documentation](`m:IEx#module-the-iex-exs-file`).
+
+  ## Auto reloading
+
+  When set to `true`, the `:auto_reload` option automatically purges
+  in-memory modules when they get invalidated by a concurrent compilation
+  happening in the Operating System.
   """
   @spec configure(keyword()) :: :ok
   def configure(options) do
@@ -516,24 +532,23 @@ defmodule IEx do
     IEx.Config.configuration()
   end
 
-  @doc """
-  Registers a function to be invoked after the IEx process is spawned.
-  """
-  @spec after_spawn(fun()) :: :ok
+  @doc false
+  @deprecated "The functionality is no longer supported"
   def after_spawn(fun) when is_function(fun) do
     IEx.Config.after_spawn(fun)
   end
 
-  @doc """
-  Returns registered `after_spawn` callbacks.
-  """
-  @spec after_spawn() :: [fun()]
+  @doc false
+  @deprecated "The functionality is no longer supported"
   def after_spawn do
     IEx.Config.after_spawn()
   end
 
   @doc """
   Returns `true` if IEx was started, `false` otherwise.
+
+  This means the IEx application was started, but not
+  that its CLI interface is running.
   """
   @spec started?() :: boolean()
   def started? do
@@ -577,9 +592,6 @@ defmodule IEx do
   @doc """
   Pries into the process environment.
 
-  When you start `iex`, IEx will set this function to be the
-  default `dbg/2` backend unless the `--no-pry` flag is given.
-
   This function is useful for debugging a particular chunk of code
   when executed by a particular process. The process becomes
   the evaluator of IEx commands and is temporarily changed to
@@ -595,6 +607,11 @@ defmodule IEx do
 
   See also `break!/4` for others ways to pry.
 
+  > #### `dbg/0` integration
+  >
+  > By calling `iex --dbg pry`, `iex` will set this function
+  > as the default backend for `dbg/0` calls.
+
   ## Examples
 
   Let's suppose you want to investigate what is happening
@@ -617,7 +634,7 @@ defmodule IEx do
   the shell will be reset and you gain access to all variables
   and the lexical scope from above:
 
-      pry(1)> map([a, b, c], &IO.inspect(&1))
+      iex(1)> map([a, b, c], &IO.inspect(&1))
       1
       2
       3
@@ -627,7 +644,7 @@ defmodule IEx do
   process can be freed by calling [`respawn/0`](`IEx.Helpers.respawn/0`), which starts a
   new IEx evaluation cycle, letting this one go:
 
-      pry(2)> respawn()
+      iex(2)> respawn()
       true
 
       Interactive Elixir - press Ctrl+C to exit (type h() ENTER for help)
@@ -661,8 +678,8 @@ defmodule IEx do
   the `iex` command and pass the `--trace` to `mix test` to avoid running
   into timeouts:
 
-      iex -S mix test --trace
-      iex -S mix test path/to/file:line --trace
+      $ iex -S mix test --trace
+      $ iex -S mix test path/to/file:line --trace
 
   """
   defmacro pry() do
@@ -706,7 +723,7 @@ defmodule IEx do
   defp __break__!(ast, module, fun, args, guards, stops, env) do
     module = Macro.expand(module, env)
 
-    unless is_atom(module) do
+    if not is_atom(module) do
       raise_unknown_break_ast!(ast)
     end
 
@@ -783,7 +800,7 @@ defmodule IEx do
   a breakpoint directly from your IEx shell. But you can set up a break
   from anywhere by using the fully qualified name `IEx.break!`.
 
-  The following sets up a breakpoint on `URI.parse/2`:
+  The following sets up a breakpoint on `URI.parse/1`:
 
       break! URI, :parse, 1
 
@@ -820,7 +837,7 @@ defmodule IEx do
   ## Macros
 
   While it is possible to set breakpoint in macros, remember that macros
-  are generally expanded at compilation time, and therfore they may never
+  are generally expanded at compilation time, and therefore they may never
   be invoked during runtime. Similarly, while patterns may be given to
   macros, macros receive ASTs as arguments, and not values. For example,
   if you try to break on a macro with the following pattern:
@@ -838,62 +855,23 @@ defmodule IEx do
   the `iex` command and pass the `--trace` to `mix test` to avoid running
   into timeouts:
 
-      iex -S mix test --trace
-      iex -S mix test path/to/file:line --trace
+      $ iex -S mix test --trace
+      $ iex -S mix test path/to/file:line --trace
 
   """
   @doc since: "1.5.0"
   @spec break!(module, atom, arity, non_neg_integer) :: IEx.Pry.id()
   defdelegate break!(module, function, arity, stops \\ 1), to: IEx.Pry
 
-  ## Callbacks
-
-  # This is a callback invoked by Erlang shell utilities
-  # when someone presses Ctrl+G and adds 's Elixir.IEx'.
-  @doc false
-  def start(opts \\ [], mfa \\ {IEx, :dont_display_result, []}) do
-    spawn(fn ->
-      case :init.notify_when_started(self()) do
-        :started -> :ok
-        _ -> :init.wait_until_started()
-      end
-
-      :ok = start_iex()
-      :ok = set_expand_fun()
-      :ok = run_after_spawn()
-      IEx.Server.run_from_shell(opts, mfa)
-    end)
-  end
-
   @doc false
   def dont_display_result, do: :"do not show this result in output"
 
-  ## Helpers
+  ## CLI
 
-  defp start_iex() do
-    {:ok, _} = Application.ensure_all_started(:iex)
-    :ok
-  end
-
-  defp set_expand_fun do
-    gl = Process.group_leader()
-
-    expand_fun =
-      if node(gl) != node() do
-        IEx.Autocomplete.remsh(node())
-      else
-        &IEx.Autocomplete.expand/1
-      end
-
-    # expand_fun is not supported by a shell variant
-    # on Windows, so we do two IO calls, not caring
-    # about the result of the expand_fun one.
-    _ = :io.setopts(gl, expand_fun: expand_fun)
-    :io.setopts(gl, binary: true, encoding: :unicode)
-  end
-
-  defp run_after_spawn do
-    _ = for fun <- Enum.reverse(after_spawn()), do: fun.()
-    :ok
+  # TODO: Remove me on Elixir v1.20+
+  @doc false
+  def start do
+    IO.warn("Use \"s iex\" instead")
+    :iex.start()
   end
 end
