@@ -18,15 +18,19 @@ defmodule IEx.Autocomplete do
     %{kind: :variable, name: "utf32"}
   ]
 
+  @alias_only_atoms ~w(alias import require)a
+  @alias_only_charlists ~w(alias import require)c
+
   @doc """
   Provides one helper function that is injected into connecting
   remote nodes to properly handle autocompletion.
   """
   def remsh(node) do
     fn e ->
-      case :rpc.call(node, IEx.Autocomplete, :expand, [e, IEx.Broker.shell()]) do
-        {:badrpc, _} -> {:no, '', []}
-        r -> r
+      try do
+        :erpc.call(node, IEx.Autocomplete, :expand, [e, IEx.Broker.shell()])
+      catch
+        _, _ -> {:no, ~c"", []}
       end
     end
   end
@@ -34,7 +38,7 @@ defmodule IEx.Autocomplete do
   @doc """
   The expansion logic.
 
-  Some of the expansion has to be use the current shell
+  Some of the expansion has to use the current shell
   environment, which is found via the broker.
   """
   def expand(code, shell \\ IEx.Broker.shell()) do
@@ -62,7 +66,11 @@ defmodule IEx.Autocomplete do
         expand_typespecs(expansion, shell, &get_module_types/1)
 
       {:dot, path, hint} ->
-        expand_dot(path, List.to_string(hint), false, shell)
+        if alias = alias_only(path, hint, code, shell) do
+          expand_aliases(List.to_string(alias), shell)
+        else
+          expand_dot(path, List.to_string(hint), false, shell)
+        end
 
       {:dot_arity, path, hint} ->
         expand_dot(path, List.to_string(hint), true, shell)
@@ -80,6 +88,9 @@ defmodule IEx.Autocomplete do
       {:local_arity, local} ->
         expand_local(List.to_string(local), true, shell)
 
+      {:local_call, local} when local in @alias_only_charlists ->
+        expand_aliases("", shell)
+
       {:local_call, local} ->
         expand_local_call(List.to_atom(local), shell)
 
@@ -93,6 +104,9 @@ defmodule IEx.Autocomplete do
       {:operator_arity, operator} ->
         expand_local(List.to_string(operator), true, shell)
 
+      {:operator_call, operator} when operator in ~w(|)c ->
+        expand_container_context(code, :expr, "", shell) || expand_local_or_var("", shell)
+
       {:operator_call, _operator} ->
         expand_local_or_var("", shell)
 
@@ -105,8 +119,8 @@ defmodule IEx.Autocomplete do
       {:struct, struct} when is_list(struct) ->
         expand_structs(List.to_string(struct), shell)
 
-      {:struct, {:dot, {:alias, struct}, ''}} when is_list(struct) ->
-        expand_structs(List.to_string(struct ++ '.'), shell)
+      {:struct, {:dot, {:alias, struct}, ~c""}} when is_list(struct) ->
+        expand_structs(List.to_string(struct ++ ~c"."), shell)
 
       # {:module_attribute, charlist}
       # :none
@@ -116,23 +130,23 @@ defmodule IEx.Autocomplete do
   end
 
   defp get_helper(expr) do
-    with [helper | rest] when helper in 'bt' <- expr,
+    with [helper | rest] when helper in ~c"bt" <- expr,
          [space_or_paren, char | _] <- squeeze_spaces(rest),
          true <-
-           space_or_paren in ' (' and
-             (char in ?A..?Z or char in ?a..?z or char in ?0..?9 or char in '_:') do
+           space_or_paren in ~c" (" and
+             (char in ?A..?Z or char in ?a..?z or char in ?0..?9 or char in ~c"_:") do
       helper
     else
       _ -> nil
     end
   end
 
-  defp squeeze_spaces('  ' ++ rest), do: squeeze_spaces([?\s | rest])
+  defp squeeze_spaces(~c"  " ++ rest), do: squeeze_spaces([?\s | rest])
   defp squeeze_spaces(rest), do: rest
 
   @doc false
   def exports(mod) do
-    if Code.ensure_loaded?(mod) and function_exported?(mod, :__info__, 1) do
+    if ensure_loaded?(mod) and function_exported?(mod, :__info__, 1) do
       mod.__info__(:macros) ++ (mod.__info__(:functions) -- [__info__: 1])
     else
       mod.module_info(:exports) -- [module_info: 0, module_info: 1]
@@ -183,9 +197,7 @@ defmodule IEx.Autocomplete do
   end
 
   defp expand_signatures([_ | _] = signatures, _shell) do
-    [head | tail] = Enum.sort(signatures, &(String.length(&1) <= String.length(&2)))
-    if tail != [], do: IO.write("\n" <> (tail |> Enum.reverse() |> Enum.join("\n")))
-    yes("", [head])
+    yes("", Enum.sort_by(signatures, &String.length/1))
   end
 
   defp expand_signatures([], shell), do: expand_local_or_var("", shell)
@@ -299,41 +311,33 @@ defmodule IEx.Autocomplete do
       for {alias, mod} <- aliases_from_env(shell),
           [name] = Module.split(alias),
           String.starts_with?(name, hint),
-          struct?(mod) and not function_exported?(mod, :exception, 1),
-          do: %{kind: :struct, name: name}
+          do: {mod, name}
 
     modules =
       for "Elixir." <> name = full_name <- match_modules("Elixir." <> hint, true),
           String.starts_with?(name, hint),
           mod = String.to_atom(full_name),
-          struct?(mod) and not function_exported?(mod, :exception, 1),
+          do: {mod, name}
+
+    all = aliases ++ modules
+    Code.ensure_all_loaded(Enum.map(all, &elem(&1, 0)))
+
+    refs =
+      for {mod, name} <- all,
+          function_exported?(mod, :__struct__, 1) and not function_exported?(mod, :exception, 1),
           do: %{kind: :struct, name: name}
 
-    format_expansion(aliases ++ modules, hint)
-  end
-
-  defp struct?(mod) do
-    Code.ensure_loaded?(mod) and function_exported?(mod, :__struct__, 1)
+    format_expansion(refs, hint)
   end
 
   defp expand_container_context(code, context, hint, shell) do
     case container_context(code, shell) do
+      {:map, map, pairs} when context == :expr ->
+        container_context_map_fields(pairs, map, hint)
+
       {:struct, alias, pairs} when context == :expr ->
-        pairs =
-          Enum.reduce(pairs, Map.from_struct(alias.__struct__), fn {key, _}, map ->
-            Map.delete(map, key)
-          end)
-
-        entries =
-          for {key, _value} <- pairs,
-              name = Atom.to_string(key),
-              if(hint == "",
-                do: not String.starts_with?(name, "_"),
-                else: String.starts_with?(name, hint)
-              ),
-              do: %{kind: :keyword, name: name}
-
-        format_expansion(entries, hint)
+        map = Map.from_struct(alias.__struct__())
+        container_context_map_fields(pairs, map, hint)
 
       :bitstring_modifier ->
         existing =
@@ -352,22 +356,52 @@ defmodule IEx.Autocomplete do
     end
   end
 
+  defp container_context_map_fields(pairs, map, hint) do
+    pairs =
+      Enum.reduce(pairs, map, fn {key, _}, map ->
+        Map.delete(map, key)
+      end)
+
+    entries =
+      for {key, _value} <- pairs,
+          name = Atom.to_string(key),
+          if(hint == "",
+            do: not String.starts_with?(name, "_"),
+            else: String.starts_with?(name, hint)
+          ),
+          do: %{kind: :keyword, name: name}
+
+    format_expansion(entries, hint)
+  end
+
   defp container_context(code, shell) do
     case Code.Fragment.container_cursor_to_quoted(code) do
       {:ok, quoted} ->
         case Macro.path(quoted, &match?({:__cursor__, _, []}, &1)) do
           [cursor, {:%{}, _, pairs}, {:%, _, [{:__aliases__, _, aliases}, _map]} | _] ->
-            with {pairs, [^cursor]} <- Enum.split(pairs, -1),
-                 alias = value_from_alias(aliases, shell),
-                 true <- Keyword.keyword?(pairs) and struct?(alias) do
-              {:struct, alias, pairs}
-            else
+            container_context_struct(cursor, pairs, aliases, shell)
+
+          [
+            cursor,
+            pairs,
+            {:|, _, _},
+            {:%{}, _, _},
+            {:%, _, [{:__aliases__, _, aliases}, _map]} | _
+          ] ->
+            container_context_struct(cursor, pairs, aliases, shell)
+
+          [cursor, pairs, {:|, _, [{variable, _, nil} | _]}, {:%{}, _, _} | _] ->
+            container_context_map(cursor, pairs, variable, shell)
+
+          [cursor, {special_form, _, [cursor]} | _] when special_form in @alias_only_atoms ->
+            :alias_only
+
+          [cursor | tail] ->
+            case remove_operators(tail, cursor) do
+              [{:"::", _, [_, _]}, {:<<>>, _, [_ | _]} | _] -> :bitstring_modifier
               _ -> nil
             end
 
-          [cursor, {:"::", _, [_, cursor]}, {:<<>>, _, [_ | _]} | _] ->
-            :bitstring_modifier
-
           _ ->
             nil
         end
@@ -377,8 +411,46 @@ defmodule IEx.Autocomplete do
     end
   end
 
+  defp remove_operators([{op, _, [_, previous]} = head | tail], previous) when op in [:-],
+    do: remove_operators(tail, head)
+
+  defp remove_operators(tail, _previous),
+    do: tail
+
+  defp container_context_struct(cursor, pairs, aliases, shell) do
+    with {pairs, [^cursor]} <- Enum.split(pairs, -1),
+         alias = value_from_alias(aliases, shell),
+         true <-
+           Keyword.keyword?(pairs) and ensure_loaded?(alias) and
+             function_exported?(alias, :__struct__, 1) do
+      {:struct, alias, pairs}
+    else
+      _ -> nil
+    end
+  end
+
+  defp container_context_map(cursor, pairs, variable, shell) do
+    with {pairs, [^cursor]} <- Enum.split(pairs, -1),
+         {:ok, map} when is_map(map) <- value_from_binding([variable], shell),
+         true <- Keyword.keyword?(pairs) do
+      {:map, map, pairs}
+    else
+      _ -> nil
+    end
+  end
+
   ## Aliases and modules
 
+  defp alias_only(path, hint, code, shell) do
+    with {:alias, alias} <- path,
+         [] <- hint,
+         :alias_only <- container_context(code, shell) do
+      alias ++ [?.]
+    else
+      _ -> nil
+    end
+  end
+
   defp expand_aliases(all, shell) do
     case String.split(all, ".") do
       [hint] ->
@@ -470,7 +542,7 @@ defmodule IEx.Autocomplete do
   end
 
   defp no do
-    {:no, '', []}
+    {:no, ~c"", []}
   end
 
   ## Helpers
@@ -613,7 +685,7 @@ defmodule IEx.Autocomplete do
 
   defp hidden_fun?({name, arity}, docs) do
     case Enum.find(docs, &match?({{_, ^name, ^arity}, _, _, _, _}, &1)) do
-      nil -> hd(Atom.to_charlist(name)) == ?_
+      nil -> match?([?_ | _], Atom.to_charlist(name))
       {_, _, _, :hidden, _} -> true
       {_, _, _, _, _} -> false
     end
