@@ -1,41 +1,44 @@
 defmodule IEx.Broker do
   @moduledoc false
-  @name IEx.Broker
+  @name __MODULE__
 
   @type take_ref :: {takeover_ref :: reference(), server_ref :: reference()}
+  @type shell :: pid | nil
 
   use GenServer
 
   ## Shell API
 
   @doc """
-  Finds the IEx server running inside `:user_drv`, on this node exclusively.
+  Finds the IEx server.
   """
-  @spec shell :: pid | nil
+  @spec shell :: shell()
+  # TODO: Use shell:whereis() from Erlang/OTP 26+.
   def shell() do
-    # Locate top group leader when using the "new shell".
     if user = Process.whereis(:user) do
-      case :group.interfaces(user) do
-        # Old or no shell
-        [] ->
-          nil
-
-        # Get current group from user_drv
-        [user_drv: user_drv] ->
-          case :user_drv.interfaces(user_drv) do
-            [] -> nil
-            [current_group: group] -> :group.interfaces(group)[:shell]
-          end
+      if user_drv = get_from_dict(user, :user_drv) do
+        if group = get_from_dict(user_drv, :current_group) do
+          get_from_dict(group, :shell)
+        end
       end
     end
   end
 
+  defp get_from_dict(pid, key) do
+    with {:dictionary, dictionary} <- Process.info(pid, :dictionary),
+         {^key, value} <- List.keyfind(dictionary, key, 0) do
+      value
+    else
+      _ -> nil
+    end
+  end
+
   @doc """
   Finds the evaluator and server running inside `:user_drv`, on this node exclusively.
   """
-  @spec evaluator :: {evaluator :: pid, server :: pid} | nil
-  def evaluator() do
-    if pid = shell() do
+  @spec evaluator(shell()) :: {evaluator :: pid, server :: pid} | nil
+  def evaluator(pid \\ shell()) do
+    if pid do
       {:dictionary, dictionary} = Process.info(pid, :dictionary)
       {dictionary[:evaluator], pid}
     end
@@ -44,7 +47,7 @@ defmodule IEx.Broker do
   ## Broker API
 
   def start_link(_) do
-    GenServer.start_link(__MODULE__, :ok, name: @name)
+    GenServer.start_link(@name, :ok, name: @name)
   end
 
   @doc """
@@ -62,27 +65,26 @@ defmodule IEx.Broker do
 
   The broker's PID is needed to support remote shells.
   """
-  @spec respond(pid, take_ref, boolean()) :: :ok | {:error, :refused | :already_accepted}
-  def respond(broker_pid, take_ref, true) do
-    GenServer.call(broker_pid, {:accept, take_ref, Process.group_leader()})
+  @spec respond(pid, take_ref, integer(), boolean()) ::
+          :ok | {:error, :refused | :already_accepted}
+  def respond(broker_pid, take_ref, counter, true) do
+    GenServer.call(broker_pid, {:accept, take_ref, Process.group_leader(), counter})
   end
 
-  def respond(broker_pid, take_ref, false) do
+  def respond(broker_pid, take_ref, _counter, false) do
     GenServer.call(broker_pid, {:refuse, take_ref})
   end
 
   @doc """
   Client requests a takeover.
   """
-  @spec take_over(binary, keyword) ::
-          {:ok, server :: pid, group_leader :: pid} | {:error, :no_iex | :refused}
-  def take_over(identifier, opts) do
+  @spec take_over(binary, iodata, keyword) ::
+          {:ok, server :: pid, group_leader :: pid, counter :: integer}
+          | {:error, :no_iex | :refused | atom()}
+  def take_over(location, whereami, opts) do
     case GenServer.whereis(@name) do
-      nil ->
-        {:error, :no_iex}
-
-      _pid ->
-        GenServer.call(@name, {:take_over, identifier, opts}, :infinity)
+      nil -> {:error, :no_iex}
+      _pid -> GenServer.call(@name, {:take_over, location, whereami, opts}, :infinity)
     end
   end
 
@@ -99,7 +101,7 @@ defmodule IEx.Broker do
   end
 
   @impl true
-  def handle_call({:take_over, identifier, opts}, {_, ref} = from, state) do
+  def handle_call({:take_over, location, whereami, opts}, {_, ref} = from, state) do
     case servers(state) do
       [] ->
         {:reply, {:error, :no_iex}, state}
@@ -107,7 +109,7 @@ defmodule IEx.Broker do
       servers ->
         server_refs =
           for {server_ref, server_pid} <- servers do
-            send(server_pid, {:take_over, self(), {ref, server_ref}, identifier, opts})
+            send(server_pid, {:take_over, self(), {ref, server_ref}, location, whereami, opts})
             server_ref
           end
 
@@ -122,13 +124,13 @@ defmodule IEx.Broker do
     {:reply, :ok, state}
   end
 
-  def handle_call({:accept, {ref, _server_ref}, group_leader}, {server, _}, state) do
+  def handle_call({:accept, {ref, _server_ref}, group_leader, counter}, {server, _}, state) do
     case pop_in(state.takeovers[ref]) do
       {nil, state} ->
         {:reply, {:error, :already_accepted}, state}
 
       {{from, _}, state} ->
-        GenServer.reply(from, {:ok, server, group_leader})
+        GenServer.reply(from, {:ok, server, group_leader, counter})
         {:reply, :ok, state}
     end
   end
@@ -175,8 +177,11 @@ defmodule IEx.Broker do
 
   defp local_or_remote_shell() do
     Enum.find_value([node() | Node.list()], fn node ->
-      server = :rpc.call(node, IEx.Broker, :shell, [])
-      if is_pid(server), do: server
+      try do
+        :erpc.call(node, IEx.Broker, :shell, [])
+      catch
+        _, _ -> nil
+      end
     end)
   end
 end
