@@ -97,7 +97,7 @@ defmodule IEx.Introspection do
     case open_mfa(module, :__info__, 1) do
       {source, nil, _} -> open(source)
       {_, tuple, _} -> open(tuple)
-      :error -> puts_error("Could not open: #{inspect(module)}. Module is not available.")
+      {:error, reason} -> puts_error("Could not open #{inspect(module)}, #{reason}")
     end
 
     dont_display_result()
@@ -107,14 +107,14 @@ defmodule IEx.Introspection do
     case open_mfa(module, function, :*) do
       {_, _, nil} ->
         puts_error(
-          "Could not open: #{inspect(module)}.#{function}. Function/macro is not available."
+          "Could not open #{inspect(module)}.#{function}, function/macro is not available"
         )
 
       {_, _, tuple} ->
         open(tuple)
 
-      :error ->
-        puts_error("Could not open: #{inspect(module)}.#{function}. Module is not available.")
+      {:error, reason} ->
+        puts_error("Could not open #{inspect(module)}.#{function}, #{reason}")
     end
 
     dont_display_result()
@@ -125,16 +125,14 @@ defmodule IEx.Introspection do
     case open_mfa(module, function, arity) do
       {_, _, nil} ->
         puts_error(
-          "Could not open: #{inspect(module)}.#{function}/#{arity}. Function/macro is not available."
+          "Could not open #{inspect(module)}.#{function}/#{arity}, function/macro is not available"
         )
 
       {_, _, tuple} ->
         open(tuple)
 
-      :error ->
-        puts_error(
-          "Could not open: #{inspect(module)}.#{function}/#{arity}. Module is not available."
-        )
+      {:error, reason} ->
+        puts_error("Could not open #{inspect(module)}.#{function}/#{arity}, #{reason}")
     end
 
     dont_display_result()
@@ -143,7 +141,7 @@ defmodule IEx.Introspection do
   def open({file, line}) when is_binary(file) and is_integer(line) do
     cond do
       not File.regular?(file) ->
-        puts_error("Could not open: #{inspect(file)}. File is not available.")
+        puts_error("Could not open #{inspect(file)}, file is not available.")
 
       editor = System.get_env("ELIXIR_EDITOR") || System.get_env("EDITOR") ->
         command =
@@ -174,12 +172,19 @@ defmodule IEx.Introspection do
   end
 
   defp open_mfa(module, fun, arity) do
-    with {:module, _} <- Code.ensure_loaded(module),
-         source when is_list(source) <- module.module_info(:compile)[:source] do
-      source = rewrite_source(module, source)
-      open_abstract_code(module, fun, arity, source)
-    else
-      _ -> :error
+    case Code.ensure_loaded(module) do
+      {:module, _} ->
+        case module.module_info(:compile)[:source] do
+          [_ | _] = source ->
+            source = rewrite_source(module, source)
+            open_abstract_code(module, fun, arity, source)
+
+          _ ->
+            {:error, "source code is not available"}
+        end
+
+      _ ->
+        {:error, "module is not available"}
     end
   end
 
@@ -259,7 +264,11 @@ defmodule IEx.Introspection do
     case Code.ensure_loaded(module) do
       {:module, _} ->
         case Code.fetch_docs(module) do
-          {:docs_v1, _, _, format, %{} = doc, metadata, _} ->
+          # TODO remove once we require Erlang/OTP 27+
+          {:docs_v1, _, :erlang, "application/erlang+html", _, _, _} = erl_docs ->
+            :shell_docs.render(module, erl_docs) |> IO.puts()
+
+          {:docs_v1, _, _, format, doc, metadata, _} when is_map(doc) or doc == :none ->
             print_doc([inspect(module)], [], format, doc, metadata)
 
           {:docs_v1, _, _, _, _, _, _} ->
@@ -279,7 +288,7 @@ defmodule IEx.Introspection do
   def h({module, function}) when is_atom(module) and is_atom(function) do
     case Code.ensure_loaded(module) do
       {:module, _} ->
-        {_language, _format, docs} = get_docs(module, [:function, :macro])
+        {_language, _format, docs, _} = get_docs(module, [:function, :macro])
 
         exports =
           cond do
@@ -293,6 +302,7 @@ defmodule IEx.Introspection do
               module.module_info(:exports)
           end
           |> Enum.sort()
+          |> Enum.dedup()
 
         result =
           for {^function, arity} <- exports,
@@ -370,10 +380,15 @@ defmodule IEx.Introspection do
   end
 
   defp h_mod_fun_arity(mod, fun, arity) when is_atom(mod) do
-    {language, format, docs} = get_docs(mod, [:function, :macro])
+    {language, format, docs, docs_v1} = get_docs(mod, [:function, :macro])
     spec = get_spec(mod, fun, arity)
 
     cond do
+      # TODO remove once we require Erlang/OTP 27+
+      language == :erlang and format == "application/erlang+html" ->
+        print_erlang_doc(mod, fun, arity, docs_v1)
+        :ok
+
       doc_tuple = find_doc_with_content(docs, fun, arity) ->
         print_fun(mod, language, format, doc_tuple, spec)
         :ok
@@ -407,23 +422,23 @@ defmodule IEx.Introspection do
   end
 
   defp has_type?(mod, fun) do
-    {_, _, docs} = get_docs(mod, [:type])
+    {_, _, docs, _} = get_docs(mod, [:type])
     Enum.any?(docs, &match?({_, ^fun, _}, elem(&1, 0)))
   end
 
   defp has_type?(mod, fun, arity) do
-    {_, _, docs} = get_docs(mod, [:type])
+    {_, _, docs, _} = get_docs(mod, [:type])
     Enum.any?(docs, &match?({_, ^fun, ^arity}, elem(&1, 0)))
   end
 
   defp get_docs(mod, kinds) do
     case Code.fetch_docs(mod) do
-      {:docs_v1, _, language, format, _, _, docs} ->
+      {:docs_v1, _, language, format, _, _, docs} = docs_v1 ->
         docs = for {{kind, _, _}, _, _, _, _} = doc <- docs, kind in kinds, do: doc
-        {language, format, docs}
+        {language, format, docs, docs_v1}
 
       {:error, _} ->
-        {nil, nil, nil}
+        {nil, nil, nil, nil}
     end
   end
 
@@ -563,7 +578,7 @@ defmodule IEx.Introspection do
   end
 
   defp get_callback_docs(mod, filter) do
-    {_, format, docs} = get_docs(mod, [:callback, :macrocallback])
+    {_, format, docs, _} = get_docs(mod, [:callback, :macrocallback])
 
     case Typespec.fetch_callbacks(mod) do
       :error ->
@@ -620,7 +635,6 @@ defmodule IEx.Introspection do
   defp format_callback({{kind, name, _arity}, specs}) do
     Enum.map(specs, fn spec ->
       Typespec.spec_to_quoted(name, spec)
-      |> Macro.prewalk(&drop_macro_env/1)
       |> format_typespec(kind, 0)
     end)
   end
@@ -649,50 +663,62 @@ defmodule IEx.Introspection do
     format_typespec(callbacks, :optional_callbacks, 0)
   end
 
-  defp drop_macro_env({name, meta, [{:"::", _, [_, {{:., _, [Macro.Env, :t]}, _, _}]} | args]}),
-    do: {name, meta, args}
-
-  defp drop_macro_env(other), do: other
-
   @doc """
   Prints the types for the given module and type documentation.
   """
   def t(module) when is_atom(module) do
-    case Typespec.fetch_types(module) do
-      :error ->
-        no_beam(module)
+    case :code.get_doc(module) do
+      # TODO remove once we require Erlang/OTP 27+
+      {:ok, {:docs_v1, _, :erlang, "application/erlang+html", _, _, _} = erl_docs} ->
+        :shell_docs.render_type(module, erl_docs) |> IO.puts()
 
-      {:ok, []} ->
-        types_not_found(inspect(module))
+      _ ->
+        case Typespec.fetch_types(module) do
+          :error ->
+            no_beam(module)
 
-      {:ok, types} ->
-        types
-        |> Enum.sort_by(fn {_, {name, _, args}} -> {name, length(args)} end)
-        |> Enum.each(&(&1 |> format_type() |> IO.puts()))
+          {:ok, []} ->
+            types_not_found(inspect(module))
+
+          {:ok, types} ->
+            types
+            |> Enum.sort_by(fn {_, {name, _, args}} -> {name, length(args)} end)
+            |> Enum.each(&(&1 |> format_type() |> IO.puts()))
+        end
     end
 
     dont_display_result()
   end
 
   def t({module, type}) when is_atom(module) and is_atom(type) do
-    case Typespec.fetch_types(module) do
-      :error ->
-        no_beam(module)
+    case get_docs(module, [:type]) do
+      # TODO remove once we require Erlang/OTP 27+
+      {:erlang, "application/erlang+html", _, erl_docs} ->
+        case :shell_docs.render_type(module, type, erl_docs) do
+          {:error, :type_missing} -> types_not_found_or_private("#{inspect(module)}.#{type}")
+          iodata -> IO.puts(iodata)
+        end
 
-      {:ok, types} ->
-        types
-        |> Enum.filter(&match?({kind, {^type, _, _}} when kind in [:type, :opaque], &1))
-        |> Enum.sort_by(fn {_, {name, _, args}} -> {name, length(args)} end)
-        |> case do
-          [] ->
-            types_not_found_or_private("#{inspect(module)}.#{type}")
+      {_, format, docs, _} ->
+        case Typespec.fetch_types(module) do
+          :error ->
+            no_beam(module)
 
-          types ->
-            Enum.map(types, fn {_, {_, _, args}} = typespec ->
-              module
-              |> type_doc(type, length(args), typespec)
-              |> print_typespec()
-            end)
+          {:ok, types} ->
+            types
+            |> Enum.filter(&match?({kind, {^type, _, _}} when kind in [:type, :opaque], &1))
+            |> Enum.sort_by(fn {_, {name, _, args}} -> {name, length(args)} end)
+            |> case do
+              [] ->
+                types_not_found_or_private("#{inspect(module)}.#{type}")
+
+              types ->
+                Enum.map(types, fn {_, {_, _, args}} = typespec ->
+                  type
+                  |> type_doc(length(args), typespec, format, docs)
+                  |> print_typespec()
+                end)
+            end
         end
     end
 
@@ -700,26 +726,36 @@ defmodule IEx.Introspection do
   end
 
   def t({module, type, arity}) when is_atom(module) and is_atom(type) and is_integer(arity) do
-    case Typespec.fetch_types(module) do
-      :error ->
-        no_beam(module)
+    case get_docs(module, [:type]) do
+      # TODO remove once we require Erlang/OTP 27+
+      {:erlang, "application/erlang+html", _, erl_docs} ->
+        case :shell_docs.render_type(module, type, arity, erl_docs) do
+          {:error, :type_missing} -> types_not_found_or_private("#{inspect(module)}.#{type}")
+          chardata -> IO.puts(chardata)
+        end
 
-      {:ok, types} ->
-        types
-        |> Enum.find(
-          &match?(
-            {kind, {^type, _, args}} when kind in [:type, :opaque] and length(args) == arity,
-            &1
-          )
-        )
-        |> case do
-          nil ->
-            types_not_found_or_private("#{inspect(module)}.#{type}")
+      {_, format, docs, _} ->
+        case Typespec.fetch_types(module) do
+          :error ->
+            no_beam(module)
 
-          typespec ->
-            module
-            |> type_doc(type, arity, typespec)
-            |> print_typespec()
+          {:ok, types} ->
+            types
+            |> Enum.find(
+              &match?(
+                {kind, {^type, _, args}} when kind in [:type, :opaque] and length(args) == arity,
+                &1
+              )
+            )
+            |> case do
+              nil ->
+                types_not_found_or_private("#{inspect(module)}.#{type}")
+
+              typespec ->
+                type
+                |> type_doc(arity, typespec, format, docs)
+                |> print_typespec()
+            end
         end
     end
 
@@ -731,9 +767,7 @@ defmodule IEx.Introspection do
     dont_display_result()
   end
 
-  defp type_doc(module, type, arity, typespec) do
-    {_, format, docs} = get_docs(module, [:type])
-
+  defp type_doc(type, arity, typespec, format, docs) do
     if docs do
       {_, _, _, content, metadata} = Enum.find(docs, &match?({:type, ^type, ^arity}, elem(&1, 0)))
       {format, format_type(typespec), content, metadata}
@@ -797,6 +831,20 @@ defmodule IEx.Introspection do
   defp translate_doc(%{"en" => doc}), do: doc
   defp translate_doc(_), do: nil
 
+  defp print_erlang_doc(mod, fun, arity, docs) do
+    heading = Exception.format_mfa(mod, fun, arity)
+    opts = IEx.Config.ansi_docs()
+
+    case :shell_docs.render(mod, fun, arity, docs) do
+      {:error, :function_missing} ->
+        docs_not_found("#{inspect(mod)}.#{fun}")
+
+      chardata ->
+        IO.ANSI.Docs.print_headings([heading], opts)
+        IO.puts(chardata)
+    end
+  end
+
   defp no_beam(module) do
     case Code.ensure_loaded(module) do
       {:module, _} ->
