@@ -79,9 +79,11 @@ defmodule IEx.HelpersTest do
     end
 
     test "errors when setting up a breakpoint with invalid guard" do
-      assert_raise CompileError, ~r"cannot find or invoke local is_whatever/1", fn ->
-        break!(URI.decode_query(_, map) when is_whatever(map))
-      end
+      assert capture_io(:stderr, fn ->
+               assert_raise CompileError, fn ->
+                 break!(URI.decode_query(_, map) when is_whatever(map))
+               end
+             end) =~ "cannot find or invoke local is_whatever/1"
     end
 
     test "errors when setting up a break with no beam" do
@@ -140,9 +142,10 @@ defmodule IEx.HelpersTest do
   end
 
   describe "open" do
+    @describetag :requires_source
     @iex_helpers "iex/lib/iex/helpers.ex"
     @elixir_erl "elixir/src/elixir.erl"
-    @lists_erl "#{:code.lib_dir(:stdlib, :src)}/lists.erl"
+    @lists_erl Application.app_dir(:stdlib, "src/lists.erl")
     @httpc_erl "src/http_client/httpc.erl"
     @editor System.get_env("ELIXIR_EDITOR")
 
@@ -233,35 +236,35 @@ defmodule IEx.HelpersTest do
     end
 
     test "errors if module is not available" do
-      assert capture_iex("open(:unknown)") == "Could not open: :unknown. Module is not available."
+      assert capture_iex("open(:unknown)") == "Could not open :unknown, module is not available"
     end
 
     test "errors if module.function is not available" do
       assert capture_iex("open(:unknown.unknown)") ==
-               "Could not open: :unknown.unknown. Module is not available."
+               "Could not open :unknown.unknown, module is not available"
 
       assert capture_iex("open(:elixir.unknown)") ==
-               "Could not open: :elixir.unknown. Function/macro is not available."
+               "Could not open :elixir.unknown, function/macro is not available"
 
       assert capture_iex("open(:lists.unknown)") ==
-               "Could not open: :lists.unknown. Function/macro is not available."
+               "Could not open :lists.unknown, function/macro is not available"
 
       assert capture_iex("open(:httpc.unknown)") ==
-               "Could not open: :httpc.unknown. Function/macro is not available."
+               "Could not open :httpc.unknown, function/macro is not available"
     end
 
     test "errors if module.function/arity is not available" do
       assert capture_iex("open(:unknown.start/10)") ==
-               "Could not open: :unknown.start/10. Module is not available."
+               "Could not open :unknown.start/10, module is not available"
 
       assert capture_iex("open(:elixir.start/10)") ==
-               "Could not open: :elixir.start/10. Function/macro is not available."
+               "Could not open :elixir.start/10, function/macro is not available"
 
       assert capture_iex("open(:lists.reverse/10)") ==
-               "Could not open: :lists.reverse/10. Function/macro is not available."
+               "Could not open :lists.reverse/10, function/macro is not available"
 
       assert capture_iex("open(:httpc.request/10)") ==
-               "Could not open: :httpc.request/10. Function/macro is not available."
+               "Could not open :httpc.request/10, function/macro is not available"
     end
 
     test "errors if module is in-memory" do
@@ -286,8 +289,8 @@ defmodule IEx.HelpersTest do
     end
 
     test "errors when given {file, line} is not available" do
-      assert capture_iex("open({~s[foo], 3})") ==
-               "Could not open: \"foo\". File is not available."
+      assert capture_iex("open({~s[foo], 3})") =~
+               "Could not open \"foo\", file is not available"
     end
 
     defp maybe_trim_quotes(string) do
@@ -315,6 +318,11 @@ defmodule IEx.HelpersTest do
   describe "runtime_info" do
     test "shows VM information" do
       assert "\n## System and architecture" <> _ = capture_io(fn -> runtime_info() end)
+
+      assert "\n## Loaded OTP applications" <> _ =
+               capture_io(fn -> runtime_info([:applications]) end)
+
+      assert "\n## Memory allocators" <> _ = capture_io(fn -> runtime_info([:allocators]) end)
     end
   end
 
@@ -325,9 +333,81 @@ defmodule IEx.HelpersTest do
       assert help =~ "Welcome to Interactive Elixir"
     end
 
-    test "prints non-Elixir module specs" do
-      assert capture_io(fn -> h(:timer.sleep() / 1) end) =~
-               "@spec sleep(time) :: :ok when time: timeout()"
+    @tag :erlang_doc
+    test "prints Erlang module documentation" do
+      captured = capture_io(fn -> h(:timer) end)
+      assert captured =~ "This module provides useful functions related to time."
+    end
+
+    @tag :erlang_doc
+    test "prints Erlang module function specs" do
+      captured = capture_io(fn -> h(:timer.sleep() / 1) end)
+
+      # TODO remove once we require Erlang/OTP 27+
+      if System.otp_release() < "27" do
+        assert captured =~ ":timer.sleep/1"
+        assert captured =~ "-spec sleep(Time) -> ok when Time :: timeout()."
+      else
+        assert captured =~ "sleep(Time)"
+        assert captured =~ "@spec sleep(time) :: :ok when time: timeout()"
+      end
+    end
+
+    @tag :erlang_doc
+    test "handles non-existing Erlang module function" do
+      captured = capture_io(fn -> h(:timer.baz() / 1) end)
+
+      # TODO remove once we require Erlang/OTP 27+
+      if System.otp_release() < "27" do
+        assert captured =~ "No documentation for :timer.baz was found"
+      else
+        assert captured =~ "No documentation for :timer.baz/1 was found"
+      end
+    end
+
+    @tag :erlang_doc
+    # TODO remove once we require Erlang/OTP 27+
+    @tag skip: System.otp_release() < "27"
+    test "erlang -moduledoc" do
+      filename = "sample.erl"
+
+      code = ~s'''
+      -module(sample).
+      -moduledoc "Module documentation.".
+      '''
+
+      with_file(filename, code, fn ->
+        assert c(filename, ".") == [:sample]
+
+        help = capture_iex("h(:sample)")
+        assert help =~ "Module documentation."
+      end)
+    after
+      cleanup_modules([:sample])
+    end
+
+    @tag :erlang_doc
+    # TODO remove once we require Erlang/OTP 27+
+    @tag skip: System.otp_release() < "27"
+    test "erlang -doc" do
+      filename = "sample.erl"
+
+      code = ~s'''
+      -module(sample).
+      -export([hello/0]).
+
+      -doc "Function documentation.".
+      hello() -> world.
+      '''
+
+      with_file(filename, code, fn ->
+        assert c(filename, ".") == [:sample]
+
+        help = capture_iex("h(:sample.hello)")
+        assert help =~ "Function documentation."
+      end)
+    after
+      cleanup_modules([:sample])
     end
 
     test "prints module documentation" do
@@ -784,7 +864,7 @@ defmodule IEx.HelpersTest do
 
       assert capture_io(fn -> b(NoMix.run()) end) == "Could not load module NoMix, got: nofile\n"
 
-      assert capture_io(fn -> b(Exception.message() / 1) end) ==
+      assert capture_io(fn -> b(Exception.message() / 1) end) =~
                "@callback message(t()) :: String.t()\n\n"
 
       assert capture_io(fn -> b(:gen_server.handle_cast() / 2) end) =~
@@ -912,7 +992,6 @@ defmodule IEx.HelpersTest do
     test "prints type information" do
       assert "@type t() ::" <> _ = capture_io(fn -> t(Enum.t()) end)
       assert capture_io(fn -> t(Enum.t()) end) == capture_io(fn -> t(Enum.t() / 0) end)
-      assert "@type child_spec() ::" <> _ = capture_io(fn -> t(:supervisor.child_spec()) end)
       assert capture_io(fn -> t(URI.t()) end) == capture_io(fn -> t(URI.t() / 0) end)
     end
 
@@ -990,6 +1069,54 @@ defmodule IEx.HelpersTest do
     after
       cleanup_modules([TypeSample])
     end
+
+    @tag :erlang_doc
+    test "prints all types in Erlang module" do
+      captured = capture_io(fn -> t(:queue) end)
+
+      # TODO remove once we require Erlang/OTP 27+
+      if System.otp_release() < "27" do
+        assert captured =~ "-type queue() :: queue(_)"
+        assert captured =~ "-opaque queue(Item)"
+      else
+        assert captured =~ "@type queue() :: queue(_)"
+        assert captured =~ "@opaque queue(item)"
+      end
+    end
+
+    @tag :erlang_doc
+    test "prints single type from Erlang module" do
+      captured = capture_io(fn -> t(:erlang.iovec()) end)
+
+      # TODO remove once we require Erlang/OTP 27+
+      if System.otp_release() < "27" do
+        assert captured =~ "-type iovec() :: [binary()]"
+      else
+        assert captured =~ "@type iovec() :: [binary()]"
+      end
+
+      assert captured =~ "A list of binaries."
+
+      captured = capture_io(fn -> t(:erlang.iovec() / 0) end)
+
+      # TODO remove once we require Erlang/OTP 27+
+      if System.otp_release() < "27" do
+        assert captured =~ "-type iovec() :: [binary()]"
+      else
+        assert captured =~ "@type iovec() :: [binary()]"
+      end
+
+      assert captured =~ "A list of binaries."
+    end
+
+    @tag :erlang_doc
+    test "handles non-existing types from Erlang module" do
+      captured = capture_io(fn -> t(:erlang.foo()) end)
+      assert captured =~ "No type information for :erlang.foo was found or :erlang.foo is private"
+
+      captured = capture_io(fn -> t(:erlang.foo() / 1) end)
+      assert captured =~ "No type information for :erlang.foo was found or :erlang.foo is private"
+    end
   end
 
   describe "v" do
@@ -1001,6 +1128,29 @@ defmodule IEx.HelpersTest do
       assert capture_iex("1\n2\nv(2)") == capture_iex("1\n2\nv(-1)")
       assert capture_iex("1\n2\nv(2)") == capture_iex("1\n2\nv()")
     end
+
+    test "returns proper error when trying to access history out of bounds" do
+      # We evaluate 22 statements as 20 is the current limit for iex history
+      assert capture_iex("""
+             \n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n
+             v(1)
+             """) =~
+               "(RuntimeError) v(1) is out of bounds, the currently preserved history ranges from 3 to 22"
+    end
+
+    test "negative lookup works properly after crashes" do
+      capture =
+        capture_iex("""
+            \n
+            \n
+            \n
+            Process.exit(self(), :some_reason)\n
+            :target_value
+            v(-1)
+        """)
+
+      assert capture |> String.split("\n") |> List.last() == ":target_value"
+    end
   end
 
   describe "flush" do
@@ -1055,13 +1205,11 @@ defmodule IEx.HelpersTest do
   describe "import_file" do
     test "imports a file" do
       with_file("dot-iex", "variable = :hello\nimport IO", fn ->
-        capture_io(:stderr, fn ->
-          assert "** (CompileError) iex:1: undefined function variable/0" <> _ =
-                   capture_iex("variable")
-        end)
+        assert capture_io(:stderr, fn -> capture_iex("variable") end) =~
+                 "undefined variable \"variable\""
 
-        assert "** (CompileError) iex:1: undefined function puts/1" <> _ =
-                 capture_iex("puts \"hi\"")
+        assert capture_io(:stderr, fn -> capture_iex("puts \"hi\"") end) =~
+                 "undefined function puts/1"
 
         assert capture_iex("import_file \"dot-iex\"\nvariable\nputs \"hi\"") ==
                  "IO\n:hello\nhi\n:ok"
@@ -1073,13 +1221,11 @@ defmodule IEx.HelpersTest do
       dot_1 = "variable = :hello\nimport IO"
 
       with_file(["dot-iex", "dot-iex-1"], [dot, dot_1], fn ->
-        capture_io(:stderr, fn ->
-          assert "** (CompileError) iex:1: undefined function parent/0" <> _ =
-                   capture_iex("parent")
-        end)
+        assert capture_io(:stderr, fn -> capture_iex("parent") end) =~
+                 "undefined variable \"parent\""
 
-        assert "** (CompileError) iex:1: undefined function puts/1" <> _ =
-                 capture_iex("puts \"hi\"")
+        assert capture_io(:stderr, fn -> capture_iex("puts \"hi\"") end) =~
+                 "undefined function puts/1"
 
         assert capture_iex("import_file \"dot-iex\"\nvariable\nputs \"hi\"\nparent") ==
                  "IO\n:hello\nhi\n:ok\ntrue"
@@ -1102,7 +1248,7 @@ defmodule IEx.HelpersTest do
 
       with_file(["dot-iex-1", "dot-iex-2"], [dot_1, dot_2], fn ->
         assert capture_io(:stderr, fn ->
-                 assert capture_iex(":ok", [], dot_iex_path: "dot-iex-1") == ":ok"
+                 assert capture_iex(":ok", [], dot_iex: "dot-iex-1") == ":ok"
                end) =~ "dot-iex-2 was already imported, skipping circular file imports"
       end)
     end
@@ -1127,7 +1273,8 @@ defmodule IEx.HelpersTest do
   describe "iex> |> (and other binary operators)" do
     test "passes previous result to the pipe" do
       Enum.each([:~>>, :<<~, :~>, :<~, :<~>], fn op ->
-        assert capture_iex("42\n  #{op} IO.puts()") =~ "undefined function #{op}/2"
+        assert capture_io(:stderr, fn -> capture_iex("42\n  #{op} IO.puts()") end) =~
+                 "undefined function #{op}/2"
       end)
 
       assert capture_iex("+42") =~ "42"
@@ -1167,7 +1314,7 @@ defmodule IEx.HelpersTest do
     end
 
     test "handles errors" do
-      ExUnit.CaptureIO.capture_io(fn ->
+      ExUnit.CaptureIO.capture_io(:stderr, fn ->
         with_file("sample.ex", "raise \"oops\"", fn ->
           assert_raise CompileError, fn -> c("sample.ex") end
         end)
@@ -1289,13 +1436,18 @@ defmodule IEx.HelpersTest do
     @tag :capture_log
     test "loads a given module on the given nodes" do
       assert nl([node()], :lists) == {:ok, [{:nonode@nohost, :error, :sticky_directory}]}
-      assert nl(:nonexistent_module) == {:error, :nofile}
-      assert nl([:nosuchnode@badhost], Enum) == {:ok, [{:nosuchnode@badhost, :badrpc, :nodedown}]}
       assert nl([node()], Enum) == {:ok, [{:nonode@nohost, :loaded, Enum}]}
+
+      assert nl(:nonexistent_module) == {:error, :nofile}
+
+      assert nl([:nosuchnode@badhost], Enum) ==
+               {:ok, [{:nosuchnode@badhost, :badrpc, :noconnection}]}
     end
   end
 
   describe "r" do
+    @describetag :requires_source
+
     test "raises when reloading a nonexistent module" do
       assert_raise ArgumentError, "could not load nor find module: :nonexistent_module", fn ->
         r(:nonexistent_module)
@@ -1326,7 +1478,7 @@ defmodule IEx.HelpersTest do
                    Sample.run()
                  end
                end) =~
-                 "redefining module Sample (current version loaded from ./Elixir.Sample.beam)"
+                 "redefining module Sample (current version loaded from Elixir.Sample.beam)"
       end)
     after
       # Clean up old version produced by the r helper
@@ -1356,8 +1508,8 @@ defmodule IEx.HelpersTest do
 
   describe "pid/1,3" do
     test "builds a PID from string" do
-      assert inspect(pid("0.32767.3276")) == "#PID<0.32767.3276>"
-      assert inspect(pid("0.5.6")) == "#PID<0.5.6>"
+      assert inspect(pid("0.32767.0")) == "#PID<0.32767.0>"
+      assert inspect(pid("0.5.0")) == "#PID<0.5.0>"
 
       assert_raise ArgumentError, fn ->
         pid("0.6.-6")
@@ -1373,8 +1525,8 @@ defmodule IEx.HelpersTest do
     end
 
     test "builds a PID from integers" do
-      assert inspect(pid(0, 32767, 3276)) == "#PID<0.32767.3276>"
-      assert inspect(pid(0, 5, 6)) == "#PID<0.5.6>"
+      assert inspect(pid(0, 32767, 0)) == "#PID<0.32767.0>"
+      assert inspect(pid(0, 5, 0)) == "#PID<0.5.0>"
 
       assert_raise FunctionClauseError, fn ->
         pid(0, 6, -6)
@@ -1453,7 +1605,7 @@ defmodule IEx.HelpersTest do
       defstruct []
 
       defimpl IEx.Info do
-        def info(_), do: [{"A", "it's A"}, {:b, "it's :b"}, {'c', "it's 'c'"}]
+        def info(_), do: [{"A", "it's A"}, {:b, "it's :b"}, {~c"c", "it's 'c'"}]
       end
     end
 
