@@ -3,7 +3,7 @@ defmodule Mix do
   Mix is a build tool that provides tasks for creating, compiling,
   and testing Elixir projects, managing its dependencies, and more.
 
-  ## Mix.Project
+  ## `Mix.Project`
 
   The foundation of Mix is a project. A project can be defined by using
   `Mix.Project` in a module, usually placed in a file named `mix.exs`:
@@ -36,7 +36,7 @@ defmodule Mix do
   The best way to get started with your first project is by calling
   `mix new my_project` from the command line.
 
-  ## Mix.Task
+  ## `Mix.Task`
 
   Tasks are what make Mix extensible.
 
@@ -108,6 +108,17 @@ defmodule Mix do
 
       {:some_test_dependency, "~> 1.0", only: :test}
 
+  When running Mix via the command line, you can configure the default
+  environment or the preferred environment per task via the `def cli`
+  function in your `mix.exs`. For example:
+
+      def cli do
+        [
+          default_env: :local,
+          preferred_envs: [docs: :docs]
+        ]
+      end
+
   The environment can be read via `Mix.env/0`.
 
   ## Targets
@@ -116,16 +127,30 @@ defmodule Mix do
   project needs to compile to different architectures and some of the
   dependencies are only available to some of them. By default, the target
   is `:host` but it can be set via the `MIX_TARGET` environment variable.
+
+  When running Mix via the command line, you can configure the default
+  target or the preferred target per task via the `def cli` function
+  in your `mix.exs`. For example:
+
+      def cli do
+        [
+          default_target: :local,
+          preferred_targets: [docs: :docs]
+        ]
+      end
+
   The target can be read via `Mix.target/0`.
 
   ## Configuration
 
-  Mix allows you configure the application environment of your application
+  Mix allows you to configure the application environment of your application
   and of your dependencies. See the `Application` module to learn more about
   the application environment. On this section, we will focus on how to configure
   it at two distinct moments: build-time and runtime.
 
-  > Note: The application environment is discouraged for libraries. See Elixir's
+  > #### Avoiding the application environment {: .warning}
+  >
+  > The application environment is discouraged for libraries. See Elixir's
   > [Library Guidelines](https://hexdocs.pm/elixir/library-guidelines.html) for
   > more information.
 
@@ -151,8 +176,11 @@ defmodule Mix do
   machine before your code compiles. This can be an issue if the machine
   compiling your code does not have access to all environment variables
   used to run your code, as loading the config above will fail due to the
-  missing environment variable. Luckily, Mix also provides runtime
-  configuration, which should be preferred and we will see next.
+  missing environment variable. Furthermore, even if the environment variable
+  is set, changing the environment variable will require a full recompilation
+  of your application by calling `mix compile --force` (otherwise your project
+  won't start). Luckily, Mix also provides runtime configuration, which is
+  preferred in such cases and we will see next.
 
   ### Runtime configuration
 
@@ -189,19 +217,31 @@ defmodule Mix do
         defp aliases do
           [
             c: "compile",
-            hello: &hello/1
+            hello: &hello/1,
+            paid_task: &paid_task/1
           ]
         end
 
         defp hello(_) do
           Mix.shell().info("Hello world")
         end
+
+        defp paid_task(_) do
+          Mix.Task.run("paid.task", [
+            "first_arg",
+            "second_arg",
+            "--license-key",
+            System.fetch_env!("SOME_LICENSE_KEY")
+          ])
+        end
       end
 
-  In the example above, we have defined two aliases. One is `mix c`
-  which is a shortcut for `mix compile`. The other is named
-  `mix hello`, which is the equivalent to the `Mix.Tasks.Hello`
-  we have defined in the [Mix.Task section](#module-mix-task).
+  In the example above, we have defined three aliases. One is `mix c`
+  which is a shortcut for `mix compile`. Another is named
+  `mix hello` and the third is named `mix paid_task`, which executes
+  the code inside a custom function to invoke the `paid.task` task
+  with several arguments, including one pulled from an environment
+  variable.
 
   Aliases may also be lists, specifying multiple tasks to be run
   consecutively:
@@ -221,19 +261,13 @@ defmodule Mix do
   Where `&clean_extra/1` would be a function in your `mix.exs`
   with extra cleanup logic.
 
-  Arguments given to the alias will be appended to the arguments of
-  the last task in the list. Except when overriding an existing task.
-  In this case, the arguments will be given to the original task,
-  in order to preserve semantics. For example, in the `:clean` alias
-  above, the arguments given to the alias will be passed to "clean"
-  and not to `clean_extra/1`.
+  If the alias is overriding an existing task, the arguments given
+  to the alias will be forwarded to the original task in order to
+  preserve semantics. Otherwise arguments given to the alias are
+  appended to the arguments of the last task in the list.
 
-  Aliases defined in the current project do not affect its dependencies
-  and aliases defined in dependencies are not accessible from the
-  current project.
-
-  Aliases can be used very powerfully to also run Elixir scripts and
-  shell commands, for example:
+  Another use case of aliases is to run Elixir scripts and shell
+  commands, for example:
 
       # priv/hello1.exs
       IO.puts("Hello One")
@@ -257,10 +291,15 @@ defmodule Mix do
   then `mix cmd` to execute a command line shell script. This shows how
   powerful aliases mixed with Mix tasks can be.
 
-  Mix tasks are designed to run only once. This prevents the same task
-  from being executed multiple times. For example, if there are several tasks
-  depending on `mix compile`, the code will be compiled once. Tasks can
-  be executed again if they are explicitly reenabled using `Mix.Task.reenable/1`:
+  One common pitfall of aliases comes when trying to invoke the same task
+  multiple times. Mix tasks are designed to run only once. This prevents
+  the same task from being executed multiple times. For example, if there
+  are several tasks depending on `mix compile`, the code will be compiled
+  only once.
+
+  Similarly, `mix format` can only be invoked once. So if you have an alias
+  that attempts to invoke `mix format` multiple times, it won't work unless
+  it is explicitly reenabled using `Mix.Task.reenable/1`:
 
       another_alias: [
         "format --check-formatted priv/hello1.exs",
@@ -270,46 +309,63 @@ defmodule Mix do
       ]
 
   Some tasks are automatically reenabled though, as they are expected to
-  be invoked multiple times. They are: `mix cmd`, `mix do`, `mix loadconfig`,
-  `mix profile.cprof`, `mix profile.eprof`, `mix profile.fprof`, `mix run`,
-  and `mix xref`.
+  be invoked multiple times, such as: `mix cmd`, `mix do`, `mix xref`, etc.
 
-  It is worth mentioning that some tasks, such as in the case of the
-  `mix format` command in the example above, can accept multiple files so it
-  could be rewritten as:
-
-      another_alias: ["format --check-formatted priv/hello1.exs priv/hello2.exs"]
+  Finally, aliases defined in the current project do not affect its
+  dependencies and aliases defined in dependencies are not accessible
+  from the current project, with the exception of umbrella projects.
+  Umbrella projects will run the aliases of its children when the
+  umbrella project itself does not define said alias and there is no
+  task with said name.
 
   ## Environment variables
 
-  Several environment variables can be used to modify Mix's behaviour.
+  Several environment variables can be used to modify Mix's behavior.
 
   Mix responds to the following variables:
 
     * `MIX_ARCHIVES` - specifies the directory into which the archives should be installed
       (default: `~/.mix/archives`)
-    * `MIX_BUILD_ROOT` - sets the root directory where build artifacts
-      should be written to. For example, "_build". If `MIX_BUILD_PATH` is set, this
-      option is ignored.
-    * `MIX_BUILD_PATH` - sets the project `Mix.Project.build_path/0` config. This option
-      must always point to a subdirectory inside a temporary directory. For instance,
-      never "/tmp" or "_build" but "_build/PROD" or "/tmp/PROD", as required by Mix
-    * `MIX_DEPS_PATH` - sets the project `Mix.Project.deps_path/0` config for the current project (default: `deps`)
+
+    * `MIX_BUILD_PATH` - sets the project `Mix.Project.build_path/0` config.
+      This option must always point to a subdirectory inside a temporary directory.
+      For instance, never "/tmp" or "_build" but "_build/PROD" or "/tmp/PROD", as
+      required by Mix. This environment variable is used mostly by external build
+      tools. For your CI servers, you likely want to use `MIX_BUILD_ROOT` below.
+
+    * `MIX_BUILD_ROOT` - sets the root directory where build artifacts should be
+      written to. For example, "_build". If `MIX_BUILD_PATH` is set, this option
+      is ignored.
+
     * `MIX_DEBUG` - outputs debug information about each task before running it
+
+    * `MIX_DEPS_PATH` - sets the project `Mix.Project.deps_path/0` config for the
+      current project (default: `deps`)
+
     * `MIX_ENV` - specifies which environment should be used. See [Environments](#module-environments)
-    * `MIX_TARGET` - specifies which target should be used. See [Targets](#module-targets)
+
     * `MIX_EXS` - changes the full path to the `mix.exs` file
+
     * `MIX_HOME` - path to Mix's home directory, stores configuration files and scripts used by Mix
       (default: `~/.mix`)
-    * `MIX_INSTALL_DIR` - (since v1.12.0) specifies directory where `Mix.install/2` keeps
+
+    * `MIX_INSTALL_DIR` *(since v1.12.0)* - specifies directory where `Mix.install/2` keeps
        install cache
-    * `MIX_INSTALL_FORCE` - (since v1.13.0) runs `Mix.install/2` with empty install cache
+
     * `MIX_PATH` - appends extra code paths
+
+    * `MIX_PROFILE` - a list of comma-separated Mix tasks to profile the time spent on
+      functions by the process running the task
+
     * `MIX_QUIET` - does not print information messages to the terminal
+
     * `MIX_REBAR3` - path to rebar3 command that overrides the one Mix installs
       (default: `~/.mix/rebar3`)
+
+    * `MIX_TARGET` - specifies which target should be used. See [Targets](#module-targets)
+
     * `MIX_XDG` - asks Mix to follow the [XDG Directory Specification](https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html)
-      for its home directory and configuration files. This behaviour needs to
+      for its home directory and configuration files. This behavior needs to
       be opt-in due to backwards compatibility. `MIX_HOME` has higher preference
       than `MIX_XDG`. If none of the variables are set, the default directory
       `~/.mix` will be used
@@ -318,9 +374,44 @@ defmodule Mix do
   flags) should be set to either `1` or `true`, for example:
 
       $ MIX_DEBUG=1 mix compile
+
+  ## SSL certificates
+
+  Mix and the Hex package manager use the default operating system certificates
+  when downloading resources. In certain situations, such as when running behind
+  proxies, you want to replace those certificates.
+
+  If you simply want to change the certificates used by Mix and Hex, you may set
+  the `HEX_CACERTS_PATH` environment variable, pointing to a CA certificate file.
+  See [`mix hex.config`](https://hexdocs.pm/hex/Mix.Tasks.Hex.Config.html#module-config-keys).
+
+  From Erlang/OTP 27.2, it is also possible to change the certificates for your
+  project as a whole. To do so, you might add the following to your `config/config.exs`:
+
+      config :public_key, :cacerts_path, "/path/to/certs.pem"
+
+  You can also do so by setting the `ERL_AFLAGS` and `ERL_ZFLAGS` environment variables:
+
+  ```bash
+  ERL_AFLAGS="-public_key cacerts_path '\"/path/to/certs.pem\"'"
+  ERL_ZFLAGS="-public_key cacerts_path '\"/path/to/certs.pem\"'"
+  ```
+
+  You can verify if the configuration has been properly set by calling the following
+  inside `iex -S mix`:
+
+      Application.load(:public_key)
+      Application.get_env(:public_key, :cacerts_path)
+
+  And by loading the certificates:
+
+      :public_key.cacerts_get()
+
   """
 
   @mix_install_project __MODULE__.InstallProject
+  @mix_install_app :mix_install
+  @mix_install_app_string Atom.to_string(@mix_install_app)
 
   use Application
 
@@ -332,13 +423,22 @@ defmodule Mix do
     :ok
   end
 
-  @doc false
+  @impl true
   def start(_type, []) do
-    children = [Mix.State, Mix.TasksServer, Mix.ProjectStack]
+    Mix.Local.append_archives()
+    Mix.Local.append_paths()
+    children = [Mix.Sync.PubSub, Mix.State, Mix.TasksServer, Mix.ProjectStack]
     opts = [strategy: :one_for_one, name: Mix.Supervisor, max_restarts: 0]
     Supervisor.start_link(children, opts)
   end
 
+  @impl true
+  def stop(_data) do
+    Mix.Local.remove_archives()
+    Mix.Local.remove_paths()
+    :ok
+  end
+
   @doc """
   Returns the current Mix environment.
 
@@ -404,7 +504,7 @@ defmodule Mix do
   """
   @spec compilers() :: [atom()]
   def compilers do
-    [:yecc, :leex, :erlang, :elixir, :app]
+    [:erlang, :elixir, :app]
   end
 
   @doc """
@@ -526,6 +626,58 @@ defmodule Mix do
     Path.join(Mix.Utils.mix_home(), "escripts")
   end
 
+  @doc """
+  Ensures the given application from Erlang/OTP or Elixir and its dependencies
+  are available in the path.
+
+  Generally speaking, you should list the Erlang application dependencies under
+  the `:extra_applications` section of your `mix.exs`. This must only be used by
+  Mix tasks which wish to avoid depending on Erlang/Elixir for certain reasons.
+
+  This function does not start the given applications.
+  """
+  @doc since: "1.15.0"
+  def ensure_application!(app) when is_atom(app) do
+    ensure_application!(app, Mix.State.builtin_apps(), [], %{})
+    :ok
+  end
+
+  defp ensure_application!(app, builtin_apps, optional, seen) do
+    case seen do
+      %{^app => _} ->
+        seen
+
+      %{} ->
+        seen = Map.put(seen, app, true)
+
+        case builtin_apps do
+          %{^app => path} ->
+            Code.prepend_path(path, cache: true)
+            Application.load(app)
+
+            required = List.wrap(Application.spec(app, :applications))
+            optional = List.wrap(Application.spec(app, :optional_applications))
+
+            Enum.reduce(
+              required ++ optional,
+              seen,
+              &ensure_application!(&1, builtin_apps, optional, &2)
+            )
+
+          %{} ->
+            if app not in optional do
+              Mix.raise(
+                "The application \"#{app}\" could not be found. This may happen if your " <>
+                  "Operating System broke Erlang into multiple packages and may be fixed " <>
+                  "by installing the missing \"erlang-dev\" and \"erlang-#{app}\" packages"
+              )
+            end
+
+            seen
+        end
+    end
+  end
+
   @doc """
   Installs and starts dependencies.
 
@@ -542,9 +694,6 @@ defmodule Mix do
   This function can only be called outside of a Mix project and only with the
   same dependencies in the given VM.
 
-  **Note:** this feature is currently experimental and it may change
-  in future releases.
-
   ## Options
 
     * `:force` - if `true`, runs with empty install cache. This is useful when you want
@@ -561,21 +710,25 @@ defmodule Mix do
     * `:elixir` - if set, ensures the current Elixir version matches the given
       version requirement (Default: `nil`)
 
-    * `:system_env` (since v1.13.0) - a list or a map of system environment variable
+    * `:system_env` *(since v1.13.0)* - a list or a map of system environment variable
       names with respective values as binaries. The system environment is made part
       of the `Mix.install/2` cache, so different configurations will lead to different apps
 
-    * `:config` (since v1.13.0) - a keyword list of keyword lists with application
-      configuration to be set before the apps loaded. The configuration is part of
-      the `Mix.install/2` cache, so different configurations will lead to different
-      apps
+    * `:config` *(since v1.13.0)* - a keyword list of keyword lists of compile-time
+      configuration. The configuration is part of the `Mix.install/2` cache, so
+      different configurations will lead to different apps. For this reason, you
+      want to minimize the amount of configuration set through this option.
+      Use `Application.put_all_env/2` for setting other runtime configuration.
 
-    * `:config_path` (since v1.14.0) - path to a configuration file. If a `runtime.exs`
+    * `:config_path` *(since v1.14.0)* - path to a configuration file. If a `runtime.exs`
       file exists in the same directory as the given path, it is loaded too.
 
-    * `:lockfile` (since v1.14.0) - path to a lockfile to be used as a basis of
+    * `:lockfile` *(since v1.14.0)* - path to a lockfile to be used as a basis of
       dependency resolution.
 
+    * `:start_applications` *(since v1.15.3)* - if `true`, ensures that installed app
+      and its dependencies are started after install (Default: `true`)
+
   ## Examples
 
   Installing `:decimal` and `:jason`:
@@ -585,7 +738,7 @@ defmodule Mix do
         {:jason, "~> 1.0"}
       ])
 
-  Installing `:nx` & `:exla`, and configuring the underlying applications
+  Installing `:nx` and `:exla`, and configuring the underlying applications
   and environment variables:
 
       Mix.install(
@@ -615,10 +768,21 @@ defmodule Mix do
       Hexpm.Repo.query!("SELECT COUNT(1) from packages")
       #=> ...
 
+  The example above can be simplified by passing the application
+  name as an atom for `:config_path` and `:lockfile`:
+
+      Mix.install(
+        [
+          {:hexpm, path: "/tmp/hexpm", env: :dev},
+        ],
+        config_path: :hexpm,
+        lockfile: :hexpm
+      )
+
   ## Limitations
 
   There is one limitation to `Mix.install/2`, which is actually an Elixir
-  behaviour. If you are installing a dependency that defines a struct or
+  behavior. If you are installing a dependency that defines a struct or
   macro, you cannot use the struct or macro immediately after the install
   call. For example, this won't work:
 
@@ -629,7 +793,7 @@ defmodule Mix do
   it executes the code. This means that, by the time Elixir tries to expand
   the `%Decimal{}` struct, the dependency has not been installed yet.
 
-  Luckily this has a straightforward solution, which is move the code to
+  Luckily this has a straightforward solution, which is to move the code
   inside a module:
 
       Mix.install([:decimal])
@@ -645,6 +809,26 @@ defmodule Mix do
   The contents inside `defmodule` will only be expanded and executed
   after `Mix.install/2` runs, which means that any struct, macros,
   and imports will be correctly handled.
+
+  ## Environment variables
+
+  The `MIX_INSTALL_DIR` environment variable configures the directory that
+  caches all `Mix.install/2`. It defaults to the "mix/install" folder in the
+  default user cache of your operating system. You can use `install_project_dir/0`
+  to access the directory of an existing install (alongside other installs):
+
+      iex> Mix.install([])
+      iex> Mix.install_project_dir()
+
+  The `MIX_INSTALL_FORCE` is available since Elixir v1.13.0 and forces
+  `Mix.install/2` to discard any previously cached entry of the current install.
+
+  The `MIX_INSTALL_RESTORE_PROJECT_DIR` environment variable may be specified
+  since Elixir v1.16.2. It should point to a previous installation directory,
+  which can be obtained with `Mix.install_project_dir/0` (after calling `Mix.install/2`).
+  Using a restore dir may speed up the installation, since matching dependencies
+  do not need be refetched nor recompiled. This environment variable is ignored
+  if `:force` is enabled.
   """
   @doc since: "1.12.0"
   def install(deps, opts \\ [])
@@ -681,10 +865,25 @@ defmodule Mix do
           other
       end)
 
-    config = Keyword.get(opts, :config, [])
-    config_path = opts[:config_path] && Path.expand(opts[:config_path])
-    system_env = Keyword.get(opts, :system_env, [])
-    consolidate_protocols? = Keyword.get(opts, :consolidate_protocols, true)
+    opts =
+      Keyword.validate!(opts,
+        config: [],
+        config_path: nil,
+        consolidate_protocols: true,
+        elixir: nil,
+        force: false,
+        lockfile: nil,
+        runtime_config: [],
+        start_applications: true,
+        system_env: [],
+        verbose: false
+      )
+
+    config_path = expand_path(opts[:config_path], deps, :config_path, "config/config.exs")
+    config = Keyword.fetch!(opts, :config)
+    system_env = Keyword.fetch!(opts, :system_env)
+    consolidate_protocols? = Keyword.fetch!(opts, :consolidate_protocols)
+    start_applications? = Keyword.fetch!(opts, :start_applications)
 
     id =
       {deps, config, system_env, consolidate_protocols?}
@@ -692,57 +891,56 @@ defmodule Mix do
       |> :erlang.md5()
       |> Base.encode16(case: :lower)
 
-    force? = System.get_env("MIX_INSTALL_FORCE") in ["1", "true"] or !!opts[:force]
+    force? = System.get_env("MIX_INSTALL_FORCE") in ["1", "true"] or Keyword.fetch!(opts, :force)
 
     case Mix.State.get(:installed) do
       nil ->
         Application.put_all_env(config, persistent: true)
         System.put_env(system_env)
 
-        install_dir = install_dir(id)
+        install_project_dir = install_project_dir(id)
 
-        if opts[:verbose] do
-          Mix.shell().info("Mix.install/2 using #{install_dir}")
+        if Keyword.fetch!(opts, :verbose) do
+          Mix.shell().info("Mix.install/2 using #{install_project_dir}")
         end
 
         if force? do
-          File.rm_rf!(install_dir)
+          File.rm_rf!(install_project_dir)
         end
 
-        config = [
-          version: "0.1.0",
-          build_embedded: false,
-          build_per_environment: true,
-          build_path: "_build",
-          lockfile: "mix.lock",
-          deps_path: "deps",
+        dynamic_config = [
           deps: deps,
-          app: :mix_install,
-          erlc_paths: ["src"],
-          elixirc_paths: ["lib"],
-          compilers: [],
           consolidate_protocols: consolidate_protocols?,
           config_path: config_path
         ]
 
+        config = install_project_config(dynamic_config)
+
         started_apps = Application.started_applications()
-        :ok = Mix.Local.append_archives()
         :ok = Mix.ProjectStack.push(@mix_install_project, config, "nofile")
-        build_dir = Path.join(install_dir, "_build")
-        external_lockfile = opts[:lockfile] && Path.expand(opts[:lockfile])
+        build_dir = Path.join(install_project_dir, "_build")
+        external_lockfile = expand_path(opts[:lockfile], deps, :lockfile, "mix.lock")
 
         try do
           first_build? = not File.dir?(build_dir)
-          File.mkdir_p!(install_dir)
 
-          File.cd!(install_dir, fn ->
+          restore_dir = System.get_env("MIX_INSTALL_RESTORE_PROJECT_DIR")
+
+          if first_build? and restore_dir != nil and not force? do
+            File.cp_r(restore_dir, install_project_dir)
+            remove_dep(install_project_dir, @mix_install_app_string)
+          end
+
+          File.mkdir_p!(install_project_dir)
+
+          File.cd!(install_project_dir, fn ->
             if config_path do
               Mix.Task.rerun("loadconfig")
             end
 
             cond do
               external_lockfile ->
-                md5_path = Path.join(install_dir, "merge.lock.md5")
+                md5_path = Path.join(install_project_dir, "merge.lock.md5")
 
                 old_md5 =
                   case File.read(md5_path) do
@@ -753,10 +951,10 @@ defmodule Mix do
                 new_md5 = external_lockfile |> File.read!() |> :erlang.md5()
 
                 if old_md5 != new_md5 do
-                  lockfile = Path.join(install_dir, "mix.lock")
+                  lockfile = Path.join(install_project_dir, "mix.lock")
                   old_lock = Mix.Dep.Lock.read(lockfile)
                   new_lock = Mix.Dep.Lock.read(external_lockfile)
-                  Mix.Dep.Lock.write(lockfile, Map.merge(old_lock, new_lock))
+                  Mix.Dep.Lock.write(Map.merge(old_lock, new_lock), file: lockfile)
                   File.write!(md5_path, Base.encode64(new_md5))
                   Mix.Task.rerun("deps.get")
                 end
@@ -765,6 +963,9 @@ defmodule Mix do
                 Mix.Task.rerun("deps.get")
 
               true ->
+                # We already have a cache. If the user by any chance uninstalled Hex,
+                # we make sure it is installed back (which mix deps.get would do anyway)
+                Mix.Hex.ensure_installed?(true)
                 :ok
             end
 
@@ -781,18 +982,29 @@ defmodule Mix do
             end
           end)
 
-          for %{app: app, opts: opts} <- Mix.Dep.cached(),
-              Keyword.get(opts, :runtime, true) and Keyword.get(opts, :app, true) do
-            Application.ensure_all_started(app)
+          if start_applications? do
+            for %{app: app, opts: opts} <- Mix.Dep.cached(),
+                Keyword.get(opts, :runtime, true) and Keyword.get(opts, :app, true) do
+              Application.ensure_all_started(app)
+            end
           end
 
-          Mix.State.put(:installed, id)
+          if restore_dir do
+            remove_leftover_deps(install_project_dir)
+          end
+
+          Mix.State.put(:installed, {id, dynamic_config})
           :ok
         after
           Mix.ProjectStack.pop()
+          # Clear all tasks invoked during installation, since there
+          # is no reason to keep this in memory. Additionally this
+          # allows us to rerun tasks for the dependencies later on,
+          # such as recompilation
+          Mix.Task.clear()
         end
 
-      ^id when not force? ->
+      {^id, _dynamic_config} when not force? ->
         :ok
 
       _ ->
@@ -800,7 +1012,53 @@ defmodule Mix do
     end
   end
 
-  defp install_dir(cache_id) do
+  defp expand_path(_path = nil, _deps, _key, _), do: nil
+  defp expand_path(path, _deps, _key, _) when is_binary(path), do: Path.expand(path)
+
+  defp expand_path(app_name, deps, key, relative_path) when is_atom(app_name) do
+    app_dir =
+      case List.keyfind(deps, app_name, 0) do
+        {_, _, opts} when is_list(opts) -> opts[:path]
+        {_, opts} when is_list(opts) -> opts[:path]
+        _ -> Mix.raise("unknown dependency #{inspect(app_name)} given to #{inspect(key)}")
+      end
+
+    if !app_dir do
+      Mix.raise("#{inspect(app_name)} given to #{inspect(key)} must be a path dependency")
+    end
+
+    Path.join(app_dir, relative_path)
+  end
+
+  defp remove_leftover_deps(install_project_dir) do
+    build_lib_dir = Path.join([install_project_dir, "_build", "dev", "lib"])
+
+    deps = File.ls!(build_lib_dir)
+
+    loaded_deps =
+      for {app, _description, _version} <- Application.loaded_applications(),
+          into: MapSet.new(),
+          do: Atom.to_string(app)
+
+    # We want to keep :mix_install, but it has no application
+    loaded_deps = MapSet.put(loaded_deps, @mix_install_app_string)
+
+    for dep <- deps, not MapSet.member?(loaded_deps, dep) do
+      remove_dep(install_project_dir, dep)
+    end
+  end
+
+  defp remove_dep(install_project_dir, dep) do
+    build_lib_dir = Path.join([install_project_dir, "_build", "dev", "lib"])
+    deps_dir = Path.join(install_project_dir, "deps")
+
+    build_path = Path.join(build_lib_dir, dep)
+    File.rm_rf(build_path)
+    dep_path = Path.join(deps_dir, dep)
+    File.rm_rf(dep_path)
+  end
+
+  defp install_project_dir(cache_id) do
     install_root =
       System.get_env("MIX_INSTALL_DIR") ||
         Path.join(Mix.Utils.mix_cache(), "installs")
@@ -809,12 +1067,75 @@ defmodule Mix do
     Path.join([install_root, version, cache_id])
   end
 
+  defp install_project_config(dynamic_config) do
+    [
+      version: "0.1.0",
+      build_embedded: false,
+      build_per_environment: true,
+      build_path: "_build",
+      lockfile: "mix.lock",
+      deps_path: "deps",
+      app: @mix_install_app,
+      erlc_paths: [],
+      elixirc_paths: [],
+      compilers: [],
+      prune_code_paths: false
+    ] ++ dynamic_config
+  end
+
+  @doc false
+  def in_install_project(fun) do
+    case safe_get_installed() do
+      {id, dynamic_config} ->
+        config = install_project_config(dynamic_config)
+
+        install_project_dir = install_project_dir(id)
+
+        File.cd!(install_project_dir, fn ->
+          :ok = Mix.ProjectStack.push(@mix_install_project, config, "nofile")
+
+          try do
+            fun.()
+          after
+            Mix.ProjectStack.pop()
+          end
+        end)
+
+      nil ->
+        Mix.raise("trying to call Mix.in_install_project/1, but Mix.install/2 was never called")
+    end
+  end
+
+  @doc """
+  Returns the directory where the current `Mix.install/2` project
+  resides.
+  """
+  @doc since: "1.16.2"
+  @spec install_project_dir() :: Path.t() | nil
+  def install_project_dir() do
+    case safe_get_installed() do
+      {id, _dynamic_config} -> install_project_dir(id)
+      nil -> nil
+    end
+  end
+
   @doc """
   Returns whether `Mix.install/2` was called in the current node.
   """
   @doc since: "1.13.0"
+  @spec installed?() :: boolean()
   def installed? do
-    Mix.State.get(:installed) != nil
+    safe_get_installed() != nil
+  end
+
+  defp safe_get_installed() do
+    if mix_started?() do
+      Mix.State.get(:installed)
+    end
+  end
+
+  defp mix_started?() do
+    Process.whereis(Mix.State) != nil
   end
 
   defp stop_apps([]), do: :ok
