@@ -60,6 +60,18 @@ defmodule Mix.Dep.Converger do
     )
   end
 
+  @doc """
+  Converge without lock and accumulator updates.
+
+  Note the dependencies returned from converge are not yet loaded.
+  The relevant app keys are found under `dep.opts[:app_properties]`.
+
+  See `Mix.Dep.Loader.children/1` for options.
+  """
+  def converge(opts \\ []) do
+    converge(nil, nil, opts, &{&1, &2, &3}) |> elem(0)
+  end
+
   @doc """
   Converges all dependencies from the current project,
   including nested dependencies.
@@ -68,20 +80,23 @@ defmodule Mix.Dep.Converger do
   must return an updated dependency in case some processing
   is done.
 
+  Note the dependencies returned from converge are not yet loaded.
+  The relevant app keys are found under `dep.opts[:app_properties]`.
+
   See `Mix.Dep.Loader.children/1` for options.
   """
   def converge(acc, lock, opts, callback) do
     {deps, acc, lock} = all(acc, lock, opts, callback)
-    if remote = Mix.RemoteConverger.get(), do: remote.post_converge
+    if remote = Mix.RemoteConverger.get(), do: remote.post_converge()
     {topological_sort(deps), acc, lock}
   end
 
   defp all(acc, lock, opts, callback) do
-    main = Mix.Dep.Loader.children()
+    lock_given? = !!lock
+
+    main = Mix.Dep.Loader.children(lock_given?)
     main = Enum.map(main, &%{&1 | top_level: true})
     apps = Enum.map(main, & &1.app)
-
-    lock_given? = !!lock
     env_target = {opts[:env], opts[:target]}
 
     # If no lock was given, let's read one to fill in the deps
@@ -92,7 +107,7 @@ defmodule Mix.Dep.Converger do
     # lazily loaded, we need to check for it on every
     # iteration.
     {deps, acc, lock} =
-      all(main, apps, callback, acc, lock, env_target, fn dep ->
+      init_all(main, apps, acc, lock, callback, lock_given?, env_target, fn dep ->
         if (remote = Mix.RemoteConverger.get()) && remote.remote?(dep) do
           {:loaded, dep}
         else
@@ -108,7 +123,7 @@ defmodule Mix.Dep.Converger do
     use_remote? = !!remote and Enum.any?(deps, &remote.remote?/1)
 
     if not diverged? and use_remote? do
-      # Make sure there are no cycles before calling remote converge
+      # Make sure there are no cycles before calling the remote converger
       topological_sort(deps)
 
       # If there is a lock, it means we are doing a get/update
@@ -117,22 +132,30 @@ defmodule Mix.Dep.Converger do
       # on, there is no lock, so we won't hit this branch.
       lock = if lock_given?, do: remote.converge(deps, lock), else: lock
 
-      # Build a cache using both dep_name and dep_scm to ensure we only
-      # return :loaded for deps which have been loaded from the same source.
       cache =
         deps
         |> Enum.reject(&remote.remote?(&1))
-        |> Enum.into(%{}, &{{&1.app, &1.scm}, &1})
+        |> Enum.into(%{}, &{&1.app, &1})
 
       # In case no lock was given, we will use the local lock
       # which is potentially stale. So remote.deps/2 needs to always
       # check if the data it finds in the lock is actually valid.
       {deps, acc, lock} =
-        all(main, apps, callback, acc, lock, env_target, fn dep ->
-          if cached = cache[{dep.app, dep.scm}] do
-            {:loaded, cached}
-          else
-            {:unloaded, dep, remote.deps(dep, lock)}
+        init_all(main, apps, acc, lock, callback, lock_given?, env_target, fn dep ->
+          %{app: app, scm: scm} = dep
+
+          case cache do
+            # If the SCM matches, use the already cached dependency from first converger pass
+            %{^app => %{scm: ^scm} = cached} ->
+              {:loaded, cached}
+
+            # If SCM doesn't match, the remote converger resolved a new dependency but because
+            # there is conflict we return the new dependency as is for the converger to fail
+            %{^app => _cached} ->
+              {:loaded, dep}
+
+            %{} ->
+              {:unloaded, dep, remote.deps(dep, lock)}
           end
         end)
 
@@ -142,8 +165,9 @@ defmodule Mix.Dep.Converger do
     end
   end
 
-  defp all(main, apps, callback, rest, lock, env_target, cache) do
-    {deps, rest, lock} = all(main, [], [], apps, callback, rest, lock, env_target, cache)
+  defp init_all(main, apps, rest, lock, callback, locked?, env_target, cache) do
+    state = %{locked?: locked?, env_target: env_target, cache: cache, callback: callback}
+    {deps, _kept, _optional, rest, lock} = all(main, [], [], [], apps, [], rest, lock, state)
     deps = Enum.reverse(deps)
     # When traversing dependencies, we keep skipped ones to
     # find conflicts. We remove them now after traversal.
@@ -190,48 +214,55 @@ defmodule Mix.Dep.Converger do
   # Now, since "d" was specified in a parent project, no
   # exception is going to be raised since d is considered
   # to be the authoritative source.
-  defp all([dep | t], acc, upper_breadths, breadths, callback, rest, lock, env_target, cache) do
-    case match_deps(acc, upper_breadths, dep, env_target) do
+  defp all([dep | t], acc, kept, upper, breadths, optional, rest, lock, state) do
+    case match_deps(acc, upper, dep, state.env_target) do
       {:replace, dep, acc} ->
-        all([dep | t], acc, upper_breadths, breadths, callback, rest, lock, env_target, cache)
+        all([dep | t], acc, kept, upper, breadths, optional, rest, lock, state)
 
       {:match, acc} ->
-        all(t, acc, upper_breadths, breadths, callback, rest, lock, env_target, cache)
+        all(t, acc, kept, upper, breadths, optional, rest, lock, state)
 
       :skip ->
         # We still keep skipped dependencies around to detect conflicts.
-        # They must be rejected after every all iteration.
-        all(t, [dep | acc], upper_breadths, breadths, callback, rest, lock, env_target, cache)
+        # They must be rejected after every all iteration but they are not
+        # included in the list of kept dependencies.
+        all(t, [dep | acc], kept, upper, breadths, optional, rest, lock, state)
 
       :nomatch ->
-        {dep, rest, lock} =
-          case cache.(dep) do
+        {%{app: app, deps: deps, opts: opts} = dep, rest, lock} =
+          case state.cache.(dep) do
             {:loaded, cached_dep} ->
               {cached_dep, rest, lock}
 
             {:unloaded, dep, children} ->
-              {dep, rest, lock} = callback.(put_lock(dep, lock), rest, lock)
+              {dep, rest, lock} = state.callback.(put_lock(dep, lock), rest, lock)
 
-              Mix.Dep.Loader.with_system_env(dep, fn ->
-                # After we invoke the callback (which may actually check out the
-                # dependency), we load the dependency including its latest info
-                # and children information.
-                {Mix.Dep.Loader.load(dep, children), rest, lock}
-              end)
+              dep =
+                Mix.Dep.Loader.with_system_env(dep, fn ->
+                  Mix.Dep.Loader.load(dep, children, state.locked?)
+                end)
+
+              {dep, rest, lock}
           end
 
-        {acc, rest, lock} =
-          all(t, [dep | acc], upper_breadths, breadths, callback, rest, lock, env_target, cache)
+        # Something that we previously ruled out as an optional dependency is
+        # no longer a dependency. Add it back for traversal.
+        {no_longer_optional, optional} = Enum.split_with(optional, &(&1.app == app))
+        t = no_longer_optional ++ t
+
+        {acc, kept, optional, rest, lock} =
+          all(t, [dep | acc], [dep.app | kept], upper, breadths, optional, rest, lock, state)
+
+        # Now traverse all parent dependencies and see if we have any optional dependency.
+        {discarded, deps} = split_non_fulfilled_optional(deps, kept, opts[:from_umbrella])
 
-        umbrella? = dep.opts[:from_umbrella]
-        deps = reject_non_fulfilled_optional(dep.deps, Enum.map(acc, & &1.app), umbrella?)
         new_breadths = Enum.map(deps, & &1.app) ++ breadths
-        all(deps, acc, breadths, new_breadths, callback, rest, lock, env_target, cache)
+        all(deps, acc, kept, breadths, new_breadths, discarded ++ optional, rest, lock, state)
     end
   end
 
-  defp all([], acc, _upper, _current, _callback, rest, lock, _env_target, _cache) do
-    {acc, rest, lock}
+  defp all([], acc, kept, _upper, _current, optional, rest, lock, _state) do
+    {acc, kept, optional, rest, lock}
   end
 
   defp put_lock(%Mix.Dep{app: app} = dep, lock) do
@@ -376,12 +407,13 @@ defmodule Mix.Dep.Converger do
     apps = Enum.map(deps, & &1.app)
 
     for dep <- deps do
-      update_in(dep.deps, &reject_non_fulfilled_optional(&1, apps, dep.opts[:from_umbrella]))
+      {_discarded, deps} = split_non_fulfilled_optional(dep.deps, apps, dep.opts[:from_umbrella])
+      %{dep | deps: deps}
     end
   end
 
-  defp reject_non_fulfilled_optional(children, upper_breadths, umbrella?) do
-    Enum.reject(children, fn %Mix.Dep{app: app, opts: opts} ->
+  defp split_non_fulfilled_optional(children, upper_breadths, umbrella?) do
+    Enum.split_with(children, fn %Mix.Dep{app: app, opts: opts} ->
       opts[:optional] && app not in upper_breadths && !umbrella?
     end)
   end
