@@ -16,9 +16,15 @@ defmodule Mix.Project do
         end
       end
 
+  > #### `use Mix.Project` {: .info}
+  >
+  > When you `use Mix.Project`, it notifies Mix that a new project
+  > has been defined, so all Mix tasks use your module as a starting
+  > point.
+
   ## Configuration
 
-  In order to configure Mix, the module that `use`s `Mix.Project` should export
+  In order to configure Mix, the module that calls `use Mix.Project` should export
   a `project/0` function that returns a keyword list representing configuration
   for the project.
 
@@ -51,9 +57,6 @@ defmodule Mix.Project do
       file. See `config_files/0` for more information. Defaults to
       `"config/config.exs"`.
 
-    * `:default_task` - a string representing the default task to be run by
-      `mix` when no task is specified. Defaults to `"run"`.
-
     * `:deps` - a list of dependencies of this project. Refer to the
       documentation for the `Mix.Tasks.Deps` task for more information. Defaults
       to `[]`.
@@ -64,25 +67,56 @@ defmodule Mix.Project do
     * `:lockfile` - the name of the lockfile used by the `mix deps.*` family of
       tasks. Defaults to `"mix.lock"`.
 
-    * `:preferred_cli_env` - a keyword list of `{task, env}` tuples where `task`
-      is the task name as an atom (for example, `:"deps.get"`) and `env` is the
-      preferred environment (for example, `:test`). This option overrides what
-      is specified by the tasks with the `@preferred_cli_env` attribute (see the
-      docs for `Mix.Task`). Defaults to `[]`.
-
-    * `:preferred_cli_target` - a keyword list of `{task, target}` tuples where
-      `task` is the task name as an atom (for example, `:test`) and `target`
-      is the preferred target (for example, `:host`). Defaults to `[]`.
-
-  For more options, keep an eye on the documentation for single Mix tasks; good
-  examples are the `Mix.Tasks.Compile` task and all the specific compiler tasks
+  Mix tasks may require their own configuration inside `def project`. For example,
+  check the `Mix.Tasks.Compile` task and all the specific compiler tasks
   (such as `Mix.Tasks.Compile.Elixir` or `Mix.Tasks.Compile.Erlang`).
 
-  Note that sometimes the same configuration option is mentioned in the
-  documentation for different tasks; this is just because it's common for many
-  tasks to read and use the same configuration option (for example,
-  `:erlc_paths` is used by `mix compile.erlang`, `mix compile.yecc`, and other
-  tasks).
+  Note that different tasks may share the same configuration option. For example,
+  the `:erlc_paths` configuration is used by `mix compile.erlang`, `mix compile.yecc`,
+  and other tasks.
+
+  > #### Keep `project/0` fast {: .warning}
+  >
+  > `project/0` is called whenever your `mix.exs` is loaded, so heavy
+  > computation should be avoided. If a task requires a potentially complex
+  > configuration value, it should allow its configuration to be set to an
+  > anonymous function or similar, so that it can be invoked only when
+  > needed by the task itself.
+
+  ## CLI configuration
+
+  Mix is most often invoked from the command line. For this purpose, you may define
+  a specific `cli/0` function which customizes default values when executed from
+  the CLI. For example:
+
+      def cli do
+        [
+          default_task: "phx.server",
+          preferred_envs: [docs: :docs]
+        ]
+      end
+
+  The example above sets the default task (used by `iex -S mix` and `mix`) to
+  `phx.server`. It also sets the default environment for the "mix docs" task to
+  be "docs".
+
+  The following CLI configuration are available:
+
+    * `:default_env` - the default environment to use when none is given
+      and `MIX_ENV` is not set
+
+    * `:default_target` - the default target to use when none is given
+      and `MIX_TARGET` is not set
+
+    * `:default_task` - the default task to invoke when none is given
+
+    * `:preferred_envs` - a keyword list of `{task, env}` tuples where `task`
+      is the task name as an atom (for example, `:"deps.get"`) and `env` is the
+      preferred environment (for example, `:test`)
+
+    * `:preferred_targets` - a keyword list of `{task, target}` tuples where
+      `task` is the task name as an atom (for example, `:test`) and `target`
+      is the preferred target (for example, `:host`)
 
   ## Erlang projects
 
@@ -137,10 +171,14 @@ defmodule Mix.Project do
   # Only the top of the stack can be accessed.
   @doc false
   def push(module, file \\ nil, app \\ nil) when is_atom(module) do
-    file = file || (module && List.to_string(module.__info__(:compile)[:source]))
-    config = Keyword.merge([app: app] ++ default_config(), get_project_config(module))
+    file =
+      cond do
+        file != nil -> file
+        source = module && module.module_info(:compile)[:source] -> List.to_string(source)
+        true -> "nofile"
+      end
 
-    case Mix.ProjectStack.push(module, config, file) do
+    case Mix.ProjectStack.push(module, push_config(module, app), file) do
       :ok ->
         :ok
 
@@ -152,6 +190,34 @@ defmodule Mix.Project do
     end
   end
 
+  @preferred_envs [test: :test, "test.coverage": :test]
+
+  defp push_config(module, app) do
+    with {state_loader, task} <- Mix.ProjectStack.pop_post_config(:state_loader) do
+      config =
+        if function_exported?(module, state_loader, 0),
+          do: apply(module, state_loader, []),
+          else: []
+
+      task = String.to_atom(task || config[:default_task] || "run")
+
+      if !System.get_env("MIX_ENV") do
+        if env = config[:preferred_envs][task] || @preferred_envs[task] || config[:default_env] do
+          Mix.env(env)
+        end
+      end
+
+      if !System.get_env("MIX_TARGET") do
+        if target = config[:preferred_targets][task] || config[:default_target] do
+          Mix.target(target)
+        end
+      end
+    end
+
+    ([app: app] ++ default_config())
+    |> Keyword.merge(get_project_config(module))
+  end
+
   # Pops a project from the stack.
   @doc false
   def pop do
@@ -162,14 +228,12 @@ defmodule Mix.Project do
   @doc false
   def deps_config(config \\ config()) do
     [
-      build_embedded: config[:build_embedded],
-      build_per_environment: config[:build_per_environment],
       consolidate_protocols: false,
       consolidation_path: consolidation_path(config),
       deps_path: deps_path(config),
-      env_path: build_path(config),
+      deps_build_path: build_path(config),
       lockfile: Path.expand(config[:lockfile])
-    ]
+    ] ++ Keyword.take(config, [:build_embedded, :build_per_environment, :prune_code_paths])
   end
 
   @doc """
@@ -215,6 +279,16 @@ defmodule Mix.Project do
   @spec project_file() :: binary | nil
   defdelegate project_file(), to: Mix.ProjectStack
 
+  @doc """
+  Returns the path to the file that defines the parent umbrella project, if one.
+
+  The majority of the time, it will point to a `mix.exs` file.
+  Returns `nil` if not inside a project or not inside an umbrella.
+  """
+  @doc since: "1.15.0"
+  @spec parent_umbrella_project_file() :: binary | nil
+  defdelegate parent_umbrella_project_file(), to: Mix.ProjectStack
+
   @doc """
   Returns the project configuration.
 
@@ -264,6 +338,11 @@ defmodule Mix.Project do
   a full recompilation whenever such configuration files change.
   For this reason, the mtime is cached to avoid file system lookups.
 
+  However, for effective used of this function, you must avoid
+  comparing source files with the `config_mtime` itself. Instead,
+  store the previous `config_mtime` and compare it with the new
+  `config_mtime` in order to detect if something is stale.
+
   Note: before Elixir v1.13.0, the `mix.exs` file was also included
   in the mtimes, but not anymore. You can compute its modification
   date by calling `project_file/0`.
@@ -483,6 +562,28 @@ defmodule Mix.Project do
     traverse_deps(opts, fn %{opts: opts} -> opts[:dest] end)
   end
 
+  @doc """
+  Returns the dependencies of all dependencies as a map.
+
+  ## Options
+
+    * `:depth` - only returns dependencies to the depth level,
+      a depth of `1` will only return top-level dependencies
+    * `:parents` - starts the dependency traversal from the
+      given parents instead of the application root
+
+  ## Examples
+
+      Mix.Project.deps_tree()
+      #=> %{foo: [:bar, :baz], bar: [], baz: []}
+
+  """
+  @doc since: "1.15.0"
+  @spec deps_tree(keyword) :: %{optional(atom) => [atom]}
+  def deps_tree(opts \\ []) when is_list(opts) do
+    traverse_deps(opts, fn %{deps: deps} -> Enum.map(deps, & &1.app) end)
+  end
+
   defp traverse_deps(opts, fun) do
     all_deps = Mix.Dep.cached()
     parents = opts[:parents]
@@ -494,7 +595,7 @@ defmodule Mix.Project do
       all_deps
       |> Enum.filter(parent_filter)
       |> traverse_deps_map(fun)
-      |> traverse_deps_depth(all_deps, 1, depth || :infinity)
+      |> traverse_deps_depth(all_deps, fun, 1, depth || :infinity)
     else
       traverse_deps_map(all_deps, fun)
     end
@@ -504,21 +605,21 @@ defmodule Mix.Project do
     for %{app: app} = dep <- deps, do: {app, fun.(dep)}, into: %{}
   end
 
-  defp traverse_deps_depth(deps, _all_deps, depth, depth) do
+  defp traverse_deps_depth(deps, _all_deps, _fun, depth, depth) do
     deps
   end
 
-  defp traverse_deps_depth(parents, all_deps, depth, target_depth) do
+  defp traverse_deps_depth(parents, all_deps, fun, depth, target_depth) do
     children =
       for parent_dep <- all_deps,
           Map.has_key?(parents, parent_dep.app),
-          %{app: app, opts: opts} <- parent_dep.deps,
-          do: {app, opts[:dest]},
+          %{app: app} = dep <- parent_dep.deps,
+          do: {app, fun.(dep)},
           into: %{}
 
     case Map.merge(parents, children) do
       ^parents -> parents
-      new_parents -> traverse_deps_depth(new_parents, all_deps, depth + 1, target_depth)
+      new_parents -> traverse_deps_depth(new_parents, all_deps, fun, depth + 1, target_depth)
     end
   end
 
@@ -547,9 +648,30 @@ defmodule Mix.Project do
   @doc """
   Returns the build path for the given project.
 
-  If no configuration is given, the one for the current project is used.
+  The build path is built based on the `:build_path` configuration
+  (which defaults to `"_build"`) and a subdirectory. The subdirectory
+  is built based on two factors:
 
-  The returned path will be expanded.
+    * If `:build_per_environment` is set, the subdirectory is the value
+      of `Mix.env/0` (which can be set via `MIX_ENV`). Otherwise it is
+      set to "shared".
+
+    * If `Mix.target/0` is set (often via the `MIX_TARGET` environment
+      variable), it will be used as a prefix to the subdirectory.
+
+  Finally, the environment variables `MIX_BUILD_ROOT` and `MIX_BUILD_PATH`
+  can be used to change the result of this function. `MIX_BUILD_ROOT`
+  overwrites only the root `"_build"` directory while keeping the
+  subdirectory as is. It may be useful to change it for caching reasons,
+  typically during Continuous Integration (CI). `MIX_BUILD_PATH` overrides
+  the build path altogether and it typically used by other build tools
+  that invoke the `mix` CLI.
+
+  > #### Naming differences {: .info}
+  >
+  > Ideally the configuration option `:build_path` would be called
+  > `:build_root`, as it would fully mirror the environment variable.
+  > However, its name is preserved for backwards compatibility.
 
   ## Examples
 
@@ -567,10 +689,10 @@ defmodule Mix.Project do
   """
   @spec build_path(keyword) :: Path.t()
   def build_path(config \\ config()) do
-    System.get_env("MIX_BUILD_PATH") || config[:env_path] || env_path(config)
+    System.get_env("MIX_BUILD_PATH") || config[:deps_build_path] || do_build_path(config)
   end
 
-  defp env_path(config) do
+  defp do_build_path(config) do
     dir = System.get_env("MIX_BUILD_ROOT") || config[:build_path] || "_build"
     subdir = build_target() <> build_per_environment(config)
     Path.expand(dir <> "/" <> subdir)
@@ -617,7 +739,7 @@ defmodule Mix.Project do
   @spec manifest_path(keyword) :: Path.t()
   def manifest_path(config \\ config()) do
     app_path =
-      config[:app_path] ||
+      config[:deps_app_path] ||
         if app = config[:app] do
           Path.join([build_path(config), "lib", Atom.to_string(app)])
         else
@@ -642,7 +764,7 @@ defmodule Mix.Project do
   """
   @spec app_path(keyword) :: Path.t()
   def app_path(config \\ config()) do
-    config[:app_path] ||
+    config[:deps_app_path] ||
       cond do
         app = config[:app] ->
           Path.join([build_path(config), "lib", Atom.to_string(app)])
@@ -726,14 +848,12 @@ defmodule Mix.Project do
     source = opts[:source] || File.cwd!()
     target = app_path(config)
     File.mkdir_p!(target)
-
     target_ebin = Path.join(target, "ebin")
-    hard_copy? = config[:build_embedded]
 
     _ =
       cond do
         opts[:symlink_ebin] ->
-          _ = Mix.Utils.symlink_or_copy(hard_copy?, Path.join(source, "ebin"), target_ebin)
+          _ = Mix.Utils.symlink_or_copy(Path.join(source, "ebin"), target_ebin)
 
         match?({:ok, _}, :file.read_link(target_ebin)) ->
           _ = File.rm_rf!(target_ebin)
@@ -744,7 +864,7 @@ defmodule Mix.Project do
       end
 
     for dir <- ~w(include priv) do
-      Mix.Utils.symlink_or_copy(hard_copy?, Path.join(source, dir), Path.join(target, dir))
+      Mix.Utils.symlink_or_copy(Path.join(source, dir), Path.join(target, dir))
     end
 
     :ok
@@ -775,6 +895,48 @@ defmodule Mix.Project do
     end
   end
 
+  @doc """
+  Acquires a lock on the project build path and runs the given function.
+
+  When another process (across all OS processes) is holding the lock,
+  a message is printed and this call blocks until the lock is acquired.
+  This function can also be called if this process already has the
+  lock. In such case the function is executed immediately.
+
+  This lock is primarily useful for compiler tasks that alter the build
+  artifacts to avoid conflicts with a concurrent compilation.
+  """
+  @spec with_build_lock(keyword, (-> term())) :: term()
+  def with_build_lock(config \\ config(), fun) do
+    # To avoid duplicated compilation, we wrap compilation tasks, such
+    # as compile.all, deps.compile, compile.elixir, compile.erlang in
+    # a lock. Note that compile.all covers compile.elixir, but the
+    # latter can still be invoked directly, so we put the lock over
+    # each individual task.
+
+    build_path = build_path(config)
+
+    on_taken = fn os_pid ->
+      Mix.shell().info("Waiting for lock on the build directory (held by process #{os_pid})")
+    end
+
+    Mix.Sync.Lock.with_lock(build_path, fun, on_taken: on_taken)
+  end
+
+  @doc false
+  def with_deps_lock(config \\ config(), fun) do
+    # We wrap operations on the deps directory and on mix.lock to
+    # avoid write conflicts.
+
+    deps_path = deps_path(config)
+
+    on_taken = fn os_pid ->
+      Mix.shell().info("Waiting for lock on the deps directory (held by process #{os_pid})")
+    end
+
+    Mix.Sync.Lock.with_lock(deps_path, fun, on_taken: on_taken)
+  end
+
   # Loads mix.exs in the current directory or loads the project from the
   # mixfile cache and pushes the project onto the project stack.
   defp load_project(app, post_config) do
@@ -820,7 +982,6 @@ defmodule Mix.Project do
       build_scm: Mix.SCM.Path,
       config_path: "config/config.exs",
       consolidate_protocols: true,
-      default_task: "run",
       deps: [],
       deps_path: "deps",
       elixirc_paths: ["lib"],
@@ -828,12 +989,11 @@ defmodule Mix.Project do
       erlc_include_path: "include",
       erlc_options: [],
       lockfile: "mix.lock",
-      preferred_cli_env: [],
       start_permanent: false
     ]
   end
 
-  @private_config [:app_path, :build_scm, :env_path]
+  @private_config [:build_scm, :deps_app_path, :deps_build_path]
   defp get_project_config(nil), do: []
-  defp get_project_config(atom), do: atom.project |> Keyword.drop(@private_config)
+  defp get_project_config(atom), do: atom.project() |> Keyword.drop(@private_config)
 end
