@@ -9,34 +9,31 @@ defmodule Mix.Tasks.Cmd do
 
   For example, you can invoke an external command such as make:
 
-      mix cmd make
+      $ mix cmd make
 
   By passing the `--cd` flag before the command, you can also force
   the command to run in a specific directory:
 
-      mix cmd --cd "third-party" make
+      $ mix cmd --cd "third-party" make
 
   This task is also useful in umbrella applications to execute a command
   on each child app:
 
-      mix cmd pwd
+      $ mix cmd pwd
 
   You can limit which apps the cmd runs in by using `mix do` with the `--app`
   option:
 
-      mix do --app app1 --app app2 cmd pwd
+      $ mix do --app app1 --app app2 cmd pwd
 
-  The tasks aborts whenenever a command exits with a non-zero status.
+  The tasks aborts whenever a command exits with a non-zero status.
 
   This task is automatically re-enabled, so it can be called multiple times
   with different arguments.
 
   ## Command line options
 
-    * `--app` - limit running the command to the given app.
-      This option is currently deprecated in favor of `mix do --app`
-
-    * `--cd` - (since v1.10.4) the directory to run the command in
+    * `--cd` *(since v1.10.4)* - the directory to run the command in
 
   ## Zombie operating system processes
 
@@ -60,12 +57,15 @@ defmodule Mix.Tasks.Cmd do
   def run(args) do
     {opts, args} = OptionParser.parse_head!(args, strict: @switches)
 
-    # TODO: Deprecate `--app` flag in Elixir v1.18
     apps =
       opts
       |> Keyword.get_values(:app)
       |> Enum.map(&String.to_atom/1)
 
+    if apps != [] do
+      IO.warn("the --app in mix cmd is deprecated")
+    end
+
     if apps == [] or Mix.Project.config()[:app] in apps do
       cmd_opts = Keyword.take(opts, [:cd])
 
