@@ -9,19 +9,29 @@ defmodule Mix.Tasks.Compile.Elixir do
 
   Elixir is smart enough to recompile only files that have changed
   and their dependencies. This means if `lib/a.ex` is invoking
-  a function defined over `lib/b.ex`, whenever `lib/b.ex` changes,
-  `lib/a.ex` is also recompiled.
+  a function defined over `lib/b.ex` at compile time, whenever
+  `lib/b.ex` changes, `lib/a.ex` is also recompiled. More details
+  about dependencies between files can be found in the documentation
+  of [`mix xref`](`Mix.Tasks.Xref`).
 
-  Note it is important to recompile a file's dependencies as
-  there are often compile time dependencies between them.
+  Note Elixir considers a file as changed if its source file has
+  changed on disk since the last compilation AND its contents are
+  no longer the same.
+
+  ## `@external_resource`
+
+  If a module depends on external files, those can be annotated
+  with the `@external_resource` module attribute. If these files
+  change, the Elixir module is automatically recompiled.
 
   ## `__mix_recompile__?/0`
 
   A module may export a `__mix_recompile__?/0` function that can
   cause the module to be recompiled using custom rules. For example,
-  `@external_resource` already adds a compile-time dependency on an
-  external file, however to depend on a _dynamic_ list of files we
-  can do:
+  to recompile whenever a file is changed in a given directory, you
+  can use a combination of `@external_resource` for existing files
+  and a `__mix_recompile__?/0` check to verify when new entries are
+  added to the directory itself:
 
       defmodule MyModule do
         paths = Path.wildcard("*.txt")
@@ -46,18 +56,20 @@ defmodule Mix.Tasks.Compile.Elixir do
 
   ## Command line options
 
-    * `--verbose` - prints each file being compiled
-    * `--force` - forces compilation regardless of modification times
+    * `--all-warnings` (`--no-all-warnings`) - prints all warnings, including previous compilations
+      (default is true except on errors)
     * `--docs` (`--no-docs`) - attaches (or not) documentation to compiled modules
     * `--debug-info` (`--no-debug-info`) - attaches (or not) debug info to compiled modules
+    * `--force` - forces compilation regardless of modification times
     * `--ignore-module-conflict` - does not emit warnings if a module was previously defined
-    * `--warnings-as-errors` - treats warnings in the current project as errors and
-      return a non-zero exit status
     * `--long-compilation-threshold N` - sets the "long compilation" threshold
       (in seconds) to `N` (see the docs for `Kernel.ParallelCompiler.compile/2`)
+    * `--purge-consolidation-path-if-stale PATH` - deletes and purges modules in the
+      given protocol consolidation path if compilation is required
     * `--profile` - if set to `time`, outputs timing information of compilation steps
-    * `--all-warnings` - prints warnings even from files that do not need to be recompiled
     * `--tracer` - adds a compiler tracer in addition to any specified in the `mix.exs` file
+    * `--verbose` - prints each file being compiled
+    * `--warnings-as-errors` - exit with non-zero status if compilation has one or more warnings
 
   ## Configuration
 
@@ -65,9 +77,9 @@ defmodule Mix.Tasks.Compile.Elixir do
       Defaults to `["lib"]`.
 
     * `:elixirc_options` - compilation options that apply to Elixir's compiler.
-      See `Code.put_compiler_option/2` for a complete list of options. These
-      options are often overridable from the command line using the switches
-      above.
+      It supports many of the options above  plus the options listed in
+      `Code.put_compiler_option/2`. In case conflicting options are given,
+      the ones given through the command line are used.
 
     * `[xref: [exclude: ...]]` - a list of `module` or `{module, function, arity}`
       that should not be warned on in case on undefined modules or undefined
@@ -83,6 +95,7 @@ defmodule Mix.Tasks.Compile.Elixir do
     debug_info: :boolean,
     verbose: :boolean,
     long_compilation_threshold: :integer,
+    purge_consolidation_path_if_stale: :string,
     profile: :string,
     all_warnings: :boolean,
     tracer: :keep
@@ -97,13 +110,13 @@ defmodule Mix.Tasks.Compile.Elixir do
     dest = Mix.Project.compile_path(project)
     srcs = project[:elixirc_paths]
 
-    unless is_list(srcs) do
+    if not is_list(srcs) do
       Mix.raise(":elixirc_paths should be a list of paths, got: #{inspect(srcs)}")
     end
 
     manifest = manifest()
     base = xref_exclude_opts(project[:elixirc_options] || [], project)
-    cache_key = {base, srcs, "--no-optional-deps" in args}
+    cache_key = {base, srcs, File.cwd!(), "--no-optional-deps" in args}
 
     opts =
       base
@@ -111,19 +124,21 @@ defmodule Mix.Tasks.Compile.Elixir do
       |> tracers_opts(tracers)
       |> profile_opts()
 
-    # The Elixir compiler relies on global state in the application tracer.
-    # However, even without it, having compilations racing with other is most
-    # likely undesired, so we wrap the compiler in a lock.
-    Mix.State.lock(__MODULE__, fn ->
-      Mix.Compilers.Elixir.compile(
-        manifest,
-        srcs,
-        dest,
-        cache_key,
-        Mix.Tasks.Compile.Erlang.manifests(),
-        Mix.Tasks.Compile.Erlang.modules(),
-        opts
-      )
+    # Having compilations racing with other is most undesired,
+    # so we wrap the compiler in a lock.
+
+    with_logger_app(project, fn ->
+      Mix.Project.with_build_lock(project, fn ->
+        Mix.Compilers.Elixir.compile(
+          manifest,
+          srcs,
+          dest,
+          cache_key,
+          Mix.Tasks.Compile.Erlang.manifests(),
+          Mix.Tasks.Compile.Erlang.modules(),
+          opts
+        )
+      end)
     end)
   end
 
@@ -131,12 +146,31 @@ defmodule Mix.Tasks.Compile.Elixir do
   def manifests, do: [manifest()]
   defp manifest, do: Path.join(Mix.Project.manifest_path(), @manifest)
 
+  @impl true
+  def diagnostics do
+    dest = Mix.Project.compile_path()
+    Mix.Compilers.Elixir.diagnostics(manifest(), dest)
+  end
+
   @impl true
   def clean do
     dest = Mix.Project.compile_path()
     Mix.Compilers.Elixir.clean(manifest(), dest)
   end
 
+  # Run this operation in compile.elixir as the compiler can be called directly
+  defp with_logger_app(config, fun) do
+    app = Keyword.fetch!(config, :app)
+    logger_config_app = Application.get_env(:logger, :compile_time_application)
+
+    try do
+      Application.put_env(:logger, :compile_time_application, app)
+      fun.()
+    after
+      Application.put_env(:logger, :compile_time_application, logger_config_app)
+    end
+  end
+
   defp xref_exclude_opts(opts, project) do
     exclude = List.wrap(project[:xref][:exclude])
 
