@@ -2,7 +2,7 @@ defmodule Mix.Tasks.Compile.Protocols do
   use Mix.Task.Compiler
 
   @manifest "compile.protocols"
-  @manifest_vsn 1
+  @manifest_vsn 3
 
   @moduledoc ~S"""
   Consolidates all protocols in all paths.
@@ -49,23 +49,30 @@ defmodule Mix.Tasks.Compile.Protocols do
 
     manifest = manifest()
     output = Mix.Project.consolidation_path(config)
+    config_mtime = Mix.Project.config_mtime()
     protocols_and_impls = protocols_and_impls(config)
+    metadata = {config_mtime, protocols_and_impls}
+    {old_config_mtime, old_protocols_and_impls} = read_manifest(manifest, output)
 
     cond do
       # We need to reconsolidate all protocols whenever the dependency changes
       # because we only track protocols from the current app and from local deps.
-      opts[:force] || Mix.Utils.stale?([Mix.Project.config_mtime()], [manifest]) ->
+      #
+      # We are only interested in the compile.lock from config_mtime (which is
+      # a build artifact), so it would be fine to compare it directly against
+      # the manifest, but let's follow best practices anyway.
+      opts[:force] || config_mtime > old_config_mtime ->
         clean()
         paths = consolidation_paths()
 
         paths
         |> Protocol.extract_protocols()
-        |> consolidate(paths, output, manifest, protocols_and_impls, opts)
+        |> consolidate(paths, output, manifest, metadata, opts)
 
       protocols_and_impls ->
-        manifest
-        |> diff_manifest(protocols_and_impls, output)
-        |> consolidate(consolidation_paths(), output, manifest, protocols_and_impls, opts)
+        protocols_and_impls
+        |> diff_manifest(old_protocols_and_impls, output)
+        |> consolidate(consolidation_paths(), output, manifest, metadata, opts)
 
       true ->
         :noop
@@ -91,7 +98,7 @@ defmodule Mix.Tasks.Compile.Protocols do
   end
 
   defp protocols_and_impls(config) do
-    deps = for %{scm: scm, opts: opts} <- Mix.Dep.cached(), not scm.fetchable?, do: opts[:build]
+    deps = for %{scm: scm, opts: opts} <- Mix.Dep.cached(), not scm.fetchable?(), do: opts[:build]
 
     paths =
       if Mix.Project.umbrella?(config) do
@@ -100,11 +107,7 @@ defmodule Mix.Tasks.Compile.Protocols do
         [Mix.Project.app_path(config) | deps]
       end
 
-    Enum.flat_map(paths, fn path ->
-      manifest_path = Path.join(path, ".mix/compile.elixir")
-      compile_path = Path.join(path, "ebin")
-      Mix.Compilers.Elixir.protocols_and_impls(manifest_path, compile_path)
-    end)
+    Mix.Compilers.Elixir.protocols_and_impls(paths)
   end
 
   defp consolidation_paths do
@@ -179,7 +182,7 @@ defmodule Mix.Tasks.Compile.Protocols do
       _ ->
         # If there is no manifest or it is out of date, remove old files
         File.rm_rf(output)
-        []
+        {0, {%{}, %{}}}
     end
   end
 
@@ -189,37 +192,44 @@ defmodule Mix.Tasks.Compile.Protocols do
     File.write!(manifest, manifest_data)
   end
 
-  defp diff_manifest(manifest, new_metadata, output) do
-    modified = Mix.Utils.last_modified(manifest)
-    old_metadata = read_manifest(manifest, output)
-
+  defp diff_manifest({new_protocols, new_impls}, {old_protocols, old_impls}, output) do
     protocols =
-      for {protocol, :protocol, beam} <- new_metadata,
-          Mix.Utils.last_modified(beam) > modified,
-          remove_consolidated(protocol, output),
-          do: protocol
+      new_protocols
+      |> Enum.filter(fn {protocol, new_timestamp} ->
+        case old_protocols do
+          # There is a new version, removed the consolidated
+          %{^protocol => old_timestamp} when new_timestamp > old_timestamp ->
+            remove_consolidated(protocol, output)
+            true
 
-    protocols =
-      Enum.reduce(new_metadata -- old_metadata, Map.from_keys(protocols, true), fn
-        {_, {:impl, protocol}, _beam}, protocols ->
-          Map.put(protocols, protocol, true)
+          # Nothing changed
+          %{^protocol => _} ->
+            false
 
-        {protocol, :protocol, _beam}, protocols ->
-          Map.put(protocols, protocol, true)
+          # New protocol
+          %{} ->
+            true
+        end
       end)
+      |> Map.new()
 
-    removed_metadata = old_metadata -- new_metadata
+    protocols =
+      for {impl, protocol} <- new_impls,
+          not is_map_key(old_impls, impl),
+          do: {protocol, true},
+          into: protocols
 
     removed_protocols =
-      for {protocol, :protocol, _beam} <- removed_metadata,
-          remove_consolidated(protocol, output),
-          do: protocol
+      for {protocol, _timestamp} <- old_protocols,
+          not is_map_key(new_protocols, protocol),
+          do: remove_consolidated(protocol, output)
 
     removed_protocols = Map.from_keys(removed_protocols, true)
 
     protocols =
-      for {_, {:impl, protocol}, _beam} <- removed_metadata,
-          not Map.has_key?(removed_protocols, protocol),
+      for {impl, protocol} <- old_impls,
+          not is_map_key(new_impls, impl),
+          not is_map_key(removed_protocols, protocol),
           do: {protocol, true},
           into: protocols
 
@@ -228,5 +238,6 @@ defmodule Mix.Tasks.Compile.Protocols do
 
   defp remove_consolidated(protocol, output) do
     File.rm(Path.join(output, "#{Atom.to_string(protocol)}.beam"))
+    protocol
   end
 end
