@@ -27,6 +27,12 @@ defmodule Mix.Tasks.Deps.Unlock do
     Mix.Project.get!()
     {opts, apps, _} = OptionParser.parse(args, switches: @switches)
 
+    Mix.Project.with_deps_lock(fn ->
+      do_run(opts, apps)
+    end)
+  end
+
+  defp do_run(opts, apps) do
     cond do
       opts[:all] ->
         Mix.Dep.Lock.write(%{})
@@ -35,7 +41,7 @@ defmodule Mix.Tasks.Deps.Unlock do
         lock = Mix.Dep.Lock.read()
         unused_apps = unused_apps(lock)
 
-        unless unused_apps == [] do
+        if unused_apps != [] do
           Mix.raise("""
           Unused dependencies in mix.lock file:
 
@@ -47,7 +53,7 @@ defmodule Mix.Tasks.Deps.Unlock do
         lock = Mix.Dep.Lock.read()
         unused_apps = unused_apps(lock)
 
-        unless unused_apps == [] do
+        if unused_apps != [] do
           unlock(lock, unused_apps)
         end
 
@@ -83,7 +89,7 @@ defmodule Mix.Tasks.Deps.Unlock do
   end
 
   defp unused_apps(lock) do
-    apps = Mix.Dep.load_on_environment([]) |> Enum.map(& &1.app)
+    apps = Mix.Dep.Converger.converge([]) |> Enum.map(& &1.app)
 
     lock
     |> Map.drop(apps)
@@ -92,7 +98,7 @@ defmodule Mix.Tasks.Deps.Unlock do
   end
 
   defp unlock(lock, apps) do
-    unless apps == [] do
+    if apps != [] do
       lock |> Map.drop(apps) |> Mix.Dep.Lock.write()
 
       Mix.shell().info("""
