@@ -3,10 +3,9 @@ defmodule Mix.Tasks.Test do
 
   alias Mix.Compilers.Test, as: CT
 
-  @compile {:no_warn_undefined, [ExUnit, ExUnit.Filters]}
+  @compile {:no_warn_undefined, [IEx, ExUnit, ExUnit.Filters]}
   @shortdoc "Runs a project's tests"
   @recursive true
-  @preferred_cli_env :test
 
   @moduledoc ~S"""
   Runs the tests for a project.
@@ -18,18 +17,18 @@ defmodule Mix.Tasks.Test do
   A list of files and/or directories can be given after the task
   name in order to select the files to run:
 
-      mix test test/some/particular/file_test.exs
-      mix test test/some/particular/dir
+      $ mix test test/some/particular/file_test.exs
+      $ mix test test/some/particular/dir
 
   Tests in umbrella projects can be run from the root by specifying
   the full suite path, including `apps/my_app/test`, in which case
   recursive tests for other child apps will be skipped completely:
 
       # To run all tests for my_app from the umbrella root
-      mix test apps/my_app/test
+      $ mix test apps/my_app/test
 
       # To run a given test file on my_app from the umbrella root
-      mix test apps/my_app/test/some/particular/file_test.exs
+      $ mix test apps/my_app/test/some/particular/file_test.exs
 
   ## Understanding test results
 
@@ -37,6 +36,7 @@ defmodule Mix.Tasks.Test do
   a summary at the end, as seen below:
 
       $ mix test
+      Running ExUnit with seed: 646219, max_cases: 16
       ...
 
         1) test greets the world (FooTest)
@@ -53,8 +53,6 @@ defmodule Mix.Tasks.Test do
       Finished in 0.05 seconds (0.00s async, 0.05s sync)
       1 doctest, 11 tests, 1 failure
 
-      Randomized with seed 646219
-
   For each test, the test suite will print a dot. Failed tests
   are printed immediately in the format described in the next
   section.
@@ -86,9 +84,9 @@ defmodule Mix.Tasks.Test do
       test/foo_test.exs:5
 
   If you want to re-run only this test, all you need to do is to
-  copy the line above and past it in front of `mix test`:
+  copy the line above and paste it in front of `mix test`:
 
-      mix test test/foo_test.exs:5
+      $ mix test test/foo_test.exs:5
 
   Then we show the error message, code snippet, and general information
   about the failed test:
@@ -107,11 +105,20 @@ defmodule Mix.Tasks.Test do
 
   ## Command line options
 
+    * `--all-warnings` (`--no-all-warnings`) - prints all warnings, including previous compilations
+      (default is true except on errors)
+
+    * `-b`, `--breakpoints` *(since v1.17.0)* - sets a breakpoint at the beginning
+      of every test. The debugger goes line-by-line and can access all variables
+      and imports (but not local functions). You can press `n` for the next line
+      and `c` for the next test. This automatically sets `--trace`
+
     * `--color` - enables color in the output
 
     * `--cover` - runs coverage tool. See "Coverage" section below
 
-    * `--exclude` - excludes tests that match the filter
+    * `--exclude` - excludes tests that match the filter. This option may be given
+      several times to apply different filters, such as `--exclude ci --exclude slow`
 
     * `--exit-status` - use an alternate exit status to use when the test suite
       fails (default is 2).
@@ -126,7 +133,8 @@ defmodule Mix.Tasks.Test do
     * `--formatter` - sets the formatter module that will print the results.
       Defaults to ExUnit's built-in CLI formatter
 
-    * `--include` - includes tests that match the filter
+    * `--include` - includes tests that match the filter. This option may be given
+      several times to apply different filters, such as `--include ci --include slow`
 
     * `--listen-on-stdin` - runs tests, and then listens on stdin. It will
       re-run tests once a newline is received. See the "File system watchers"
@@ -138,6 +146,9 @@ defmodule Mix.Tasks.Test do
     * `--max-failures` - the suite stops evaluating tests when this number of test
       failures is reached. It runs all tests if omitted
 
+    * `--max-requires` - sets the maximum number of test files to compile in parallel.
+      Setting this to 1 will compile test files sequentially.
+
     * `--no-archives-check` - does not check archives
 
     * `--no-color` - disables color in the output
@@ -160,17 +171,29 @@ defmodule Mix.Tasks.Test do
 
     * `--preload-modules` - preloads all modules defined in applications
 
-    * `--profile-require` - profiles the time spent to require test files.
-      Used only for debugging. The test suite does not run.
+    * `--profile-require time` - profiles the time spent to require test files.
+      Used only for debugging. The test suite does not run
 
-    * `--raise` - raises if the test suite failed
+    * `--raise` - immediately raises if the test suite fails, instead of continuing
+      the execution of other Mix tasks
+
+    * `--repeat-until-failure` *(since v1.17.0)* - sets the number of repetitions for running
+      the suite until it fails. This is useful for debugging flaky tests within the same instance
+      of the Erlang VM. For example, `--repeat-until-failure 10000` repeats the test suite
+      up to 10000 times until the first failure. This can be combined with `--max-failures 1`
+      to immediately stop if one test fails. However, if there is any leftover global state
+      after running the tests, re-running the suite may trigger unrelated failures.
 
     * `--seed` - seeds the random number generator used to randomize the order of tests;
       `--seed 0` disables randomization so the tests in a single file will always be ran
       in the same order they were defined in
 
-    * `--slowest` - prints timing information for the N slowest tests.
-      Automatically sets `--trace` and `--preload-modules`
+    * `--slowest` - prints timing information for the N slowest tests. Includes time spent in
+     `ExUnit.Callbacks.setup/1`. Automatically sets `--trace` and `--preload-modules`
+
+    * `--slowest-modules` *(since v1.17.0)* - prints timing information for the N slowest
+      modules. Includes time spent in `ExUnit.Callbacks.setup/1`. Automatically sets
+      `--trace` and `--preload-modules`
 
     * `--stale` - runs only tests which reference modules that changed since the
       last time tests were ran with `--stale`. You can read more about this option
@@ -181,9 +204,13 @@ defmodule Mix.Tasks.Test do
     * `--trace` - runs tests with detailed reporting. Automatically sets `--max-cases` to `1`.
       Note that in trace mode test timeouts will be ignored as timeout is set to `:infinity`
 
-    * `--warnings-as-errors` - (since v1.12.0) treats warnings as errors and returns a non-zero
-      exit status. This option only applies to test files. To treat warnings as errors during
-      compilation and during tests, run:
+    * `--warnings-as-errors` *(since v1.12.0)* - treats compilation warnings (from loading the
+      test suite) as errors and returns a non-zero exit status if the test suite would otherwise
+      pass. Note that failures reported by `--warnings-as-errors` cannot be retried with the
+      `--failed` flag.
+
+      This option only applies to test files. To treat warnings as errors during compilation and
+      during tests, run:
           MIX_ENV=test mix do compile --warnings-as-errors + test --warnings-as-errors
 
   ## Configuration
@@ -194,11 +221,11 @@ defmodule Mix.Tasks.Test do
       mechanism. See the "Coverage" section for more information
 
     * `:test_elixirc_options` - the compiler options to used when
-      loading/compiling test files. By default it disables the debug chunk
-      and docs chunk
+      loading/compiling test files. By default it disables the debug chunk,
+      docs chunk, and module type inference
 
     * `:test_paths` - list of paths containing test files. Defaults to
-      `["test"]` if the `test` directory exists; otherwise, it defaults to `[]`.
+      `["test"]` if the `test` directory exists, otherwise, it defaults to `[]`.
       It is expected that all test paths contain a `test_helper.exs` file
 
     * `:test_pattern` - a pattern to load test files. Defaults to `*_test.exs`
@@ -213,7 +240,7 @@ defmodule Mix.Tasks.Test do
   explicitly enabled for the current user in the registry by running
   the following command:
 
-      reg add HKCU\Console /v VirtualTerminalLevel /t REG_DWORD /d 1
+      $ reg add HKCU\Console /v VirtualTerminalLevel /t REG_DWORD /d 1
 
   After running the command above, you must restart your current console.
 
@@ -229,13 +256,13 @@ defmodule Mix.Tasks.Test do
   Then, whenever desired, those tests could be included in the run via the
   `--include` option:
 
-      mix test --include external:true
+      $ mix test --include external:true
 
   The example above will run all tests that have the external option set to
   `true`. It is also possible to include all examples that have a given tag,
   regardless of its value:
 
-      mix test --include external
+      $ mix test --include external
 
   Note that all tests are included by default, so unless they are excluded
   first (either in the test helper or via the `--exclude` option) the
@@ -244,38 +271,56 @@ defmodule Mix.Tasks.Test do
   For this reason, Mix also provides an `--only` option that excludes all
   tests and includes only the given ones:
 
-      mix test --only external
+      $ mix test --only external
 
   Which is similar to:
 
-      mix test --include external --exclude test
+      $ mix test --include external --exclude test
 
   It differs in that the test suite will fail if no tests are executed when the `--only` option is used.
 
   In case a single file is being tested, it is possible to pass one or more specific
   line numbers to run only those given tests:
 
-      mix test test/some/particular/file_test.exs:12
+      $ mix test test/some/particular/file_test.exs:12
 
   Which is equivalent to:
 
-      mix test --exclude test --include line:12 test/some/particular/file_test.exs
+      $ mix test --exclude test --include line:12 test/some/particular/file_test.exs
 
   Or:
 
-      mix test test/some/particular/file_test.exs:12:24
+      $ mix test test/some/particular/file_test.exs:12:24
 
   Which is equivalent to:
 
-      mix test --exclude test --include line:12 --include line:24 test/some/particular/file_test.exs
+      $ mix test --exclude test --include line:12 --include line:24 test/some/particular/file_test.exs
 
   If a given line starts a `describe` block, that line filter runs all tests in it.
   Otherwise, it runs the closest test on or before the given line number.
 
   ## Coverage
 
-  The `:test_coverage` configures the coverage tool and reports generated
-  by the `--cover` flag. It accepts the following options:
+  Elixir provides built-in line-based test coverage via the `--cover` flag.
+  The test coverages shows which lines of code and in which files were executed
+  during the test run.
+
+  ### Limitations
+
+  Coverage in Elixir has the following limitations:
+
+    * Literals, such as atoms, strings, and numbers, are not traced by coverage.
+      For example, if a function simply returns `:ok`, the atom `:ok` itself is
+      never taken into account by coverage;
+
+    * Macros, such as the ones defined by `defmacro/2` and `defguard/2`, and code
+      invoked only by macros are never considered as covered, unless they are also
+      invoked during the tests themselves. That's because macros are invoked at
+      compilation time, before the test coverage instrumentation begins;
+
+  ### Configuration
+
+  The `:test_coverage` configures the coverage tool and accepts the following options:
 
     * `:output` - the output directory for cover results. Defaults to `"cover"`.
 
@@ -335,10 +380,10 @@ defmodule Mix.Tasks.Test do
   For example, to split a test suite into 4 partitions and run them, you would
   use the following commands:
 
-      MIX_TEST_PARTITION=1 mix test --partitions 4
-      MIX_TEST_PARTITION=2 mix test --partitions 4
-      MIX_TEST_PARTITION=3 mix test --partitions 4
-      MIX_TEST_PARTITION=4 mix test --partitions 4
+      $ MIX_TEST_PARTITION=1 mix test --partitions 4
+      $ MIX_TEST_PARTITION=2 mix test --partitions 4
+      $ MIX_TEST_PARTITION=3 mix test --partitions 4
+      $ MIX_TEST_PARTITION=4 mix test --partitions 4
 
   The test files are sorted upfront in a round-robin fashion. Note the partition
   itself is given as an environment variable so it can be accessed in config files
@@ -368,12 +413,12 @@ defmodule Mix.Tasks.Test do
 
   ## File-system watchers
 
-  You can integrate `mix test` with filesystem watchers through the command line
+  You can integrate `mix test` with file system watchers through the command line
   via the `--listen-on-stdin` option. For example, you can use [fswatch](https://github.com/emcrisostomo/fswatch)
   or similar to emit newlines whenever there is a change, which will cause your test
   suite to re-run:
 
-      fswatch lib test | mix test --listen-on-stdin
+      $ fswatch lib test | mix test --listen-on-stdin
 
   This can be combined with the `--stale` option to re-run only the test files that
   have changed as well as the tests that have gone stale due to changes in `lib`.
@@ -389,6 +434,8 @@ defmodule Mix.Tasks.Test do
   """
 
   @switches [
+    all_warnings: :boolean,
+    breakpoints: :boolean,
     force: :boolean,
     color: :boolean,
     cover: :boolean,
@@ -396,6 +443,7 @@ defmodule Mix.Tasks.Test do
     trace: :boolean,
     max_cases: :integer,
     max_failures: :integer,
+    max_requires: :integer,
     include: :keep,
     exclude: :keep,
     seed: :integer,
@@ -412,26 +460,31 @@ defmodule Mix.Tasks.Test do
     listen_on_stdin: :boolean,
     formatter: :keep,
     slowest: :integer,
+    slowest_modules: :integer,
     partitions: :integer,
     preload_modules: :boolean,
     warnings_as_errors: :boolean,
     profile_require: :string,
-    exit_status: :integer
+    exit_status: :integer,
+    repeat_until_failure: :integer
   ]
 
   @cover [output: "cover", tool: Mix.Tasks.Test.Coverage]
 
   @impl true
   def run(args) do
-    {opts, files} = OptionParser.parse!(args, strict: @switches)
+    {opts, files} = OptionParser.parse!(args, strict: @switches, aliases: [b: :breakpoints])
+    opts = put_manifest_file(opts)
 
     if not Mix.Task.recursing?() do
       do_run(opts, args, files)
     else
+      parent_umbrella = Path.dirname(Mix.Project.parent_umbrella_project_file())
+
       {files_in_apps_path, files_not_in_apps_path} =
         files
-        |> Enum.map(&Path.expand/1)
-        |> Enum.map(&Path.relative_to_cwd/1)
+        |> Enum.map(&Path.expand(&1, parent_umbrella))
+        |> Enum.map(&Path.relative_to(&1, parent_umbrella))
         |> Enum.split_with(&String.starts_with?(&1, "apps/"))
 
       app = Mix.Project.config()[:app]
@@ -453,21 +506,14 @@ defmodule Mix.Tasks.Test do
   end
 
   defp relative_app_file_exists?(file) do
-    {file, _} = ExUnit.Filters.parse_path(file)
+    {[file], _} = ExUnit.Filters.parse_paths([file])
     File.exists?(Path.join("../..", file))
   end
 
   defp do_run(opts, args, files) do
-    if opts[:listen_on_stdin] do
-      System.at_exit(fn _ ->
-        IO.gets(:stdio, "")
-        Mix.shell().info("Restarting...")
-        :init.restart()
-        Process.sleep(:infinity)
-      end)
-    end
+    _ = Mix.Project.get!()
 
-    unless System.get_env("MIX_ENV") || Mix.env() == :test do
+    if System.get_env("MIX_ENV") == nil && Mix.env() != :test do
       Mix.raise("""
       "mix test" is running in the \"#{Mix.env()}\" environment. If you are \
       running tests from within another command, you can either:
@@ -476,12 +522,23 @@ defmodule Mix.Tasks.Test do
 
             MIX_ENV=test mix test.another
 
-        2. set the :preferred_cli_env for a command inside "def project" in your mix.exs:
+        2. set the :preferred_envs for "def cli" in your mix.exs:
 
-            preferred_cli_env: ["test.another": :test]
+            def cli do
+              [preferred_envs: ["test.another": :test]]
+            end
       """)
     end
 
+    if opts[:listen_on_stdin] do
+      System.at_exit(fn _ ->
+        IO.gets(:stdio, "")
+        Mix.shell().info("Restarting...")
+        System.restart()
+        Process.sleep(:infinity)
+      end)
+    end
+
     # Load ExUnit before we compile anything in case we are compiling
     # helper modules that depend on ExUnit.
     Application.ensure_loaded(:ex_unit)
@@ -511,36 +568,60 @@ defmodule Mix.Tasks.Test do
     # Start the app and configure ExUnit with command line options
     # before requiring test_helper.exs so that the configuration is
     # available in test_helper.exs
-    Mix.shell().print_app
+    Mix.shell().print_app()
     app_start_args = if opts[:slowest], do: ["--preload-modules" | args], else: args
     Mix.Task.run("app.start", app_start_args)
 
     # The test helper may change the Mix.shell(), so revert it whenever we raise and after suite
     shell = Mix.shell()
+    test_elixirc_options = project[:test_elixirc_options] || []
+
+    {test_elixirc_options, opts} =
+      cond do
+        not Keyword.get(opts, :breakpoints, false) ->
+          {test_elixirc_options, opts}
+
+        IEx.started?() ->
+          {Keyword.put(test_elixirc_options, :debug_info, true), opts}
+
+        true ->
+          Mix.shell().error("you must run \"iex -S mix test\" when using -b/--breakpoints")
+          {test_elixirc_options, Keyword.delete(opts, :breakpoints)}
+      end
 
     # Configure ExUnit now and then again so the task options override test_helper.exs
     {ex_unit_opts, allowed_files} = process_ex_unit_opts(opts)
     ExUnit.configure(ex_unit_opts)
 
+    # Prepare and extract all files to require and run
     test_paths = project[:test_paths] || default_test_paths()
-    Enum.each(test_paths, &require_test_helper(shell, &1))
-    ExUnit.configure(merge_helper_opts(ex_unit_opts))
-
-    # Finally parse, require and load the files
-    test_elixirc_options = project[:test_elixirc_options] || []
-    test_files = parse_files(files, shell, test_paths)
     test_pattern = project[:test_pattern] || "*_test.exs"
     warn_test_pattern = project[:warn_test_pattern] || "*_test.ex"
 
+    {test_files, test_opts} =
+      if files != [], do: ExUnit.Filters.parse_paths(files), else: {test_paths, []}
+
+    unfiltered_test_files = Mix.Utils.extract_files(test_files, test_pattern)
+
     matched_test_files =
-      test_files
-      |> Mix.Utils.extract_files(test_pattern)
+      unfiltered_test_files
       |> filter_to_allowed_files(allowed_files)
       |> filter_by_partition(shell, partitions)
 
-    display_warn_test_pattern(test_files, test_pattern, matched_test_files, warn_test_pattern)
+    display_warn_test_pattern(test_files, test_pattern, unfiltered_test_files, warn_test_pattern)
 
-    case CT.require_and_run(matched_test_files, test_paths, test_elixirc_options, opts) do
+    try do
+      Enum.each(test_paths, &require_test_helper(shell, &1))
+      # test_opts always wins because those are given via args
+      ExUnit.configure(ex_unit_opts |> merge_helper_opts() |> Keyword.merge(test_opts))
+      CT.require_and_run(matched_test_files, test_paths, test_elixirc_options, opts)
+    catch
+      kind, reason ->
+        # Also mark the whole suite as failed
+        file = Keyword.fetch!(opts, :failures_manifest_path)
+        ExUnit.Filters.fail_all!(file)
+        :erlang.raise(kind, reason, __STACKTRACE__)
+    else
       {:ok, %{excluded: excluded, failures: failures, total: total}} ->
         Mix.shell(shell)
         cover && cover.()
@@ -567,7 +648,7 @@ defmodule Mix.Tasks.Test do
           opts[:stale] ->
             Mix.shell().info("No stale tests")
 
-          files == [] ->
+          opts[:failed] || files == [] ->
             Mix.shell().info("There are no tests to run")
 
           true ->
@@ -609,6 +690,7 @@ defmodule Mix.Tasks.Test do
   end
 
   @option_keys [
+    :breakpoints,
     :trace,
     :max_cases,
     :max_failures,
@@ -619,10 +701,12 @@ defmodule Mix.Tasks.Test do
     :formatters,
     :colors,
     :slowest,
-    :failures_manifest_file,
+    :slowest_modules,
+    :failures_manifest_path,
     :only_test_ids,
     :test_location_relative_path,
-    :exit_status
+    :exit_status,
+    :repeat_until_failure
   ]
 
   @doc false
@@ -645,40 +729,15 @@ defmodule Mix.Tasks.Test do
 
   defp merge_helper_opts(opts) do
     # The only options that are additive from app env are the excludes
-    merge_opts(opts, :exclude)
-  end
-
-  defp merge_opts(opts, key) do
-    value = List.wrap(Application.get_env(:ex_unit, key, []))
-    Keyword.update(opts, key, value, &Enum.uniq(&1 ++ value))
+    value = List.wrap(Application.get_env(:ex_unit, :exclude, []))
+    Keyword.update(opts, :exclude, value, &Enum.uniq(&1 ++ value))
   end
 
   defp default_opts(opts) do
-    # Set autorun to false because Mix
-    # automatically runs the test suite for us.
+    # Set autorun to false because Mix automatically runs the test suite for us.
     [autorun: false] ++ opts
   end
 
-  defp parse_files([], _shell, test_paths) do
-    test_paths
-  end
-
-  defp parse_files([single_file], _shell, _test_paths) do
-    # Check if the single file path matches test/path/to_test.exs:123. If it does,
-    # apply "--only line:123" and trim the trailing :123 part.
-    {single_file, opts} = ExUnit.Filters.parse_path(single_file)
-    ExUnit.configure(opts)
-    [single_file]
-  end
-
-  defp parse_files(files, shell, _test_paths) do
-    if Enum.any?(files, &match?({_, [_ | _]}, ExUnit.Filters.parse_path(&1))) do
-      raise_with_shell(shell, "Line numbers can only be used when running a single test file")
-    else
-      files
-    end
-  end
-
   defp parse_filters(opts, key) do
     if Keyword.has_key?(opts, key) do
       ExUnit.Filters.parse(Keyword.get_values(opts, key))
@@ -718,17 +777,29 @@ defmodule Mix.Tasks.Test do
 
   @manifest_file_name ".mix_test_failures"
 
+  defp put_manifest_file(opts) do
+    Keyword.put_new_lazy(
+      opts,
+      :failures_manifest_path,
+      fn -> Path.join(Mix.Project.manifest_path(), @manifest_file_name) end
+    )
+  end
+
   defp manifest_opts(opts) do
-    manifest_file = Path.join(Mix.Project.manifest_path(), @manifest_file_name)
-    opts = Keyword.put(opts, :failures_manifest_file, manifest_file)
+    manifest_file = Keyword.fetch!(opts, :failures_manifest_path)
 
     if opts[:failed] do
       if opts[:stale] do
         Mix.raise("Combining --failed and --stale is not supported.")
       end
 
-      {allowed_files, failed_ids} = ExUnit.Filters.failure_info(manifest_file)
-      {Keyword.put(opts, :only_test_ids, failed_ids), allowed_files}
+      case ExUnit.Filters.failure_info(manifest_file) do
+        {allowed_files, failed_ids} ->
+          {Keyword.put(opts, :only_test_ids, failed_ids), allowed_files}
+
+        :all ->
+          {opts, nil}
+      end
     else
       {opts, nil}
     end
@@ -747,7 +818,7 @@ defmodule Mix.Tasks.Test do
     partition = System.get_env("MIX_TEST_PARTITION")
 
     case partition && Integer.parse(partition) do
-      {partition, ""} when partition in 1..total ->
+      {partition, ""} when partition in 1..total//1 ->
         partition = partition - 1
 
         # We sort the files because Path.wildcard does not guarantee
