@@ -2,10 +2,9 @@ defmodule Mix.Tasks.Xref do
   use Mix.Task
 
   import Mix.Compilers.Elixir,
-    only: [read_manifest: 1, source: 0, source: 1, source: 2, module: 1]
+    only: [read_manifest: 1, source: 1, source: 2, module: 1]
 
   @shortdoc "Prints cross reference information"
-  @recursive true
   @manifest "compile.elixir"
 
   @moduledoc """
@@ -15,16 +14,111 @@ defmodule Mix.Tasks.Xref do
 
       $ mix xref MODE
 
-  All available modes are discussed below.
+  All available modes are discussed below, after a brief
+  introduction to xref.
 
   This task is automatically re-enabled, so you can print
   information multiple times in the same Mix invocation.
 
-  ## mix xref callers MODULE
+  ## A brief introduction to xref
 
-  Prints all callers of the given `MODULE`. Example:
+  The goal of `xref` is to analyze the dependencies between modules
+  and files. It is most commonly used to find problematic areas where
+  touching one file in a project causes a large subset of the project
+  to recompile. The most common cause of these problems are the so-called
+  "compile-connected" files. Those are files you depend on at compile-time
+  (for example, by invoking its macro or using it in the body of a module)
+  which also have their own dependencies.
 
-      $ mix xref callers MyMod
+  Therefore, if your goal is to reduce recompilations, the first step is to run:
+
+      $ mix xref graph --format stats --label compile-connected
+
+  This command will show general information about the project, but
+  focus on compile-connected dependencies. In the stats, you will see
+  the following report:
+
+      Top 10 files with most incoming dependencies:
+        * lib/livebook_web.ex (97)
+        * lib/livebook/config.ex (3)
+        * proto/lib/livebook_proto/deployment_group.pb.ex (2)
+        * lib/livebook_web/plugs/memory_provider.ex (2)
+        * proto/lib/livebook_proto/user_connected.pb.ex (1)
+
+  You can see the first file, "lib/livebook_web.ex", is depended on by 97
+  other files and, because we are using compile-connected, it also means
+  that "lib/livebook_web.ex" itself has its own dependencies. We can find
+  which files depend on "lib/livebook_web.ex" at compile time like this:
+
+      $ mix xref graph --sink lib/livebook_web.ex --label compile --only-nodes
+
+  And you can find the files lib/livebook_web.ex depends on like this:
+
+      $ mix xref graph --source lib/livebook_web.ex --only-nodes
+
+  The trouble here is precisely that, if any of the files in the latter
+  command changes, all of the files in the first command will be recompiled,
+  because compile time dependencies are transitive.
+
+  Having compile time dependencies is a common feature in Elixir projects.
+  However, the modules you depend on at compile-time must avoid dependencies
+  to modules within the same project. You can understand all of the
+  dependencies of a given file by running:
+
+      $ mix xref trace lib/livebook_web.ex
+
+  The command above will output three types of dependencies, which we
+  detail next.
+
+  ### Dependency types
+
+  Elixir tracks three types of dependencies between modules: compile,
+  exports, and runtime. If a module has a compile time dependency on
+  another module, the caller module has to be recompiled whenever the
+  callee changes (or any dependency of the callee changes).
+  Let's see an example:
+
+      # lib/a.ex
+      defmodule A do
+        @hello B.hello()
+        def hello, do: @hello
+      end
+
+      # lib/b.ex
+      defmodule B do
+        def hello, do: "hello"
+        def world, do: C.world()
+      end
+
+      # lib/c.ex
+      defmodule C do
+        def world, do: "world"
+      end
+
+  If `C.world/0` changes, `B` is marked as stale. `B` does not need to
+  be recompiled, because it depends on `C` at runtime, but anything that
+  depends on `B` at compile-time has to recompile, and that includes `A`.
+
+  Compile-time dependencies are typically added when using macros or
+  when invoking functions in the module body (outside of functions).
+  This type of transitive compile-time dependencies, such as `A`
+  depending on `C` at compile-time through `B`, are called compile-connected.
+
+  Export dependencies are compile time dependencies on the module API,
+  namely structs and its public definitions. For example, if you import
+  a module but only use its functions, it is an export dependency. If
+  you use a struct, it is an export dependency too. Export dependencies
+  are only recompiled if the module API changes. Note, however, that compile
+  time dependencies have higher precedence than exports. Therefore if
+  you import a module and use its macros, it is a compile time dependency.
+
+  Runtime dependencies are added whenever you invoke another module
+  inside a function. Modules with runtime dependencies do not have
+  to be compiled when the callee changes, unless there is a transitive
+  compile or an outdated export time dependency between them.
+
+  Over the next sections, we will explain what which `mix xref` command
+  does in detail.
 
   ## mix xref trace FILE
 
@@ -39,8 +133,9 @@ defmodule Mix.Tasks.Xref do
       $ mix xref trace lib/my_app/router.ex --label compile
 
   If you have an umbrella application, we also recommend using the
-  `--include-siblings` flag to see the dependencies on other
-  umbrella applications.
+  `--include-siblings` flag to see the dependencies from sibling
+  applications. The `trace` command is not currently supported at the
+  umbrella root.
 
   ### Example
 
@@ -61,34 +156,38 @@ defmodule Mix.Tasks.Xref do
 
       lib/b.ex:2: require A (export)
       lib/b.ex:3: call A.macro/0 (compile)
-      lib/b.ex:4: import A.macro/0 (compile)
+      lib/b.ex:4: import call A.macro/0 (compile)
       lib/b.ex:5: call A.fun/0 (compile)
-      lib/b.ex:6: call A.fun/0 (compile)
-      lib/b.ex:6: import A.fun/0 (compile)
+      lib/b.ex:6: import call A.fun/0 (compile)
       lib/b.ex:7: call A.macro/0 (compile)
       lib/b.ex:8: call A.fun/0 (runtime)
       lib/b.ex:9: struct A (export)
 
   ## mix xref graph
 
-  Prints a file dependency graph where an edge from `A` to `B` indicates
+  Emits a file dependency graph where an edge from `A` to `B` indicates
   that `A` (source) depends on `B` (sink).
 
       $ mix xref graph --format stats
 
+  For any non-small project, the output of `mix xref graph` itself, without
+  any additional flags, is not useful: once your project grows, it is hard
+  to gather actionable feedback by looking at the graph as a whole. Instead,
+  `mix xref graph` is better used as a "database", which can help you answer
+  queries about your project.
+
   The following options are accepted:
 
     * `--exclude` - path to exclude. Can be repeated to exclude multiple paths.
 
     * `--label` - only shows relationships with the given label.
-      The labels are "compile", "export" and "runtime". By default,
-      the `--label` option simply filters the printed graph to show
-      only relationships with the given label. You can pass `--only-direct`
-      to trim the graph to only the nodes that have the direct
-      relationship given by label. There is also a special label
-      called "compile-connected" that keeps only compile-time files
-      with at least one transitive dependency. See "Dependencies types"
-      section below.
+      The labels are "compile", "export" and "runtime". By default, the `--label`
+      option does not change how the graph is computed, it simply filters the
+      printed graph to show only relationships with the given label. However,
+      you can pass `--only-direct` to trim the graph to only the nodes that
+      have the direct relationship given by label. There is also a special
+      label called "compile-connected" that keeps only compile-time files with
+      at least one transitive dependency. See "Dependency types" section below.
 
     * `--group` - provide comma-separated paths to consider as a group. Dependencies
       from and into multiple files of the group are considered a single dependency.
@@ -129,25 +228,36 @@ defmodule Mix.Tasks.Xref do
       * `dot` - produces a DOT graph description in `xref_graph.dot` in the
         current directory. Warning: this will override any previously generated file
 
+    * `--output` *(since v1.15.0)* - can be set to one of
+
+      * `-` - prints the output to standard output;
+
+      * a path - writes the output graph to the given path
+
+      Defaults to `xref_graph.dot` in the current directory.
+
   The `--source` and `--sink` options are particularly useful when trying to understand
   how the modules in a particular file interact with the whole system. You can combine
   those options with `--label` and `--only-nodes` to get all files that exhibit a certain
   property, for example:
 
-      # To show all compile-time relationships
-      mix xref graph --label compile
+      # To show all compile-connected relationships
+      $ mix xref graph --label compile-connected
 
       # To get the tree that depend on lib/foo.ex at compile time
-      mix xref graph --label compile --sink lib/foo.ex
+      $ mix xref graph --label compile --sink lib/foo.ex
 
       # To get all files that depend on lib/foo.ex at compile time
-      mix xref graph --label compile --sink lib/foo.ex --only-nodes
+      $ mix xref graph --label compile --sink lib/foo.ex --only-nodes
 
       # To get all paths between two files
-      mix xref graph --source lib/foo.ex --sink lib/bar.ex
+      $ mix xref graph --source lib/foo.ex --sink lib/bar.ex
 
       # To show general statistics about the graph
-      mix xref graph --format stats
+      $ mix xref graph --format stats
+
+      # To show all cycles with at least one compile-time dependency
+      $ mix xref graph --format cycles --label compile-connected
 
   ### Understanding the printed graph
 
@@ -190,10 +300,9 @@ defmodule Mix.Tasks.Xref do
 
   The `--label compile` flag removes all non-compile dependencies. However,
   this can be misleading because having direct compile time dependencies is
-  not necessarily an issue. The biggest concern, as mentioned above, are the
-  transitive compile time dependencies. You can get all compile time
-  dependencies that cause transitive compile time dependencies by using
-  `--label compile-connected`:
+  not necessarily an issue. The biggest concern are the transitive compile
+  time dependencies. You can get all compile time dependencies that cause
+  transitive compile time dependencies by using `--label compile-connected`:
 
       $ mix xref graph --label compile-connected
       lib/a.ex
@@ -215,29 +324,41 @@ defmodule Mix.Tasks.Xref do
       lib/a.ex
       └── lib/b.ex (compile)
 
-  ### Dependencies types
+  If you have an umbrella application, we also recommend using the
+  `--include-siblings` flag to see the dependencies from sibling
+  applications. When invoked at the umbrella root, the `graph`
+  command will list all files from all umbrella children, without
+  any namespacing.
 
-  Elixir tracks three types of dependencies between modules: compile,
-  exports, and runtime. If a module has a compile time dependency on
-  another module, the caller module has to be recompiled whenever the
-  callee changes. Compile-time dependencies are typically added when
-  using macros or when invoking functions in the module body (outside
-  of functions). You can list all dependencies in a file by running
-  `mix xref trace path/to/file.ex`.
+  ### Understanding the printed cycle
 
-  Exports dependencies are compile time dependencies on the module API,
-  namely structs and its public definitions. For example, if you import
-  a module but only use its functions, it is an export dependency. If
-  you use a struct, it is an export dependency too. Export dependencies
-  are only recompiled if the module API changes. Note, however, that compile
-  time dependencies have higher precedence than exports. Therefore if
-  you import a module and use its macros, it is a compile time dependency.
+  If you run `mix xref graph --format cycle`, Elixir will print cycles
+  of shape:
 
-  Runtime dependencies are added whenever you invoke another module
-  inside a function. Modules with runtime dependencies do not have
-  to be compiled when the callee changes, unless there is a transitive
-  compile or an outdated export time dependency between them. The option
-  `--label compile-connected` can be used to find the first case.
+      Cycle of length 3:
+
+          lib/c.ex
+          lib/b.ex
+          lib/a.ex
+
+  The cycles are given in order: `c.ex` depends on `b.ex` which depends
+  on `a.ex` which depends on `c.ex`. In particular, you want to avoid
+  cycles with compile dependencies in there. You can find those cycles
+  with:
+
+      $ mix xref graph --format cycles --label compile-connected
+
+  Which may look like this:
+
+      Cycle of length 3:
+
+          lib/c.ex
+          lib/b.ex (compile)
+          lib/a.ex
+
+  This means `c.ex` depends on `b.ex` at compile time. Any compile dependency
+  in a cycle is by definition a compile-connected dependency, which must be
+  generally avoided, as explained earlier in the module documentation.
 
   ## Shared options
 
@@ -248,7 +369,8 @@ defmodule Mix.Tasks.Xref do
 
     * `--include-siblings` - includes dependencies that have `:in_umbrella` set
       to true in the current project in the reports. This can be used to find
-      callers or to analyze graphs between projects
+      callers or to analyze graphs between projects (it applies only to `trace`
+      subcommand)
 
     * `--no-compile` - does not compile even if files require compilation
 
@@ -275,7 +397,8 @@ defmodule Mix.Tasks.Xref do
     only_direct: :boolean,
     sink: :keep,
     source: :keep,
-    min_cycle_size: :integer
+    min_cycle_size: :integer,
+    output: :string
   ]
 
   @impl true
@@ -287,9 +410,11 @@ defmodule Mix.Tasks.Xref do
 
     case args do
       ["callers", module] ->
+        no_umbrella!("callers")
         handle_callers(module, opts)
 
       ["trace", file] ->
+        no_umbrella!("trace")
         handle_trace(file, opts)
 
       ["graph"] ->
@@ -312,6 +437,14 @@ defmodule Mix.Tasks.Xref do
     end
   end
 
+  defp no_umbrella!(task) do
+    if Mix.Project.umbrella?() do
+      Mix.raise(
+        "mix xref #{task} is not supported in the umbrella root. Please run it inside the umbrella applications instead"
+      )
+    end
+  end
+
   @doc """
   Returns a list of information of all the runtime function calls in the project.
 
@@ -337,7 +470,7 @@ defmodule Mix.Tasks.Xref do
         ]
   def calls(opts \\ []) do
     for manifest <- manifests(opts),
-        source(source: source, modules: modules) <- read_manifest(manifest) |> elem(1),
+        {source, source(modules: modules)} <- read_manifest(manifest) |> elem(1),
         module <- modules,
         call <- collect_calls(source, module),
         do: call
@@ -443,9 +576,10 @@ defmodule Mix.Tasks.Xref do
     module = parse_module(module)
 
     file_callers =
-      for source <- sources(opts),
-          reference = reference(module, source),
-          do: {source(source, :source), reference}
+      for manifest <- manifests(opts),
+          {source_path, source_entry} <- read_manifest(manifest) |> elem(1),
+          reference = reference(module, source_entry),
+          do: {source_path, reference}
 
     for {file, type} <- Enum.sort(file_callers) do
       Mix.shell().info([file, " (", type, ")"])
@@ -517,15 +651,20 @@ defmodule Mix.Tasks.Xref do
   ## Trace
 
   @doc false
+  def trace({:alias_reference, meta, module}, env) when env.module != module do
+    case env do
+      %{function: nil} -> add_trace(:compile, :alias, module, module, meta, env)
+      %{context: nil} -> add_trace(:runtime, :alias, module, module, meta, env)
+      %{} -> :ok
+    end
+  end
+
   def trace({:require, meta, module, _opts}, env),
     do: add_trace(require_mode(meta), :require, module, module, meta, env)
 
   def trace({:struct_expansion, meta, module, _keys}, env),
     do: add_trace(:export, :struct, module, module, meta, env)
 
-  def trace({:alias_reference, meta, module}, env) when env.module != module,
-    do: add_trace(mode(env), :alias, module, module, meta, env)
-
   def trace({:remote_function, meta, module, function, arity}, env),
     do: add_trace(mode(env), :call, module, {module, function, arity}, meta, env)
 
@@ -572,7 +711,7 @@ defmodule Mix.Tasks.Xref do
       shell.info([
         Exception.format_file_line(Path.relative_to_cwd(file), line),
         ?\s,
-        Atom.to_string(type),
+        trace_type(type),
         ?\s,
         format_module_or_mfa(module_or_mfa),
         " (#{mode})"
@@ -582,6 +721,9 @@ defmodule Mix.Tasks.Xref do
     end
   end
 
+  defp trace_type(:import), do: "import call"
+  defp trace_type(other), do: Atom.to_string(other)
+
   defp trace_label(nil), do: nil
   defp trace_label("compile"), do: :compile
   defp trace_label("export"), do: :export
@@ -601,7 +743,7 @@ defmodule Mix.Tasks.Xref do
         reduce: {file_references, %{}} do
       {file_references, aliases} ->
         group_paths
-        |> String.split(",")
+        |> String.split(",", trim: true)
         |> check_files(file_references, :group)
         |> group(file_references, aliases)
     end
@@ -688,32 +830,31 @@ defmodule Mix.Tasks.Xref do
   end
 
   defp file_references(filter, opts) do
-    module_sources =
+    module_sources_list =
       for manifest_path <- manifests(opts),
           {manifest_modules, manifest_sources} = read_manifest(manifest_path),
-          module(module: module, sources: sources) <- manifest_modules,
-          source <- sources,
-          source = Enum.find(manifest_sources, &match?(source(source: ^source), &1)),
-          do: {module, source}
+          {module, module(sources: sources)} <- manifest_modules,
+          source_path <- sources,
+          source_entry = manifest_sources[source_path],
+          do: {module, {source_path, source_entry}}
 
-    all_modules = MapSet.new(module_sources, &elem(&1, 0))
+    module_sources = Map.new(module_sources_list)
 
-    Map.new(module_sources, fn {current, source} ->
+    Map.new(module_sources_list, fn {current, {file, source_entry}} ->
       source(
         runtime_references: runtime,
         export_references: exports,
-        compile_references: compile,
-        source: file
-      ) = source
+        compile_references: compile
+      ) = source_entry
 
       compile_references =
-        modules_to_nodes(compile, :compile, current, source, module_sources, all_modules, filter)
+        modules_to_nodes(compile, :compile, current, file, module_sources, filter)
 
       export_references =
-        modules_to_nodes(exports, :export, current, source, module_sources, all_modules, filter)
+        modules_to_nodes(exports, :export, current, file, module_sources, filter)
 
       runtime_references =
-        modules_to_nodes(runtime, nil, current, source, module_sources, all_modules, filter)
+        modules_to_nodes(runtime, nil, current, file, module_sources, filter)
 
       references =
         runtime_references
@@ -725,16 +866,16 @@ defmodule Mix.Tasks.Xref do
     end)
   end
 
-  defp modules_to_nodes(_, label, _, _, _, _, filter) when filter != :all and label != filter do
+  defp modules_to_nodes(_, label, _, _, _, filter) when filter != :all and label != filter do
     %{}
   end
 
-  defp modules_to_nodes(modules, label, current, source, module_sources, all_modules, _filter) do
+  defp modules_to_nodes(modules, label, current, file, module_sources, _filter) do
     for module <- modules,
         module != current,
-        module in all_modules,
-        module_sources[module] != source,
-        do: {source(module_sources[module], :source), label},
+        {source_path, _source_entry} <- [module_sources[module]],
+        file != source_path,
+        do: {source_path, label},
         into: %{}
   end
 
@@ -782,66 +923,46 @@ defmodule Mix.Tasks.Xref do
         true -> file_references
       end
 
-    # Filter according to non direct label
-    file_references = filter(file_references, filter)
-
-    # If a label is given, remove empty root nodes
-    file_references =
-      if opts[:label] do
-        for {_, [_ | _]} = pair <- file_references, into: %{}, do: pair
-      else
-        file_references
-      end
-
-    roots =
-      if sources do
-        Enum.map(sources, &{&1, nil})
-      else
-        file_references
-        |> Map.drop(sinks || [])
-        |> Enum.map(&{elem(&1, 0), nil})
-      end
-
-    callback = fn {file, type} ->
-      children = if opts[:only_nodes], do: [], else: Map.get(file_references, file, [])
-      type = type && "(#{type})"
-      {{file, type}, Enum.sort(children)}
-    end
-
     {found, count} =
       case opts[:format] do
         "dot" ->
-          Mix.Utils.write_dot_graph!(
-            "xref_graph.dot",
-            "xref graph",
-            Enum.sort(roots),
-            callback,
-            opts
-          )
+          {roots, callback, count} =
+            roots_and_callback(file_references, filter, sources, sinks, opts)
 
-          """
-          Generated "xref_graph.dot" in the current directory. To generate a PNG:
+          path = Keyword.get(opts, :output, "xref_graph.dot")
 
-             dot -Tpng xref_graph.dot -o xref_graph.png
+          Mix.Utils.write_dot_graph!(path, "xref graph", Enum.sort(roots), callback, opts)
 
-          For more options see http://www.graphviz.org/.
-          """
-          |> String.trim_trailing()
-          |> Mix.shell().info()
+          if path != "-" do
+            png_path = (path |> Path.rootname() |> Path.basename()) <> ".png"
 
-          {:references, count_references(file_references)}
+            """
+            Generated #{inspect(path)} in the current directory. To generate a PNG:
+
+               dot -Tpng #{inspect(path)} -o #{inspect(png_path)}
+
+            For more options see http://www.graphviz.org/.
+            """
+            |> String.trim_trailing()
+            |> Mix.shell().info()
+          end
+
+          {:references, count}
 
         "stats" ->
-          print_stats(file_references, opts)
+          print_stats(file_references, filter, opts)
           {:stats, 0}
 
         "cycles" ->
-          {:cycles, print_cycles(file_references, opts)}
+          {:cycles, print_cycles(file_references, filter, opts)}
 
         other when other in [nil, "plain", "pretty"] ->
+          {roots, callback, count} =
+            roots_and_callback(file_references, filter, sources, sinks, opts)
+
           Mix.Utils.print_tree(Enum.sort(roots), callback, opts)
 
-          {:references, count_references(file_references)}
+          {:references, count}
 
         other ->
           Mix.raise("Unknown --format #{other} in mix xref graph")
@@ -850,28 +971,6 @@ defmodule Mix.Tasks.Xref do
     check_failure(found, count, opts[:fail_above])
   end
 
-  defp count_references(file_references) do
-    Enum.reduce(file_references, 0, fn {_, refs}, total -> total + length(refs) end)
-  end
-
-  defp filter_fn(file_references, :compile_connected),
-    do: fn {key, type} ->
-      type == :compile and match?([_ | _], file_references[key] || [])
-    end
-
-  defp filter_fn(_file_references, filter),
-    do: fn {_key, type} -> type == filter end
-
-  defp filter(file_references, :all), do: file_references
-
-  defp filter(file_references, filter) do
-    filter_fn = filter_fn(file_references, filter)
-
-    for {key, children} <- file_references,
-        into: %{},
-        do: {key, Enum.filter(children, filter_fn)}
-  end
-
   defp source_tree(file_references, keys) do
     keys
     |> Enum.reduce({%{}, %{}}, fn key, {acc, seen} ->
@@ -910,7 +1009,59 @@ defmodule Mix.Tasks.Xref do
     end)
   end
 
-  defp print_stats(references, opts) do
+  defp roots_and_callback(file_references, filter, sources, sinks, opts) do
+    # Filter according to non direct label
+    file_references = transitive_filter(file_references, filter)
+
+    # If a label is given, remove empty root nodes
+    file_references =
+      if opts[:label] do
+        for {_, [_ | _]} = pair <- file_references, into: %{}, do: pair
+      else
+        file_references
+      end
+
+    roots =
+      if sources do
+        Enum.map(sources, &{&1, nil})
+      else
+        file_references
+        |> Map.drop(sinks || [])
+        |> Enum.map(&{elem(&1, 0), nil})
+      end
+
+    callback = fn {file, type} ->
+      children = if opts[:only_nodes], do: [], else: Map.get(file_references, file, [])
+      type = type && "(#{type})"
+      {{file, type}, Enum.sort(children)}
+    end
+
+    {roots, callback, count_references(file_references)}
+  end
+
+  defp count_references(file_references) do
+    Enum.reduce(file_references, 0, fn {_, refs}, total -> total + length(refs) end)
+  end
+
+  defp transitive_filter_fn(file_references, :compile_connected),
+    do: fn {key, type} ->
+      type == :compile and match?([_ | _], file_references[key] || [])
+    end
+
+  defp transitive_filter_fn(_file_references, filter),
+    do: fn {_key, type} -> type == filter end
+
+  defp transitive_filter(file_references, :all), do: file_references
+
+  defp transitive_filter(file_references, filter) do
+    filter_fn = transitive_filter_fn(file_references, filter)
+
+    for {key, children} <- file_references,
+        into: %{},
+        do: {key, Enum.filter(children, filter_fn)}
+  end
+
+  defp print_stats(references, filter, opts) do
     with_digraph(references, fn graph ->
       shell = Mix.shell()
 
@@ -925,11 +1076,11 @@ defmodule Mix.Tasks.Xref do
       shell.info("Compile dependencies: #{counters.compile} (edges)")
       shell.info("Exports dependencies: #{counters.export} (edges)")
       shell.info("Runtime dependencies: #{counters.nil} (edges)")
-      shell.info("Cycles: #{length(cycles(graph, opts))}")
+      shell.info("Cycles: #{length(cycles(graph, filter, opts))}")
 
       outgoing =
         references
-        |> Enum.map(fn {file, _} -> {:digraph.out_degree(graph, file), file} end)
+        |> Enum.map(fn {file, _} -> {out_stats_filter(references, graph, file, filter), file} end)
         |> Enum.sort(:desc)
         |> Enum.take(10)
 
@@ -938,7 +1089,7 @@ defmodule Mix.Tasks.Xref do
 
       incoming =
         references
-        |> Enum.map(fn {file, _} -> {:digraph.in_degree(graph, file), file} end)
+        |> Enum.map(fn {file, _} -> {in_stats_filter(references, graph, file, filter), file} end)
         |> Enum.sort(:desc)
         |> Enum.take(10)
 
@@ -947,6 +1098,32 @@ defmodule Mix.Tasks.Xref do
     end)
   end
 
+  defp out_stats_filter(_references, graph, file, :all), do: :digraph.out_degree(graph, file)
+
+  defp out_stats_filter(references, graph, file, filter) do
+    filter_fn = transitive_filter_fn(references, filter)
+
+    graph
+    |> :digraph.out_neighbours(file)
+    |> Enum.count(fn v ->
+      {_edge, _v1, _v2, label} = :digraph.edge(graph, {file, v})
+      filter_fn.({file, label})
+    end)
+  end
+
+  defp in_stats_filter(_references, graph, file, :all), do: :digraph.in_degree(graph, file)
+
+  defp in_stats_filter(references, graph, file, filter) do
+    filter_fn = transitive_filter_fn(references, filter)
+
+    graph
+    |> :digraph.in_neighbours(file)
+    |> Enum.count(fn v ->
+      {_edge, _v1, _v2, label} = :digraph.edge(graph, {v, file})
+      filter_fn.({file, label})
+    end)
+  end
+
   defp with_digraph(references, callback) do
     graph = :digraph.new()
 
@@ -956,7 +1133,7 @@ defmodule Mix.Tasks.Xref do
       end
 
       for {file, deps} <- references, {dep, label} <- deps do
-        :digraph.add_edge(graph, file, dep, label)
+        :digraph.add_edge(graph, {file, dep}, file, dep, label)
       end
 
       callback.(graph)
@@ -965,28 +1142,56 @@ defmodule Mix.Tasks.Xref do
     end
   end
 
-  defp cycles(graph, opts) do
+  defp cycles(graph, filter, opts) do
+    # Vertices order in cyclic_strong_components/1 return is arbitrary and changes between
+    # OTP versions, sorting is necessary to make the output stable across versions.
     cycles =
       graph
       |> :digraph_utils.cyclic_strong_components()
-      |> Enum.reduce([], &inner_cycles(graph, &1, &2))
+      |> Enum.reduce([], &inner_cycles(graph, Enum.sort(&1), &2))
       |> Enum.map(&{length(&1), &1})
 
-    if min = opts[:min_cycle_size], do: Enum.filter(cycles, &(elem(&1, 0) > min)), else: cycles
+    cycles =
+      if min = opts[:min_cycle_size] do
+        Enum.filter(cycles, &(elem(&1, 0) > min))
+      else
+        cycles
+      end
+
+    # :compile_connected is the same
+    if cycle_fn = cycle_filter_fn(filter) do
+      Enum.filter(cycles, fn {_length, cycle} -> Enum.any?(cycle, cycle_fn) end)
+    else
+      cycles
+    end
   end
 
+  # In cycles, a compile connected is compile
+  defp cycle_filter_fn(:all), do: nil
+  defp cycle_filter_fn(:compile_connected), do: cycle_filter_fn(:compile)
+  defp cycle_filter_fn(filter), do: fn {_node, type} -> type == filter end
+
   defp inner_cycles(_graph, [], acc), do: acc
 
   defp inner_cycles(graph, [v | vertices], acc) do
     cycle = :digraph.get_cycle(graph, v)
-    inner_cycles(graph, vertices -- cycle, [cycle | acc])
+    inner_cycles(graph, vertices -- cycle, [label_cycle(cycle, graph) | acc])
   end
 
-  defp print_cycles(references, opts) do
+  defp label_cycle([from, to | cycle], graph) do
+    {_edge, _v1, _v2, label} = :digraph.edge(graph, {from, to})
+    [{to, label} | label_cycle([to | cycle], graph)]
+  end
+
+  defp label_cycle([_from], _graph) do
+    []
+  end
+
+  defp print_cycles(references, filter, opts) do
     with_digraph(references, fn graph ->
       shell = Mix.shell()
 
-      case graph |> cycles(opts) |> Enum.sort(:desc) do
+      case graph |> cycles(filter, opts) |> Enum.sort(:desc) do
         [] ->
           shell.info("No cycles found")
           0
@@ -997,8 +1202,9 @@ defmodule Mix.Tasks.Xref do
           for {length, cycle} <- cycles do
             shell.info("Cycle of length #{length}:\n")
 
-            for node <- cycle do
-              shell.info("    " <> node)
+            for {node, type} <- cycle do
+              type = if type, do: " (#{type})", else: ""
+              shell.info("    " <> node <> type)
             end
 
             shell.info("")
@@ -1011,12 +1217,6 @@ defmodule Mix.Tasks.Xref do
 
   ## Helpers
 
-  defp sources(opts) do
-    for manifest <- manifests(opts),
-        source() = source <- read_manifest(manifest) |> elem(1),
-        do: source
-  end
-
   defp apps(opts) do
     siblings =
       if opts[:include_siblings] do
@@ -1032,12 +1232,18 @@ defmodule Mix.Tasks.Xref do
 
   defp manifests(opts) do
     siblings =
-      if opts[:include_siblings] do
-        for %{scm: Mix.SCM.Path, opts: opts} <- Mix.Dep.cached(),
-            opts[:in_umbrella],
-            do: Path.join([opts[:build], ".mix", @manifest])
-      else
-        []
+      cond do
+        Mix.Project.umbrella?() ->
+          for %{opts: opts} <- Mix.Dep.Umbrella.cached(),
+              do: Path.join([opts[:build], ".mix", @manifest])
+
+        opts[:include_siblings] ->
+          for %{scm: Mix.SCM.Path, opts: opts} <- Mix.Dep.cached(),
+              opts[:in_umbrella],
+              do: Path.join([opts[:build], ".mix", @manifest])
+
+        true ->
+          []
       end
 
     [Path.join(Mix.Project.manifest_path(), @manifest) | siblings]
