@@ -7,7 +7,7 @@ defmodule Mix.Utils do
   @doc """
   Gets the Mix home.
 
-  It uses the the locations `MIX_HOME`, `XDG_DATA_HOME/mix`,
+  It uses the locations `MIX_HOME`, `XDG_DATA_HOME/mix`,
   `~/.mix` with decreasing priority.
 
   Developers should only store entries in the
@@ -225,6 +225,7 @@ defmodule Mix.Utils do
   @doc """
   Prints n files are being compiled with the given extension.
   """
+  def compiling_n(0, _ext), do: :ok
   def compiling_n(1, ext), do: Mix.shell().info("Compiling 1 file (.#{ext})")
   def compiling_n(n, ext), do: Mix.shell().info("Compiling #{n} files (.#{ext})")
 
@@ -327,6 +328,8 @@ defmodule Mix.Utils do
 
   The callback will be invoked for each node and it
   must return a `{printed, children}` tuple.
+
+  If `path` is `-`, prints the output to standard output.
   """
   @spec write_dot_graph!(
           Path.t(),
@@ -338,7 +341,13 @@ defmodule Mix.Utils do
         when node: term()
   def write_dot_graph!(path, title, nodes, callback, _opts \\ []) do
     {dot, _} = build_dot_graph(make_ref(), nodes, MapSet.new(), callback)
-    File.write!(path, ["digraph ", quoted(title), " {\n", dot, "}\n"])
+    contents = ["digraph ", quoted(title), " {\n", dot, "}\n"]
+
+    if path == "-" do
+      IO.write(contents)
+    else
+      File.write!(path, contents)
+    end
   end
 
   defp build_dot_graph(_parent, [], seen, _callback), do: {[], seen}
@@ -435,10 +444,7 @@ defmodule Mix.Utils do
     |> Enum.map_join(".", &Macro.camelize/1)
   end
 
-  @doc """
-  Symlinks or copy with the option to force a hard copy.
-  See `symlink_or_copy/2`.
-  """
+  @deprecated "Use symlink_or_copy/2"
   def symlink_or_copy(hard_copy?, source, target) do
     if hard_copy? do
       if File.exists?(source) do
@@ -460,13 +466,19 @@ defmodule Mix.Utils do
   Expects source and target to be absolute paths as it generates
   a relative symlink.
   """
+  def symlink_or_copy(path, path) do
+    :ok
+  end
+
   def symlink_or_copy(source, target) do
     if File.exists?(source) do
       # Relative symbolic links on Windows are broken
       link =
         case :os.type() do
           {:win32, _} -> source
-          _ -> make_relative_path(source, target)
+          # We are needing the relative path to the parent dir since we are doing
+          # a symlink
+          _ -> Path.relative_to(source, Path.dirname(target), force: true)
         end
 
       case File.read_link(target) do
@@ -500,7 +512,7 @@ defmodule Mix.Utils do
           do_symlink_or_copy(source, target, link)
 
         {:error, _} ->
-          unless File.dir?(target) do
+          if not File.dir?(target) do
             File.rm_rf!(target)
           end
 
@@ -530,21 +542,6 @@ defmodule Mix.Utils do
     end
   end
 
-  # Make a relative path between the two given paths.
-  # Expects both paths to be fully expanded.
-  defp make_relative_path(source, target) do
-    do_make_relative_path(Path.split(source), Path.split(target))
-  end
-
-  defp do_make_relative_path([h | t1], [h | t2]) do
-    do_make_relative_path(t1, t2)
-  end
-
-  defp do_make_relative_path(source, target) do
-    base = List.duplicate("..", max(length(target) - 1, 0))
-    Path.join(base ++ source)
-  end
-
   @doc """
   Opens and reads content from either a URL or a local file system path.
 
@@ -644,6 +641,10 @@ defmodule Mix.Utils do
   end
 
   defp read_httpc(path) do
+    Mix.ensure_application!(:public_key)
+    Mix.ensure_application!(:ssl)
+    Mix.ensure_application!(:inets)
+
     {:ok, _} = Application.ensure_all_started(:ssl)
     {:ok, _} = Application.ensure_all_started(:inets)
 
@@ -651,16 +652,33 @@ defmodule Mix.Utils do
     # the effects of using an HTTP proxy to this function
     {:ok, _pid} = :inets.start(:httpc, profile: :mix)
 
-    headers = [{'user-agent', 'Mix/#{System.version()}'}]
+    headers = [{~c"user-agent", ~c"Mix/#{System.version()}"}]
     request = {:binary.bin_to_list(path), headers}
 
+    # allow override of system CA certs to support running on managed networks
+    # using an SSL proxy etc. Piggy back on Hex defined environment variable
+    # rather than creating a new one, as these are almost always going to be
+    # set and used together.
+    cacert_opt =
+      case System.get_env("HEX_CACERTS_PATH") do
+        nil -> {:cacerts, :public_key.cacerts_get()}
+        file -> {:cacertfile, file}
+      end
+
+    # Use the system certificates
+    ssl_options = [
+      cacert_opt,
+      verify: :verify_peer,
+      customize_hostname_check: [match_fun: :public_key.pkix_verify_hostname_match_fun(:https)]
+    ]
+
     # We are using relaxed: true because some servers is returning a Location
     # header with relative paths, which does not follow the spec. This would
     # cause the request to fail with {:error, :no_scheme} unless :relaxed
     # is given.
     #
     # If a proxy environment variable was supplied add a proxy to httpc.
-    http_options = [relaxed: true] ++ proxy_config(path)
+    http_options = [relaxed: true, ssl: ssl_options] ++ proxy_config(path)
 
     # Silence the warning from OTP as we verify the contents
     level = Logger.level()
