@@ -27,21 +27,11 @@ defmodule Mix.Tasks.DepsPathTest do
     end
   end
 
-  @tag apps: [:raw_sample]
-  test "does not mark for compilation on get/update" do
-    Mix.Project.push(DepsApp)
-
-    in_fixture("deps_status", fn ->
-      Mix.Tasks.Deps.Get.run(["--all"])
-      refute File.exists?("custom/raw_repo/.fetch")
-    end)
-  end
-
   @tag apps: [:raw_sample]
   test "compiles and runs even if lock does not match" do
-    Mix.Project.push(DepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(DepsApp)
+
       Mix.Dep.Lock.write(%{raw_repo: "abcdef"})
       Mix.Tasks.Run.run(["-e", "Mix.shell().info RawRepo.hello()"])
       assert_received {:mix_shell, :info, ["==> raw_repo"]}
@@ -51,9 +41,9 @@ defmodule Mix.Tasks.DepsPathTest do
 
   @tag apps: [:raw_sample]
   test "uses the name of the app, not the path basename" do
-    Mix.Project.push(MismatchDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(MismatchDepsApp)
+
       Mix.Tasks.Deps.Compile.run([])
       assert File.exists?("_build/dev/lib/cooked_repo/ebin")
     end)
