@@ -57,6 +57,18 @@ defmodule Mix.Tasks.DepsTest do
     end
   end
 
+  defmodule RawRepoDep do
+    def project do
+      [
+        app: :raw_sample,
+        version: "0.1.0",
+        deps: [
+          {:raw_repo, "0.1.0", path: "custom/raw_repo"}
+        ]
+      ]
+    end
+  end
+
   ## deps
 
   test "prints list of dependencies and their status" do
@@ -270,36 +282,6 @@ defmodule Mix.Tasks.DepsTest do
     end)
   end
 
-  test "does not load deps with --no-load-deps" do
-    in_fixture("deps_status", fn ->
-      Mix.Project.push(SuccessfulDepsApp)
-
-      # Start from scratch!
-      File.rm_rf("_build")
-
-      Mix.Tasks.Deps.Compile.run([])
-      Mix.Tasks.Deps.Loadpaths.run([])
-      assert File.exists?("_build/dev/lib/ok/ebin/ok.app")
-      assert File.exists?("_build/dev/lib/ok/priv/sample")
-
-      Mix.Tasks.Compile.run([])
-      assert to_charlist(Path.expand("_build/dev/lib/ok/ebin/")) in :code.get_path()
-      assert File.exists?("_build/dev/lib/sample/ebin/sample.app")
-
-      # Remove the deps without build_path
-      Mix.ProjectStack.post_config(deps: [])
-      Mix.State.clear_cache()
-      Mix.Project.pop()
-      Mix.Project.push(SuccessfulDepsApp)
-      Code.delete_path("_build/dev/lib/ok/ebin")
-
-      Mix.Tasks.Deps.Loadpaths.run(["--no-load-deps"])
-      refute to_charlist(Path.expand("_build/dev/lib/ok/ebin/")) in :code.get_path()
-      assert File.exists?("_build/dev/lib/ok/ebin/ok.app")
-      assert File.exists?("_build/dev/lib/sample/ebin/sample.app")
-    end)
-  end
-
   ## deps.unlock
 
   test "unlocks all deps", context do
@@ -414,18 +396,6 @@ defmodule Mix.Tasks.DepsTest do
 
   ## Deps environment
 
-  defmodule DepsEnvApp do
-    def project do
-      [
-        app: :raw_sample,
-        version: "0.1.0",
-        deps: [
-          {:raw_repo, "0.1.0", path: "custom/raw_repo"}
-        ]
-      ]
-    end
-  end
-
   defmodule CustomDepsEnvApp do
     def project do
       [
@@ -440,7 +410,7 @@ defmodule Mix.Tasks.DepsTest do
 
   test "sets deps env to prod by default" do
     in_fixture("deps_status", fn ->
-      Mix.Project.push(DepsEnvApp)
+      Mix.Project.push(RawRepoDep)
 
       Mix.Tasks.Deps.Update.run(["--all"])
       assert_received {:mix_shell, :info, [":raw_repo env is prod"]}
@@ -537,7 +507,7 @@ defmodule Mix.Tasks.DepsTest do
 
   test "fails on diverged dependencies on get/update" do
     in_fixture("deps_status", fn ->
-      Mix.Project.push(ConflictDepsApp)
+      Mix.Project.push(ConflictDepsApp, "mix.exs")
 
       assert_raise Mix.Error, fn ->
         Mix.Tasks.Deps.Loadpaths.run([])
@@ -608,11 +578,11 @@ defmodule Mix.Tasks.DepsTest do
     end)
   end
 
-  @overriding_msg "  the dependency git_repo in mix.exs is overriding"
+  @overriding_msg "  the dependency git_repo in custom/deps_repo/mix.exs is overriding"
 
   test "fails on diverged dependencies even when optional" do
     in_fixture("deps_status", fn ->
-      Mix.Project.push(ConvergedDepsApp)
+      Mix.Project.push(ConvergedDepsApp, "custom/deps_repo/mix.exs")
 
       File.write!("custom/deps_repo/mix.exs", """
       defmodule DepsRepo do
@@ -700,15 +670,15 @@ defmodule Mix.Tasks.DepsTest do
 
   test "converged dependencies errors if not overriding" do
     in_fixture("deps_status", fn ->
-      Mix.Project.push(NonOverriddenDepsApp)
+      Mix.Project.push(NonOverriddenDepsApp, "custom_mix.exs")
 
       assert_raise Mix.Error, fn ->
         Mix.Tasks.Deps.Loadpaths.run([])
       end
 
       receive do
-        {:mix_shell, :error, ["  the dependency git_repo in mix.exs" <> _ = msg]} ->
-          assert msg =~ "In mix.exs:"
+        {:mix_shell, :error, ["  the dependency git_repo in custom_mix.exs" <> _ = msg]} ->
+          assert msg =~ "In custom_mix.exs:"
 
           assert msg =~
                    "{:git_repo, \"0.1.0\", [env: :prod, git: #{inspect(fixture_path("git_repo"))}]}"
@@ -775,6 +745,41 @@ defmodule Mix.Tasks.DepsTest do
     end)
   end
 
+  test "checks if compile env changed" do
+    in_fixture("deps_status", fn ->
+      Mix.Project.push(RawRepoDep)
+      Mix.Tasks.Deps.Loadpaths.run([])
+      assert_receive {:mix_shell, :info, ["Generated raw_repo app"]}
+      assert Application.spec(:raw_repo, :vsn)
+
+      File.mkdir_p!("config")
+
+      File.write!("config/config.exs", """
+      import Config
+      config :raw_repo, :compile_env, :new_value
+      """)
+
+      Application.unload(:raw_repo)
+      Mix.ProjectStack.pop()
+      Mix.Task.clear()
+      Mix.Project.push(RawRepoDep)
+      purge([RawRepo])
+      Mix.Tasks.Loadconfig.load_compile("config/config.exs")
+
+      Mix.Tasks.Deps.run([])
+
+      assert_receive {:mix_shell, :info,
+                      ["  the dependency build is outdated, please run \"mix deps.compile\""]}
+
+      Mix.Tasks.Deps.Loadpaths.run([])
+
+      assert_receive {:mix_shell, :info, ["Generated raw_repo app"]}
+      assert Application.spec(:raw_repo, :vsn)
+    end)
+  after
+    Application.delete_env(:raw_repo, :compile_env, persistent: true)
+  end
+
   defmodule NonCompilingDeps do
     def project do
       [
